/*
 * Authors:  Joe Habermann July 19, 1996
 *           Thomas M. Ruwart June 8, 1998
 *           Jon Kappler April, 26, 1999
 *           David Porter Feb. 2001
 *           Laboratory for Computational Science & Engineering
 *           University of Minnesota
 *
 */
#include "stdafx.h"

#include <windows.h>         /* must include this before GL/gl.h */
#include <sys/types.h>
#include <time.h>
#include <sys/timeb.h>
#include <GL/gl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define YES 1
#define NO  0


static HDC hDC;				/* device context */
static HPALETTE hPalette = 0;			/* custom palette (if needed) */
static GLboolean animate = GL_TRUE;		/* animation flag */

/* win32 didn't have gettimeofday, so I added one, jk */
static void gettimeofday( struct _timeb *tv, struct _timeb *t2 )
{
	struct _timeb tbuf;

	_ftime( &tbuf );
	tv->time = tbuf.time;
	tv->millitm = tbuf.millitm;
	return;
}

static double timediff(struct _timeb *tim1, struct _timeb *tim2)
{
    int  sec, usec;
	double t;

    sec = tim2->time - tim1->time;
    usec = tim2->millitm - tim1->millitm;
    if (usec < 0) {
        sec--;
        usec = tim2->millitm + 1000 - tim1->millitm;
    }
//    return(((double)(sec * 1000000 + usec))/1000000);
//	sec = sec + (usec / 1000.0);
	t = usec;
	t = t / 1000.0;
	t = t + (double) sec;
    return t;
}

LONG WINAPI
WindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{ 
    static PAINTSTRUCT ps;

    switch(uMsg) {
    case WM_PAINT: 
//	display();
	BeginPaint(hWnd, &ps);
	EndPaint(hWnd, &ps);
	return 0;

    case WM_SIZE:
	glViewport(0, 0, LOWORD(lParam), HIWORD(lParam));
	PostMessage(hWnd, WM_PAINT, 0, 0);
	return 0;

    case WM_CHAR:
	switch (wParam) {
	case 27:			/* ESC key */
	    PostQuitMessage(0);
	    break;
	case ' ':
	    animate = !animate;
	    break;
	}
	return 0;

    case WM_ACTIVATE:
	if (IsIconic(hWnd))
	    animate = GL_FALSE;
	else
	    animate = GL_TRUE;
	return 0;

    case WM_PALETTECHANGED:
	if (hWnd == (HWND)wParam)
	    break;
	/* fall through to WM_QUERYNEWPALETTE */

    case WM_QUERYNEWPALETTE:
	if (hPalette) {
	    UnrealizeObject(hPalette);
	    SelectPalette(hDC, hPalette, FALSE);
	    RealizePalette(hDC);
	    return TRUE;
	}
	return FALSE;

    case WM_CLOSE:
	PostQuitMessage(0);
	return 0;
    }

    return DefWindowProc(hWnd, uMsg, wParam, lParam); 
} 

HWND
CreateOpenGLWindow(char* title, int x, int y, int width, int height, 
		   BYTE type, DWORD flags, int depth, HWND hLocalWnd)
{
    int         n, pf;
    HWND        hWnd;
    WNDCLASS    wc;
    LOGPALETTE* lpPal;
    PIXELFORMATDESCRIPTOR pfd;
    static HINSTANCE hInstance = 0;
	FILE *fp;

    if (hLocalWnd == NULL) {
	    /* only register the window class once - use hInstance as a flag. */
		if (!hInstance) {
			hInstance = GetModuleHandle(NULL);
			wc.style         = CS_OWNDC;
			wc.lpfnWndProc   = (WNDPROC)WindowProc;
			wc.cbClsExtra    = 0;
			wc.cbWndExtra    = 0;
			wc.hInstance     = hInstance;
			wc.hIcon         = LoadIcon(NULL, IDI_WINLOGO);
			wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
			wc.hbrBackground = NULL;
			wc.lpszMenuName  = NULL;
	//		wc.lpszClassName = (const unsigned short *)"OpenGL";
			wc.lpszClassName = "OpenGL";

			if (!RegisterClass(&wc)) {
				fp = fopen("C:/DEBUG_HVR.txt", "a+");
				fprintf(fp, "RegisterClass failed\n");
				fclose(fp);
				return NULL;
			}
		}

		hWnd = CreateWindow("OpenGL", title, WS_POPUP,
				x, y, width, height, NULL, NULL, hInstance, NULL);
	} else {
		hWnd = hLocalWnd;
	}

//    hWnd = CreateWindow("OpenGL", title, WS_OVERLAPPEDWINDOW |
//			WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
//			x, y, width, height, NULL, NULL, hInstance, NULL);

    if (hWnd == NULL) {
		fp = fopen("C:/DEBUG_HVR.txt", "a+");
		fprintf(fp, "hWnd = hLocalWnd or CreateWindow   produced NULL!! \n");
		fclose(fp);
		return NULL;
    }




/*
------------------------------------------
PIXELFORMATDESCRIPTOR pfd = { 
    sizeof(PIXELFORMATDESCRIPTOR),   // size of this pfd 
    1,                     // version number 
    PFD_DRAW_TO_WINDOW |   // support window 
    PFD_SUPPORT_OPENGL |   // support OpenGL 
    PFD_DOUBLEBUFFER,      // double buffered 
    PFD_TYPE_RGBA,         // RGBA type 
    24,                    // 24-bit color depth 
    0, 0, 0, 0, 0, 0,      // color bits ignored 
    0,                     // no alpha buffer 
    0,                     // shift bit ignored 
    0,                     // no accumulation buffer 
    0, 0, 0, 0,            // accum bits ignored 
    32,                    // 32-bit z-buffer 
    0,                     // no stencil buffer 
    0,                     // no auxiliary buffer 
    PFD_MAIN_PLANE,        // main layer 
    0,                     // reserved 
    0, 0, 0                // layer masks ignored 
}; 
HDC  hdc; 
int  iPixelFormat; 
 
// get the best available match of pixel format for the device context  
iPixelFormat = ChoosePixelFormat(hdc, &pfd); 
 
// make that the pixel format of the device context 
SetPixelFormat(hdc, iPixelFormat, &pfd); 

------------------------------------------

*/


    hDC = ::GetDC(hWnd);

    /* there is no guarantee that the contents of the stack that become
       the pfd are zeroed, therefore _make sure_ to clear these bits. */
    memset(&pfd, 0, sizeof(pfd));
    pfd.nSize        = sizeof(pfd);
    pfd.nVersion     = 1;
    pfd.dwFlags      = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | flags;
    pfd.iPixelType   = type;
    pfd.cColorBits   = depth; // 32 is probably the best choice

    pf = ChoosePixelFormat(hDC, &pfd);
    if (pf == 0) {
		fp = fopen("C:/DEBUG_HVR.txt", "a+");
		fprintf(fp, "CoosePixelFormat failed \n");
		fclose(fp);
		return NULL;
    } 
 
    if (SetPixelFormat(hDC, pf, &pfd) == FALSE) {

		DWORD dwError = 0;
		dwError = GetLastError();
		char cStr[256];

//		LPVOID lpMsgBuf;
		FormatMessage( 
			/* FORMAT_MESSAGE_ALLOCATE_BUFFER | */
			FORMAT_MESSAGE_FROM_SYSTEM | 
			FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL,
			dwError,
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
			/* (LPTSTR) &lpMsgBuf, */
			cStr,
			256,
			NULL 
		);
		// Process any inserts in lpMsgBuf.
		// ...
		// Display the string.
		fp = fopen("C:/DEBUG_HVR.txt", "a+");
		fprintf(fp, "SetPixelFormat failed \n");
		fprintf(fp, "--> ERROR NUMBER IS: %lu\n",  dwError);
		fprintf(fp, "--> ERROR STRING IS: %s\n\n", cStr);

		fclose(fp);
		return NULL;
    } 

    DescribePixelFormat(hDC, pf, sizeof(PIXELFORMATDESCRIPTOR), &pfd);

    if (pfd.dwFlags & PFD_NEED_PALETTE ||
	pfd.iPixelType == PFD_TYPE_COLORINDEX) {

	n = 1 << pfd.cColorBits;
	if (n > 256) n = 256;

	lpPal = (LOGPALETTE*)malloc(sizeof(LOGPALETTE) +
				    sizeof(PALETTEENTRY) * n);
	if (lpPal == NULL)
	{
		fp = fopen("C:/DEBUG_HVR.txt", "a+");
		fprintf(fp, "malloc of lpPal failed \n");
		fclose(fp);
		return NULL;
	}
	memset(lpPal, 0, sizeof(LOGPALETTE) + sizeof(PALETTEENTRY) * n);
	lpPal->palVersion = 0x300;
	lpPal->palNumEntries = n;

	GetSystemPaletteEntries(hDC, 0, n, &lpPal->palPalEntry[0]);
    
	/* if the pixel type is RGBA, then we want to make an RGB ramp,
	   otherwise (color index) set individual colors. */
	if (pfd.iPixelType == PFD_TYPE_RGBA) {
	    int redMask = (1 << pfd.cRedBits) - 1;
	    int greenMask = (1 << pfd.cGreenBits) - 1;
	    int blueMask = (1 << pfd.cBlueBits) - 1;
	    int i;

	    /* fill in the entries with an RGB color ramp. */
	    for (i = 0; i < n; ++i) {
		lpPal->palPalEntry[i].peRed = 
		    (((i >> pfd.cRedShift)   & redMask)   * 255) / redMask;
		lpPal->palPalEntry[i].peGreen = 
		    (((i >> pfd.cGreenShift) & greenMask) * 255) / greenMask;
		lpPal->palPalEntry[i].peBlue = 
		    (((i >> pfd.cBlueShift)  & blueMask)  * 255) / blueMask;
		lpPal->palPalEntry[i].peFlags = 0;
	    }
	} else {
	    lpPal->palPalEntry[0].peRed = 0;
	    lpPal->palPalEntry[0].peGreen = 0;
	    lpPal->palPalEntry[0].peBlue = 0;
	    lpPal->palPalEntry[0].peFlags = PC_NOCOLLAPSE;
	    lpPal->palPalEntry[1].peRed = 255;
	    lpPal->palPalEntry[1].peGreen = 0;
	    lpPal->palPalEntry[1].peBlue = 0;
	    lpPal->palPalEntry[1].peFlags = PC_NOCOLLAPSE;
	    lpPal->palPalEntry[2].peRed = 0;
	    lpPal->palPalEntry[2].peGreen = 255;
	    lpPal->palPalEntry[2].peBlue = 0;
	    lpPal->palPalEntry[2].peFlags = PC_NOCOLLAPSE;
	    lpPal->palPalEntry[3].peRed = 0;
	    lpPal->palPalEntry[3].peGreen = 0;
	    lpPal->palPalEntry[3].peBlue = 255;
	    lpPal->palPalEntry[3].peFlags = PC_NOCOLLAPSE;
	}

	hPalette = CreatePalette(lpPal);
	if (hPalette) {
	    SelectPalette(hDC, hPalette, FALSE);
	    RealizePalette(hDC);
	}

	free(lpPal);
    }

    ReleaseDC(hWnd, hDC);
//ReleaseDC(hDC, hWnd);

	if(hWnd == NULL) {
		fp = fopen("C:/DEBUG_HVR.txt", "a+");
		fprintf(fp, "hWnd was set to NULL, somehow, by the end of CreateOpenGLWindow \n");
		fclose(fp);
	}

    return hWnd;
}    

void deleteGLPallet()
{
	DeleteObject(hPalette);
}

#if 0

static    HDC hDC;				/* device context */
static    HGLRC hRC;				/* opengl context */
static    HWND  hWnd;				/* window */
static    MSG   msg;				/* message */


    int depth = 0;
    int sizex = 0, sizey = 0; 
	BYTE  colortype;
	DWORD buffertype;
	char s[4096];

/************************************************************
 *  Here's how to create an OpenGL window using wgl         *
 ************************************************************/
	*s = '\0';
	strcat(s,"DHPs Stupid Test");
	sizex = 512;
	sizey = 512;
	colortype = PFD_TYPE_RGBA;
//	buffertype = PFD_DOUBLEBUFFER;
	buffertype = 0;
	depth = 24;
    hWnd = CreateOpenGLWindow(s, 0, 0, sizex, sizey, colortype, buffertype, depth);
    if (hWnd == NULL) { exit(1); }
	hDC = GetDC(hWnd);
	hRC = wglCreateContext(hDC);
	wglMakeCurrent(hDC, hRC);

	ShowWindow(hWnd, SW_SHOW);
	UpdateWindow(hWnd);
/*
    glViewport(0, 0, sizex, sizey);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho (0, sizex, 0, sizey, -1.0, 1.0);
    glMatrixMode(GL_MODELVIEW);
    glDisable(GL_TEXTURE_1D);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_DEPTH_TEST);
*/
/************************************************************
 *  Here's how to bring it down cleanly                     *
 ************************************************************/
	wglMakeCurrent(hDC, NULL);
	wglDeleteContext(hRC);
	DestroyWindow(hWnd);
	ReleaseDC(hWnd, hDC);
	if (hPalette) { DeleteObject(hPalette); }

#endif