
#include <string.h>
#include <stdio.h>

#include <math.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>

#include <stdlib.h>
#include <limits.h>
#include <errno.h>
#include <linux/stat.h>


#define MAXVECLEGTH   10000
float xvec[MAXVECLEGTH];
float yvec[MAXVECLEGTH];
float zvec[MAXVECLEGTH];
float vvec[MAXVECLEGTH];

float dxinv;
float xmin0, xmax0, ymin0, ymax0, zmin0, zmax0;
int nx0, ny0, nz0, nxb, nyb, nzb;

float *vvx, *vvy, *vvz;

float xsc, ysc, zsc, timetot;
float dx1, dx2, dx3, dy1, dy2, dy3, dz1, dz2, dz3;
int  ns1, ns2, ns3, maxout;
char  cInFile[256];


int main(int argc, char* argv[])
{
  int i,j, nbytes, iDir, iEnd, length;
  int   nvecs;
  char cOutFile[256], cTmp[256];
  float x0, y0, z0, xsmin, ysmin, zsmin, dt, ds;
  float xs, ys, zs, shift, dir;
  float dx, xs0, ys0, zs0;
  int jmid, jmin, jmax, idir;
  int i1,i2,i3, istop;
  float dsout,vvmin,tsum,vx,vy,vz,vv, snext, ssum;


  void readfieldformat();
  void readfield(char *cFile, int nbytes, float *fbuf);
  float vinterp(float xx, float yy, float zz, float *vv);
  FILE *fp;


 /*--------------------------------------------------------
  * Parse command line
  */
  if(argc != 2) {
    printf("usage: getvec <path_to_Bx_file>\n");
    return 0;
  }
  sscanf(argv[1], "%s", cInFile);

 /*--------------------------------------------------------
  * Parse Input file name
  */
  iDir  = 1;
  iEnd = 0;
  for(i=  0; i<256; i++) { if(cInFile[i] ==  '/') { iDir = i+2; } }
  for(i=255; i>= 0; i--) { if(cInFile[i] == '\0') { iEnd = i;   } }
  printf("cInFile[iEnd-3] = -->%s<--\n", &cInFile[iEnd-3]);
  if(strcmp(&cInFile[iEnd-3], "aaa") == 0) { iEnd = iEnd - 3; }

 /*--------------------------------------------------------
  * Generate output file name
  */
  sprintf(cOutFile, "%s", cInFile);
  cOutFile[iDir] = '_';
  sprintf(&cOutFile[iEnd], ".vec");
  printf("OUtPUt-->%s<--\n", cOutFile);

 /*--------------------------------------------------------
  * Read dimensions of problem and format of input file(s)
  */
  readfieldformat();


 /*--------------------------------------------------------
  * Allocate space & read vector fields
  */
  nbytes = 4 * nx0 * ny0 * nz0;
  vvx = (float *)malloc(nbytes);
  vvy = (float *)malloc(nbytes);
  vvz = (float *)malloc(nbytes);

  cInFile[iDir] = 'x';   readfield(cInFile, nbytes, vvx);
  cInFile[iDir] = 'y';   readfield(cInFile, nbytes, vvy);
  cInFile[iDir] = 'z';   readfield(cInFile, nbytes, vvz);



 /*--------------------------------------------------------
  * Generate vector objects
  * 1) generate list of starting points
  *    Option 1) uniform sampling in subvolume
  *    Option 2) preferentially from points of large (vx,vy,vz)
  *    will start with option 1
  *    - specify: (xc,yc,zc), ns_i, dx_i,dy_i,dz_i   where _i={1,2,3}
  * 2) generate point strip from each starting point (x,y,z)
  *    - specify:
  *       timetot      sum of all dtime for each strip
  *       maxout       maximum number of output points
  *       dsout        distance between output points   
  */

  xsc = 0.0;    ns1 =  1;    dx1 = 1.0; dy1 = 0.0; dz1 = 0.0;
  ysc = 0.0;    ns2 =  1;    dx2 = 0.0; dy2 = 1.0; dz2 = 0.0;
  zsc = 0.0;    ns3 = 30;    dx3 = 0.0; dy3 = 0.0; dz3 = 1.0;
  timetot = 500.0;
  maxout  = 1000;

  // read_start_points();

  dx     = (xmax0 - xmin0) / (float)nx0;
  ds     = 0.2 * dx;   // will be dt*|V|
  dsout  = 1.0 * dx;
  shift = 0.5 * dx;
  vvmin = 0.01 * ds / timetot;
  dxinv = 1.0 / dx;

  printf("dx=%f  ds=%f  dsout=%f\n", dx, ds, dsout);

  xs0 = xsc - 0.5*(dx1*(float)(ns1-1)+dx2*(float)(ns2-1)+dx3*(float)(ns3-1));
  ys0 = ysc - 0.5*(dy1*(float)(ns1-1)+dy2*(float)(ns2-1)+dy3*(float)(ns3-1));
  zs0 = zsc - 0.5*(dz1*(float)(ns1-1)+dz2*(float)(ns2-1)+dz3*(float)(ns3-1));

  jmid = MAXVECLEGTH / 2;
  nvecs = ns1 * ns2 * ns3;
  if(nvecs > 0) {
    fp = fopen(cOutFile, "w");
    fprintf(fp, "#ntubes         %d\n", nvecs);

    for(i3=0; i3<ns3; i3++) {
    for(i2=0; i2<ns2; i2++) {
    for(i1=0; i1<ns1; i1++) {

      for(idir = -1; idir < 2; idir += 2) {

        dir = (float)idir;

        // Initialize the point
        tsum = 0.0;
        ssum = 0.0;
        if(idir < 0) { snext =     0;   j = jmid-1; }
        else         { snext = dsout;   j = jmid; }

        xs = xs0 + dx1*(float)i1 + dx2*(float)i2 + dx3*(float)i3;
        ys = ys0 + dy1*(float)i1 + dy2*(float)i2 + dy3*(float)i3;
        zs = zs0 + dz1*(float)i1 + dz2*(float)i2 + dz3*(float)i3;

        while(tsum < timetot) {

          // Interpolate (face centered) fields for the point
          vx = dir * vinterp(xs      , ys-shift, zs-shift, vvx);
          vy = dir * vinterp(xs-shift, ys      , zs-shift, vvy);
          vz = dir * vinterp(xs-shift, ys-shift, zs      , vvz);
          vv = sqrt(vx*vx + vy*vy + vz*vz);

          // Report the point
          if(ssum >= snext) {
            xvec[j] = xs;
            yvec[j] = ys;
            zvec[j] = zs;
            vvec[j] = vv;
            snext += dsout;
            j = j + idir;
          }

          // Stop if the point goes outside the volume or vv too small
          // Otherwise, itterate the point.
          istop = 0;
          if(vv < vvmin) { istop = 1; }
          if(xs < xmin0  ||  xs > xmax0) { istop = 1; }
          if(ys < ymin0  ||  ys > ymax0) { istop = 1; }
          if(zs < zmin0  ||  zs > zmax0) { istop = 1; }

          if(istop == 0) {
            dt = ds / vv;
            xs += vx*dt;
            ys += vy*dt;
            zs += vz*dt;
            tsum += dt;
            ssum += ds;
          } else {
            tsum = timetot + 1.0;
          }
        }
        if(ssum > snext-dsout) {
          xvec[j] = xs;
          yvec[j] = ys;
          zvec[j] = zs;
          vvec[j] = vv;
          j = j + idir;
        }
        if(idir < 0) { jmin = j+1; }
        else         { jmax = j; }
      }
      length = jmax - jmin;


      // Output strip of points
      fprintf(fp, "#tubelength     %d\n", length);
      for(j=jmin; j<jmax; j++) {
        fprintf(fp, "%f  %f  %f  %f\n", xvec[j], yvec[j], zvec[j], vvec[j]);
      }
    }}}
    fclose(fp);
    printf("closed\n");
  }

  return 0;
}



void readfieldformat()
{
  FILE *fp;
  char cline[256], ckey[256];

  fp = fopen("fieldformat", "r");
  while(fgets(cline, 256, fp) != NULL) {
    sscanf(cline, "%s", ckey);
    if(strcmp(ckey,"xcoord")==0){
      sscanf(cline,"%s %f %f %d", ckey, &xmin0, &xmax0, &nx0);
    }
    if(strcmp(ckey,"ycoord")==0){
      sscanf(cline,"%s %f %f %d", ckey, &ymin0, &ymax0, &ny0);
    }
    if(strcmp(ckey,"zcoord")==0){
      sscanf(cline,"%s %f %f %d", ckey, &zmin0, &zmax0, &nz0);
    }
    if(strcmp(ckey,"blockdim")==0){
      sscanf(cline,"%s %d %d %d", ckey, &nxb, &nyb, &nzb);
    }
  }
  fclose(fp);

/*
  printf("XYZmin:  %f %f %f\n", xmin0, ymin0, zmin0);
  printf("XYZmax:  %f %f %f\n", xmax0, ymax0, zmax0);
  printf("X0b: %d %d\n", nx0, nxb);
  printf("Y0b: %d %d\n", ny0, nyb);
  printf("Z0b: %d %d\n", nz0, nzb);
*/
}

void readfield(char *cFile, int nbytes, float *fbuf)
{
  int fd, ngot;

  printf("INPUT -->%s<--\n", cFile);

  fd = open(cFile, O_RDONLY);
  if(fd < 0) {
    fprintf(stderr,"Open Failed, cFile: -->%s<--\n", cFile);
    exit(0);
  }
  ngot = read(fd, fbuf, nbytes);
  if(ngot < nbytes) {
    fprintf(stderr,"Short read: cFile=%s  iNbytes=%d  ngot=%d\n",
            cFile,nbytes,ngot);
    exit(0);
  }
  close(fd);
}

float vinterp(float xx, float yy, float zz, float *vv)
{
  float xind, yind, zind;
  int ix0, ix1, iy0, iy1, iz0, iz1;
  float fracx1,fracy1,fracz1,fracx0,fracy0,fracz0, value;
  int k000,k100,k010,k110,k001,k101,k011,k111;

  xind = dxinv * (xx - xmin0);
  yind = dxinv * (yy - ymin0);
  zind = dxinv * (zz - zmin0);

  ix0 = (int)xind;  if(ix0 < 0) { ix0 = 0; }   if(ix0 >= nx0) { ix0 = nx0-1; }
  iy0 = (int)yind;  if(iy0 < 0) { iy0 = 0; }   if(iy0 >= ny0) { iy0 = ny0-1; }
  iz0 = (int)zind;  if(iz0 < 0) { iz0 = 0; }   if(iz0 >= nz0) { iz0 = nz0-1; }

  ix1 = ix0+1;  if(ix1 >= nx0) { ix1 = ix0; }
  iy1 = iy0+1;  if(iy1 >= ny0) { iy1 = iy0; }
  iz1 = iz0+1;  if(iz1 >= nz0) { iz1 = iz0; }

  fracx1 = xind - (float)ix0;
  fracy1 = yind - (float)iy0;
  fracz1 = zind - (float)iz0;
  fracx0 = 1.0 - fracx1;
  fracy0 = 1.0 - fracy1;
  fracz0 = 1.0 - fracz1;

  k000 = ix0 + nx0*(iy0 + ny0*iz0);
  k100 = ix1 + nx0*(iy0 + ny0*iz0);
  k010 = ix0 + nx0*(iy1 + ny0*iz0);
  k110 = ix1 + nx0*(iy1 + ny0*iz0);
  k001 = ix0 + nx0*(iy0 + ny0*iz1);
  k101 = ix1 + nx0*(iy0 + ny0*iz1);
  k011 = ix0 + nx0*(iy1 + ny0*iz1);
  k111 = ix1 + nx0*(iy1 + ny0*iz1);

  value = fracx0*fracy0*fracz0*vv[k000]
        + fracx1*fracy0*fracz0*vv[k100]
        + fracx0*fracy1*fracz0*vv[k010]
        + fracx1*fracy1*fracz0*vv[k110]
        + fracx0*fracy0*fracz1*vv[k001]
        + fracx1*fracy0*fracz1*vv[k101]
        + fracx0*fracy1*fracz1*vv[k011]
        + fracx1*fracy1*fracz1*vv[k111] ;

  return value;
}

/* ------------------------------------------------------------------------
 * Specify:
 *   Edit fields: (n1,n2,n3)
 *   % of vertical view spanned by sample in max. n direction
 *   Radio buttons: (1,2,3) along (x,y,z) or
 *                  (horizontal, up, eye-center), or
 *                  Fill box
 *   Buttons: 1+, 1-, 2+, 2-, 3+, 3-
 *   Edit fields: timetot, maxout
 *   Button:  save from this dialog
 *   Button:  load to this dialog
 */
/*
void read_start_points()
{
  FILE *fp;
  char cline[256], ckey[256];

  xsc = 0.0;    ns1 =  1;    dx1 = 1.0; dy1 = 0.0; dz1 = 0.0;
  ysc = 0.0;    ns2 =  1;    dx2 = 0.0; dy2 = 1.0; dz2 = 0.0;
  zsc = 0.0;    ns3 = 30;    dx3 = 0.0; dy3 = 0.0; dz3 = 1.0;
  timetot = 500.0;
  maxout  = 1000;

  fp = fopen("startpoints", "r");
  while(fgets(cline, 256, fp) != NULL) {
    sscanf(cline, "%s", ckey);
    if(strcmp(ckey,"center")==0){
      sscanf(cline,"%s %f %f %f", ckey, &xsc, &ysc, &zsc);
    }
    if(strcmp(ckey,"dir1")==0){
      sscanf(cline,"%s %d %f %f %f", ckey, &ns1, &dx1, &dy1, &dz1);
    }
    if(strcmp(ckey,"dir2")==0){
      sscanf(cline,"%s %d %f %f %f", ckey, &ns2, &dx2, &dy2, &dz2);
    }
    if(strcmp(ckey,"dir3")==0){
      sscanf(cline,"%s %d %f %f %f", ckey, &ns3, &dx3, &dy3, &dz3);
    }
    if(strcmp(ckey,"limits")==0){
      sscanf(cline,"%s %d %f", ckey, &maxout, &timetot);
    }
  fclose(fp);
}
*/
