/* hvr_cl.cpp : Defines the entry point for the console application. */
/* Last Updated: October 21, 2006 */

#include "StdAfx.h"
#include "hvr_work.h"
#include <math.h>
#include <signal.h>

#ifdef linux
  #include <stdlib.h>
#endif

Chvr_work hvr;
float fSeamAngle = 0.0f;
float fStereoEyeDisp = 0.0f;
int   iStereoEyeChange = 0;
char cTimerOutFile[256] = "none";

struct LUTrgba {
    int iNalphaKnots, iNrgbKnots, iIndexAlpha[32], iIndexRGB[32];
    float fAlpha[32], fRed[32], fGrn[32], fBlu[32];
};

struct RenderFlags {
    int iPaletted, iResident, iDoubleBuffered, iTex3d, iAutoSwapBuffers, iShowColorBar;
    int iSelectKnots, iSelALPknot, iSelRGBknot;
};

struct Frustum {
    int isNew, iWidth, iHeight;
    float fPanNorm[3], fPanUp[3], fPanCent[3], fWidth;
};

char vecfile[256]="none",      vecfilelast[256]="none";
char cvecloclr[256]="#008000", cvecloclrLast[256]="none";
char cvechiclr[256]="#FFFFFF", cvechiclrLast[256]="none";
float fVecRadius = 0.0;
float fVecRadiusLast = 0.0;

char cOptions[102400], cTimerOutputFile[256];
int iTimersOn=0;

/* define this if taking input from keyfile as opposed to stdin */
//#define KEY_FROM_FILE

#define DEFAULT_WIDTH 1280
#define DEFAULT_HEIGHT 1024
#define STDERR 2

FILE *fpdebug;

int main(int argc, char* argv[])
{
    int load_commands_from_stream(FILE *fp,  struct Renderparams *Rp,
                                  struct LUTrgba *la, struct RenderFlags *rf,
                                  struct Frustum *fr                          );
    void rend_setdefaults(struct Renderparams *rp );
    void setstandardfrustum(int iWidth, int iHeight, float fFOVangle);
    void setpanelfrustum(struct Renderparams rp);
    void RestoreOptions();
    Renderparams rp, rplast, rpsterio;
    LUTrgba la;
    RenderFlags rf;
    Frustum fr;
    FILE *fp;
    static int iFirst = 1;
    int i = 0;
    float fEDX,fEDY,fEDZ, fFactor;
    char cRadius[256];
    char cKnots[256];
    
    int height = 0;
    int width = 0;
    float fov = 0.0;
    int DoubleBuffered = 0;
    int Paletted = 1;
    int Resident = 1;
    int Tex_3d = 1;
    int SectX = 0;
    int SectY = 0;
    int iAutoSwapBuffers = 1;
    int iShowColorBar = 0;
    int  iChangedFldClr;
    
    //fpdebug = fopen("hvrdebug.txt", "a");
#ifdef KEY_FROM_FILE
    /* Validate command line */
    if(argc < 2) {
        printf("Usage: hvr_cl.exe [keyfile]\n");
        return 0;
    }

    /* Open input stream */
    fp = fopen(argv[1], "r");
    if(fp == NULL) {
        printf("Failed to open inputfile = %s\n", argv[1]);
        return 0;
    }

#else
    fp = stdin;
#endif

    rend_setdefaults(&rp);
    rf.iDoubleBuffered  = DoubleBuffered;
    rf.iPaletted        = Paletted;
    rf.iResident        = Resident;
    rf.iTex3d           = Tex_3d;
    rf.iAutoSwapBuffers = iAutoSwapBuffers;
    rf.iShowColorBar    = iShowColorBar;
    rf.iSelectKnots     = 0;
    fr.isNew            = 0; 

    if(argc > 1) {
        for(i=2;i<argc;i++) {
            if(strcmp(argv[i-1],"-display")==0) {
                if(strcmp(argv[i],"window")==0) {
                    rp.offscreen = FALSE;
                }
            }
        }
    }    

    /* Read in rendering instructions and do them while rendering is requested */
    while(load_commands_from_stream(fp, &rp, &la, &rf, &fr) == 1) {

        if(   strcmp(cvecloclr, cvecloclrLast) != 0
           || strcmp(cvechiclr, cvechiclrLast) != 0) { iChangedFldClr = 1; }
        else                                         { iChangedFldClr = 0; }

        if(   iFirst == 1
           || fVecRadius != fVecRadiusLast
           || strcmp(vecfile, vecfilelast) != 0
           || iChangedFldClr   == 1
           || rplast.width     != rp.width
           || rplast.height    != rp.height
           || rplast.fovangle  != rp.fovangle
           || rplast.nsectx    != rp.nsectx
           || rplast.nsecty    != rp.nsecty
           || rplast.isectx    != rp.isectx
           || rplast.isecty    != rp.isecty
           || rplast.nsubsectx != rp.nsubsectx
           || rplast.nsubsecty != rp.nsubsecty
           || rplast.subsectx  != rp.subsectx
           || rplast.subsecty  != rp.subsecty
           || fr.isNew         == 1
           || iStereoEyeChange == 1            )
        {
          if(fr.isNew == 1) {
            /* Only for explicitely set (custum) frusta */
            hvr.SetFrustum(fr.fPanCent, fr.fPanUp,  fr.fPanNorm,
                           fr.iWidth,   fr.iHeight, fr.fWidth   );
          } else {
             setpanelfrustum(rp);
/* was this:
            if(rp.nsectx == 1 && rp.nsubsectx == 1 &&
               rp.nsecty == 1 && rp.nsubsecty == 1     ) {
                setstandardfrustum(rp.width,rp.height,rp.fovangle);
            } else {
                setpanelfrustum(rp);
            }
*/
          }
          rplast.width     = rp.width;
          rplast.height    = rp.height;
          rplast.fovangle  = rp.fovangle;
          rplast.nsectx    = rp.nsectx;
          rplast.nsecty    = rp.nsecty;
          rplast.isectx    = rp.isectx;
          rplast.isecty    = rp.isecty;
          rplast.nsubsectx = rp.nsubsectx;
          rplast.nsubsecty = rp.nsubsecty;
          rplast.subsectx  = rp.subsectx;
          rplast.subsecty  = rp.subsecty;
          fr.isNew         = 0;
          iStereoEyeChange = 0;
        }

        if(iFirst == 1) {
            if(!rp.offscreen) {
                hvr.SetOption("Offscreen","FALSE");
            }
            hvr.StartUp();
        }

        RestoreOptions();

        if((rf.iDoubleBuffered != DoubleBuffered) || (iFirst == 1)) {
            if(rf.iDoubleBuffered == 1) { 
                if(!rp.offscreen)
                    hvr.SetOption("NRenderingBuffers", "2");
            }
            else if(rf.iDoubleBuffered == 0) {
                if(!rp.offscreen)
                    hvr.SetOption("NRenderingBuffers", "1");
            }
            DoubleBuffered = rf.iDoubleBuffered;
        }

        if(rf.iTex3d != Tex_3d) {
            if(rf.iTex3d == 1)
                hvr.SetOption("3dtextures", "TRUE");
            else {
                hvr.SetOption("3dtextures", "FALSE");
                rf.iResident = 0;
            }
            Tex_3d = rf.iTex3d;
        }


        if(strcmp(vecfile, vecfilelast) != 0) {
           hvr.SetOption("vecfile", vecfile);
           sprintf(vecfilelast, "%s", vecfile);
        }
        if(fVecRadius != fVecRadiusLast) {
           sprintf(cRadius, "%f", fVecRadius);
           hvr.SetOption("vecradius", cRadius);
           fVecRadiusLast = fVecRadius;
        }
        if(iChangedFldClr != 0) {
           hvr.SetOption("vecloclr", cvecloclr);
           hvr.SetOption("vechiclr", cvechiclr);
           sprintf(cvecloclrLast, "%s", cvecloclr);
           sprintf(cvechiclrLast, "%s", cvechiclr);
        }
        
        if(rf.iPaletted != Paletted) {
            if(rf.iPaletted == 1) { 
                hvr.SetOption("UsePalet", "TRUE");
            }
            else if(rf.iPaletted == 0) { 
                hvr.SetOption("UsePalet", "FALSE");
                rf.iResident = 0;
            }
            Paletted = rf.iPaletted;
        }

        if(rf.iResident != Resident) {
            if(rf.iResident == 1) { 
                if((rf.iTex3d == 1) && (rf.iPaletted == 1)) {
                    //hvr.SetOption("AutoCache","FALSE");
                    hvr.SetOption("UseResTex","TRUE");
                }
                
            }
            else if(rf.iResident == 2) { 
                if((rf.iTex3d == 1) && (rf.iPaletted == 1))
                    //hvr.SetOption("AutoCache","TRUE");
                    hvr.SetOption("UseResTex","TRUE");
            }
            else if(rf.iResident == 0) { 
                //hvr.SetOption("AutoCache","FALSE");
                hvr.SetOption("UseResTex","FALSE"); 
            }
            Resident = rf.iResident;
        }

        if(rf.iAutoSwapBuffers != iAutoSwapBuffers) {
            if(rf.iAutoSwapBuffers == 1) { 
                hvr.SetOption("AutoSwapBuffers", "1");
            }
            if(rf.iAutoSwapBuffers == 0) { 
                hvr.SetOption("AutoSwapBuffers", "0");
            }
            iAutoSwapBuffers = rf.iAutoSwapBuffers;
        }

        if(rf.iShowColorBar != iShowColorBar) {
            if(rf.iShowColorBar == 1) { 
                hvr.SetOption("ShowColorBar", "1");
            }
            if(rf.iShowColorBar == 0) { 
                hvr.SetOption("ShowColorBar", "0");
            }
            iShowColorBar = rf.iShowColorBar;
        }

        if(rf.iSelectKnots == 1) {
           sprintf(cKnots, "%d %d\n", rf.iSelALPknot, rf.iSelRGBknot);
           hvr.SetOption("SelectKnots", cKnots);
           rf.iSelectKnots = 0;
        }


        if(iTimersOn == 1) {
          hvr.SetOption("TimersOn", "1");
          hvr.SetOption("TimerOutputFile", cTimerOutputFile);
        }

        if(fStereoEyeDisp == 0.0f) {
          hvr.RestorKey(&rp);
        } else {
          memcpy(&rpsterio.offscreen, &rp.offscreen, sizeof(Renderparams));
          fEDX = (rp.centerY - rp.eyeY) * rp.upZ - (rp.centerZ - rp.eyeZ) * rp.upY;
          fEDY = (rp.centerZ - rp.eyeZ) * rp.upX - (rp.centerX - rp.eyeX) * rp.upZ;
          fEDZ = (rp.centerX - rp.eyeX) * rp.upY - (rp.centerY - rp.eyeY) * rp.upX;
          fFactor   = fStereoEyeDisp / sqrt(fEDX*fEDX + fEDY*fEDY + fEDZ*fEDZ);
          rpsterio.eyeX    += fEDX * fFactor;
          rpsterio.eyeY    += fEDY * fFactor;
          rpsterio.eyeZ    += fEDZ * fFactor;

          rpsterio.centerX += fEDX * fFactor;
          rpsterio.centerY += fEDY * fFactor;
          rpsterio.centerZ += fEDZ * fFactor;

          /* No modification to Center vector --> eyes cross @ Center location */
          /* Eye & Center vectors changed equally --> eyes cross infinity */
          hvr.RestorKey(&rpsterio);
        }

        hvr.SetLUTirgb(la.iNrgbKnots, la.iIndexRGB, la.fRed, la.fGrn, la.fBlu);
        hvr.SetLUTia(la.iNalphaKnots, la.iIndexAlpha, la.fAlpha);
        hvr.SetDoRender();

        if(iTimersOn == 1) { hvr.SetOption("TimersOn", "0");  iTimersOn = 0; }

       iFirst = 0;
    }

#ifdef KEY_FROM_FILE
    fclose(fp);
#endif

    hvr.ClearIt();

    //fclose(fpdebug);

    return 0;
}

void RestoreOptions()
{
    int i = 0, j, iDone = 0;
    int iLen = strlen(cOptions);
    char cKey[256], cVal[256];

    while(i < iLen) {
     // Parse next key value pair
      sscanf(&cOptions[i], "%s", cKey);
      i += strlen(cKey) + 1;
      j = 0;
      while(cOptions[i] != '\n') { cVal[j] = cOptions[i]; i++; j++; }
      cVal[j] = '\0';
      i++;

      // Send option to hvr object if not an idle status
      if(strcmp(cKey, "IDLE") != 0) {
         hvr.SetOption(cKey, cVal);
      }
  }
  sprintf(cOptions, "");
}


void setstandardfrustum(int iWidth, int iHeight, float fFOVangle)
{
    float fHalfHeight, fHalfWidth, fAspectRatio;
    float fPanNorm[3], fPanUp[3], fPanCent[3], fWidth;
    float Pid180 = (float)(3.141592654/180.);

    fPanNorm[0] = 0.;  fPanUp[0] = 0.;  fPanCent[0] = 0.;
    fPanNorm[1] = 1.;  fPanUp[1] = 0.;  fPanCent[1] = 1.;
    fPanNorm[2] = 0.;  fPanUp[2] = 1.;  fPanCent[2] = 0.;    
    
    fAspectRatio = (float)iWidth / (float)iHeight;
    fHalfHeight  = (float)tan( Pid180 * fFOVangle * 0.5 );
    fHalfWidth   = fHalfHeight * fAspectRatio;
    fWidth       = (float)2.0 * fHalfWidth;

    hvr.SetFrustum(fPanCent, fPanUp, fPanNorm, iWidth, iHeight, fWidth);
}

void setpanelfrustum(Renderparams rp)
{
  float fHalfHeight, fPanWidth, fPanHeight, fHalfWidth, fRadius, fAngle;
  float Pid180, half, one, zero;
  float fPanNorm[3], fPanUp[3], fPanCent[3];
  
  int   iPanelWidth  = rp.width;
  int   iPanelHeight = rp.height;
  int   ix           = rp.isectx - 1;
  int   iz           = rp.isecty - 1;
  int   iNPanelHoriz = rp.nsectx;
  int   iNPanelVert  = rp.nsecty;
  float fFullFOV     = rp.fovangle;

  /* subsections */
  int isubx          = rp.subsectx - 1;
  int isuby          = rp.subsecty - 1;
  int nSubPanelHoriz = rp.nsubsectx;
  int nSubPanelVert  = rp.nsubsecty;
  float fPanHoriz[3], fShiftHoriz, fShiftVert;

  //fprintf(stderr, "at top of setpanelfrustum\n");
  //fprintf(fpdebug, "at top of setpanelfrustum\n");

 /***********************************************************************
  * Assume iPanelWidth,iPanelHeight are the dimensions of the subpanel
  * which would be the entire panel if nSubPanelHoriz=nSubPanelVert=1
  */

  /* Useful constants */
  half   = (float)0.5;
  one    = (float)1.0;
  zero   = (float)0.0;
  Pid180 = (float)3.141592654 / (float)180.;
  
 /***********************************************************************
  * 1) generate (fPanCent,fPanUp,fPanNorm,iPanelWidth,iPanelHeight,fPanWidth)
  *    assuming subpanel is entire panel
  */
  fPanUp[0] = zero;
  fPanUp[1] = zero;
  fPanUp[2] = one;

  fHalfHeight = (float)tan( Pid180 * fFullFOV * half );   // Assumes panel is at unit dist. from eye
  fPanHeight  = (float)2.0 * fHalfHeight / (float)iNPanelVert;
  fPanWidth   = fPanHeight * (float)(iPanelWidth*nSubPanelHoriz) / (float)(iPanelHeight*nSubPanelVert);
  fHalfWidth  = half * fPanWidth;
  
  fPanCent[2] = fPanHeight * ((float)iz - half * ((float)iNPanelVert - 1));
  fPanNorm[2] = zero;
  
  if(fSeamAngle < 0.00001) {
    // Flat wall case
    fPanCent[0] = fPanWidth * ((float)ix - half * ((float)iNPanelHoriz - 1));
    fPanCent[1] = one;
    fPanNorm[0] = zero;
    fPanNorm[1] = one;
  } else {
    //Curved wall case:
    fRadius = (float)0.5 * fPanWidth / (float)tan( Pid180 * fSeamAngle * (float)0.5 );
    if(0.999 < fRadius  &&  fRadius < 1.001) { fRadius = one; }
    fAngle  = Pid180 * fSeamAngle * ((float)ix - half * (float)(iNPanelHoriz - 1));
    fPanCent[0] = fRadius * (float)sin(fAngle);
    fPanCent[1] = fRadius * (float)cos(fAngle) - fRadius + one;
    fPanNorm[0] = (float)sin(fAngle);
    fPanNorm[1] = (float)cos(fAngle);
  }

  fPanHoriz[0] =  fPanNorm[1];
  fPanHoriz[1] = -fPanNorm[0];
  fPanHoriz[2] =  zero;

 /***********************************************************************
  * 2) If more than one subpanel in full panel, then resize & shift
  */
  if (nSubPanelHoriz > 1 || nSubPanelVert > 1) {
    fPanWidth  = fPanWidth  / nSubPanelHoriz;
    fPanHeight = fPanHeight / nSubPanelVert;

    fShiftHoriz  = fPanWidth  * (isubx - half*(float)(nSubPanelHoriz-1));
    fShiftVert   = fPanHeight * (isuby - half*(float)(nSubPanelVert -1));
    fPanCent[0] += fShiftHoriz * fPanHoriz[0];
    fPanCent[1] += fShiftHoriz * fPanHoriz[1];
    fPanCent[2] += fShiftVert  * fPanUp[2];
  }

 /***********************************************************************
  * 3) If there is a stereo eye displacment, adjust center of
  *    view frustum relative to eye position
  */
  if(fStereoEyeDisp > 0.0f) {
    fPanCent[0] -= fStereoEyeDisp * fPanHoriz[0];   
    fPanCent[1] -= fStereoEyeDisp * fPanHoriz[1];   
    fPanCent[2] -= fStereoEyeDisp * fPanHoriz[2];   
  }

  hvr.SetFrustum(fPanCent, fPanUp, fPanNorm, iPanelWidth, iPanelHeight, fPanWidth);
  //fprintf(fpdebug, "fPanCent = %f %f %f\n", fPanCent[0],fPanCent[1],fPanCent[2]);
}

void rend_setdefaults( struct Renderparams *rp )
{
  strncpy( (char *)rp->cmapname, "default", 8 );
  strncpy( (char *)rp->amapname, "default", 8 );
  rp->opacity = 1.0;
  rp->nx = rp->ny = rp->nz = 0;
  *rp->bobname = 0;
  rp->hpos = rp->vpos = 0;
  rp->nsectx = rp->nsecty = 1;
  rp->isectx = rp->isecty = 1;
  rp->width = DEFAULT_WIDTH;
  rp->height = DEFAULT_HEIGHT;
  *rp->imagename = 0;
  *rp->moviename = 0;
  rp->fovangle = (float)35.3084;
  rp->eyeX = -5.0;
  rp->eyeY = 0.;
  rp->eyeZ = 0.;
  rp->centerX = rp->centerY = rp->centerZ= 0;
  rp->upX = rp->upY = 0.;
  rp->upZ = 1.0;
  rp->viewdefined[0] = rp->viewdefined[1] =  rp->viewdefined[2] =
    rp->viewdefined[2] = 0;
  rp->my_near = (float)3.27;      /* unit cube for any orientation */
  rp->my_far  = (float)5.73;
  rp->softdistance = (float)0.;
  rp->nplanes    = (float)4.0;
  rp->maxplanes  = (float)4.0;
  rp->boxcolor = -1;    /* that is, none */
  rp->ifauxclip[0]= rp->ifauxclip[1]= rp->ifauxclip[2]= rp->ifauxclip[3]=
  rp->ifauxclip[4]= rp->ifauxclip[5]= 0;
  rp->nfpk = 1;
  rp->nrpk = 0.;
  rp->ntpf = 0;
  rp->offscreen = TRUE;
  sprintf((char *)rp->imagename, "none");
  sprintf(cOptions, "");

  /* subsections */
  rp->subsectx = 1;
  rp->subsecty = 1;
  rp->nsubsectx = 1;
  rp->nsubsecty = 1;

  rp->icoltolerance = 6;
  rp->angletolerance = 0.;
  rp->vrtolerance = -1.;
}

static int match( char *token, char *word, int tokenlength )
{
  int n,i;
  n = strlen(word);
  if ( n>tokenlength ) n= tokenlength;
  //if strncmp(token[tokenlength-1],
  
  i = !strncmp(token,word,n);
  return i;
}

/*int rend_load_from_file(char *cInputFile, struct Renderparams *Rp, struct LUTrgba *la)*/

#define SUCCESS 0 /* Success */
#define X_FAILED 1 /* Could not connect to X Server */
#define VISUAL_FAILED 2 /* Could not choose visual */
#define HV_FAILED 3 /* Could not open HV File */
#define FBO_FAILED 4 /* Fatal FBO or Renderbuffer error */
#define CACHE_FAILED 5 /* Fatal caching scheme error */
#define CG_FAILED 6 /* Fatal Cg Error */
#define LOG_FAILED 7 /* Logfile failed to open */
#define GLEW_FAILED 8 /* GLEW failed to init */
#define IPRINT_FAILED 9 /* iprint failed */
#define STDOUT 1

int load_commands_from_stream(FILE *fp,
                              struct Renderparams *Rp,
                              struct LUTrgba      *la,
                              struct RenderFlags  *rf,
                              struct Frustum      *fr  )
{
  bool done = FALSE;
  //char line[256];
  char *keyword, ckey[256];
  static char *seps = " ,;\t\n\r";
  int nc, ioff, igotalloptions;
  int i, iNalpha=0, iNrgb=0, iCount, iWillSleep;
  static int iDebug=0;
  int status = SUCCESS;
#ifdef linux
  double getmillis_();
  static double fLastTime = -1.0;
#else
  static unsigned long ulLastTime = (unsigned long)0;
#endif

  iCount = 0;

//  while ( fgets(line,200,fp) != NULL  &&  iCount < 100) {
  while(!done) {
      char line[256];

      if(fgets(line,200,fp) == NULL) {
          done = TRUE;
      }
      else if(strlen(line) > 1)
      {
            if(iDebug ==1) { fprintf(stderr, "%3d: %s", iCount, line); }
            iCount++;

            //if ( strncmp(line, "# NEXT KEY",10) == 0 ) { return(1); }
            if ( line[0] == '#' ) continue;

            nc = strlen(line);
            keyword = strtok( line,seps );
            nc = strlen(keyword);
            if (!nc) continue;

            /*if ( match(keyword,"offscreen",nc) ) {
              Rp->offscreen = TRUE;
            }
            else if ( match(keyword,"window",nc) ) {
              Rp->offscreen = FALSE;
            }*/
            /*else if ( match(keyword,"color",nc) ) {
              strncpy( (char *)Rp->cmapname, strtok(NULL,seps),80 );
            }
            else if ( match(keyword,"alpha",nc) ) {
              strncpy( (char *)Rp->amapname, strtok(NULL,seps),80 );
            }*/
            else if ( match(keyword,"beginoptions",nc) ) {
              igotalloptions = 0;
              while(igotalloptions == 0) {
                if(fgets(line,200,fp) != NULL) { 
                  if(sscanf(line, "%s", ckey) > 0) {
                    if( strcmp(ckey,"endoptions") != 0 ) {
                      ioff = strlen(cOptions);
                      sprintf(&cOptions[ioff], "%s", line);
                    } else { igotalloptions = 1; }
                  } else { igotalloptions = 1; }
                } else { igotalloptions = 1; }
              }
              // RestoreOptions();
            } 
            else if ( match(keyword,"opacity",nc) ) {
              Rp->opacity = (float)atof( strtok(NULL,seps) );
            } 
            /*else if ( match(keyword,"dimensions",nc) ) {
              Rp->nx = atoi( strtok(NULL,seps) );
              Rp->ny = atoi( strtok(NULL,seps) );
              Rp->nz = atoi( strtok(NULL,seps) );
            }
            else if ( match(keyword,"file",nc) ) {
              strncpy( (char *)Rp->bobname, strtok(NULL,seps),80 );
            }*/
            else if ( match(keyword,"hvfile",nc) ) {
              strncpy( (char *)Rp->hvname, strtok(NULL,seps),80 );
            }
            else if ( match(keyword,"vecfile",nc) ) {
              strncpy( vecfile, strtok(NULL,seps),80 );
            }
            else if ( match(keyword,"vecradius",nc) ) {
              fVecRadius = atof( strtok(NULL,seps) );
            }
            else if ( match(keyword,"veclohiclr",nc) ) {
              strncpy( cvecloclr, strtok(NULL,seps),80 );
              strncpy( cvechiclr, strtok(NULL,seps),80 );
            }
            else if ( match(keyword,"imagefile",nc) ) {
              strncpy( (char *)Rp->imagename, strtok(NULL,seps),80 );
            }
            else if ( match(keyword,"ptofile",nc) ) {
              strncpy( cTimerOutputFile, strtok(NULL,seps),80 );
              iTimersOn = 1;
            }
            /*else if ( match(keyword,"moviefile",nc) ) {
              strncpy( (char *)Rp->moviename, strtok(NULL,seps),80 );
            }*/
/*
struct Frustum { int isNew, iWidth, iHeight;
                 float fPanNorm[3], fPanUp[3], fPanCent[3], fWidth; };
*/
            else if ( match(keyword,"frustum",nc) ) {
              fr->isNew       = 1;
              fr->iWidth      = atoi(strtok(NULL,seps) );
              fr->iHeight     = atoi(strtok(NULL,seps) );
              fr->fWidth      = atof(strtok(NULL,seps) );
              fr->fPanNorm[0] = atof(strtok(NULL,seps) );
              fr->fPanNorm[1] = atof(strtok(NULL,seps) );
              fr->fPanNorm[2] = atof(strtok(NULL,seps) );
              fr->fPanUp[0] = atof(strtok(NULL,seps) );
              fr->fPanUp[1] = atof(strtok(NULL,seps) );
              fr->fPanUp[2] = atof(strtok(NULL,seps) );
              fr->fPanCent[0] = atof(strtok(NULL,seps) );
              fr->fPanCent[1] = atof(strtok(NULL,seps) );
              fr->fPanCent[2] = atof(strtok(NULL,seps) );
            }
            else if ( match(keyword,"section",nc) ) {
              Rp->isectx = atoi(strtok(NULL,seps) );
              Rp->isecty = atoi(strtok(NULL,seps) );
              Rp->nsectx = atoi(strtok(NULL,seps) );
              Rp->nsecty = atoi(strtok(NULL,seps) );
            }
            else if ( match(keyword,"subsection",nc) ) {
              Rp->subsectx = atoi(strtok(NULL,seps) );
              Rp->subsecty = atoi(strtok(NULL,seps) );
              Rp->nsubsectx = atoi(strtok(NULL,seps) );
              Rp->nsubsecty = atoi(strtok(NULL,seps) );
            }
            else if ( match(keyword,"seamangle",nc) ) {
                fSeamAngle = atof(strtok(NULL,seps) );
            }
            else if ( match(keyword,"steriodisp",nc) ) {
                fStereoEyeDisp = atof(strtok(NULL,seps) );
                iStereoEyeChange = 1;
            }
            /*else if ( match(keyword,"imageposition",nc) ) {
              Rp->hpos = atoi(strtok(NULL,seps) );
              Rp->vpos = atoi(strtok(NULL,seps) );
            }*/
            else if ( match(keyword,"imagesize",nc) ) {
              Rp->width = atoi( strtok(NULL,seps) );
              Rp->height = atoi( strtok(NULL,seps) );
            } 
            else if ( match(keyword,"clipping",nc) ) {
              Rp->my_near =  (float)atof( strtok(NULL,seps) );
              Rp->my_far =  (float)atof( strtok(NULL,seps) );
            } 
            /*else if ( match(keyword,"softdistance",nc) ) {
              Rp->softdistance =  (float)atof( strtok(NULL,seps) );
            }*/ 
            else if ( match(keyword,"nplanes",nc) ) {
              Rp->nplanes    = (float)atof( strtok(NULL,seps)) ;
              Rp->maxplanes = (float)atof( strtok(NULL,seps)) ;
            }
            /*else if ( match(keyword,"boxcolor",nc) ) {
              Rp->boxcolor = atoi( strtok(NULL,seps)) ;
            }*/
            else if ( match(keyword,"fov",nc) ) {
              Rp->fovangle = (float)atof( strtok(NULL,seps) ) ;
            }
            else if ( (match(keyword,"eye",nc)) ||
            (match(keyword, "eye", nc)) ){
              Rp->eyeX = (float)atof( strtok(NULL,seps) );
              Rp->eyeY = (float)atof( strtok(NULL,seps) );
              Rp->eyeZ = (float)atof( strtok(NULL,seps) );
              Rp->viewdefined[0] = 1;
            }
            else if ( match(keyword,"center",nc) ) {
              Rp->centerX = (float)atof( strtok(NULL,seps) );
              Rp->centerY = (float)atof( strtok(NULL,seps) );
              Rp->centerZ = (float)atof( strtok(NULL,seps) );
              Rp->viewdefined[1] = 1;
            }
            else if ( match(keyword,"up",nc) ) {
              Rp->upX = (float)atof( strtok(NULL,seps) );
              Rp->upY = (float)atof( strtok(NULL,seps) );
              Rp->upZ = (float)atof( strtok(NULL,seps) );
              Rp->viewdefined[2] = 1;
            }
            /*else if ( match(keyword,"eyedirection",nc) ) {
              Rp->eyedirX = (float)atof( strtok(NULL,seps) );
              Rp->eyedirY = (float)atof( strtok(NULL,seps) );
              Rp->eyedirZ = (float)atof( strtok(NULL,seps) );
              Rp->viewdefined[3] = 1;
            }*/
            else if ( match(keyword,"tolerance",nc) ) {
              Rp->icoltolerance  = atoi( strtok(NULL,seps) );
              Rp->angletolerance = (float)atof( strtok(NULL,seps) );
              Rp->vrtolerance = (float)atof( strtok(NULL,seps) );
            }
            else if ( match(keyword,"ifauxclip",nc) ) {
                i = atoi( strtok(NULL,seps) );
                Rp->ifauxclip[i] = atoi( strtok(NULL,seps) );
            }
            else if ( match(keyword,"auxclip",nc) ) {
                i = atoi( strtok(NULL,seps) );
                Rp->auxclip[i][0] = (float)atof( strtok(NULL,seps) );
                Rp->auxclip[i][1] = (float)atof( strtok(NULL,seps) );
                Rp->auxclip[i][2] = (float)atof( strtok(NULL,seps) );
                Rp->auxclip[i][3] = (float)atof( strtok(NULL,seps) );
            }
            else if ( match(keyword,"frameperkey",nc) ) {
              Rp->nfpk = atoi( strtok(NULL,seps) );
            }
            else if ( match(keyword,"timeperframe",nc) ) {
              Rp->ntpf = atoi( strtok(NULL,seps) );
            }
            else if ( match(keyword,"rotperkey",nc) ) {
              Rp->nrpk = (float)atof( strtok(NULL,seps) );
            }
            else if ( match(keyword,"Anot",nc) ) {
              la->iIndexAlpha[iNalpha] = atoi( strtok(NULL,seps) );
              la->fAlpha[iNalpha]      = (float)atof( strtok(NULL,seps) );
              iNalpha++;
              la->iNalphaKnots = iNalpha;
            }
            else if ( match(keyword,"Cnot",nc) ) {
              la->iIndexRGB[iNrgb] = atoi( strtok(NULL,seps) );
              la->fRed[iNrgb]      = (float)atof( strtok(NULL,seps) );
              la->fGrn[iNrgb]      = (float)atof( strtok(NULL,seps) );
              la->fBlu[iNrgb]      = (float)atof( strtok(NULL,seps) );
              iNrgb++;
              la->iNrgbKnots = iNrgb;
            }
            else if ( match(keyword,"timer",nc) ) {
               #ifdef linux
                double timenow = getmillis_();
                double dtime   = timenow - fLastTime;
                if(fLastTime < 0.0) { fprintf(stderr,"time = %f\n", (float)timenow); }
                else                { fprintf(stderr,"time = %f   dtime=%f\n", (float)timenow, (float)dtime); }
                fLastTime = timenow;
               #else
                float timenow, dtime;
                unsigned long ulTime = GetTickCount();
                timenow = (float)ulTime / (float)1000.0;
                dtime   = (float)(ulTime - ulLastTime) / (float)1000.0;
                //if(ulLastTime == 0.0) { printf("time = %f\n", timenow); }
                //else                  { printf("time = %f   dtime=%f\n", timenow, dtime); }
                if(ulLastTime == 0.0) { fprintf(stderr,"\n"); }
                else                  { fprintf(stderr,"time = %f\n", dtime); }
                ulLastTime = ulTime;
               #endif
            }
            else if ( match(keyword,"doublebuffered",nc) ) { 
                rf->iDoubleBuffered = atoi(strtok(NULL,seps)); }
            else if ( match(keyword,"caching",nc) ) {
                rf->iResident = atoi(strtok(NULL,seps)); }
            else if ( match(keyword,"paletted",      nc) ) { 
                rf->iPaletted = atoi(strtok(NULL,seps)); }
            else if ( match(keyword,"tex3d",      nc) ) { 
                rf->iTex3d = atoi(strtok(NULL,seps)); }
            else if ( match(keyword,"autoswapbuffers",      nc) ) { 
                rf->iAutoSwapBuffers = atoi(strtok(NULL,seps)); }

            else if ( match(keyword,"colorbar",      nc) ) { 
                rf->iShowColorBar = atoi(strtok(NULL,seps)); }
            else if ( match(keyword,"selknots",      nc) ) { 
                rf->iSelALPknot  = atoi(strtok(NULL,seps));
                rf->iSelRGBknot  = atoi(strtok(NULL,seps));
                rf->iSelectKnots = 1;
           } else if ( match(keyword,"sleep",nc) ) {
                iWillSleep = atoi( strtok(NULL,seps) );
                _sleep(iWillSleep);
            }
            else if ( match(keyword,"debug",nc) ) {
                iDebug = atoi( strtok(NULL,seps) );
            }
            else if ( match(keyword,"swapbuffers",nc) ) {
                iWillSleep = atoi( strtok(NULL,seps) );
                write(STDOUT,&status,4);
                usleep(iWillSleep);
#ifdef linux
                hvr.SetOption("SwapBuffers","0");
#else
                hvr.SetOption(L"SwapBuffers",L"0");
#endif
            }
            else if ( match(keyword,"writeimagetostdout",nc) ) {
                i = atoi( strtok(NULL,seps) );
#ifdef linux
                if(i==1) { hvr.SetOption("WriteImageToSTDout","1"); }
                else     { hvr.SetOption("WriteImageToSTDout","0"); }
#else
                if(i==1) { hvr.SetOption(L"WriteImageToSTDout",L"1"); }
                else     { hvr.SetOption(L"WriteImageToSTDout",L"0"); }
#endif
            }
            else if ( match(keyword,"render",nc) ) { return(1); }
            else if ( match(keyword,"exit",nc) ) { done = TRUE; }
            else { 
                if(iDebug > 0) { fprintf(stderr,"***** '%s' undefined keyword ignored\n",keyword); }
            }
        }
  } /* end while */

  return(0); 
}

int read_inputs(char *cInputFile, Renderparams *rp)
{
    FILE *fp;
    char str[256];
    int iCount = 0;

    fp = fopen(cInputFile, "r");
    if(fp == NULL) { return -1; }

    while(fgets(str, 256, fp) != NULL) {
        if(strncmp(str, "hvfile", 6) == 0) { sscanf(str, "hvfile%s", rp->hvname); }
    }

    fclose(fp);

    return 1;
}
