// hvr_work.h : Declaration of the Chvr_work
//
// Author:   David H. Porter
//           Laboratory for Computational Science & Engineering
//           University of Minnesota
//
/***************************************************************************************
Program:  HVR_SERVER  --  Hierarchical Volume Rendering Component Object Moduel
Copyright (C) 2002  David H. Porter

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
***************************************************************************************/

#ifndef __HVR_WORK_H_
#define __HVR_WORK_H_

// dhp111 #include "resource.h"       // main symbols
//#include <gl\gl.h>
//#include <gl\glu.h>

#include <GL/glx.h>
#include <GL/gl.h>
#include <unistd.h>


//  # nclude "C:\include\GL\glut.h"

struct HvrInput {    // This structure contains all of the input needed for the hvr code

		int iNewHV;				// (1/0) = (Is/Is Not) new HVfile
		char sHVname[1024];

		int iNewFrustum;		// (1/0) = (Is/Is Not) new frustum
		int iWidth;				// Pixel resolution of panel in horizontal
		int iHeight;			// Pixel resolution of panel in vertical
		float fPanCent[3];
		float fPanNorm[3];
		float fPanUp[3];
		float fHalfWidth;		// Half horizontal width of panel in object space
		float fHalfHeight;		// Half vertical height of panel in object space
		float fFovAngle;		// field of view, in the vertical direction
		float fAspectRatio;		// horizontal to vertical aspect ratio (=iHeight/iWidth=fHalfHeight/fHalfWidth)

		int iNewLUT;			// (1/0) = (Is/Is Not) new color/alpha Look Up Table
		char sCmapname[100];	// Name of color LUT file
		char sAmapname[100];	// Name of alpha LUT file

		int iNewView;			// (1/0) = (Is/Is Not) new view parameters
		float fEye[3];			// position of eye in object space
		float fCent[3];			// position of center of view in object space
		float fUp[3];			// up vector in opject space
		float fNearClip;		// distance to near clipping plane in object spece
		float fFarrClip;		// distance fo far clipping plane in object space
		int   ifauxclip[6];		// flag for enabled auxiliary clipping planes
		float auxclip[6][4];	// plane eqn Ax + By + Cz + D = 0 for each

		float opacity;		/* opacity scaling, 1.0 is a good default */
		float nplanes;		/* number of planes rendered per voxel at the far clipping distance */
		float maxplanes;	/* max number of planes rendered per voxel */
		int   tolerance;	/* max color level error allowed with no scan down  */
		float angtol;		/* max angle allowed with no scan down              */
		float viewres;		/* dx_brick > farclip * vrtolerance => no scan down */
};

/////////////////////////////////////////////////////////////////////////////
// Chvr_work
/* dhp111
class ATL_NO_VTABLE Chvr_work : 
	public CComObjectRootEx<CComSingleThreadModel>,
	public CComCoClass<Chvr_work, &CLSID_hvr_work>,
	public Ihvr_work
*/
class Chvr_work
{
public:
	Chvr_work();
	void init(void);
    HDC hDC;				/* device context */
    HGLRC hRC;				/* opengl context */
    HWND  hWnd;				/* window */
	HWND m_hLocalWnd;

private:
	float m_fScreenRatio;
	int  m_iNTicks;
	int  m_iTicks[256];
	char m_sTicks[256][64];
	int  m_iXNTicks;
	int  m_iXTicks[256];
	char m_sXTicks[256][64];
	int  m_iYNTicks;
	int  m_iYTicks[256];
	char m_sYTicks[256][64];
	int m_iShowColorBar;
        int m_iSelALPknot;
        int m_iSelRGBknot;
	int m_iNRendBuffs;
	int m_iNBoundaries;
	int m_iNBoundariesRep[128];
	int m_iBufferedIO;
	char m_cOldHVname[1024];
	bool m_bColorTableIsAvailable;
	int m_iTimersOn;
	char m_sTimerFile[256];
	char m_cOldCmap[1024];
	float m_fNearFar;
	float m_fNearLimit;
	int m_iStopRendering;
	HvrInput m_ShvrIn;			// GUI writes input here
	HvrInput m_ShvrUse;			// Rendering code uses ShvrUse, after copying info from ShvrIn
	int m_iRender;
	float m_flClearRGB[3];
	int m_iAlphaIsSet;
	int m_iNrgbNots;
	int m_iNalpNots;
	int m_iRGBind[32], m_iALPind[32];
	float m_fRED[32], m_fGRN[32], m_fBLU[32], m_fALP[32];
	float frgba[256][4];
//	unsigned long m_lrgba[256];
	unsigned int m_lrgba[256];
	float opacity;		/* opacity scaling, 1.0 is a good default */

	int  Nx,Ny,Nz;	    /* Full brick size */
	char m_cImageName[100];	/* name of image file w/o pannel number */
	int moviemode;      /* 0 for singleframe image files (old behavior) */
                        /* In this case, imagename must be defined  */
                        /* >0 diskmovie frame number.  */
                        /* <0 -sharedmemory frame number. */

                        /* viewing parameters, all relative to a
                           canonical -1:+1 coordinate, where the
                           X-dimension of the brick defines -1:+1 */
  int   viewdefined[4];     /* flag for eye/center/up/eyedir defined */

  int   ifauxclip[6];    /* flag for enabled auxiliary clipping planes */
  float auxclip[6][4];   /* plane eqn Ax + By + Cz + D = 0 for each  */

  float softdistance;
  float nplanes;	/* number of planes rendered per voxel at the
			           far clipping distance */
  float maxplanes;	/* max number of planes rendered per voxel */
  int   boxcolor;	/* or -1 if no outline */

  int   tolerance;		/* max color level error allowed with no scan down  */
  float angtol;			/* max angle allowed with no scan down              */
  float viewres;		/* dx_brick > farclip * vrtolerance => no scan down */



// ====================================================================================
// Rendering variables

#define NPMAX 100			/* Maximum number of pannels */

GLuint texname[256];		/* GL 2D textures */
char *rgbpannel[NPMAX];		/* RGB pannels */
int pannels_initted;		/* malloc rgppannels only once */
int ipcur;					/* current pannel loaded into rgbpannel */
int isubcur;				/* current sub-brick loaded into bobsub */
int ireload;				/* flag to reload brick in 3dtex code   */
int npannel;				/* number of pannels (in PW_CONFIGURATION) */
GLfloat Aspect;				/* Output image aspect ratio = horizontal/vertical */
float Pid180;				/* Radian to degree conversion factor */

	
	void SetDefaultFrustum(void);
	void SetDefaultView(void);
	void CopyHvrInputs(void);
	void draw_scene();
	void colorbar();

	/*
no	bool DoRender();
no	void SetClearColor(float f_red, float f_grn, float f_blu);
	*/

public:
	void SpawnWork();

/* dhp111
DECLARE_REGISTRY_RESOURCEID(IDR_HVR_WORK)
DECLARE_NOT_AGGREGATABLE(Chvr_work)

DECLARE_PROTECT_FINAL_CONSTRUCT()

BEGIN_COM_MAP(Chvr_work)
	COM_INTERFACE_ENTRY(Ihvr_work)
END_COM_MAP()
*/

// Ihvr_work
public:
	void GLUTdraw();
	void initialize_vars();
	void ClearOption(wchar_t wcClearOp[256]);
	void AddOption();
	void SaveOption();
	float m_fScaleFactors[3];
	char m_cStatusRequest[256];
	int m_iAutoSwapBuffers;
	void DoIO();
	char m_sTileO[256];
	char m_sStats[256];
	void DoPutPix();
	void DoGetPix();
	void DoSetOption();
	int m_iBlockOnRender;

	STDMETHOD(PutPix)(/*[in]*/ long lSize, /*[in, size_is(lSize)]*/ unsigned char *cPix);
	STDMETHOD(GetPix)(/*[out]*/ struct BinaryDataHVR *pData);
	STDMETHOD(GetStatus)(/*[out, string]*/ wchar_t **wcStatus);
	STDMETHOD(SetOption)(/*[in, string]*/ wchar_t *wcKey, /*[in, string]*/ wchar_t *wcValue);
	STDMETHOD(GetLUT32ia)(/*[out]*/ int *N, /*[out, size_is(32)]*/ int *ind, /*[out, size_is(32)]*/ float *alp);
	STDMETHOD(GetLUT32irgb)(/*[out]*/ int *N, /*[out, size_is(32)]*/ int *ind, /*[out, size_is(32)]*/ float *red, /*[out, size_is(32)]*/ float *grn, /*[out, size_is(32)]*/ float *blu);
	STDMETHOD(SetLUTia)(/*[in]*/ int N, /*[in, size_is(N)]*/ int *ind, /*[in, size_is(N)]*/ float *alpha);
	STDMETHOD(SetLUTirgb)(/*[in]*/ int N, /*[in, size_is(N)]*/ int *ind, /*[in, size_is(N)]*/ float *red, /*[in, size_is(N)]*/ float *grn, /*[in, size_is(N)]*/ float *blu);
	STDMETHOD(Set_hWnd)(/*[in]*/ int i_hWnd);
	STDMETHOD(SetNearFarClip)(/*[in]*/ float fNear, /*[in]*/ float fFar);
	STDMETHOD(SetOpacity)(/*[in]*/ float fOpacity);
	STDMETHOD(AddFrustum)(/*[in, size_is(3)]*/ float *center, /*[in, size_is(3)]*/ float *up, /*[in, size_is(3)]*/ float *normal, /*[in]*/ int iWidth, /*[in]*/ int iHeight, /*[in]*/ float fPanWidth);
	STDMETHOD(SetCurrentWorkingDir)(unsigned char *cDir);
	STDMETHOD(CopyPixToFile)(unsigned char *cFile);
	STDMETHOD(SetLUT)(/*[in]*/ unsigned char *sAmap, /*[in]*/ unsigned char *sCmap);
	STDMETHOD(ZoomEyePoint)(/*[in]*/ int delx, /*[in]*/ int dely, /*[in]*/ int iMouseMode);
	STDMETHOD(StopRendering)();
	STDMETHOD(SetRenderQuality)(/*[in]*/ int iPixErr, /*[in]*/ float fAngleError, /*[in]*/ float fSizeError);
	STDMETHOD(SetFrustum)(/*[in, size_is(3)]*/ float *center, /*[in, size_is(3)]*/ float *up, /*[in, size_is(3)]*/ float *normal, /*[in]*/ int iWidth, /*[in]*/ int iHeight, /*[in]*/ float fPanWidth);
	STDMETHOD(Set_HVname)(/*[in]*/ unsigned char* cHVnew);
	STDMETHOD(RestorKey)(/*[in]*/ struct Renderparams *SKey);
	STDMETHOD(SaveKey)(/*[out]*/ struct Renderparams *SKey);
	STDMETHOD(IncrementEyePoint)(/*[in]*/ int delx, /*[in]*/ int dely, /*[in]*/ int iMouseMode);
	STDMETHOD(SetDoRender)();
	STDMETHOD(StartUp)();
	STDMETHOD(ClearIt)();
};

#endif //__HVR_WORK_H_
