#include <string.h>
#include <stdio.h>


/* OpenGL Headers */
#define GL_GLEXT_PROTOTYPES
#include <GL/glx.h>
#include <X11/extensions/xf86vmode.h>
#include <X11/keysym.h>
#include <X11/extensions/Xdbe.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>

#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <linux/stat.h>

/************************************************
 *       Defined constants and macros
 ************************************************/
#define _sleep(n) usleep(n*1000)
#define FIFO_FILE       "RDFIFO"

/**************************************************************************** 
 *             Globale variables
 ****************************************************************************/
Display *dpyHVR;
XVisualInfo *viHVR;
Colormap cmapHVR;
XSetWindowAttributes swaHVR;
Window winHVR;
GLXContext cxHVR;
XEvent eventHVR;

unsigned char *ucBuf = NULL;
unsigned char *ucPix;
int iNx, iNy, iNbytes, iSleep;
char cFile[256], cDisplay[256];
int fd_hvr_in, fd_hvr_out;


/**************************************************************************** 
 *                       Prototypes
 ****************************************************************************/
void open_window();
void read_pix_from_file_to_buf();
void read_pix_from_hvr_to_display();
void draw_pix_from_buf();
void close_window();


/**************************************************************************** 
 *                     Utility routines
 ****************************************************************************/
Bool WaitForNotify(Display *d, XEvent *e, char *arg)
{ return (e->type == MapNotify) && (e->xmap.window == (Window)arg); }


/**************************************************************************** 
 *                       Main routine
 ****************************************************************************/
int main(int argc, char* argv[])
{
    int iostat, idodisplay;
    int i, iVal, iValsend, ngot, iarg, nextarg;
    int iexit, icount, iexecret, err;
    FILE *fp, *fp_hvr_in;
    char cnewline[256], key[256];
    char byte = '\0';
    pid_t processID;
    int fds[4];
    char* aargv[2];

   /********************************************************************
    *             Initialize                          
    */

    //         Defaults
    idodisplay = 0;
    iNx = 512;
    iNy = 512;
    iSleep = 5000;
    sprintf(cDisplay, "localhost:10.0");


    //         Parse command line
    iarg = 1;
    while(iarg < argc) {
      nextarg = iarg+1;
      if(strcmp(argv[iarg], "-nxy") == 0) {
        sscanf(argv[iarg+1], "%d", &iNx);
        sscanf(argv[iarg+2], "%d", &iNy);
        nextarg += 2;
      }
      if(strcmp(argv[iarg], "-nx") == 0) {
        sscanf(argv[iarg+1], "%d", &iNx);
        nextarg++;
      }
      if(strcmp(argv[iarg], "-ny") == 0) {
        sscanf(argv[iarg+1], "%d", &iNy);
        nextarg++;
      }
      if(strcmp(argv[iarg], "-file") == 0) {
        sscanf(argv[iarg+1], "%s", cFile);
        idodisplay = 1;
        nextarg++;
      }
      if(strcmp(argv[iarg], "-display") == 0) {
        sscanf(argv[iarg+1], "%s", cDisplay);
        nextarg++;
      }
      if(strcmp(argv[iarg], "-sleep") == 0) {
        sscanf(argv[iarg+1], "%d", &iSleep);
        iSleep *= 1000;
        nextarg++;
      }

      iarg = nextarg;
    }

    //         Allocate buffer
    iNbytes = 3*iNx*iNy + 4;
    ucBuf = (unsigned char *)malloc(iNbytes);
    if(ucBuf == NULL) {
      fprintf(stdout, "ucBuf allocation FAILED\n");
      exit(0);
    }



    if(idodisplay == 1) {
     /******************************************************************
      * Display one raw frame and exit: driven by command line
      */
      open_window();
      read_pix_from_file_to_buf();
      draw_pix_from_buf();
      close_window();
      exit(0);
    }



   /********************************************************************
    * Interacrive rendering session: driven by stdin
    */
    iSleep = 0;

    open_window();
    for(i=0; i<iNbytes; i++) { ucBuf[i] = (unsigned char)0; }
    draw_pix_from_buf();

    aargv[0] = (char *)malloc(sizeof(char) * 11);
    sprintf(aargv[0], "%s", "hvr_tester");
    aargv[1] = NULL;

    if(pipe(&fds[0]) != 0) { printf("1st pipe creation failed\n");  exit(0); }
    if(pipe(&fds[2]) != 0) { printf("2nd pipe creation failed\n");  exit(0); }

    processID = fork();

    if (processID < 0) {
      err = errno;
      printf("fork failed.  %s.", strerror(err));
      exit(0);
    } 

    if (processID == 0) {
      while (dup2(fds[1], STDOUT_FILENO) < 0) {
        err = errno;
        if (err != EINTR) {
          fp = fopen("RD_DUP2_ERROR", "w");
          fprintf(fp,"dup2 failed on STDOUT_FILENO.  %s.", strerror(err));
          fclose(fp);
          exit(0);
        }
      }
      // STDOUT_FILENO=fds[1] in exected process, so fds[1] is not needed
      // Cclose(fds[1]);

      while (dup2(fds[2], STDIN_FILENO) < 0) {
        err = errno;
        if (err != EINTR) {
          fp = fopen("RD_DUP2_ERROR", "w");
          fprintf(fp, "dup2 failed on STDIN_FILENO.  %s.", strerror(err));
          fclose(fp);
          exit(0);
        }
      }
      // STDIN_FILENO=fds[2] in exected process, so fds[2] is not needed
      // Cclose(fds[2]);

      iValsend = 123;
      write(STDOUT_FILENO, &iValsend, 4);
      execvp(aargv[0], aargv);

      fp = fopen("DEBUG_execvp", "w");
      fprintf(fp, "execvp failed!\n");
      fclose(fp);
    }

    fd_hvr_out = fds[0];
    fd_hvr_in  = fds[3];

    fp_hvr_in = fdopen(fd_hvr_in, "w");
    if(fp_hvr_in == NULL) {
      fprintf(stderr, "Failed to create file stream fp_hvr_in\n");
      exit(0);
    }

    // Get one int from hvr_tester, means it is up and ready
    iostat = read(fd_hvr_out, &iVal, 4);



   /********************************************************************
    * Main event loop: get input stdin
    */
    icount = 0;
    iexit = 0;
    while(iexit == 0  &&  icount < 100) {
      ngot = 3;
      if(fgets(cnewline, 256, stdin) != NULL) {
        ngot = sscanf(cnewline, "%s", key);

        if(strcmp(key, "display") == 0) {         // Display a Pixel file
          iostat = fprintf(fp_hvr_in, "%s", cnewline);
          fflush(fp_hvr_in);
          read_pix_from_hvr_to_display();
        }

      } else {
        iexit = 1;
      }
      icount++;
    }


   /********************************************************************
    * Shutdown: Make hvr_tester exit, close pipes, close window, & exit
    */
    iostat = fprintf(fp_hvr_in, "exit\n");
    fflush(fp_hvr_in);
    iostat = read(fd_hvr_out, &iVal, 4);
    fclose(fp_hvr_in);
    close(fd_hvr_in);
    close_window();
}

void read_pix_from_hvr_to_display()
{
  int iVal, iostat, ngot;
  read(fd_hvr_out, &iVal, 4);
  if(iVal == 0) {
    ngot = read(fd_hvr_out, ucBuf, 3*iNx*iNy);
    draw_pix_from_buf();
  }
}

void open_window()
{
    int attributeListDbl[] = { GLX_DOUBLEBUFFER, GLX_RGBA, None };
    int attributeListSgl[] = { GLX_RGBA, None };

    // int attributeListDbl[] = { GLX_DOUBLEBUFFER, GL_RGB, None };
    // int attributeListSgl[] = {GLX_RGBA, GLX_RED_SIZE, 4, GLX_GREEN_SIZE, 4,
    //                           GLX_BLUE_SIZE, 4, None};
    int iNbuffers;

    iNbuffers = 1;
    dpyHVR = XOpenDisplay(cDisplay);
    if(dpyHVR == NULL) {
        fprintf(stdout, "Could not XOpenDisplay(), on --%s<--\n", cDisplay);
        exit(0);
    }

    if(iNbuffers == 1) {
      viHVR = glXChooseVisual(dpyHVR, DefaultScreen(dpyHVR), attributeListSgl);
      if (viHVR == NULL) {
          fprintf(stdout,"could not glXChooseVisual single buffered\n");
          exit(0);
      }
    } else {
      viHVR = glXChooseVisual(dpyHVR, DefaultScreen(dpyHVR), attributeListDbl);
      if (viHVR == NULL) {
          fprintf(stdout,"could not glXChooseVisual double buffered\n");
          exit(0);
      }
    }

    cxHVR = glXCreateContext(dpyHVR, viHVR, 0, GL_TRUE);
    cmapHVR = XCreateColormap(dpyHVR, RootWindow(dpyHVR, viHVR->screen),
                                viHVR->visual, AllocNone);
    swaHVR.colormap = cmapHVR;
    swaHVR.border_pixel = 0;
    swaHVR.event_mask = StructureNotifyMask;
    winHVR  =  XCreateWindow(dpyHVR,  RootWindow(dpyHVR, viHVR->screen),
                             0, 0, iNx, iNy,
                             0, viHVR->depth, InputOutput, viHVR->visual,
                             CWBorderPixel|CWColormap|CWEventMask, &swaHVR);

    XMapWindow(dpyHVR, winHVR);
    XIfEvent(dpyHVR, &eventHVR, WaitForNotify, (char*)winHVR);
    glXMakeCurrent(dpyHVR, winHVR, cxHVR);


    glMatrixMode(GL_TEXTURE);
    glDisable(GL_TEXTURE_3D);
    glDisable(GL_BLEND);
    //glEnable(GL_TEXTURE_2D);
}

void read_pix_from_file_to_buf()
{
    int fd, ngot;
    fd = open(cFile, O_RDONLY);
    if(fd < 0) {
      fprintf(stdout,"Open Failed, cFile: -->%s<--\n", cFile);
      exit(0);
    }
    ngot = read(fd, ucBuf, iNbytes);
    if(ngot < iNbytes) {
      fprintf(stdout,"Short read: cFile=%s  iNbytes=%d  ngot=%d\n", cFile,iNbytes,ngot);
      exit(0);
    }
    close(fd);
}

void draw_pix_from_buf()
{
    ucPix = &ucBuf[4];
    glDrawPixels(iNx, iNy, GL_RGB, GL_UNSIGNED_BYTE, ucPix);
    glFlush();
    glFinish();

    _sleep(iSleep);
}

unsigned long sol[1280*1024];

void draw_pix_from_buf_ack()
{
  float fZero = (float)0.0;
  float fOne  = (float)1.0;
  int i, i3;

  ucPix = &ucBuf[4];
  i3 = 0;
  for(i=0; i<iNx*iNy; i++) {
      sol[i] =   (unsigned long)ucPix[i3  ]
             + (((unsigned long)ucPix[i3+1])<<8 )
             + (((unsigned long)ucPix[i3+2])<<16);
      i3 += 3;
  }

  glColor4d((double)1.0, (double)1.0, (double)1.0, (double)1.0);

  glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,iNx,iNy,0,GL_RGBA,GL_UNSIGNED_BYTE,sol);
  glBegin(GL_QUADS);
  glTexCoord2f(fZero, fZero);    glVertex2i(  0,   0);
  glTexCoord2f(fOne , fZero);    glVertex2i(511,   0);
  glTexCoord2f(fOne , fOne );    glVertex2i(511, 511);
  glTexCoord2f(fZero, fOne );    glVertex2i(  0, 511);
  glEnd();

  glFlush();
  glFinish();
}


void close_window()
{
    if (dpyHVR != NULL) XCloseDisplay(dpyHVR);
    if (ucBuf != NULL) free(ucBuf);
}

