#include <string.h>
#include <stdio.h>


/* OpenGL Headers */
#define GL_GLEXT_PROTOTYPES
#include <GL/glx.h>
#include <X11/extensions/xf86vmode.h>
#include <X11/keysym.h>
#include <X11/extensions/Xdbe.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>

#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <linux/stat.h>

/************************************************
 *       Defined constants and macros
 ************************************************/
#define _sleep(n) usleep(n*1000)
#define FIFO_FILE       "RDFIFO"

/************************************************
 *             Globale variables
 ************************************************/
Display *dpyHVR;
XVisualInfo *viHVR;
Colormap cmapHVR;
XSetWindowAttributes swaHVR;
Window winHVR;
GLXContext cxHVR;
XEvent eventHVR;

unsigned char *ucBuf = NULL;
unsigned char *ucPix;

/************************************************
 *              Utility routines
 ************************************************/
Bool WaitForNotify(Display *d, XEvent *e, char *arg)
{ return (e->type == MapNotify) && (e->xmap.window == (Window)arg); }

/************************************************
 *               Main routine
 ************************************************/
int main(int argc, char* argv[])
{
    int attributeListDbl[] = { GLX_DOUBLEBUFFER, GLX_RGBA, None };
    int attributeListSgl[] = { GLX_RGBA, None };
    int iNx, iNy, iNbuffers, iostat;
    int i, iNbytes, iVal, iValsend, ngot;
    int fd, iexit, icount, iexecret, err, fd_hvr_in, fd_hvr_out;
    FILE *fp, *fp_hvr_in;
    char cnewline[256], key[256];


    char byte = '\0';
    pid_t processID;
    int fds[4];
    char* aargv[2];


    aargv[0] = (char *)malloc(sizeof(char) * 11);
    sprintf(aargv[0], "%s", "hvr_tester");
    aargv[1] = NULL;

    if(pipe(&fds[0]) != 0) { printf("1st pipe creation failed\n");  exit(0); }
    if(pipe(&fds[2]) != 0) { printf("2nd pipe creation failed\n");  exit(0); }
    printf("Both pipes created\n");
    for(i=0; i<4; i++) { printf("fds[%d] = %d\n", i, fds[i]); }

    processID = fork();
    printf("Forked: processID = %d\n", processID);

    if (processID < 0) {
      err = errno;
      printf("fork failed.  %s.", strerror(err));
      exit(0);
    } 

    if (processID == 0) {
      while (dup2(fds[1], STDOUT_FILENO) < 0) {
        err = errno;
        if (err != EINTR) {
          fp = fopen("RD_DUP2_ERROR", "w");
          fprintf(fp,"dup2 failed on STDOUT_FILENO.  %s.", strerror(err));
          fclose(fp);
          exit(0);
        }
      }
      // Cclose(fds[1]);  STDOUT_FILENO=fds[1] in exected process, so fds[1] is not needed

      while (dup2(fds[2], STDIN_FILENO) < 0) {
        err = errno;
        if (err != EINTR) {
          fp = fopen("RD_DUP2_ERROR", "w");
          fprintf(fp, "dup2 failed on STDIN_FILENO.  %s.", strerror(err));
          fclose(fp);
          exit(0);
        }
      }
      // Cclose(fds[2]);  STDIN_FILENO=fds[2] in exected process, so fds[2] is not needed

      iValsend = 123;
      write(STDOUT_FILENO, &iValsend, 4);
      execvp(aargv[0], aargv);

      fp = fopen("DEBUG_execvp", "w");
      fprintf(fp, "execvp failed!\n");
      fclose(fp);
    }

    printf("processID = %d\n", processID);
    for(i=0; i<4; i++) { printf("fds[%d] = %d\n", i, fds[i]); }

    fd_hvr_out = fds[0];
    fd_hvr_in  = fds[3];

    fp_hvr_in = fdopen(fd_hvr_in, "w");
    if(fp_hvr_in == NULL) {
      printf("Failed to create file stream fp_hvr_in\n");
      exit(0);
    }

    iNx = 512;
    iNy = 512;
    iNbytes = 3*iNx*iNy + 4;
    ucBuf = (unsigned char *)malloc(iNbytes);
    fprintf(stdout, "ucBuf allocated & filled\n");

    // Get one int from hvr_tester, means it is up and ready
    iostat = read(fd_hvr_out, &iVal, 4);
    fprintf(stdout, "Read:  iostat=%d   iVal=%d\n", iostat, iVal);



    icount = 0;
    iexit = 0;
    while(iexit == 0  &&  icount < 100) {
      ngot = 3;
      if(fgets(cnewline, 256, stdin) != NULL) {
        ngot = sscanf(cnewline, "%s", key);
        iostat = fprintf(fp_hvr_in, "%s\n", key);
        fflush(fp_hvr_in);
        fprintf(stdout, "fprintf to fp_hvr_in: iostat=%d\n", iostat);
        iostat = read(fd_hvr_out, &iVal, 4);
        fprintf(stdout, "Read:  iostat=%d   iVal=%d\n", iostat, iVal);
      } else {
        iexit = 1;
      }
      icount++;
    }

    // This should make hvr_tester exit
    iostat = fprintf(fp_hvr_in, "exit\n");
    fflush(fp_hvr_in);
    fprintf(stdout, "fprintf: iostat=%d\n", iostat);
    iostat = read(fd_hvr_out, &iVal, 4);
    fprintf(stdout, "Read:  iostat=%d   iVal=%d\n", iostat, iVal);

    fclose(fp_hvr_in);
    close(fd_hvr_in);

    iexit = 1;
    if(iexit == 1) { exit(0); }
//-----------------------------------------------------------------------------
    printf("Should not have gotten here\n");




    icount = 0;
    iexit = 0;
    while(iexit == 0  &&  icount < 100) {
      ngot = 3;
      if(fgets(cnewline, 256, stdin) != NULL) {
        ngot = sscanf(cnewline, "%s", key);
        printf("(rd:-->%s<--)\n", key);
        fflush(stdout);
      } else {
        iexit = 1;
      }
      icount++;
    }

    iNbuffers = 1;

    fprintf(stdout, "(rd:exitting)\n");

    if(iNx > 1) { exit(0); }


    iVal = 255;

    dpyHVR = XOpenDisplay(":10.0");
    if(dpyHVR == NULL) {
        fprintf(stdout, "Could not XOpenDisplay()\n");
        exit(0);
    }


    fprintf(stdout, "Display opened\n");

    if(iNbuffers == 1) {
      viHVR = glXChooseVisual(dpyHVR, DefaultScreen(dpyHVR), attributeListSgl);
      if (viHVR == NULL) {
          fprintf(stdout,"could not glXChooseVisual single buffered\n");
          exit(0);
      }
    } else {
      viHVR = glXChooseVisual(dpyHVR, DefaultScreen(dpyHVR), attributeListDbl);
      if (viHVR == NULL) {
          fprintf(stdout,"could not glXChooseVisual double buffered\n");
          exit(0);
      }
    }

    cxHVR = glXCreateContext(dpyHVR, viHVR, 0, GL_TRUE);
    cmapHVR = XCreateColormap(dpyHVR, RootWindow(dpyHVR, viHVR->screen),
                                viHVR->visual, AllocNone);
    swaHVR.colormap = cmapHVR;
    swaHVR.border_pixel = 0;
    swaHVR.event_mask = StructureNotifyMask;
    winHVR  =  XCreateWindow(dpyHVR,  RootWindow(dpyHVR, viHVR->screen),
                             0, 0, iNx, iNy,
                             0, viHVR->depth, InputOutput, viHVR->visual,
                             CWBorderPixel|CWColormap|CWEventMask, &swaHVR);

    fprintf(stdout, "Window Created\n");

    XMapWindow(dpyHVR, winHVR);
    XIfEvent(dpyHVR, &eventHVR, WaitForNotify, (char*)winHVR);
    glXMakeCurrent(dpyHVR, winHVR, cxHVR);

    fprintf(stdout, "Window mapped & made current\n");

    glMatrixMode(GL_TEXTURE);
    glDisable(GL_TEXTURE_3D);
    glDisable(GL_BLEND);

    fd = open("aa.out", O_RDONLY);
    // ngot = lseek(fd, offset, SEEK_SET);
    ngot = read(fd, ucBuf, iNbytes);
    close(fd);

    ucPix = &ucBuf[4];
    glDrawPixels(iNx, iNy, GL_RGB, GL_UNSIGNED_BYTE, ucPix);
    glFlush();
    glFinish();

    fprintf(stdout, "Pixels drawn & flushed\n");

    _sleep(5000);

    fprintf(stdout, "Sleep done\n");

   // Close window
   if (dpyHVR != NULL) XCloseDisplay(dpyHVR);
   if (ucBuf != NULL) free(ucBuf);


    fprintf(stdout, "Window closed\n");
}

