#include <string.h>
#include <stdio.h>


/* OpenGL Headers */
#define GL_GLEXT_PROTOTYPES
#include <GL/glx.h>
#include <X11/extensions/xf86vmode.h>
#include <X11/keysym.h>
#include <X11/extensions/Xdbe.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>

#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <linux/stat.h>

/************************************************
 *       Defined constants and macros
 ************************************************/
#define _sleep(n) usleep(n*1000)
#define FIFO_FILE       "RDFIFO"

/**************************************************************************** 
 *             Globale variables
 ****************************************************************************/
Display *dpyHVR;
XVisualInfo *viHVR;
Colormap cmapHVR;
XSetWindowAttributes swaHVR;
Window winHVR;
GLXContext cxHVR;
XEvent eventHVR;

unsigned char *ucBuf = NULL;
unsigned char *ucPix;
int iNx, iNy, iNbytes, iSleep;
char cFile[256], cDisplay[256];
int fd_hvr_in, fd_hvr_out;

int   idebug;
int   ishowcolorbar = 0;
int   iSelALPknot = -1,iSelRGBknot = -1;
int   naknots, ncknots, iaknot[256], icknot[256];
float aknot[256], rknot[256], gknot[256], bknot[256];
float opacity, fov, eye[3], cent[3], up[3], clip[2];
int   ifauxclip;
float auxclip[6][4];
float nplanes[2];
int   indextolerance;
float angletolerance, sizetolerance;
int   ipaletted;
char  hvfile[256];
char  vecfile[256];
float vecradius=0.0;
char  cFldClrLo[256], cFldClrHi[256];
int   nKeyLines;
char  cKeyLines[100][80];
int   movieframe = -1;
int   keysonly   = 0;
char  cMovieRoot[256];

int iusedefaultcmap=0;
char current_lut_filter[256]="";

FILE *fp_hvr_in;


/**************************************************************************** 
 *                       Prototypes
 ****************************************************************************/
void open_window();
void read_pix_from_file_to_buf();
void read_pix_from_hvr_to_display();
void draw_pix_from_buf();
void close_window();
int read_keyframe_file(char *cKeyFile);
int difview();
int difsubvol();
int calcfield(char *cCalcFld);
void gen_keyframe_lines();
void load_hv(char *cnewline);
void load_vec(char *cnewline);
void load_key(char *cnewline);
void save_key(char *cnewline);
void save_image(char *cnewline);
void start_movie(char *cnewline);
void end_movie();
void do_ecandup(char *cnewline);
void do_trackball(char *cnewline);
void do_shift(char *cnewline);
void do_setcent(char *cnewline);
void do_clip(char *cnewline);
void do_auxclip_set(char *cnewline);
void do_auxclip_off(char *cnewline);
void set_default_keyinfo();
void do_edit_cmap(char *cnewline);
void do_field_lines(char *cnewline);
void render_and_display();
void load_lut_list(char *cLutFilter);
void do_default_lut();


int file_shift(char *file, int shift);
int gen_filter(char *file, char *filter);
int gen_file_list(char *filter);
int get_ith_file(int iFile, char *cFile);
int get_file_index(char *file);


/**************************************************************************** 
 *                     A few useful constants
 ****************************************************************************/
float halfpi, flog2;


/**************************************************************************** 
 *                     Utility routines
 ****************************************************************************/
Bool WaitForNotify(Display *d, XEvent *e, char *arg)
{ return (e->type == MapNotify) && (e->xmap.window == (Window)arg); }


/**************************************************************************** 
 *                       Main routine
 ****************************************************************************/
int main(int argc, char* argv[])
{
    int iostat;
    int i, iVal, iValsend, ngot, iarg, nextarg;
    int iexit, icount, iexecret, err;
    FILE *fp;
    char cnewline[256], key[256], cKeyFile[256], cKeyLine[256];
    char ccarg[256];
    float farg1, farg2, farg3;
    int nrot, nshift, ntimes;

    char byte = '\0';
    pid_t processID;
    int fds[4];
    char* aargv[2];

   /********************************************************************
    *             Initialize                          
    */

    // set useful constants
    halfpi = 2.0 * atan(1.0);
    flog2  = log(2.0);

    //         Defaults
    iNx = 512;
    iNy = 512;
    iSleep = 0;
    sprintf(cDisplay, "localhost:10.0");


    //         Parse command line
    iarg = 1;
    while(iarg < argc) {
      nextarg = iarg+1;
      if(strcmp(argv[iarg], "-nxy") == 0) {
        sscanf(argv[iarg+1], "%d", &iNx);
        sscanf(argv[iarg+2], "%d", &iNy);
        nextarg += 2;
      }
      if(strcmp(argv[iarg], "-display") == 0) {
        sscanf(argv[iarg+1], "%s", cDisplay);
        nextarg++;
      }
      if(strcmp(argv[iarg], "-sleep") == 0) {
        sscanf(argv[iarg+1], "%d", &iSleep);
        iSleep *= 1000;
        nextarg++;
      }

      iarg = nextarg;
    }

    //   Allocate buffer
    iNbytes = 3*iNx*iNy + 4;
    ucBuf = (unsigned char *)malloc(iNbytes);
    if(ucBuf == NULL) {
      fprintf(stdout, "ucBuf allocation FAILED\n");
      exit(0);
    }

    //   Open X11 Display
    open_window();
    for(i=0; i<iNbytes; i++) { ucBuf[i] = (unsigned char)0; }
    draw_pix_from_buf();


   /********************************************************************
    * Spawn hvr_devel with pipes to stdin & stdout   
    */
    aargv[0] = (char *)malloc(sizeof(char) * 14);
    sprintf(aargv[0], "%s", "hvr_devel.exe");
    aargv[1] = NULL;

    if(pipe(&fds[0]) != 0) { printf("1st pipe creation failed\n");  exit(0); }
    if(pipe(&fds[2]) != 0) { printf("2nd pipe creation failed\n");  exit(0); }

    processID = fork();

    if (processID < 0) {
      err = errno;
      printf("fork failed.  %s.", strerror(err));
      exit(0);
    } 

    if (processID == 0) {
      while (dup2(fds[1], STDOUT_FILENO) < 0) {
        err = errno;
        if (err != EINTR) {
          fp = fopen("RD_DUP2_ERROR", "w");
          fprintf(fp,"dup2 failed on STDOUT_FILENO.  %s.", strerror(err));
          fclose(fp);
          exit(0);
        }
      }
      // STDOUT_FILENO=fds[1] in exected process, so fds[1] is not needed
      // Cclose(fds[1]);

      while (dup2(fds[2], STDIN_FILENO) < 0) {
        err = errno;
        if (err != EINTR) {
          fp = fopen("RD_DUP2_ERROR", "w");
          fprintf(fp, "dup2 failed on STDIN_FILENO.  %s.", strerror(err));
          fclose(fp);
          exit(0);
        }
      }
      // STDIN_FILENO=fds[2] in exected process, so fds[2] is not needed
      // Cclose(fds[2]);

      iValsend = 123;
      write(STDOUT_FILENO, &iValsend, 4);
      execvp(aargv[0], aargv);

      fp = fopen("DEBUG_execvp", "w");
      fprintf(fp, "execvp failed!\n");
      fclose(fp);
    }

    fd_hvr_out = fds[0];
    fd_hvr_in  = fds[3];

    fp_hvr_in = fdopen(fd_hvr_in, "w");
    if(fp_hvr_in == NULL) {
      fprintf(stderr, "Failed to create file stream fp_hvr_in\n");
      exit(0);
    }

    // Get one int from hvr_devel, means it is up and ready
    iostat = read(fd_hvr_out, &iVal, 4);



   /********************************************************************
    * Main event loop: get input stdin
    */
    set_default_keyinfo();

    icount = 0;
    iexit = 0;
    while(iexit == 0) {
      ngot = 3;
      if(fgets(cnewline, 256, stdin) != NULL) {
        ngot = sscanf(cnewline, "%s", key);

        if(strcmp(key, "loadhv"  ) == 0) { load_hv(cnewline); }
        if(strcmp(key, "loadvec" ) == 0) { load_vec(cnewline); }

        if(strcmp(key, "loadkey" ) == 0) { load_key(cnewline); }
        if(strcmp(key, "loadluts") == 0) { load_key(cnewline); }
        if(strcmp(key, "loaddata") == 0) { load_key(cnewline); }
        if(strcmp(key, "loadview") == 0) { load_key(cnewline); }
        if(strcmp(key, "loadqlty") == 0) { load_key(cnewline); }

        if(strcmp(key, "savekey" ) == 0) { save_key(cnewline); }
        if(strcmp(key, "saveluts") == 0) { save_key(cnewline); }
        if(strcmp(key, "savedata") == 0) { save_key(cnewline); }
        if(strcmp(key, "saveview") == 0) { save_key(cnewline); }
        if(strcmp(key, "saveqlty") == 0) { save_key(cnewline); }
        if(strcmp(key, "saveimage") == 0) { save_image(cnewline); }
        if(strcmp(key, "movie_start") == 0) { start_movie(cnewline); }
        if(strcmp(key, "movie_end") == 0) { end_movie(); }

        if(strncmp(key, "dorender",3) == 0) { render_and_display(); }

        if(strcmp(key, "ecandup") == 0) { do_ecandup(cnewline); }
        if(strcmp(key, "trackball") == 0) { do_trackball(cnewline); }
        if(strcmp(key, "shift") == 0) { do_shift(cnewline); }
        if(strcmp(key, "center") == 0) { do_setcent(cnewline); }
        if(strcmp(key, "clipcpm") == 0) { do_clip(cnewline); }
        if(strcmp(key, "clip_aux_set") == 0) { do_auxclip_set(cnewline); }
        if(strcmp(key, "clip_aux_off") == 0) { do_auxclip_off(cnewline); }
        if(strncmp(key, "cmap",4) == 0) { do_edit_cmap(cnewline); }
        if(strncmp(key, "fld",3) == 0) { do_field_lines(cnewline); }

        if(strcmp(key, "trackball100") == 0) {
          for(i=0; i<100; i++) { do_trackball(cnewline); }
        }

        if(strcmp(key, "timeshift") == 0) {
          ngot = sscanf(cnewline, "timeshift %d %d", &nshift, &ntimes);
          if(ngot == 2) {
            for(i=0; i<ntimes; i++) {
              if(file_shift(hvfile, nshift) != 0) {
                render_and_display();
                printf("hvfile = %s\n", hvfile);
              }
            }
          }
        }

        if(strcmp(key, "trackball90") == 0) {
          ngot = sscanf(cnewline, "%s %f %f %f", ccarg, &farg1, &farg2, &farg3);
          nrot = 0;
          if(int(0.001+farg1) > nrot) { nrot = int(0.001+farg1); }
          if(int(0.001-farg1) > nrot) { nrot = int(0.001-farg1); }
          if(int(0.001+farg2) > nrot) { nrot = int(0.001+farg2); }
          if(int(0.001-farg2) > nrot) { nrot = int(0.001-farg2); }
          for(i=0; i<nrot; i++) { do_trackball(cnewline); }
        }

        if(strcmp(key, "display") == 0) {         // Display a Pixel file
          iostat = fprintf(fp_hvr_in, "%s", cnewline);
          fflush(fp_hvr_in);
          read_pix_from_hvr_to_display();
        }

      } else {
        iexit = 1;
      }
      icount++;
    }



   /********************************************************************
    * Shutdown: Make hvr_devel exit, close pipes, close window, & exit
    */
    iostat = fprintf(fp_hvr_in, "exit\n");
    fflush(fp_hvr_in);
    iostat = read(fd_hvr_out, &iVal, 4);
    fclose(fp_hvr_in);
    close(fd_hvr_in);
    close_window();
}

void read_pix_from_hvr_to_display()
{
  int iVal, iostat, ngot;
  read(fd_hvr_out, &iVal, 4);
  if(iVal == 0) {
    ngot = read(fd_hvr_out, ucBuf, 3*iNx*iNy);
    draw_pix_from_buf();
  } else {
    fprintf(stderr, "HVR ERROR: %d\n", iVal);
  }
}

void open_window()
{
    int attributeListDbl[] = { GLX_DOUBLEBUFFER, GLX_RGBA, None };
    int attributeListSgl[] = { GLX_RGBA, None };

    // int attributeListDbl[] = { GLX_DOUBLEBUFFER, GL_RGB, None };
    // int attributeListSgl[] = {GLX_RGBA, GLX_RED_SIZE, 4, GLX_GREEN_SIZE, 4,
    //                           GLX_BLUE_SIZE, 4, None};
    int iNbuffers;

    iNbuffers = 1;
    dpyHVR = XOpenDisplay(cDisplay);
    if(dpyHVR == NULL) {
        fprintf(stdout, "Could not XOpenDisplay(), on --%s<--\n", cDisplay);
        exit(0);
    }

    if(iNbuffers == 1) {
      viHVR = glXChooseVisual(dpyHVR, DefaultScreen(dpyHVR), attributeListSgl);
      if (viHVR == NULL) {
          fprintf(stdout,"could not glXChooseVisual single buffered\n");
          exit(0);
      }
    } else {
      viHVR = glXChooseVisual(dpyHVR, DefaultScreen(dpyHVR), attributeListDbl);
      if (viHVR == NULL) {
          fprintf(stdout,"could not glXChooseVisual double buffered\n");
          exit(0);
      }
    }

    cxHVR = glXCreateContext(dpyHVR, viHVR, 0, GL_TRUE);
    cmapHVR = XCreateColormap(dpyHVR, RootWindow(dpyHVR, viHVR->screen),
                                viHVR->visual, AllocNone);
    swaHVR.colormap = cmapHVR;
    swaHVR.border_pixel = 0;
    swaHVR.event_mask = StructureNotifyMask;
    winHVR  =  XCreateWindow(dpyHVR,  RootWindow(dpyHVR, viHVR->screen),
                             0, 0, iNx, iNy,
                             0, viHVR->depth, InputOutput, viHVR->visual,
                             CWBorderPixel|CWColormap|CWEventMask, &swaHVR);

    XMapWindow(dpyHVR, winHVR);
    XIfEvent(dpyHVR, &eventHVR, WaitForNotify, (char*)winHVR);
    glXMakeCurrent(dpyHVR, winHVR, cxHVR);


    glMatrixMode(GL_TEXTURE);
    glDisable(GL_TEXTURE_3D);
    glDisable(GL_BLEND);
    //glEnable(GL_TEXTURE_2D);
}


void set_default_keyinfo()
{
  naknots = 2;
  iaknot[0]=  0;  aknot[0]=0.0f;
  iaknot[1]=255;  aknot[1]=1.0f;

  ncknots = 2;
  icknot[0]=  0;  rknot[0]=0.0f;  gknot[0]=0.0f;  bknot[0]=0.0f;
  icknot[1]=255;  rknot[1]=1.0f;  gknot[1]=1.0f;  bknot[1]=1.0f;

  idebug  = 0;
  opacity = 1.0f;
  fov     = 35.308399f;
  eye[0]  = 0.0f;  eye[1]  = 0.0f;   eye[2]  = 1.0f;
  cent[0] = 0.0f;  cent[1] = 0.0f;   cent[2] = 0.0f;
  up[0]   = 0.0f;  up[1]   = 1.0f;   up[2]   = 0.0f;
  nplanes[0] = 1.0f;   nplanes[1] = 1.0f;
  clip[0]    = 0.1f;   clip[1]    = 100.0f;
  ifauxclip = 0;
  indextolerance = 5;
  angletolerance = 0.0000001f;
  sizetolerance  = 1.0f;
  ipaletted      = 1;
  sprintf(hvfile, "none");
  sprintf(vecfile, "none");
  vecradius = 0.0;
  sprintf(cFldClrLo, "#008000");
  sprintf(cFldClrHi, "#FFFFFF");
}

int read_keyframe_file(char *cKeyFile)
{
  char c[256], cc[256], ll[256];
  int ngot, iresetcolors, iresetalpha;
  int iaux, i, iDot;
  float auxx, auxy, auxz, auxd;
  FILE *fp;

  iresetcolors = 0;
  iresetalpha  = 0;

  for(i=0; i<strlen(cKeyFile); i++) {
    if(cKeyFile[i] == '.') iDot = i;
  }
  if(strcmp(&cKeyFile[iDot], ".key" ) == 0) { ifauxclip = 0; }
  if(strcmp(&cKeyFile[iDot], ".view") == 0) { ifauxclip = 0; }

  fp = fopen(cKeyFile, "r");
  if(fp == NULL) {
     printf("Failed to open: %s\n", cKeyFile);
     return -1;
   }

  while(fgets(ll, 256, fp) != NULL) {
    ngot = sscanf(ll, "%s", cc);

    if(strcmp(cc, "Anot:") == 0) {
      if(iresetalpha == 0) { iresetalpha = 1;   naknots = 0; }
      sscanf(ll, "%s %d %f", c, &iaknot[naknots], &aknot[naknots]);
      naknots++;
    }
    if(strcmp(cc, "Cnot:") == 0) {
      if(iresetcolors == 0) { iresetcolors = 1;   ncknots = 0; }
      ngot = sscanf(ll, "%s %d %f %f %f",
             c, &icknot[ncknots], &rknot[ncknots], &gknot[ncknots], &bknot[ncknots]);
      ncknots++;
    }
    if(strcmp(cc,"debug")== 0) { sscanf(ll,"%s %d", c, &idebug); }
    if(strcmp(cc,"opacity")== 0) { sscanf(ll,"%s %f", c, &opacity); }
    if(strcmp(cc,"hvfile") == 0) { sscanf(ll,"%s %s", c, hvfile); }
    if(strcmp(cc,"vecfile") == 0) { sscanf(ll,"%s %s", c, vecfile); }
    if(strcmp(cc,"vecradius") == 0) { sscanf(ll,"%s %f", c, &vecradius); }
    if(strcmp(cc,"veclohiclr") == 0) { sscanf(ll,"%s %s %s", c, cFldClrLo, cFldClrHi); }
    if(strcmp(cc,"fov"   ) == 0) { sscanf(ll,"%s %f", c, &fov); }
    if(strcmp(cc,"eye"   ) == 0) { sscanf(ll,"%s %f %f %f", c,&eye[0],&eye[1],&eye[2]); }
    if(strcmp(cc,"center") == 0) { sscanf(ll,"%s %f %f %f", c,&cent[0],&cent[1],&cent[2]); }
    if(strcmp(cc,"up"    ) == 0) { sscanf(ll,"%s %f %f %f", c,&up[0],&up[1],&up[2]); }
    if(strcmp(cc,"clipping") == 0) { sscanf(ll,"%s %f, %f", c,&clip[0],&clip[1]); }
    if(strcmp(cc,"auxclip" ) == 0) {
      sscanf(ll,"%s %d %f %f %f %f", c, &iaux, &auxx, &auxy, &auxz, &auxd);
      auxclip[iaux][0] = auxx;
      auxclip[iaux][1] = auxy;
      auxclip[iaux][2] = auxz;
      auxclip[iaux][3] = auxd;
      ifauxclip = 2;
    }
    if(strcmp(cc,"nplanes" ) == 0) { sscanf(ll,"%s %f, %f", c,&nplanes[0],&nplanes[1]); }
    if(strcmp(cc,"tolerance" ) == 0) {
      sscanf(ll,"%s %d %f %f", c,&indextolerance, &angletolerance, &sizetolerance);
    }
    if(strcmp(cc,"paletted" ) == 0) { sscanf(ll,"%s %d", c, &ipaletted); }
  }
  fclose(fp);

  return 0;
}

int gen_luts_lines(int i)
{
  int j;
  for(j=0; j<naknots; j++) {
    sprintf(cKeyLines[i], "Anot:       %3d %f\n", iaknot[j], aknot[j]);  i++;
  }
  for(j=0; j<ncknots; j++) {
    sprintf(cKeyLines[i], "Cnot:       %3d %f %f %f\n",
                          icknot[j],rknot[j],gknot[j],bknot[j]);         i++;
  }
  sprintf(cKeyLines[i], "opacity     %f\n", opacity);                    i++;
  sprintf(cKeyLines[i], "colorbar    %d\n", ishowcolorbar);              i++;
  sprintf(cKeyLines[i], "selknots    %d %d\n", iSelALPknot,iSelRGBknot); i++;
  return i;
}
int gen_data_lines(int i)
{
  sprintf(cKeyLines[i], "hvfile      %s\n", hvfile);                  i++;
  sprintf(cKeyLines[i], "vecfile     %s\n", vecfile);                 i++;
  sprintf(cKeyLines[i], "vecradius   %f\n", vecradius);               i++;
  sprintf(cKeyLines[i], "veclohiclr  %s %s\n", cFldClrLo, cFldClrHi); i++;
  return i;
}
int gen_view_lines(int i)
{
  int j;
  sprintf(cKeyLines[i], "fov         %f\n", fov);                          i++;
  sprintf(cKeyLines[i], "eye         %f %f %f\n", eye[0], eye[1], eye[2]); i++;
  sprintf(cKeyLines[i], "center      %f %f %f\n",cent[0],cent[1],cent[2]); i++;
  sprintf(cKeyLines[i], "up          %f %f %f\n",  up[0],  up[1],  up[2]); i++;
  sprintf(cKeyLines[i], "clipping    %f, %f\n",  clip[0],clip[1]);         i++;
  if(ifauxclip == 0) {
    sprintf(cKeyLines[i], "ifauxclip   0 0\n");  i++;
  } else {
    for(j=0; j<6; j++) { sprintf(cKeyLines[i], "ifauxclip   %d 2\n", j);  i++; }
    for(j=0; j<6; j++) { sprintf(cKeyLines[i], "auxclip     %d %f %f %f %f\n",
                         j, auxclip[j][0], auxclip[j][1], auxclip[j][2], auxclip[j][3]);  i++; }
  }
  return i;
}
int gen_qlty_lines(int i)
{
  sprintf(cKeyLines[i], "imagesize   %d %d\n", iNx, iNy);                 i++;
  sprintf(cKeyLines[i], "imagefile   none\n");                            i++;
  sprintf(cKeyLines[i], "nplanes     %f, %f\n", nplanes[0], nplanes[1]);  i++;
  sprintf(cKeyLines[i], "tolerance   %d %f %f\n",
                          indextolerance,angletolerance,sizetolerance);   i++;
  sprintf(cKeyLines[i], "paletted    %d\n",ipaletted);                    i++;
  return i;
}

void gen_keyframe_lines()
{

  nKeyLines = 0;
  if(idebug == 1) {
   sprintf(cKeyLines[nKeyLines], "debug       1\n");  nKeyLines++;
  }
  nKeyLines = gen_luts_lines(nKeyLines);
  nKeyLines = gen_data_lines(nKeyLines);
  nKeyLines = gen_view_lines(nKeyLines);
  nKeyLines = gen_qlty_lines(nKeyLines);
  sprintf(cKeyLines[nKeyLines], "render\n");          nKeyLines++;
}

void gen_keyframe_lines0()
{
  int j,i;
  i = 0;

  // Debug
  if(idebug == 1) { sprintf(cKeyLines[i], "debug       1\n"); i++; }

  // LUT
  for(j=0; j<naknots; j++) {
    sprintf(cKeyLines[i], "Anot:       %3d %f\n", iaknot[j], aknot[j]);
    i++;
  }
  for(j=0; j<ncknots; j++) {
    sprintf(cKeyLines[i], "Cnot:       %3d %f %f %f\n", icknot[j],rknot[j],gknot[j],bknot[j]);
    i++;
  }
  sprintf(cKeyLines[i], "opacity     %f\n", opacity);                           i++;

  // Data
  sprintf(cKeyLines[i], "hvfile      %s\n", hvfile);                            i++;
  sprintf(cKeyLines[i], "vecfile     %s\n", vecfile);                           i++;
  sprintf(cKeyLines[i], "vecradius   %f\n", vecradius);                         i++;
  sprintf(cKeyLines[i], "veclohiclr  %s %s\n", cFldClrLo, cFldClrHi);           i++;

  // View
  sprintf(cKeyLines[i], "fov         %f\n", fov);                               i++;
  sprintf(cKeyLines[i], "eye         %f %f %f\n",  eye[0],  eye[1],  eye[2]);   i++;
  sprintf(cKeyLines[i], "center      %f %f %f\n", cent[0], cent[1], cent[2]);   i++;
  sprintf(cKeyLines[i], "up          %f %f %f\n",   up[0],   up[1],   up[2]);   i++;
  sprintf(cKeyLines[i], "clipping    %f, %f\n",    clip[0], clip[1]);           i++;
  if(ifauxclip == 0) {
    sprintf(cKeyLines[i], "ifauxclip   0 0\n");  i++;
  } else {
    for(j=0; j<6; j++) { sprintf(cKeyLines[i], "ifauxclip   %d 1\n", j);  i++; }
    for(j=0; j<6; j++) { sprintf(cKeyLines[i], "auxclip     %d %f %f %f %f\n",
                         j, auxclip[j][0], auxclip[j][1], auxclip[j][2], auxclip[j][3]);  i++; }
  }

  // Image quality
  sprintf(cKeyLines[i], "imagesize   %d %d\n", iNx, iNy);                       i++;
  sprintf(cKeyLines[i], "imagefile   none\n");                                  i++;
  sprintf(cKeyLines[i], "nplanes     %f, %f\n",    nplanes[0], nplanes[1]);     i++;
  sprintf(cKeyLines[i], "tolerance   %d %f %f\n",
                          indextolerance,angletolerance,sizetolerance);         i++;
  sprintf(cKeyLines[i], "paletted    %d\n",ipaletted);                          i++;

  // Finish/Render
  sprintf(cKeyLines[i], "render\n");                                            i++;

  nKeyLines = i;
}

/*
Anot:         0 0.000000
Anot:        64 0.000000
Anot:       120 0.000000
Anot:       136 0.000000
Anot:       192 1.000000
Anot:       255 1.000000
Cnot:         0 1.000000 1.000000 0.000000
Cnot:        64 1.000000 1.000000 0.000000
Cnot:       120 1.000000 0.000000 0.000000
Cnot:       136 0.000000 0.000000 1.000000
Cnot:       192 0.000000 1.000000 1.000000
Cnot:       255 0.000000 1.000000 1.000000
opacity     1.000000
hvfile      /mnt/global/dhp/data/rgb/g02vrad0400.hv
imagesize   512 512
imagefile   none
fov         35.308399
eye         0.00  6.00  0.00
center      0.00  0.00  0.00
up          0.00  0.00  1.00
clipping    0.1, 14.0
nplanes     1.000000, 1.000000
tolerance   5 0.000001 1.000000
paletted    1
render
*/

void read_pix_from_file_to_buf()
{
    int fd, ngot;
    fd = open(cFile, O_RDONLY);
    if(fd < 0) {
      fprintf(stdout,"Open Failed, cFile: -->%s<--\n", cFile);
      exit(0);
    }
    ngot = read(fd, ucBuf, iNbytes);
    if(ngot < iNbytes) {
      fprintf(stdout,"Short read: cFile=%s  iNbytes=%d  ngot=%d\n", cFile,iNbytes,ngot);
      exit(0);
    }
    close(fd);
}

void draw_pix_from_buf()
{
    ucPix = &ucBuf[0];
    glDrawPixels(iNx, iNy, GL_RGB, GL_UNSIGNED_BYTE, ucPix);
    glFlush();
    glFinish();

    //_sleep(iSleep);  sleeps for iSleep millisec
}

unsigned long sol[1280*1024];

void draw_pix_from_buf_ack()
{
  float fZero = (float)0.0;
  float fOne  = (float)1.0;
  int i, i3;

  ucPix = &ucBuf[4];
  i3 = 0;
  for(i=0; i<iNx*iNy; i++) {
      sol[i] =   (unsigned long)ucPix[i3  ]
             + (((unsigned long)ucPix[i3+1])<<8 )
             + (((unsigned long)ucPix[i3+2])<<16);
      i3 += 3;
  }

  glColor4d((double)1.0, (double)1.0, (double)1.0, (double)1.0);

  glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,iNx,iNy,0,GL_RGBA,GL_UNSIGNED_BYTE,sol);
  glBegin(GL_QUADS);
  glTexCoord2f(fZero, fZero);    glVertex2i(  0,   0);
  glTexCoord2f(fOne , fZero);    glVertex2i(511,   0);
  glTexCoord2f(fOne , fOne );    glVertex2i(511, 511);
  glTexCoord2f(fZero, fOne );    glVertex2i(  0, 511);
  glEnd();

  glFlush();
  glFinish();
}

void close_window()
{
    if (dpyHVR != NULL) XCloseDisplay(dpyHVR);
    if (ucBuf != NULL) free(ucBuf);
}





/* --------------------------------------------------------------
 *  Utility functions
 *  Use of these would be very ineffient if they were inner most loops.
 *  These are only a few times per frame and take no time compared
 *  to any of the volume rendering, system calls, or network xfers. 
 *  Redability and resuability are more important here.
 */
void cross(float *axb, float *a, float *b)
{
  axb[0] = a[1]*b[2] - a[2]*b[1];
  axb[1] = a[2]*b[0] - a[0]*b[2];
  axb[2] = a[0]*b[1] - a[1]*b[0];
}

float fnorm(float *v) { return (float)sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]); } 
float vecdot(float *a, float *b) { return (a[0]*b[0] + a[1]*b[1] + a[2]*b[2]); }


void render_and_display()
{
  int i, ierr, ilen; 
  float fOldNearClip, clipmin, ec[3];
  char str[256], cLutFilter[256], cKeyFile[256];
  FILE *fp;

  // Validate keyframe info: complain and return if no good.
  if(strcmp(hvfile, "none") == 0) {
    fprintf(stderr, "HV file is not set\n");
    return;
  } 

  // Use default LUT
  if(iusedefaultcmap == 1) {
    ierr = gen_filter(hvfile, cLutFilter);
    ilen = strlen(cLutFilter);
    sprintf(&cLutFilter[ilen-3], ".lut");
    if(strcmp(current_lut_filter,cLutFilter) != 0) {
      load_lut_list(cLutFilter);
      sprintf(current_lut_filter,"%s",cLutFilter);
    }
    do_default_lut();
  }

  // Generate keyframe lines.
  // Limit near clip to be > 0 for final rendering only.
  fOldNearClip = clip[0];
  for(i=0; i<3; i++) { ec[i] = eye[i] - cent[i]; }
  clipmin = 0.05f * fnorm(ec);
  if(clip[0] < clipmin) { clip[0] = clipmin; }
  gen_keyframe_lines();
  clip[0] = fOldNearClip;

  if(keysonly == 0) {
    // Send keyframe lines to hvr_cl & get image back
    for(i=0; i<nKeyLines; i++) { fprintf(fp_hvr_in, "%s", cKeyLines[i]); }
    fflush(fp_hvr_in);
    read_pix_from_hvr_to_display();
  }

  if(movieframe >= 0) {
    if(keysonly == 0) {
      sprintf(str, "xxx %s%04d", cMovieRoot, movieframe);
      printf("save image command line: %s\n", str);
      save_image(str);
    } else {
      sprintf(cKeyFile, "%s%04d.key", cMovieRoot, movieframe);
      printf("Keyfile=%s  hvfile=%s   eye=%f %f %f\n", cKeyFile, hvfile, eye[0],eye[1],eye[2]);
      fp = fopen(cKeyFile, "w");
      for(i=0; i<nKeyLines; i++) { fprintf(fp, "%s", cKeyLines[i]); }
      fclose(fp);
    }
    movieframe++;
  }
}

/* --------------------------------------------------------------
 *  New eye and up vectors based input.
 *  Center point remains unchainged .
 */
void do_ecandup(char *cnewline)
{
  char key[256];
  float ecsubfac, ecdir[3], updir[3], ec[3], ecscale, upscale, vecmag;
  int ngot, i;

  ngot = sscanf(cnewline, "%s %f %f %f %f %f %f", key,
         &ecdir[0], &ecdir[1], &ecdir[2], &updir[0], &updir[1], &updir[2]);

  // Scaling factor for ec
  vecmag = fnorm(ecdir);
  if(vecmag == 0.0f) {
    printf("do_ecandup: new ec direction is 0\n");
    printf("No change to view.\n");
    return;
  }
  for(i=0; i<3; i++) { ec[i] = eye[i] - cent[i]; }
  ecscale = fnorm(ec) / vecmag;

  // Orthogonalize new up for new ec, & get normalization factor
  ecsubfac = vecdot(updir, ecdir) / (vecmag*vecmag);
  for(i=0; i<3; i++) { updir[i] -= ecsubfac * ecdir[i]; }
  vecmag = fnorm(up);
  if(vecmag < 0.000001f) {
    printf("do_ecandup: new up dir is 0 or || to new ec direction.\n");
    printf("No change to view.\n");
    return;
  }
  upscale = 1.0 / vecmag;

  // New eye location and up vector
  for(i=0; i<3; i++) { eye[i] = ecscale * ecdir[i] + cent[i]; }
  for(i=0; i<3; i++) { up[i]  = upscale * updir[i]; }

  render_and_display();
}

/* --------------------------------------------------------------
 *  Generate new eye and up vectors based on
 *  tracball rotated, and scaled (in or out), eye-center vector
 *  Center point remains unchainged .
 */
void do_trackball(char *cnewline)
{
  char key[256];
  float hrot, vrot, rdelta, ecmagorig, upfac, ecsubfac, hfac, vfac, rfac;
  float hrot1, vrot1, rdelta1, rot2, rotinv, rotmag;
  float ec[3], dh[3], dv[3], ecmagnew;
  int ngot, i;

  ngot = sscanf(cnewline, "%s %f %f %f", key, &hrot1, &vrot1, &rdelta1);

  // Default & backward compatibility
  hrot   = hrot1;
  vrot   = vrot1;
  rdelta = rdelta1;

  // If |rot| > 1: interpret as number of increments to rotate 90 deg.
  rot2 = hrot1*hrot1 + vrot1*vrot1;
  if(rot2 > 1.0) {
    rotinv = 1.0 / sqrt(rot2);
    rotmag  = tan(halfpi * rotinv);
    hrot   = hrot1 * rotinv * rotmag;
    vrot   = vrot1 * rotinv * rotmag;
  }

  // If |delta| > 1.0: interpret as number of increments to zoom factor of 2
  if(rdelta1*rdelta1 > 1.0) { rdelta = flog2/rdelta1; }

  // Original eye-center vector
  for(i=0; i<3; i++) { ec[i] = eye[i] - cent[i]; }
  ecmagorig = fnorm(ec);

  // "Horizontal" and "Vertical" displacements
  cross(dh, up, ec);
  cross(dv, ec, dh);

  // Normalization factors for displancements
  hfac = hrot * ecmagorig / fnorm(dh);
  vfac = vrot * ecmagorig / fnorm(dv);

  // New eye-center vector
  for(i=0; i<3; i++) { ec[i] += hfac*dh[i] + vfac*dv[i]; }
  ecmagnew =  exp(rdelta) * ecmagorig;
  rfac = ecmagnew / fnorm(ec);
  for(i=0; i<3; i++) { ec[i] *= rfac; }

  // New eye location
  for(i=0; i<3; i++) { eye[i] = cent[i] + ec[i]; }

  // New normalized up vector, which is orthoganal to (eye-center)
  ecsubfac = vecdot(up, ec) / vecdot(ec, ec);
  for(i=0; i<3; i++) { up[i] -= ecsubfac * ec[i]; }
  upfac = 1.0 / fnorm(up);
  for(i=0; i<3; i++) { up[i] *= upfac; }

  //  Near and Far cliping planes are fixed relative to the center point.
  clip[1] += ecmagnew - ecmagorig;
  clip[0] += ecmagnew - ecmagorig;

  render_and_display();
}








/* --------------------------------------------------------------
 *  Set new near and far clipping planes based on distance from
 * center point
 */
void do_clip(char *cnewline)
{
  int   ngot, i;
  char  key[256];
  float centnear, centfar, ecmag, ec[3];

  ngot = sscanf(cnewline, "%s %f %f %f", key, &centnear, &centfar);

  // eye-center vector
  for(i=0; i<3; i++) { ec[i] = eye[i] - cent[i]; }
  ecmag = fnorm(ec);

  // New near and far clipping planes
  clip[0] = ecmag - centnear;
  clip[1] = ecmag + centfar ;

  render_and_display();
}

/* --------------------------------------------------------------
 *  Set object oriented "auxclip" based on center point and
 *  distance to eye point
 */
void do_auxclip_set(char *cnewline)
{
  int   ngot, i;
  char  key[256];
  float auxhalf, ecmag, ec[3];

  // eye-center vector
  for(i=0; i<3; i++) { ec[i] = eye[i] - cent[i]; }
  ecmag = fnorm(ec);
  auxhalf = 0.2 * ecmag;

  // Set aux clipping planes
  ifauxclip = 2;
  auxclip[0][0] = -1.0;  auxclip[0][1] =  0.0;  auxclip[0][2] =  0.0;  auxclip[0][3] = cent[0]-auxhalf;
  auxclip[1][0] =  1.0;  auxclip[1][1] =  0.0;  auxclip[1][2] =  0.0;  auxclip[1][3] = cent[0]+auxhalf;
  auxclip[2][0] =  0.0;  auxclip[2][1] = -1.0;  auxclip[2][2] =  0.0;  auxclip[2][3] = cent[1]-auxhalf;
  auxclip[3][0] =  0.0;  auxclip[3][1] =  1.0;  auxclip[3][2] =  0.0;  auxclip[3][3] = cent[1]+auxhalf;
  auxclip[4][0] =  0.0;  auxclip[4][1] =  0.0;  auxclip[4][2] = -1.0;  auxclip[4][3] = cent[2]-auxhalf;
  auxclip[5][0] =  0.0;  auxclip[5][1] =  0.0;  auxclip[5][2] =  1.0;  auxclip[5][3] = cent[2]+auxhalf;

  render_and_display();
}


/* --------------------------------------------------------------
 *  Turn off object oriented clipping "auxclip"
 */
void do_auxclip_off(char *cnewline)
{
  ifauxclip = 0;
  render_and_display();
}

/* --------------------------------------------------------------
 *  Set Set center to specified location and
 *  shift eye to maintian same eye-cent
 */
void do_setcent(char *cnewline)
{
  char key[256];
  float ec[3], centnew[3];
  int ngot, i;

  ngot = sscanf(cnewline,"%s %f %f %f",key,&centnew[0],&centnew[1],&centnew[2]);

  // Shift eye and center vectors by same amount
  for(i=0; i<3; i++) { ec[i]   = eye[i] - cent[i]; }
  for(i=0; i<3; i++) { cent[i] = centnew[i]; }
  for(i=0; i<3; i++) { eye[i]  = centnew[i] + ec[i]; }

  render_and_display();
}

/* --------------------------------------------------------------
 *  Generate new eye and center vectors based on translations
 *  Up vector remains unchainged .
 */
void do_shift(char *cnewline)
{
  char key[256];
  float hshift, vshift, rshift, shift, ecmagorig, hfac, vfac;
  float hshift1, vshift1, rshift1, shift2, shiftmag;
  float ec[3], dh[3], dv[3];
  int ngot, i;

  ngot = sscanf(cnewline, "%s %f %f %f", key, &hshift1, &vshift1, &rshift1);

  // Original eye-center vector
  for(i=0; i<3; i++) { ec[i] = eye[i] - cent[i]; }
  ecmagorig = fnorm(ec);

  // If size of shift > 1: inerpret as # of increments to shift by |eye-cent|
  shift2 = hshift1*hshift1+vshift1*vshift1+rshift1*rshift1;
  if(shift2 > 1.0) {
    hshift = 0.25 * hshift1 / shift2;
    vshift = 0.25 * vshift1 / shift2;
    rshift = 0.25 * rshift1 / shift2;
  } else {
    hshift = hshift1;
    vshift = vshift1;
    rshift = rshift1;
  }

  // "Horizontal" and "Vertical" displacements
  cross(dh, up, ec);
  cross(dv, ec, dh);

  // Normalization factors for displancements
  hfac = hshift * ecmagorig / fnorm(dh);
  vfac = vshift * ecmagorig / fnorm(dv);

  // Shifted eye and center vectors
  for(i=0; i<3; i++) {
     shift = hfac*dh[i] + vfac*dv[i] + rshift*ec[i];
     eye[i]  += shift;
     cent[i] += shift;
  }

  render_and_display();
}

void load_hv(char *cnewline)
{
  char key[256];
  int ngot;
  ngot = sscanf(cnewline, "%s %s", key, hvfile);
  render_and_display();
}

void load_vec(char *cnewline)
{
  char key[256];
  int ngot;
  ngot = sscanf(cnewline, "%s %s", key, vecfile);
  render_and_display();
}

void load_key(char *cnewline)
{
  char cKeyFile[256], key[256];
  int ngot;
  ngot = sscanf(cnewline, "%s %s", key, cKeyFile);
  if(read_keyframe_file(cKeyFile) < 0) { return; }
  render_and_display();
}

void save_key(char *cnewline)
{
  char cKeyFile[256], key[256];
  int ngot, i;
  FILE *fp;

  ngot = sscanf(cnewline, "%s %s", key, cKeyFile);

  if(strcmp(key, "savekey" ) ==0) { gen_keyframe_lines(); }
  if(strcmp(key, "saveluts") ==0) { nKeyLines = gen_luts_lines(0); }
  if(strcmp(key, "saveview") ==0) { nKeyLines = gen_view_lines(0); }
  if(strcmp(key, "savedata") ==0) { nKeyLines = gen_data_lines(0); }
  if(strcmp(key, "saveqlty") ==0) { nKeyLines = gen_qlty_lines(0); }

  fp = fopen(cKeyFile, "w");
  for(i=0; i<nKeyLines; i++) { fprintf(fp, "%s", cKeyLines[i]); }
  fclose(fp);
}

void save_image(char *cnewline)
{
  char cFile[256], key[256], cRootName[256], cCommand[256];
  int ngot, nbytes, nwrite, fd, iDot, i;
  FILE *fp;

  // Get root name of output file
  for(i=0; i<256; i++) { cRootName[i] == ' '; }
  ngot = sscanf(cnewline, "%s %s", key, cRootName);
  iDot = -1;
  for(i=0; i<256; i++) { if(cRootName[i] == '.') { iDot = i; }}
  if(iDot > 0) { cRootName[iDot] = '\0'; }

  // Write out raw image file
  sprintf(cFile, "%s.raw", cRootName);
  fd = open(cFile, O_WRONLY|O_CREAT, 0644 );
  if(fd < 0) {
    fprintf(stderr,"Save Image: Open failed on file -->%s<--\n", cFile);
    return;
  }
  nbytes = 3*iNx*iNy;
  nwrite = write(fd, ucBuf, nbytes);
  if(nwrite < nbytes) {
    fprintf(stderr,"Short write: cFile=%s  NtoWrite=%d  Nwritten=%d\n",
            cFile,nbytes,nwrite);
    return;
  }
  close(fd);

  // Convert raw image file into jpeg format
  sprintf(cCommand, "raw2jpg %d %d %s\n", iNx, iNy, cRootName);
  system(cCommand);

}

/* ---------------------------------------------------------------
 * Get root name for movie & initialize movie frame number
 */
void start_movie(char *cnewline)
{
  int ngot, iDot, i;
  char key[256];

  for(i=0; i<256; i++) { cMovieRoot[i] == ' '; }
  ngot = sscanf(cnewline, "%s %s", key, cMovieRoot);
  iDot = -1;
  for(i=0; i<256; i++) { if(cMovieRoot[i] == '.') { iDot = i; }}
  if(iDot > 0) { cMovieRoot[iDot] = '\0'; }
  if(strcmp(&cMovieRoot[iDot+1], "keys") == 0) { keysonly = 1; }

  movieframe = 0;

  // Debug
  if(keysonly == 0) printf("Root Name For Movie: %s\n", cMovieRoot);
  else              printf("Root Name For Keys : %s\n", cMovieRoot);
}

/* ---------------------------------------------------------------
 * Turn movie frames off
 */
void end_movie() { movieframe = -1; keysonly = 0; }

/* ---------------------------------------------------------------
 * Here begins the edit color & alpha map code
 */

int imin(int i, int j) { if(i<j) { return i; } else { return j; } }
int imax(int i, int j) { if(i>j) { return i; } else { return j; } }

void do_edit_cmap(char *cl)
{
  char key[256],c[256], filter[256], cLutFile[256], cLine[256];
  int ngot, i, indmin, indmax, ired, igrn, iblu, ierr, ilen;
  int i1, i2, i3;
  FILE *fp;

  ngot = sscanf(cl, "%s", key);

  if(strcmp(key,"cmap_show") ==0) { ishowcolorbar = 1 - ishowcolorbar; }
  if(strcmp(key,"cmap_edit") ==0) {
    sscanf(cl,"%s %d", c, &i);
    if(i == 0) { iSelALPknot = -1;   iSelRGBknot = -1;   ishowcolorbar = 0; }
    else       { iSelALPknot =  0;   iSelRGBknot =  0;   ishowcolorbar = 1; }
  }
  if(strcmp(key,"cmap_aknotpp") ==0) { iSelALPknot = imin(iSelALPknot+1, naknots-1); }
  if(strcmp(key,"cmap_aknotmm") ==0) { iSelALPknot = imax(iSelALPknot-1, 0); }
  if(strcmp(key,"cmap_cknotpp") ==0) { iSelRGBknot = imin(iSelRGBknot+1, ncknots-1); }
  if(strcmp(key,"cmap_cknotmm") ==0) { iSelRGBknot = imax(iSelRGBknot-1, 0); }

  if(strcmp(key,"cmap_move_a") ==0) { sscanf(cl, "%s %d", c, &i);
    if(iSelALPknot == 0  ||  iSelALPknot == naknots-1) { return; }
    indmin = iaknot[iSelALPknot-1]+1;
    indmax = iaknot[iSelALPknot+1]-1;
    iaknot[iSelALPknot] = imax(indmin, imin(indmax, iaknot[iSelALPknot]+i));
  }
  if(strcmp(key,"cmap_move_c") ==0) { sscanf(cl, "%s %d", c, &i);
    if(iSelRGBknot == 0  ||  iSelRGBknot == ncknots-1) { return; }
    indmin = icknot[iSelRGBknot-1]+1;
    indmax = icknot[iSelRGBknot+1]-1;
    icknot[iSelRGBknot] = imax(indmin, imin(indmax, icknot[iSelRGBknot]+i));
  }

  if(strcmp(key,"cmap_set_a") ==0) { sscanf(cl, "%s %f", c, &aknot[iSelALPknot]); }
  if(strcmp(key,"cmap_set_c") ==0) { sscanf(cl, "%s %f %f %f", c, &rknot[iSelRGBknot],
                                                                  &gknot[iSelRGBknot],
                                                                  &bknot[iSelRGBknot]); }
  if(strcmp(key,"cmap_set_cc") ==0) {
    sscanf(cl, "cmap_set_cc #%02x%02x%02x", &ired, &igrn, &iblu);
    rknot[iSelRGBknot] = (float)ired / 255.0f;
    gknot[iSelRGBknot] = (float)igrn / 255.0f;
    bknot[iSelRGBknot] = (float)iblu / 255.0f;
  }

  if(strcmp(key,"cmap_cknotadd") ==0) {
    i1 = iSelRGBknot;
    if(i1 == ncknots-1) { i1--; }
    i2 = i1 + 1;
    if(icknot[i2]-icknot[i1] < 2) { return; }
    ncknots++;
    for(i=ncknots-1; i>i2; i--) { icknot[i] = icknot[i-1];
                                   rknot[i] =  rknot[i-1];
                                   gknot[i] =  gknot[i-1];
                                   bknot[i] =  bknot[i-1]; }
     i3 = i2 + 1;
     icknot[i2] = (icknot[i1] + icknot[i3]) / 2;
      rknot[i2] = ( rknot[i1] +  rknot[i3]) / 2.0f;
      gknot[i2] = ( gknot[i1] +  gknot[i3]) / 2.0f;
      bknot[i2] = ( bknot[i1] +  bknot[i3]) / 2.0f;
      iSelRGBknot = i2;
  }
  if(strcmp(key,"cmap_cknotdel") ==0) {
    if(iSelRGBknot == 0  ||  iSelRGBknot == ncknots-1) { return; }
    for(i=iSelRGBknot; i<ncknots; i++) { icknot[i] = icknot[i+1];
                                          rknot[i] =  rknot[i+1];
                                          gknot[i] =  gknot[i+1];
                                          bknot[i] =  bknot[i+1]; }
    ncknots--;
  }
  if(strcmp(key,"cmap_aknotadd") ==0) {
    i1 = iSelALPknot;
    if(i1 == naknots-1) { i1--; }
    i2 = i1 + 1;
    if(iaknot[i2]-iaknot[i1] < 2) { return; }
    naknots++;
    for(i=naknots-1; i>i2; i--) { iaknot[i] = iaknot[i-1];
                                   aknot[i] =  aknot[i-1]; }
     i3 = i2 + 1;
     iaknot[i2] = (iaknot[i1] + iaknot[i3]) / 2;
      aknot[i2] = ( aknot[i1] +  aknot[i3]) / 2.0f;
      iSelALPknot = i2;
  }
  if(strcmp(key,"cmap_aknotdel") ==0) {
    if(iSelALPknot == 0  ||  iSelALPknot == naknots-1) { return; }
    for(i=iSelALPknot; i<naknots; i++) { iaknot[i] = iaknot[i+1];  aknot[i] = aknot[i+1]; }
    naknots--;
  }

  // Default LUTs
/*
  void save_key(char *cnewline)
  char cKeyFile[256], key[256];
  ngot = sscanf(cnewline, "%s %s", key, cKeyFile);
  if(strcmp(key, "saveluts") ==0) { nKeyLines = gen_luts_lines(0); }
*/

  if(strcmp(key,"cmap_f_default") == 0 || strcmp(key,"cmap_tf_default") == 0) {
    if(strcmp(key,"cmap_f_default") == 0) {
      ierr = gen_filter(hvfile, cLutFile);
      ilen = strlen(cLutFile);
      for(i=0; i<ilen-3; i++) { if(cLutFile[i] == '?') { cLutFile[i] = '0'; }}
    } else {
      sprintf(cLutFile, "%s", hvfile);
      ilen = strlen(cLutFile);
    }
    sprintf(&cLutFile[ilen-3], ".lut");
    sprintf(cLine, "saveluts   %s\n", cLutFile);
    save_key(cLine);
    printf("Saved current LUT in file: %s\n", cLutFile);

    // set LUT file filter to empty, which will force a reload the next time default LUTs are required.
    sprintf(current_lut_filter, "");
    ierr = gen_file_list(current_lut_filter);
    return;  // No need to re-render
  }

  if(strcmp(key,"cmap_usedefault") == 0) {
    sscanf(cl, "%s %d", c, &iusedefaultcmap);
    if(iusedefaultcmap != 1) {
      // set LUT file filter to empty, which will force a reload the next time default LUTs are required.
      sprintf(current_lut_filter, "");
      ierr = gen_file_list(current_lut_filter);
      return;  // No need to re-render
    }
  }
/*
cmap_usedefault    0 (default)
cmap_usedefault    1
cmap_f_default
cmap_tf_default
*/

//    int   naknots, ncknots, iaknot[256], icknot[256];
//    float aknot[256], rknot[256], gknot[256], bknot[256];

   render_and_display();
}

#define MAXLUTS 100
int   nLUTs, index_lut_list[MAXLUTS];
int   naknots_list[MAXLUTS],     ncknots_list[MAXLUTS];
int    iaknot_list[MAXLUTS][256], icknot_list[MAXLUTS][256];
float   aknot_list[MAXLUTS][256],  rknot_list[MAXLUTS][256];
float   gknot_list[MAXLUTS][256],  bknot_list[MAXLUTS][256];

void load_lut_list(char *cLutFilter)
{
  int i, iFile, ierr, knot, index, iread;
  char cLutFile[256];

  nLUTs = gen_file_list(cLutFilter);
  printf("NLutFiles = %d\n", nLUTs);
  i = 0;
  for(iFile=0; iFile<nLUTs; iFile++) {
    ierr = get_ith_file(iFile, cLutFile);
    printf("LUT File: %s", cLutFile);
    index = get_file_index(cLutFile);
    iread = read_keyframe_file(cLutFile);
    if(index < 0) printf(" -- bad index\n");
    if(index >= 0 && iread == 0) {
      index_lut_list[i] = index;
      naknots_list[i] = naknots;
      ncknots_list[i] = ncknots;
      for(knot=0; knot<naknots; knot++) {
        iaknot_list[i][knot] = iaknot[knot];
         aknot_list[i][knot] =  aknot[knot];
      }
      for(knot=0; knot<ncknots; knot++) {
        icknot_list[i][knot] = icknot[knot];
         rknot_list[i][knot] =  rknot[knot];
         gknot_list[i][knot] =  gknot[knot];
         bknot_list[i][knot] =  bknot[knot];
      }
      printf("    index = %4d\n", index);
      i++;
    }
  }
  nLUTs = i;
}

void do_default_lut()
{
  int index, i, k, i0, i1, mindist, mydist, index0, index1, indexdel;
  float fac0, fac1;
  if(nLUTs < 1) return;

  // Get index (=sequence number) of current hv file
  index = get_file_index(hvfile);
  // printf("do_default_lut: file=%s  index=%d\n", hvfile, index);

  // Find closest index of LUT in the list of default LUTs
  i0 = -1;
  mindist   = 100000;
  for(i=0; i<nLUTs; i++) { 
    mydist = index - index_lut_list[i];
    if(mydist < 0) { mydist = -mydist; }
    if(mydist < mindist) { i0 = i; mindist = mydist; }
  }

  // Get index of LUT on other side
  i1 = i0;
  if(index > index_lut_list[i0] && i0 < nLUTs-1) i1++;
  if(index < index_lut_list[i0] && i0 >       0) i1--;
  // printf("do_default_lut: i0=%d  i1=%d  nLUTs=%d\n", i0, i1, nLUTs);

  // If number of knots differ, just take closest LUT
  if(ncknots_list[i0] != ncknots_list[i1] ||
     naknots_list[i0] != naknots_list[i1]   ) i1 = i0;
    
  // Interpolate knot location and values between the two bounding LUTs
  fac0 = 1.0;
  fac1 = 0.0;
  index0 = index_lut_list[i0];
  index1 = index_lut_list[i1];
  indexdel = index1 - index0;
  if(indexdel != 0) {
    fac0 = (float)(index1 - index) / (float)indexdel;
    if(fac0 < 0.0) fac0 = -fac0;
    fac1 = 1.0 - fac0;
  }
  // printf("do_default_lut: fac0=%f  fac1=%f\n", fac0, fac1);

  naknots = naknots_list[i0];
  ncknots = ncknots_list[i0];
  for(k=0; k<naknots; k++) {
    iaknot[k] = (int)(fac0*(float)iaknot_list[i0][k]+fac1*(float)iaknot_list[i1][k]);
     aknot[k] =       fac0*        aknot_list[i0][k]+fac1*        aknot_list[i1][k] ;
  }
  for(k=0; k<ncknots; k++) {
    icknot[k] = (int)(fac0*(float)icknot_list[i0][k]+fac1*(float)icknot_list[i1][k]);
     rknot[k] =       fac0*        rknot_list[i0][k]+fac1*        rknot_list[i1][k] ;
     gknot[k] =       fac0*        gknot_list[i0][k]+fac1*        gknot_list[i1][k] ;
     bknot[k] =       fac0*        bknot_list[i0][k]+fac1*        bknot_list[i1][k] ;
  }
/*
  printf("index: ");
  for(k=0; k<naknots; k++) printf("%5d", iaknot[k]);
  printf("\n");
  printf("alpha: ");
  for(k=0; k<naknots; k++) printf("%5.2f", aknot[k]);
  printf("\n");
*/

  return;
}

/* ---------------------------------------------------------------
 * Here begins the getvec code
 */

/* 
 * Specify:
 *   Edit fields: (n1,n2,n3)
 *   % of vertical view spanned by sample in max. n direction
 *   Radio buttons: (1,2,3) along (x,y,z) or
 *                  (horizontal, up, eye-center), or
 *   Edit fields: timetot, maxout
 *
 *   Buttons: 1+, 1-, 2+, 2-, 3+, 3-
 *   Button:  save from this dialog
 *   Button:  load to this dialog
 */


/*
flddim     1 1 1          ns1,ns2,ns3
fldmax     1000           maxout
fldprcnt   50             fldsampprcnt
fldtim     500            timetot
flddir     EYE            cFldSampDir
fldfile    Bxj00a0123aaa  cFldFile
fldthresh  0.0            vlow
*/
int   ns1=0, ns2=0, ns3=0, maxout=0;
float fldsampprcnt, xsc, ysc, zsc;
float timetot;
char  cFldSampDir[32];
char  cFldFile[256];
float vlow=0.0;
char  cFldColor[256];
int   iVecNameLoop=0, nVecNameLoop=3;
int   iSetSampCent=1;
float dx1=1.0f, dx2=0.0f, dx3=0.0f;
float dy1=0.0f, dy2=1.0f, dy3=0.0f;
float dz1=0.0f, dz2=0.0f, dz3=1.0f;

void do_field_lines(char *cl)
{
  int doFieldLines();
  char key[256],c[256];
  int ngot, i;
  FILE *fp;
  float fldshiftx, fldshifty, fldshiftz;
  float ec[3], ecmag, pi, del;

  ngot = sscanf(cl, "%s", key);

  if(strcmp(key,"flddim"    ) ==0) { sscanf(cl,"%s %d %d %d",c,&ns1,&ns2,&ns3); }
  if(strcmp(key,"fldrad"    ) ==0) { sscanf(cl,"%s %f", c, &vecradius); }
  if(strcmp(key,"fldmax"    ) ==0) { sscanf(cl,"%s %d", c, &maxout); }
  if(strcmp(key,"fldprcnt"  ) ==0) { sscanf(cl,"%s %f", c, &fldsampprcnt); }
  if(strcmp(key,"fldtim"    ) ==0) { sscanf(cl,"%s %f", c, &timetot); }
  if(strcmp(key,"flddir"    ) ==0) { sscanf(cl,"%s %s", c, cFldSampDir); }
  if(strcmp(key,"fldsetcent") ==0) { sscanf(cl,"%s %d", c, &iSetSampCent); }
  if(strcmp(key,"fldfile"   ) ==0) { sscanf(cl,"%s %s", c, cFldFile); }
  if(strcmp(key,"fldthresh" ) ==0) { sscanf(cl,"%s %f", c, &vlow); }
  if(strcmp(key,"fldcolor"  ) ==0) { sscanf(cl,"%s %s", c, cFldColor); }
  if(strcmp(key,"fldlohiclr") ==0) { sscanf(cl,"%s %s %s", c, cFldClrLo, cFldClrHi); }
  if(strcmp(key,"fldgen"  ) ==0) {
    if(iSetSampCent == 1) {
      xsc = cent[0];
      ysc = cent[1];
      zsc = cent[2];
    }
    if(doFieldLines() < 0) { return; }
    render_and_display();
  }
  if(strcmp(key,"fldshift") ==0) {
    if(ns1 == 0) {
      printf("Generate Field Lines Before Shifting Them\n");
      return;
    }
    sscanf(cl,"%s %f %f %f",c,&fldshiftx,&fldshifty,&fldshiftz);
    for(i=0; i<3; i++) { ec[i] = eye[i] - cent[i]; }
    ecmag = fnorm(ec);
    pi = acos(-1.0);
    del = 2.0 * (ecmag/100.0f) * sin(fov * pi / 360.0);
    xsc += del * fldshiftx;
    ysc += del * fldshifty;
    zsc += del * fldshiftz;
    if(doFieldLines() < 0) { return; }
    render_and_display();
  }
}


#define MAXVECLEGTH   10000
float xvec[MAXVECLEGTH];
float yvec[MAXVECLEGTH];
float zvec[MAXVECLEGTH];
float vvec[MAXVECLEGTH];

float dxinv;
float xmin0, xmax0, ymin0, ymax0, zmin0, zmax0;
int nx0, ny0, nz0, nxb, nyb, nzb;

float *vvx, *vvy, *vvz, *vvclr;

int doFieldLines()
{
  static int nbyteslast=0;
  static char cFldFileLast[256]=" ";
  static char cFldColorLast[256]=" ";
  int i,j, nbytes, iDir, iXYZ, iEnd, length;
  int   nvecs, mvecs, nsmax;
  char cOutFile[256], cTmp[256], cFormatFile[256];
  float x0, y0, z0, xsmin, ysmin, zsmin, dt, ds;
  float xs, ys, zs, shift, dir;
  float dx, xs0, ys0, zs0;
  int jmid, jmin, jmax, idir;
  int i1,i2,i3, istop;
  float dsout,vvmin,tsum,vx,vy,vz,vv, snext, ssum;
  float fSampSpan, del, ecmag, pi, ec[3], dh[3], dv[3];
  float ecfac, dhfac, dvfac;

  int readfieldformat(char *cFormatFile);
  int readfield(char *cFile, int nbytes, float *fbuf);
  float vinterp(float xx, float yy, float zz, float *vv);
  float value, delx, dely, delz, distscale;
  FILE *fp;


 /*--------------------------------------------------------
  * Parse Input file or field name
  */
  iDir  = -1;
  iEnd = 0;
  for(i=  0; i<256; i++) { if(cFldFile[i] ==  '/') { iDir = i; } }
  for(i=255; i>= 0; i--) { if(cFldFile[i] == '\0') { iEnd = i;   } }
  iXYZ = iDir + 2;
  for(i=0; i<iEnd; i++) { if(cFldFile[i] == 'x') { iXYZ = i;   } }
  for(i=0; i<iEnd; i++) { if(cFldFile[i] == '?') { iXYZ = i;   } }
  if(strcmp(&cFldFile[iEnd-3], "aaa") == 0) { iEnd = iEnd - 3; }
  sprintf(cTmp, "%s", cFldFile);
  if(iDir < 0) { cTmp[     0] = '\0'; }
  else         { cTmp[iDir+1] = '\0'; }
  sprintf(cFormatFile, "%sfieldformat", cTmp);

 /*--------------------------------------------------------
  * Generate field files from adump if requested
  */
  if(strncmp(cFldFile,"calc",4) == 0 &&
     (difsubvol() > 0 || strcmp(cFldFile,cFldFileLast) != 0)) {
    cFldFile[iXYZ] = 'x';   if(calcfield(cFldFile) < 0) { return -1; }
    cFldFile[iXYZ] = 'y';   if(calcfield(cFldFile) < 0) { return -1; }
    cFldFile[iXYZ] = 'z';   if(calcfield(cFldFile) < 0) { return -1; }

    sprintf(cFldFileLast, " ");
  }
  if(strncmp(cFldFile,"calc",4) == 0) {
    sprintf(cFormatFile, "subvol_format");
  }

 /*--------------------------------------------------------
  * Generate output file name
  */
  sprintf(cTmp, "%s", cFldFile);
  cTmp[iXYZ] = '_';
  cTmp[iEnd] = '\0';
  sprintf(cOutFile, "%s%c.vec", cTmp, (char)((int)'A'+iVecNameLoop));
  iVecNameLoop = (iVecNameLoop+1)%nVecNameLoop;
  // printf("Output vector file: -->%s<--\n", cOutFile);


 /*--------------------------------------------------------
  * Read dimensions of problem and format of input file(s)
  */
  if(readfieldformat(cFormatFile) < 0) { return -1; }


 /*--------------------------------------------------------
  * Allocate space & read vector and scalar (color) fields as needed
  */
  nbytes = 4 * nx0 * ny0 * nz0;
  if(nbytes > nbyteslast) {
    if(nbyteslast > 0) {
      free(vvx);
      free(vvy);
      free(vvz);
      free(vvclr);
    }
    vvx = (float *)malloc(nbytes);
    vvy = (float *)malloc(nbytes);
    vvz = (float *)malloc(nbytes);
    vvclr = (float *)malloc(nbytes);
    nbyteslast = nbytes;
  }

  if(strcmp(cFldFile,cFldFileLast) != 0) {
    cFldFile[iXYZ] = 'x';   if(readfield(cFldFile, nbytes, vvx) < 0) { return -1; }
    cFldFile[iXYZ] = 'y';   if(readfield(cFldFile, nbytes, vvy) < 0) { return -1; }
    cFldFile[iXYZ] = 'z';   if(readfield(cFldFile, nbytes, vvz) < 0) { return -1; }
    cFldFile[iXYZ] = 'x';
    sprintf(cFldFileLast, "%s", cFldFile);
  }
  if(strcmp(cFldColor,cFldColorLast) != 0 && cFldColor[0] == '/') {
    if(readfield(cFldColor, nbytes, vvclr) < 0) { return -1; }
    sprintf(cFldColorLast, "%s", cFldColor);
  }



 /*--------------------------------------------------------
  * Generate vector objects
  * 1) generate list of starting points
  *    Option 1) uniform sampling in subvolume
  *    Option 2) preferentially from points of large (vx,vy,vz)
  *    will start with option 1
  *    - specify: (xc,yc,zc), ns_i, dx_i,dy_i,dz_i   where _i={1,2,3}
  * 2) generate point strip from each starting point (x,y,z)
  *    - specify:
  *       timetot      sum of all dtime for each strip
  *       maxout       maximum number of output points
  *       dsout        distance between output points   
  */

  // Generate magnitude of displacement (del) between starting points
  // Based on field of view and distance from eye to center point
  for(i=0; i<3; i++) { ec[i] = eye[i] - cent[i]; }
  ecmag = fnorm(ec);
  pi = acos(-1.0);
  fSampSpan = 2.0 * ecmag * (fldsampprcnt/100.0f) * sin(fov * pi / 360.0);
  nsmax = ns1;
  if(nsmax < ns2) { nsmax = ns2; }
  if(nsmax < ns3) { nsmax = ns3; }
  del = 0.0f;
  if(nsmax > 1) { del = fSampSpan / (float)(nsmax-1); }

  if(strcmp(cFldSampDir,"XYZ") == 0) {
    dx1 = del;   dy1 = 0.0;   dz1 = 0.0;
    dx2 = 0.0;   dy2 = del;   dz2 = 0.0;
    dx3 = 0.0;   dy3 = 0.0;   dz3 = del;
  }
  if(strcmp(cFldSampDir,"EYE") == 0) {
    // "Horizontal" and "Vertical" displacements
    cross(dh, up, ec);
    cross(dv, ec, dh);

    // Normalization factors for displancements
    dhfac = del / fnorm(dh);
    dvfac = del / fnorm(dv);
    ecfac = del / ecmag;

    dx1 = dhfac * dh[0];   dy1 = dhfac * dh[1];   dz1 = dhfac * dh[2];
    dx2 = dvfac * dv[0];   dy2 = dvfac * dv[1];   dz2 = dvfac * dv[2];
    dx3 = ecfac * ec[0];   dy3 = ecfac * ec[1];   dz3 = ecfac * ec[2];
  }



  dx     = (xmax0 - xmin0) / (float)nx0;
  ds     = 0.2 * dx;   // will be dt*|V|
  dsout  = 1.0 * dx;
  // shift = 0.5 * dx;
  shift = 0.0;   // fields are no longer face centered
  vvmin = 0.01 * ds / timetot;
  dxinv = 1.0 / dx;

  xs0 = xsc - 0.5*(dx1*(float)(ns1-1)+dx2*(float)(ns2-1)+dx3*(float)(ns3-1));
  ys0 = ysc - 0.5*(dy1*(float)(ns1-1)+dy2*(float)(ns2-1)+dy3*(float)(ns3-1));
  zs0 = zsc - 0.5*(dz1*(float)(ns1-1)+dz2*(float)(ns2-1)+dz3*(float)(ns3-1));

  jmid = MAXVECLEGTH / 2;
  nvecs = ns1 * ns2 * ns3;
  if(nvecs > 0) {
    mvecs = 0;
    for(i3=0; i3<ns3; i3++) {
    for(i2=0; i2<ns2; i2++) {
    for(i1=0; i1<ns1; i1++) {
      xs = xs0 + dx1*(float)i1 + dx2*(float)i2 + dx3*(float)i3;
      ys = ys0 + dy1*(float)i1 + dy2*(float)i2 + dy3*(float)i3;
      zs = zs0 + dz1*(float)i1 + dz2*(float)i2 + dz3*(float)i3;
      vx = vinterp(xs      , ys-shift, zs-shift, vvx);
      vy = vinterp(xs-shift, ys      , zs-shift, vvy);
      vz = vinterp(xs-shift, ys-shift, zs      , vvz);
      vv = sqrt(vx*vx + vy*vy + vz*vz);
      if(vv >= vlow) { mvecs++; }
    }}}
  }

  nvecs = mvecs;
  if(nvecs > 0) {
    sprintf(vecfile, "%s", cOutFile);
    fp = fopen(cOutFile, "w");
    fprintf(fp, "#ntubes         %d\n", nvecs);

    for(i3=0; i3<ns3; i3++) {
    for(i2=0; i2<ns2; i2++) {
    for(i1=0; i1<ns1; i1++) {
    xs = xs0 + dx1*(float)i1 + dx2*(float)i2 + dx3*(float)i3;
    ys = ys0 + dy1*(float)i1 + dy2*(float)i2 + dy3*(float)i3;
    zs = zs0 + dz1*(float)i1 + dz2*(float)i2 + dz3*(float)i3;
    vx = vinterp(xs      , ys-shift, zs-shift, vvx);
    vy = vinterp(xs-shift, ys      , zs-shift, vvy);
    vz = vinterp(xs-shift, ys-shift, zs      , vvz);
    vv = sqrt(vx*vx + vy*vy + vz*vz);
    if(vv >= vlow) {

      for(idir = -1; idir < 2; idir += 2) {

        dir = (float)idir;

        // Initialize the point
        tsum = 0.0;
        ssum = 0.0;
        if(idir < 0) { snext =     0;   j = jmid-1; }
        else         { snext = dsout;   j = jmid; }

        xs = xs0 + dx1*(float)i1 + dx2*(float)i2 + dx3*(float)i3;
        ys = ys0 + dy1*(float)i1 + dy2*(float)i2 + dy3*(float)i3;
        zs = zs0 + dz1*(float)i1 + dz2*(float)i2 + dz3*(float)i3;

        while(tsum < timetot) {

          // Interpolate (face centered) fields for the point
          // Zone centered: shift=0.0
          // Face centered: shift=0.5*dx
          vx = dir * vinterp(xs      , ys-shift, zs-shift, vvx);
          vy = dir * vinterp(xs-shift, ys      , zs-shift, vvy);
          vz = dir * vinterp(xs-shift, ys-shift, zs      , vvz);
          vv = sqrt(vx*vx + vy*vy + vz*vz);
          value = vv;
          if(cFldColor[0] == '/') {
            // Interpolate zone centered value for color field
            value = vinterp(xs, ys, zs, vvclr);
          }

          // Report the point
          if(ssum >= snext) {
            xvec[j] = xs;
            yvec[j] = ys;
            zvec[j] = zs;
            vvec[j] = value;
            snext += dsout;
            j = j + idir;
          }

          // Stop if the point goes outside the volume or vv too small
          // Otherwise, itterate the point.
          istop = 0;
          if(vv < vvmin) { istop = 1; }
          if(xs < xmin0  ||  xs > xmax0) { istop = 1; }
          if(ys < ymin0  ||  ys > ymax0) { istop = 1; }
          if(zs < zmin0  ||  zs > zmax0) { istop = 1; }

          if(istop == 0) {
            dt = ds / vv;
            xs += vx*dt;
            ys += vy*dt;
            zs += vz*dt;
            tsum += dt;
            ssum += ds;
          } else {
            tsum = timetot + 1.0;
          }
        }
        if(ssum > snext-dsout) {
          xvec[j] = xs;
          yvec[j] = ys;
          zvec[j] = zs;
          vvec[j] = value;
          j = j + idir;
        }
        if(idir < 0) { jmin = j+1; }
        else         { jmax = j; }
      }
      length = jmax - jmin;

      if(strncmp(cFldColor,"Direction", 9) == 0) {
        distscale = 20.0 / (xmax0 - xmin0);
        ssum = 0.0;
        vvec[jmin] = ssum;
        for(j=jmin+1; j<jmax; j++) {
          delx = xvec[j] - xvec[j-1];
          dely = yvec[j] - yvec[j-1];
          delz = zvec[j] - zvec[j-1];
          ssum += distscale * sqrt(delx*delx + dely*dely + delz*delz);
          if(ssum > 1.0) { ssum = 0.0; }
          vvec[j] = ssum;
        }
      }

      // Output strip of points
      fprintf(fp, "#tubelength     %d\n", length);
      for(j=jmin; j<jmax; j++) {
        fprintf(fp, "%f  %f  %f  %f\n", xvec[j], yvec[j], zvec[j], vvec[j]);
      }
    }   // end of (vv >= vlow)
    }}}
    fclose(fp);
  }

  return 0;
}

int calcfield(char *cCalcFld)
{
  char cRoot[256], cFld[32];
  FILE *fp;

  sprintf(cFld, "%s", &cCalcFld[4]);
  sprintf(cRoot, "%s", hvfile);
  cRoot[10] = '\0';

  if(ifauxclip == 2) {
    // 012345678901234
    // esp03-0123-000
    fp = fopen("get_bof.in", "w");
    fprintf(fp, "%s-000\n", cRoot);
    fprintf(fp, "%s\n", cCalcFld);
    fprintf(fp, "subvol_format\n");
    fprintf(fp, "%s\n", cFld);
    fprintf(fp, "%f %f\n", auxclip[0][3], auxclip[1][3]);
    fprintf(fp, "%f %f\n", auxclip[2][3], auxclip[3][3]);
    fprintf(fp, "%f %f\n", auxclip[4][3], auxclip[5][3]);
    fprintf(fp, "  1\n");
    fclose(fp);
  }
printf("calcfield: %s\n", cCalcFld);
  system("get_bof");
printf("done\n");

/*
eps03-0131-000
calcBx
subvol_format
Bx
  -1.545497   1.545497
  -2.897807   0.193187
  -1.545497   1.545497
  1
*/

  return 0;
}

int difsubvol()
{
  static int   if_last_auxclip=999;
  static float last_auxclip[6][4];
  int           i, j, isdif;

  isdif = 0;
  if(ifauxclip != if_last_auxclip) { isdif = 1; }

  if(isdif == 0  && ifauxclip != 0) { for(i=0; i<6; i++) { for(j=0; j<4; j++) {
    if(last_auxclip[i][j] != auxclip[i][j]) { isdif = 1; }
  }}}

  if(isdif == 1) {
    if_last_auxclip = ifauxclip;
    for(i=0; i<6; i++) { for(j=0; j<4; j++) { last_auxclip[i][j] = auxclip[i][j]; }}
  }

  return isdif;
}

int difview()
{
  static char  cLastView[100][80];
  static int  nLastView=0;
  int         nNewView, i, isdif;

  isdif = 0;
  nNewView = gen_view_lines(0);
  if(nNewView != nLastView) { isdif = 1; }
  if(isdif == 0) {
    for(i=0; i<nNewView; i++) {
      if(strcmp(cLastView[i],cKeyLines[i]) != 0) { isdif = 1; }
    }
  }
  if(isdif == 1) {
    for(i=0; i<nNewView; i++) { sprintf(cLastView[i], cKeyLines[i]); }
    nLastView = nNewView;
  }

  return isdif;
}

int readfieldformat(char *cFormatFile)
{
  FILE *fp;
  char cline[256], ckey[256];

  fp = fopen(cFormatFile, "r");
  if(fp == NULL) {
    printf("Could not open file: %s\n", cFormatFile);
    return -1;
  }
  while(fgets(cline, 256, fp) != NULL) {
    sscanf(cline, "%s", ckey);
    if(strcmp(ckey,"xcoord")==0){
      sscanf(cline,"%s %f %f %d", ckey, &xmin0, &xmax0, &nx0);
    }
    if(strcmp(ckey,"ycoord")==0){
      sscanf(cline,"%s %f %f %d", ckey, &ymin0, &ymax0, &ny0);
    }
    if(strcmp(ckey,"zcoord")==0){
      sscanf(cline,"%s %f %f %d", ckey, &zmin0, &zmax0, &nz0);
    }
    if(strcmp(ckey,"blockdim")==0){
      sscanf(cline,"%s %d %d %d", ckey, &nxb, &nyb, &nzb);
    }
  }
  fclose(fp);

  return 0;

/*
  printf("XYZmin:  %f %f %f\n", xmin0, ymin0, zmin0);
  printf("XYZmax:  %f %f %f\n", xmax0, ymax0, zmax0);
  printf("X0b: %d %d\n", nx0, nxb);
  printf("Y0b: %d %d\n", ny0, nyb);
  printf("Z0b: %d %d\n", nz0, nzb);
*/
}

int readfield(char *cFile, int nbytes, float *fbuf)
{
  int fd, ngot;

  // printf("INPUT -->%s<--\n", cFile);

  fd = open(cFile, O_RDONLY);
  if(fd < 0) {
    fprintf(stderr,"readfield: Open failed on file -->%s<--\n", cFile);
    return -1;
  }
  ngot = read(fd, fbuf, nbytes);
  if(ngot < nbytes) {
    fprintf(stderr,"Short read: cFile=%s  iNbytes=%d  ngot=%d\n",
            cFile,nbytes,ngot);
    return -1;
  }
  close(fd);

  return 0;
}

float vinterp(float xx, float yy, float zz, float *vv)
{
  float xind, yind, zind;
  int ix0, ix1, iy0, iy1, iz0, iz1;
  float fracx1,fracy1,fracz1,fracx0,fracy0,fracz0, value;
  int k000,k100,k010,k110,k001,k101,k011,k111;

  xind = dxinv * (xx - xmin0);
  yind = dxinv * (yy - ymin0);
  zind = dxinv * (zz - zmin0);

  ix0 = (int)xind;  if(ix0 < 0) { ix0 = 0; }   if(ix0 >= nx0) { ix0 = nx0-1; }
  iy0 = (int)yind;  if(iy0 < 0) { iy0 = 0; }   if(iy0 >= ny0) { iy0 = ny0-1; }
  iz0 = (int)zind;  if(iz0 < 0) { iz0 = 0; }   if(iz0 >= nz0) { iz0 = nz0-1; }

  ix1 = ix0+1;  if(ix1 >= nx0) { ix1 = ix0; }
  iy1 = iy0+1;  if(iy1 >= ny0) { iy1 = iy0; }
  iz1 = iz0+1;  if(iz1 >= nz0) { iz1 = iz0; }

  fracx1 = xind - (float)ix0;
  fracy1 = yind - (float)iy0;
  fracz1 = zind - (float)iz0;
  fracx0 = 1.0 - fracx1;
  fracy0 = 1.0 - fracy1;
  fracz0 = 1.0 - fracz1;

  k000 = ix0 + nx0*(iy0 + ny0*iz0);
  k100 = ix1 + nx0*(iy0 + ny0*iz0);
  k010 = ix0 + nx0*(iy1 + ny0*iz0);
  k110 = ix1 + nx0*(iy1 + ny0*iz0);
  k001 = ix0 + nx0*(iy0 + ny0*iz1);
  k101 = ix1 + nx0*(iy0 + ny0*iz1);
  k011 = ix0 + nx0*(iy1 + ny0*iz1);
  k111 = ix1 + nx0*(iy1 + ny0*iz1);

  value = fracx0*fracy0*fracz0*vv[k000]
        + fracx1*fracy0*fracz0*vv[k100]
        + fracx0*fracy1*fracz0*vv[k010]
        + fracx1*fracy1*fracz0*vv[k110]
        + fracx0*fracy0*fracz1*vv[k001]
        + fracx1*fracy0*fracz1*vv[k101]
        + fracx0*fracy1*fracz1*vv[k011]
        + fracx1*fracy1*fracz1*vv[k111] ;

  return value;
}

