/* Latest Working Version of HVR
   Last Updated:  November 25, 2006
   Works for both Windows (Glut) and Linux (GLX)
   Supports 2D and 3D Textures, both RGBA and Paletted
   Uses a Cg Fragment Shader for Paletted Texture Implementation
   Supports Texture Caching for Paletted 3D Textures
   Toggles Texture Caching On/Off as necessary automatically on the fly
   Chooses Texture Caching Scheme automatically on the fly
   Supports Offscreen Rendering using FBO's and Renderbuffers
   All texture options can be changed on the fly */

// hvr_work.cpp : Implementation of Chvr_work
////////////////////////////////////////////////////////////////////////////////
//
// Author:   David H. Porter
//           Laboratory for Computational Science & Engineering
//           University of Minnesota
//
////////////////////////////////////////////////////////////////////////////////
/***************************************************************************************
Program:  HVR_SERVER  --  Hierarchical Volume Rendering Component Object Moduel
Copyright (C) 2002  David H. Porter

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
***************************************************************************************/

#include "StdAfx.h"

#ifdef _DEBUG
#undef THIS_FILE
 char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#ifndef linux
#define WINDOWS
#endif

#ifdef WINDOWS
#include <stdio.h>
#include <string.h>
#include <sys/timeb.h>
#include <process.h>
/* OpenGL Headers */
#define GLEW_STATIC
#define GL_GLEXT_PROTOTYPES
#include <gl\glew.h>
#include <gl\glut.h>
/* Cg Headers */
#include <Cg\cg.h>
#include <Cg\cgGL.h>
#define USE_GLUT
#endif

#ifdef linux
/* OpenGL Headers */
#define GL_GLEXT_PROTOTYPES
#include <GL/glew.h>
#include <GL/glx.h>
/* Cg Headers */
#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include <X11/extensions/xf86vmode.h>
#include <X11/keysym.h>
#include <X11/extensions/Xdbe.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>
#define USE_XGL
#endif

#include <stdlib.h>
#include <limits.h>
#include <math.h>
#include "hvr_work.h"
#include "TimeStamp.h"
#include <errno.h>
#include <signal.h>

#define MAXNODES 40000
#define MAXTEX 8192
#define MAXFBO 5
#define MEGABYTE 1048576
#define SHORTNAP  1
 
//#define DO_ANNOTATION
#define NEWTEX
// #define PBO         // dhp2  jan29: this was comment out (2009 jan 29)
#define WRITE_STD_OUT
//#define TEST_ALLBLACK 

#ifdef NEWTEX
int old_max_nxsb2 = 0;
int old_max_nysb2 = 0;
int old_max_nzsb2 = 0;
int tex_mem_avail = 200;
#endif

#ifdef PBO
GLuint texBuffer;
void *pboMemory;
#define BUFFER_OFFSET(i) ((char *)NULL + (i))
#endif
 
/* define if writing image to stdout */
#ifdef WRITE_STD_OUT
 #define STDIN  0
 #define STDOUT 1
 #define STDERR 2
 #ifdef linux
   # define SET_BINARY_MODE(handle) ((void)0) 
 #else
   # include <io.h>
   # include <fcntl.h>
   # define SET_BINARY_MODE(handle) setmode(handle, O_BINARY)
 #endif
#endif
 
FILE* error_log = NULL;
#define SUCCESS 0 /* Success */
#define X_FAILED 1 /* Could not connect to X Server */
#define VISUAL_FAILED 2 /* Could not choose visual */
#define HV_FAILED 3 /* Could not open HV File */
#define FBO_FAILED 4 /* Fatal FBO or Renderbuffer error */
#define CACHE_FAILED 5 /* Fatal caching scheme error */
#define CG_FAILED 6 /* Fatal Cg Error */
#define LOG_FAILED 7 /* Logfile failed to open */
#define GLEW_FAILED 8 /* GLEW failed to init */
#define IPRINT_FAILED 9 /* iprint failed */
#define HV_FILE_BAD 10 /* HV File bad */
#define SIGTERM_EXIT 11 /* Exitting because of a sigterm */
#define CLIENT_EXIT 12 /* Exitting because calling routine call ClearIt */
int status = SUCCESS;
bool am_exiting = FALSE;
 
//#define DEBUG_TEX
#ifdef DEBUG_TEX
#define CHECK_RESIDENCY
//#define DEBUG_CG       // dhp1
#define MAX_TEX 8190
#endif

#define INTELLIGENT_OVERWRITE
//#define AUTO_CACHE
//#define TEMP_TEST
//#define BGRA_LUT
//#define TEST_FBO
//#define FORCE_TEXSCHEME 1
//#define DEBUG_NEW_TREE 1

/* Texture Caching Scheme Efficiency */
#define TCS_Eff_Min 75 /* Reload cache if efficiency is below this % */
//#define PRINT_TCS_EFF
//#define DEBUG_TCS /* Print TCS Log */
int nBricksInVolume = 0; /* # of bricks in the entire volume */
int nBricksInCache = 0; /* # of bricks currently cached in texture memory */
int nBricksFitInCache = 0; /* # of bricks which can fit in texture memory */

/* FBO-Renderbuffer */
GLuint fbo[MAXFBO];
GLuint rendbuff;
int fb_count = 0;
int rb_width = 256;
int rb_height = 256;
bool size_change = FALSE;
bool Renderbuffer;
bool FBOs_created = FALSE;
int iWriteImageToSTDout = 1;

bool Tex3d;
int TexMode;
int Prev_TexMode;

/* Resident Textures */
bool ResidentTextures;
bool AutoCaching;
GLuint texture[MAXTEX];
#ifdef NEWTEX
GLboolean residences[MAXTEX];
#endif
int bricktex[MAXTEX]; /* Given brick #, returns associated texture # */
int texbrick[MAXTEX]; /* Given texture #, returns associated brick # */
int TexScheme = 0;
int count = 0;
int hv_count = 0;
int loopcount = 0;
int end_count = 0;
int test_count = 0;
int texmem = 0;
#ifdef INTELLIGENT_OVERWRITE
    int prev_count = -1;
    bool all_in_use = FALSE;
#else
    int texmem_previous = 0;
#endif
bool texmem_full = FALSE;
bool new_brick = TRUE;
bool overwrite = FALSE;
bool reload = FALSE;
int max_nxsb2, max_nysb2, max_nzsb2;

/* Amount of video memory for textures */
#ifdef WINDOWS
int max_tex_mem = 220*MEGABYTE; 
#else
int max_tex_mem = 470*MEGABYTE;
#endif

int max_tex_needed;

#ifdef WINDOWS
/* OpenGL rendering thread is the calling thread */
int iSpawnRenderThread = 1; 
#else
int iSpawnRenderThread = 0;
#endif

LPVOID pHVRforOpenGL;

/* Cg Declarations */
bool Precompile = TRUE;
bool CgInit = FALSE;
CGcontext Context;
CGprogram fProgram;
CGprofile fProfile = CG_PROFILE_FP30;
CGparameter textureHandle, textureHandle2;
CGparameter alphaHandle;
CGbool is_context = FALSE;

/* Cg Fragment Program Source Code */
const char *CGsource = 
"struct TexLookup {\n"
"    float4 color : COLOR;\n"
"};\n"
"\n"    
"TexLookup main(TexLookup IN,\n"
"                float3 texCoord : TEXCOORD0,\n"
"                uniform sampler3D data,\n"
"                uniform sampler1D LUT)\n"          
"{\n"
"    TexLookup OUT;\n"
"    float index = tex3D(data, texCoord).r;\n"
"    OUT.color = tex1D(LUT, index);\n"
"    OUT.color.a *= IN.color.a;\n"    
"    return OUT;\n"
"}\n";

const char *CGsource_2d = 
"struct TexLookup_2D {\n"
"    float4 color : COLOR;\n"
"};\n"
"\n"    
"TexLookup_2D main(TexLookup_2D IN,\n"
"                float2 texCoord : TEXCOORD0,\n"
"                uniform sampler2D data,\n"
"                uniform sampler1D LUT)\n"          
"{\n"
"    TexLookup_2D OUT;\n"
"    float index = tex2D(data, texCoord).r;\n"
"    OUT.color = tex1D(LUT, index);\n"
"    OUT.color.a *= IN.color.a;\n"    
"    return OUT;\n"
"}\n";

/* Cg Fragment Program Precompiled Object Code */
const char *CGobject =
"!!FP1.0\n"
"# cgc version 1.4.0000, build date Sep 26 2005 21:50:31\n"
"# command line args: -profile fp30\n"
"# source file: TexLookup.cg\n"
"#vendor NVIDIA Corporation\n"
"#version 1.0.02\n"
"#profile fp30\n"
"#program main\n"
"#semantic main.data\n"
"#semantic main.LUT\n"
"#var float4 IN.color : $vin.COLOR : COL0 : 0 : 1\n"
"#var float3 texCoord : $vin.TEXCOORD0 : TEX0 : 1 : 1\n"
"#var sampler3D data :  : texunit 0 : 2 : 1\n"
"#var sampler1D LUT :  : texunit 1 : 3 : 1\n"
"#var float4 main.color : $vout.COLOR : COL : -1 : 1\n"
"TEX   R0.x, f[TEX0], TEX0, 3D;\n"
"TEX   R0, R0, TEX1, 1D;\n"
"MULR  o[COLR].w, R0, f[COL0];\n"
"MOVR  o[COLR].xyz, R0;\n"
"END\n"
"# 4 instructions, 1 R-regs, 0 H-regs\n";

const char *CGobject_2d =
"!!FP1.0\n"
"# cgc version 1.4.0000, build date Sep 26 2005 21:50:31\n"
"# command line args: -profile fp30\n"
"# source file: TexLookup_2D.cg\n"
"#vendor NVIDIA Corporation\n"
"#version 1.0.02\n"
"#profile fp30\n"
"#program main\n"
"#semantic main.data\n"
"#semantic main.LUT\n"
"#var float4 IN.color : $vin.COLOR : COL0 : 0 : 1\n"
"#var float2 texCoord : $vin.TEXCOORD0 : TEX0 : 1 : 1\n"
"#var sampler2D data :  : texunit 0 : 2 : 1\n"
"#var sampler1D LUT :  : texunit 1 : 3 : 1\n"
"#var float4 main.color : $vout.COLOR : COL : -1 : 1\n"
"TEX   R0, f[TEX0], TEX0, 2D;\n"
"TEX   R0, R0, TEX1, 1D;\n"
"MULR  o[COLR].w, R0, f[COL0];\n"
"MOVR  o[COLR].xyz, R0;\n"
"END\n"
"# 4 instructions, 1 R-regs, 0 H-regs\n";

SIZE_T dwMaxSize = 512*512*512;   /* was 128*128*128 */
unsigned char *bobsubX;
unsigned char *bobsubY;
unsigned char *bobsubZ;
unsigned char *bobsubZ0;
unsigned char *bobOX;
unsigned char *bobOY;
unsigned char *bobOZ;
bool bBobsAllocated = FALSE;
unsigned char *ucGPix = NULL;
bool bSetBobX, bSetBobY;
HvrInput global_ShvrIn; /* GUI writes input here */

#ifdef WINDOWS
long sol[8192*4096];
#else
unsigned int  sol[512*512*512];   /* was 256*256*256 */
#endif

//HDC hDC; /* device context */
//HGLRC hRC;    /* opengl context */
//HWND  hWnd; /* window */

#ifdef USE_XGL
Display *dpyHVR;
XVisualInfo *viHVR;
Colormap cmapHVR;
XSetWindowAttributes swaHVR;
Window winHVR;
GLXContext cxHVR;
XEvent eventHVR;
#endif

//MSG msg;
int iFirst_Test = 1;
int iGLwindowIsMade = 0;
int iNClients = 0;
long iNRendering = 0;
int iAmUp = 0;
int iStopRendering = 0;
int iDoThis = 0;
wchar_t wcErrorStatus[256];

#define MAX_OPTIONS  1000
wchar_t wcLocalStatus[256*MAX_OPTIONS];
wchar_t wcKey[256], wcValue[256];
wchar_t wcOpKeys[MAX_OPTIONS][256], wcOpVals[MAX_OPTIONS][256];
int iOpNumber=0;

/* INPUTS for bounding box and floor */
float fTickSpacing = (float)0.2;
float fFloorSize = (float)5.0;
int iFloorRed = 0;
int iFloorGrn = 0;
int iFloorBlu = 0;
int iFloorAlp = 0;
int iBoxRed =   0;
int iBoxGrn =   0;
int iBoxBlu =   0;
int iBoxAlp =   0;
int iPriorityRequest = 0;

/*-------------------------------------------------------
 * VECTOR DATA  (see routines setvectors & drawsurfaces)
 */
#define MAXPTS   500000
#define MAXOBJ   100000

char cVecFile[256] = "none";
char cVecFileOld[256] = "nonelast";
char cVecLoClr[256] = "#008000";
char cVecHiClr[256] = "#FFFFFF";

float vecradius=0.0;        // Common radius of tubes or spheres
float vecvalmin,vecvalmax;  // Min & Max valus at vector points
int   nvecobj=0;            // # of vector objects
int   ntubefacets=10;        // # of facets going around a tube
int   iptstart[MAXOBJ];     // 1st pt in strip
int   iptend[MAXOBJ];       // 1 past last pt in strip
char  cobjtype[MAXOBJ][16]; // Type of object
float xpt[MAXPTS];          // X-coordinate of pt
float ypt[MAXPTS];          // Y-coordinate of pt
float zpt[MAXPTS];          // Z-coordinate of pt
float vpt[MAXPTS];          // Value of pt

int ntotvecobj=0;
int nvecobjs[MAXNODES];
int istartvecobj[MAXNODES];

int   mynormalsareset[MAXPTS];
float xundir1[MAXPTS];      // 1st orthoganal direction
float yundir1[MAXPTS];
float zundir1[MAXPTS];
float xundir2[MAXPTS];      // 2nd orthoganal direction
float yundir2[MAXPTS];
float zundir2[MAXPTS];
float ccos[100], ssin[100];

float xcenteye, ycenteye, zcenteye;
/*-------------------------------------------------------*/

#ifdef linux
Chvr_work *savethis;
#endif

CTimeStamp tsEvent;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Chvr_work::Chvr_work()
{    
    iAmUp = 1;
    initialize_vars();
    pannels_initted = 0;

    if(!bBobsAllocated) {
#ifdef linux
        bobsubX  = (unsigned char *)malloc( dwMaxSize );
        bobsubY  = (unsigned char *)malloc( dwMaxSize );
        bobsubZ0 = (unsigned char *)malloc( dwMaxSize );
        bobOX    = (unsigned char *)malloc( dwMaxSize );
        bobOY    = (unsigned char *)malloc( dwMaxSize );
        bobOZ    = (unsigned char *)malloc( dwMaxSize );
#else
        bobsubX  = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
        bobsubY  = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
        bobsubZ0 = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
        bobOX    = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
        bobOY    = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
        bobOZ    = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
#endif
        bBobsAllocated = TRUE;
    }
}

//void Chvr_work::SIGTERM_Handler(int signum) {
//    exit HVR(status);
//}

void mySwapBuffers()
{
#ifdef USE_WGL
    SwapBuffers(hDC);
#endif
#ifdef USE_GLUT
    glutSwapBuffers();
#endif
#ifdef USE_XGL
    glXSwapBuffers(dpyHVR, winHVR);
#endif
}

void myCreateFont()
{
#ifdef USE_WGL
    SelectObject(hDC,CreateFont(0,0,0,0,FW_DONTCARE,FALSE,FALSE,FALSE,
                 ANSI_CHARSET,OUT_DEFAULT_PRECIS,CLIP_DEFAULT_PRECIS,
                 DEFAULT_QUALITY,FF_ROMAN,NULL));
    wglUseFontBitmaps (hDC, 0, 255, 1000);
    glListBase (1000);
    iFirst_Test = 0;
#endif
}

void Chvr_work::initialize_vars()
{
    int i, j;
    void setstr(char*, char*);

    m_iNRendBuffs = 2; /* double or single buffered */
    m_bColorTableIsAvailable = TRUE; /* Paletted Textures */
    ResidentTextures = TRUE; /* Resident Textures */
    Tex3d = TRUE; /* 3d Textures */
    Renderbuffer = TRUE; /* Offscreen Rendering */

    #ifdef AUTO_CACHE
    AutoCaching = TRUE;
#else
    AutoCaching = FALSE;
#endif

    TexMode = 0;
    Prev_TexMode = TexMode;
    /* Tex Modes:
       4 = 3d paletted w/ caching 
       3 = 3d paletted w/o caching
       2 = 3d rgba
       1 = 2d paletted
       0 = 2d rgba
    */
    
    isubcur = -1; /* current sub-brick loaded into bobsub */
    ireload = 1;  /* flag to reload brick in 3dtex code   */
    Pid180 = (float)(3.141592654/180.);
    m_flClearRGB[0] = (float)0.;
    m_flClearRGB[1] = (float)0.;
    m_flClearRGB[2] = (float)0.;
    m_iRender = 0;
    m_ShvrIn.iNewHV      = 1;
    m_ShvrIn.iNewLUT     = 1;
    m_ShvrIn.iNewView    = 1;
    m_ShvrIn.iNewFrustum = 1;
    sprintf(m_sTimerFile, "HVR_timers.pto");
    sprintf(m_cStatusRequest, "RENDERING");
    sprintf(m_sStats, "none");
    m_iTimersOn = 0;
    m_iStopRendering = 0;
    m_iBufferedIO = 0;
    m_iAutoSwapBuffers = 1;
    m_iShowColorBar = 0;
    m_iSelALPknot   = -1;
    m_iSelRGBknot   = -1;
    m_iNTicks = 0;
    m_iXNTicks = 0;
    m_iYNTicks = 0;
    iStopRendering = 0;
    m_iBlockOnRender = 1;
    m_fScreenRatio = (float)1.0;

    SetDefaultFrustum();
    SetDefaultView(); 

    setstr(m_ShvrIn.sCmapname, "default");
    setstr(m_ShvrIn.sAmapname, "default");
    sprintf(m_cOldHVname, "none");

    for(i=0; i<256; i++) {
        for(j=0; j<4; j++) { frgba[i][j] = (float)i; }
        m_lrgba[i] = i + 256*(i + 256*(i + 256*i));
    }
    m_iNrgbNots = 2;
    m_iNalpNots = 2;
    m_iRGBind[0] = 0;
    m_iALPind[0] = 0;
    m_fRED[0]    = (float)0.;
    m_fGRN[0]    = (float)0.;
    m_fBLU[0]    = (float)0.;
    m_fALP[0]    = (float)0.;
    m_iRGBind[1] = 255;
    m_iALPind[1] = 255;
    m_fRED[1]    = (float)1.;
    m_fGRN[1]    = (float)1.;
    m_fBLU[1]    = (float)1.;
    m_fALP[1]    = (float)1.;

    m_fScaleFactors[0] = (float)1.0;
    m_fScaleFactors[1] = (float)1.0;
    m_fScaleFactors[2] = (float)1.0;

    opacity   = 1.0;
    tolerance = 5;                /* max color level error allowed with no scan down  */
    angtol    = (float)0.02;    /* max angle allowed with no scan down              */
    viewres   = (float)8.0;        /* dx_brick > vrtolerance*dx_irep1 => no scan down  */
    nplanes   = 1;
    maxplanes = 1;

    m_ShvrIn.opacity = opacity;
    m_ShvrIn.tolerance = tolerance;
    m_ShvrIn.angtol = angtol;
    m_ShvrIn.viewres = viewres;
    m_ShvrIn.nplanes = nplanes;
    m_ShvrIn.maxplanes = maxplanes;

    softdistance = 0.;
#ifdef linux
    m_hLocalWnd = 0;
#else
    m_hLocalWnd = NULL;
#endif
    setstr(m_cOldCmap, "none");
    m_iAlphaIsSet = 0;

    fTickSpacing = (float)0.2;
    fFloorSize = (float)5.0;
    iFloorRed = 0;
    iFloorGrn = 0;
    iFloorBlu = 0;
    iFloorAlp = 0;
    iBoxRed =   0;
    iBoxGrn =   0;
    iBoxBlu =   0;
    iBoxAlp =   0;

#ifdef WINDOWS
    swprintf(wcErrorStatus, L"OK");
#else
    swprintf(wcErrorStatus, "OK");
#endif
}

int iOpenGL_WindowIsOpenned = 0;
int iIO_Thread_Is_Up = 0;
int iIO_Read = 0;
int iN_IO_Threads=1;
long i_IO_Thread=0;
int iThreadIsOnCache[100];
char cOtherDirs[100][256];
char cLastDir[256] = "";

int     iNRequests;
int     iCacheStatusIsReady;
int     iTile_of_iReq[MAXNODES];
int     iSorted_Req_of_iReq[MAXNODES];
char    cFileName_of_iReq[MAXNODES][256];
__int64 i64FileOffset_of_iReq[MAXNODES];
DWORD   dwSize_of_iReq[MAXNODES];

int            iNCacheMB = 360;
//int            iNCacheMB = 800;
int            iTile_of_iCache[MAXNODES];
int            iUsed_of_iCache[MAXNODES];
unsigned char  *lpBuffer_of_iCache[MAXNODES];
long           lBufferMin_iCache[MAXNODES];
long           lBufferMax_iCache[MAXNODES];
int            iCache_of_iTile[MAXNODES];
unsigned char  *BobCache;
int            iBobCacheAllocated = 0;

static void exitHVR(int status) {
#ifdef DEBUG_NEW_TREE
fprintf(stderr, "in exitHVR: status = %d\n", status);
#endif
    if(!am_exiting) {
        am_exiting = TRUE;
    #ifdef WRITE_STD_OUT
        write(STDOUT,&status,4);
    #endif
        if(error_log != NULL)
            fclose(error_log);
    #ifdef linux
        if(cgIsContext(Context)) {
            cgDestroyProgram(fProgram);
            cgDestroyContext(Context);
        }
        if(FBOs_created)
            glDeleteFramebuffersEXT(MAXFBO, fbo);
        if (dpyHVR != NULL)
            XCloseDisplay(dpyHVR);
        if (ucGPix != NULL)
            free(ucGPix);
        if (BobCache != NULL)
            free(BobCache);
    /*#else
        if (ucGPix != NULL)
            VirtualFree(ucGPix, 3*rb_width*rb_height, MEM_DECOMMIT | MEM_RELEASE);
        if (BobCache != NULL)
            VirtualFree(BobCache, dwSize, MEM_DECOMMIT | MEM_RELEASE);*/
    #endif
        exit(status);
    }
}

void SIGTERM_Handler(int signum) {
#ifdef DEBUG_NEW_TREE
fprintf(stderr, "exitting in SIGTERM_Handler\n");
#endif
    exitHVR(SIGTERM_EXIT);
}

STDMETHODIMP Chvr_work::ClearIt()
{
    DWORD dhpReadFromFileNoBuf(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes);

    if(iSpawnRenderThread == 0) {  /* >>>> */
#ifdef USE_GLUT
        //printf("ClearIt: GLUT with iSpawnRenderThread ==0 not implemented\n");
#endif
#ifdef USE_WGL
        /* Shut down OpenGL environment IF last client */
        iNClients--;
        if(iNClients == 0) {
            void deleteGLPallet();

            wglMakeCurrent(hDC, NULL);
            wglDeleteContext(hRC);
            if(m_hLocalWnd == NULL) { DestroyWindow(hWnd); }
            deleteGLPallet();

            isubcur = -1;         /* current sub-brick loaded into bobsub */
            sprintf(m_cOldHVname, "none");
            iGLwindowIsMade = 0;
            dhpReadFromFileNoBuf(lpBuffer_of_iCache[0], "close", 0, 0);
        }
#endif
    } else {
        /* Only one thread can do any of the OpenGL calls, including shut down.
           Hence, set flag iDoThis so that OpenGL thread will shut down.
           See routine SpawnWork */
        if(error_log != NULL) {
            iDoThis = 2;
            while(iDoThis == 2) { _sleep(10); }
        }
    }
    exitHVR(CLIENT_EXIT);

    return 0;
}

#ifdef USE_GLUT
void GLUTdisplay()
{
    // if gl window is already up, do not spawn a new one
    // if(iGLwindowIsMade == 1) return;
    // Draw it
    Chvr_work* my_work = (Chvr_work*)pHVRforOpenGL;
    my_work->GLUTdraw();
}
#endif
void OpenAWindow(void *pParam)
{
    /* Only do the body of this function ONCE per instance of the host program */
    //    if(iOpenGL_WindowIsOpenned == 1) return;
    //    iOpenGL_WindowIsOpenned = 1;
    if(iGLwindowIsMade == 1) return;
    Chvr_work* my_work = (Chvr_work*)pParam;
    my_work->SpawnWork();
}

void StartIOThread(void *pParam)
{
    /* Only do the body of this function ONCE per instance of the host program */
    // if(iOpenGL_WindowIsOpenned == 1) return;
    // iOpenGL_WindowIsOpenned = 1;
    if(iIO_Thread_Is_Up == 1) return;
    Chvr_work* my_work = (Chvr_work*)pParam;
    my_work->DoIO();
}

void ResetCacheStatus()
{
    iCacheStatusIsReady = 0;
    iNRequests = 0;
}

void FlushCacheBuffers()
{
    int i;

    /* Invalidate all cache buffers */
    for(i=0; i<MAXNODES; i++) {
        iTile_of_iCache[i] = -1;
        iCache_of_iTile[i] = -1;
        iUsed_of_iCache[i] =  0;
    }
}

void RequestCacheIO(int iTile, LPCTSTR lpFileName, __int64 offset, DWORD nbytes)
{
    iTile_of_iReq[iNRequests] = iTile;
    sprintf(cFileName_of_iReq[iNRequests], "%s", lpFileName);
    i64FileOffset_of_iReq[iNRequests] = offset;
    dwSize_of_iReq[iNRequests] = nbytes;
    iNRequests++;
}

LPVOID lpGetCacheBuffer(int iTile)
{
    /* Wait for cache status arrays to be ready */
    while(iCacheStatusIsReady == 0   &&  iIO_Read == 1) { 
        _sleep(SHORTNAP);
    }

    /* Wait for cache buffer to be read in */
    while(iCache_of_iTile[iTile] < 0) {
        if(iCache_of_iTile[iTile] == -10) {
           iStopRendering = 1;
           iIO_Read = 2;
        }
        if(iStopRendering == 1) { return BobCache; }
        _sleep(SHORTNAP);
    }

    /* Return pointer to cache buffer */
    return lpBuffer_of_iCache[iCache_of_iTile[iTile]];
}

#ifdef linux
/* This is potentially dangerous if we ever run with more
   than one IO thread. will need to implement a mutex here */
long InterlockedIncrement(long *i)
{
#ifdef IDEBUG
    printf("Inside InterlockedIncrement: before bump : i = %d\n", (int)*i);
#endif
    (*i)++;
#ifdef IDEBUG
    printf("Inside InterlockedIncrement: after  bump : i = %d\n", (int)*i);
#endif
    return *i;
}
#endif
void Chvr_work::DoIO()
{
    int i, ij, iWrap, isHere, isFree, iOverlap, iTile, iRequest, jRequest, iPrinted;
    int iCache, iCache0, iBegin;
    int iPreviousOverFlow = 0, iMyThread, iNConsecutive, iCacheMin;
    int iNSort, iMax, iChunk, iReqMin;
    __int64 iTileTmp[128];
    long lBufferOffset, lNextBufferOffset, lNeedTotalBufferSize, lUnit;
    char cFileName[256], cTemp[256];
    DWORD  dhpReadFromFileSync(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes);
    DWORD dhpReadFromFileNoBuf(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes);
    DWORD iMaxRequest;
    DWORD dwGot;
/*    DWORD dhpReadFromOpenHandle(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes, int iClose); */
//    char cDebugFile[256];
//    FILE *fpp;

#ifdef IDEBUG
    printf("DoIO top, i_IO_Thread = %d\n", (int)i_IO_Thread);
#endif

    iMyThread = (int)InterlockedIncrement(&i_IO_Thread) - 1;

#ifdef IDEBUG
    printf("DoIO: after InterlockedIncrement, iMyThread   = %d\n", iMyThread);
    printf("DoIO: after InterlockedIncrement, i_IO_Thread = %d\n", (int)i_IO_Thread);
#endif

/*    sprintf(cDebugFile, "F:\\IO_Thread_%d.txt", iMyThread);
    fpp = fopen(cDebugFile, "w");
    fprintf(fpp, "Thread %d starting\n", iMyThread);
    fclose(fpp);
    fpp = fopen(cDebugFile, "a+");
    fprintf(fpp, "Thread %d started\n", iMyThread);
    fclose(fpp); */

    /* Invalidate all cache buffers */
    iCache = 0;
    lBufferOffset = (long)0;
    iWrap = 0;
    if(iMyThread == 0) {
        for(i=0; i<MAXNODES; i++) {
            iTile_of_iCache[i] = -1;
            iCache_of_iTile[i] = -1;
            iUsed_of_iCache[i] =  0;
        }
    }
    
    /* Set cache sizes */
    lUnit = (long)64 * (long)1024;
    SIZE_T  dwSize   = (SIZE_T)iNCacheMB * (SIZE_T)1024 * (SIZE_T)1024;

    if(iMyThread == 0  && iBobCacheAllocated == 0) {
        /* Allocate memory */
#ifdef linux
        BobCache = (unsigned char *)malloc(dwSize);
#else
        BobCache = (unsigned char *)VirtualAlloc(NULL, dwSize, MEM_COMMIT, PAGE_READWRITE);
#endif
        iBobCacheAllocated = 1;
    }
#ifdef IDEBUG
    printf("DoIO: after allocation of BobCache iMyThread = %d\n", iMyThread);
#endif

    /* Wait for all threads to get here */
    while((int)i_IO_Thread < iN_IO_Threads) {
#ifdef IDEBUG
        printf("DoIO: waiting all io thread to get here iN_IO_Threads = %d\n", iN_IO_Threads);
        _sleep(1000);
#else
        _sleep(SHORTNAP);
#endif
    }
    if(iMyThread == 0) { iIO_Thread_Is_Up = 1; }
    while(iIO_Thread_Is_Up != 1) { _sleep(SHORTNAP); }

    /* Wait here until there is something to do
       This wait is also at the end of the following while loop */
    while(iIO_Thread_Is_Up == 1  &&  iIO_Read == 0) { 
#ifdef IDEBUG
        printf("DoIO: waiting for something to do iIO_Read = \n", iIO_Read);
        _sleep(1000);
#else
        _sleep(SHORTNAP);
#endif
    }

#ifdef IDEBUG
    printf("DoIO: after wait for something to do\n");
#endif

    while(iIO_Thread_Is_Up == 1) {

        iPrinted = 0;

        /* If iTile_of_iCache[0] does not point to any tile, then either the buffer
           cache has been reset elsewhere or the cache wrapped on the previous pass.
           This means we are starting from scratch: reset local counters and all 
           cache pointers */
        if(iTile_of_iCache[0] < 0) {
            iCache = 0;
            lBufferOffset = (long)0;
            iWrap = 0;
            if(iMyThread == 0) {
                for(i=0; i<MAXNODES; i++) {
                    iTile_of_iCache[i] = -1;
                    iCache_of_iTile[i] = -1;
                    iUsed_of_iCache[i] = 0;
                }
            }
        }

        /* Count how much total buffer memroy will be needed in this pass assuming
           we keep previously read buffers Note: only need to keep adding until 
           needed buffer size exceeds allocated buffer size */
        lNeedTotalBufferSize = lBufferOffset;
        for(iRequest = 0; iRequest < iNRequests; iRequest++) {
            isHere = 0;
            iTile = iTile_of_iReq[iRequest];
            iCache0 = iCache_of_iTile[iTile];
            if(0 <= iCache0  &&  iCache0 < iCache) { if(iTile == iTile_of_iCache[iCache0]) { isHere = 1; }}
            if(isHere == 0  &&  lNeedTotalBufferSize <= (long)dwSize) {
                lNeedTotalBufferSize += lUnit*(1 + ((long)dwSize_of_iReq[iRequest]-1)/lUnit);
            }
        }

        iMaxRequest = 0;
        for(iRequest = 0; iRequest < iNRequests; iRequest++) {
          if(iMaxRequest < dwSize_of_iReq[iRequest]) { iMaxRequest = dwSize_of_iReq[iRequest]; }
        }

        /* If previous + new requests will run out of cache on this trip,
           wipe it all clean. Otherwise, mark all cache buffers as used */
        if(lNeedTotalBufferSize > (long)dwSize  ||  iWrap == 1) {
            iCache = 0;
            lBufferOffset = 0;
            iWrap = 0;
            if(iMyThread == 0) {
                for(i=0; i<MAXNODES; i++) {
                    iTile_of_iCache[i] = -1;
                    iCache_of_iTile[i] = -1;
                    iUsed_of_iCache[i] = 0;
                } 
            }
        } else {
            if(iMyThread == 0) { for(i=0; i < iCache; i++) { iUsed_of_iCache[i] = 1; } }
        }

        /* Sort the IO requests in blocks of iNSort;
           Tile order in hv file is given by i64FileOffset_of_iReq
           Assumes all tiles are in one file */
        if(iMyThread == 0) {
            iNSort = 64;
            if(iMaxRequest > 4194304) { iNSort = 1; }
            for(iChunk=0; iChunk<iNRequests; iChunk += iNSort) {
                iMax = iChunk+iNSort;
                if(iMax > iNRequests) { iMax = iNRequests; }
                for(i=iChunk; i<iMax; i++) { iTileTmp[i-iChunk] = i64FileOffset_of_iReq[i]; }
                for(i=iChunk; i<iMax; i++) {
                    iReqMin = iChunk;
                    for(ij=iChunk; ij<iMax; ij++) {
                      if(iTileTmp[ij-iChunk] < iTileTmp[iReqMin-iChunk]) { iReqMin = ij; }
                    }
                    iSorted_Req_of_iReq[i] = iReqMin;
                    iTileTmp[iReqMin-iChunk] = (__int64)2000000000000;
                }
            }
/*
            sprintf(cDebugFile, "E:\\HVR\\IO_SORT.txt");
            fpp = fopen(cDebugFile, "w");
            fprintf(fpp, "start\n");
            fprintf(fpp, "iNRequests = %d\n", iNRequests);
            for(i=0; i<iNRequests; i++) {
                fprintf(fpp, "i16=%2d   iReq=%3d   iSortReq=%3d   Offset=%ld\n",
                    i%iNSort, i, iSorted_Req_of_iReq[i], i64FileOffset_of_iReq[iSorted_Req_of_iReq[i]]);
            }
            fprintf(fpp, "---------------------------------------------------------------\n");
            fclose(fpp);
*/
        }

        /* Tell everyone that cache status is ready */
        if(iMyThread == 0) {
            for(i=0; i<iN_IO_Threads; i++) { iThreadIsOnCache[i] = 0; }
            i_IO_Thread = 0;
            iCacheStatusIsReady = 1;
        } else {
            while(iCacheStatusIsReady != 1  &&  iIO_Read == 1) { _sleep(SHORTNAP); }
        }

        /* Read in data */
        iNConsecutive = 4;
        jRequest = 0;
        while(iIO_Read == 1  &&  jRequest < iNRequests) {
            iRequest = iSorted_Req_of_iReq[jRequest];

            /* First, make sure that no IO thread is to far ahead of the others */
            if(iN_IO_Threads > 1) {
                iThreadIsOnCache[iMyThread] = iCache;
                iCacheMin = -1000;
                while(iCacheMin < iCache-2*iNConsecutive*iN_IO_Threads  &&  iIO_Read == 1) {
                    iCacheMin = iThreadIsOnCache[0];
                    for(i=1; i<iN_IO_Threads; i++) {
                        if(iCacheMin > iThreadIsOnCache[i]) { iCacheMin = iThreadIsOnCache[i]; }
                    }
                    if(iCacheMin < iCache-2*iNConsecutive*iN_IO_Threads) { _sleep(SHORTNAP); }
                }
            }

            /* Determine if data is already cached */
            isHere = 0;
            iTile = iTile_of_iReq[iRequest];
            iCache0 = iCache_of_iTile[iTile];
            if(0 <= iCache0  &&  iCache0 < iCache) { if(iTile == iTile_of_iCache[iCache0]) { isHere = 1; }}

            /* Only read in tile if it is not already in cache */
            if(isHere == 0) {
                /* Generate range of buffer memory [lBufferOffset, lNextBufferOffset) for next read */
                lNextBufferOffset = lBufferOffset + lUnit*(1 + ((long)dwSize_of_iReq[iRequest]-1)/lUnit);
                if(lNextBufferOffset > (long)dwSize) {
                    lBufferOffset = (long)0;
                    lNextBufferOffset = lUnit*(1 + ((long)dwSize_of_iReq[iRequest]-1)/lUnit);
                    iWrap = 1;
                }
                /* Wait for this range of memory to be free */
                isFree = 0;
                while(isFree == 0  &&  iIO_Read == 1) {
                    isFree = 1;
                    for(i=0; i<iCache; i++) {
                        if(iUsed_of_iCache[i] == 1) {
                            iOverlap = 1;
                            if(lNextBufferOffset <  lBufferMin_iCache[i]) { iOverlap = 0; }
                            if(    lBufferOffset >= lBufferMax_iCache[i]) { iOverlap = 0; }
                            if(iOverlap == 1) { isFree = 0; }
                        }
                    }
                    if(isFree == 0) { _sleep(SHORTNAP); }
                }

                if(iIO_Read == 1) {

                    if((jRequest/iNConsecutive)%iN_IO_Threads == iMyThread) {

                        /* Invalidate the cache buffers which previously used this memory */
// -----------------------------------------------------------------------------------------------
// I dont think that this is necessary
//                        for(i=0; i<iCache; i++) {
//                            iOverlap = 1;
//                            if(lNextBufferOffset <  lBufferMin_iCache[i]) { iOverlap = 0; }
//                            if(    lBufferOffset >= lBufferMax_iCache[i]) { iOverlap = 0; }
//                            if(iOverlap == 1) {
//                                iTile0 = iTile_of_iCache[i];
//                                iTile_of_iCache[i] = -1;
//                                if(iTile0 > 0) { iCache_of_iTile[iTile0] = -1; }
//                            }
//                        }
// -----------------------------------------------------------------------------------------------

                        /* Read data into this cache buffer */
                        lpBuffer_of_iCache[iCache] = &BobCache[lBufferOffset];
                        sprintf(cFileName, "%s", cFileName_of_iReq[iRequest]);
                        if(iMyThread > 0) {
                            for(i=0; i < (int)strlen(cFileName); i++) {
                                if(cFileName[i] == '/'  ||  cFileName[i] == '\\') {
                                    iBegin = i+1;
                                }
                            }
                            sprintf(cTemp, cFileName);
                            sprintf(cFileName, "%s%s", cOtherDirs[iMyThread], &cTemp[iBegin]);
                            //cFileName[0] += iMyThread;
                        }

                        /*fpp = fopen(cDebugFile, "a+");
                        fprintf(fpp, "Thread %d reading %s  iReq=%3d  iCache=%3d  offset=%ld\n",
                            iMyThread, cFileName, iRequest, iCache, lBufferOffset);
                        fclose(fpp); */

                        if(m_iBufferedIO != 1) {
                            dwGot = dhpReadFromFileSync(lpBuffer_of_iCache[iCache],
                                                        cFileName,
                                                        i64FileOffset_of_iReq[iRequest],
                                                        dwSize_of_iReq[iRequest]);
                        } else {
                            dwGot = dhpReadFromFileNoBuf(lpBuffer_of_iCache[iCache],
                                                         cFileName,
                                                         i64FileOffset_of_iReq[iRequest],
                                                         dwSize_of_iReq[iRequest]);
                        }

                        /* Validate new cache buffers in this location */

//fprintf(stderr, "YYY: iRequest=%d   dwGot=%d  dwSize_of_iReq=%d\n",
//iRequest, dwGot, dwSize_of_iReq[iRequest]);

                        if(dwGot == dwSize_of_iReq[iRequest]) {
                            iTile_of_iCache[iCache] = iTile;
                            iCache_of_iTile[iTile] = iCache;
                            lBufferMin_iCache[iCache] = lBufferOffset;
                            lBufferMax_iCache[iCache] = lNextBufferOffset;
                            iUsed_of_iCache[iCache] = 1;
                        } else {
                            /* Read Error: stop the reads and stop the rendering */
#ifdef DEBUG_NEW_TREE
fprintf(stderr, "Read from hv file failed:\n");
fprintf(stderr, "   iTile=%d   iCache=%d   iRequest=%d\n", iTile, iCache, iRequest);
fprintf(stderr, "   HV file read from   : %s\n", cFileName);
fprintf(stderr, "   size of read request: %d\n", (int)dwSize_of_iReq[iRequest]);
fprintf(stderr, "   offset into file    : %lld\n", i64FileOffset_of_iReq[iRequest]);
fprintf(stderr, "   Got this many bytes : %d\n", (int)dwGot);
#endif
                            status = HV_FILE_BAD;
                            iCache_of_iTile[iTile] = -10;

#ifdef    WINDOWS
                            swprintf(wcErrorStatus, L"ERROR: Read from %hs failed", cFileName);
#else
                            swprintf(wcErrorStatus, "ERROR: Read from %hs failed", cFileName);
#endif
                        }
                    }       /* End of (iRequest/iNConsecutive)%iN_IO_Threads == iMyThread)  IF-BLOCK */
                    lBufferOffset = lNextBufferOffset;
                    iCache++;
                }
            }                /* End of (isHere == 0) IF-BLOCK */
            jRequest++;
        }                    /* End of (iIO_Read == 1  &&  jRequest < iNRequests) while loop */

        /* Wait for all of the IO threads to finish before moving on */
        i = (int)InterlockedIncrement(&i_IO_Thread);
        while(i_IO_Thread < iN_IO_Threads) { _sleep(SHORTNAP); }

        if(iMyThread == 0) {
            i_IO_Thread = 0;
            iIO_Read = 0;
            iNRequests = 0;
        } else {
            while(iIO_Read != 0) { _sleep(SHORTNAP); }
        }

        /* Wait here until there is something to do
           This wait is also just before the while loop */
        while(iIO_Thread_Is_Up == 1  &&  iIO_Read == 0) { _sleep(SHORTNAP); }

    }                        /* End of (iIO_Thread_Is_Up == 1) WHILE-LOOP */

    /* Wait for all of the IO threads to exit loop before signalling that the
       io threads can no longer do any thing */
    i = (int)InterlockedIncrement(&i_IO_Thread);
    while(i_IO_Thread < iN_IO_Threads) { _sleep(SHORTNAP); }

    /* Tell everyone that IO thread has exitted iCache */
    if(iMyThread == 0) {
        iIO_Thread_Is_Up = 0;
        i_IO_Thread = 0;
    }
}

#ifdef USE_XGL


Bool WaitForNotify(Display *d, XEvent *e, char *arg)
{ return (e->type == MapNotify) && (e->xmap.window == (Window)arg); }

#endif

void check_FBO_status()
{
    GLenum status;

    status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
    switch(status) {
        case GL_FRAMEBUFFER_COMPLETE_EXT:
            break; /* framebuffer complete */
        case GL_FRAMEBUFFER_UNSUPPORTED_EXT: 
            fprintf(error_log,"FBO ERROR: unsupported\n");
            exitHVR(FBO_FAILED);
            break;
        case GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT: 
            fprintf(error_log,"FBO ERROR: dimensions \n");
            exitHVR(FBO_FAILED); 
            break; 
        case GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT: 
            fprintf(error_log,"FBO ERROR: formats \n");
            exitHVR(FBO_FAILED); 
            break; 
        case GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT: 
            fprintf(error_log,"FBO ERROR: draw buffer \n");
            exitHVR(FBO_FAILED); 
            break; 
        case GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT: 
            fprintf(error_log,"FBO ERROR: read buffer \n");
            exitHVR(FBO_FAILED); 
            break; 
        case GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT: 
            fprintf(error_log,"FBO ERROR: incomplete attachment \n");
            exitHVR(FBO_FAILED); 
            break; 
        case GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT: 
            fprintf(error_log,"FBO ERROR: missing attachment \n");
            exitHVR(FBO_FAILED); 
            break;
        default: 
            fprintf(error_log,"FBO ERROR\n");
            exitHVR(FBO_FAILED); 
    }
}

void new_FBO()
{    
    #ifdef DEBUG_TEX
        #ifdef WRITE_STD_OUT
            //fprintf(error_log,"Creating new Framebuffer Object\n");
        #else
            printf("Creating new Framebuffer Object\n");
        #endif
    #endif
    
    /* Bind Framebuffer Object (FBO) */
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,fbo[fb_count]);
    
    /* Create, bind and define Renderbuffer Object */
    glGenRenderbuffersEXT(1, &rendbuff);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, rendbuff);
    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,
                             GL_RGB8,
                             rb_width,
                             rb_height);

    /* Attach Renderbuffer Object to FBO */
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,
                                 GL_COLOR_ATTACHMENT0_EXT, 
                                 GL_RENDERBUFFER_EXT,
                                 rendbuff);
    
    /* Make sure FBO is "framebuffer complete" */
    check_FBO_status();

    /* Set GL Viewport */
    glViewport(0,0,rb_width,rb_height);

    /* Make FBO Current Rendering Context */
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT,fbo[fb_count]);

    /* Manage FBO's */
    fb_count++;
    if(fb_count==MAXFBO) {
        glDeleteFramebuffersEXT(MAXFBO-1, fbo);
        glGenFramebuffersEXT(MAXFBO-1, fbo);
        fb_count = 0;
    }
}

void Chvr_work::SpawnWork()
{
    DWORD dhpReadFromFileNoBuf(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes);

#ifdef USE_GLUT
    int argc;
    char *argv[100];
    pHVRforOpenGL = this;
    argc = 2;
    argv[0] = "hvr";
    argv[1] = "tri";
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
    glutCreateWindow("LCSE Hierarchical Volume Renderer");
    
    int err = glewInit();
    if (GLEW_OK != err) {
        fprintf(error_log,(char*)glewGetErrorString(err));
        exitHVR(GLEW_FAILED);
    }

    ucGPix = (unsigned char *)VirtualAlloc(NULL,3*m_ShvrIn.iWidth*m_ShvrIn.iWidth,MEM_COMMIT,PAGE_READWRITE);
    
    if(Renderbuffer) {
        rb_width = m_ShvrIn.iWidth;
        rb_height = m_ShvrIn.iWidth;
        glGenFramebuffersEXT(MAXFBO, fbo);
        FBOs_created = TRUE;
        new_FBO();
        init();
        GLUTdisplay();
        iGLwindowIsMade = 1;
    }
    else {
        glutReshapeWindow(m_ShvrIn.iWidth, m_ShvrIn.iHeight);
        init();
        glutDisplayFunc(GLUTdisplay);
        iGLwindowIsMade = 1;
        glutMainLoop();
        return;
    }

#else
#ifdef USE_XGL
    /*int  attributeListSgl[] = {GLX_RGBA,GLX_RED_SIZE,8,GLX_GREEN_SIZE,
                                     8,GLX_BLUE_SIZE,8,GLX_ALPHA_SIZE,8};*/
    /* int  attributeListDbl[] = { GLX_RGBA, GLX_DOUBLEBUFFER,
       GLX_RED_SIZE, 1, GLX_GREEN_SIZE, 1, GLX_BLUE_SIZE, 1, GLX_ALPHA_SIZE, 1 }; */
    int attributeListDbl[] = { GLX_DOUBLEBUFFER, GLX_RGBA, None };
    int attributeListSgl[] = { GLX_RGBA, None };

    //dpyHVR = XOpenDisplay(":1");
                         dpyHVR = XOpenDisplay(":1");
    if(dpyHVR == NULL) { dpyHVR = XOpenDisplay(":0"); }
    if(dpyHVR == NULL) { dpyHVR = XOpenDisplay(NULL); }

    if(dpyHVR == NULL) {
        fprintf(error_log,"Could not XOpenDisplay()\n");
        exitHVR(X_FAILED);
    }
    
    if(Renderbuffer) {
        viHVR = glXChooseVisual(dpyHVR, DefaultScreen(dpyHVR), attributeListSgl);
        if (viHVR == NULL) {
            fprintf(error_log,"could not glXChooseVisual single buffered\n");
            exitHVR(VISUAL_FAILED);
        }
    } else {
        if(m_iNRendBuffs == 1) {
            viHVR = glXChooseVisual(dpyHVR, DefaultScreen(dpyHVR), attributeListSgl);
            if (viHVR == NULL) {
                fprintf(error_log,"could not glXChooseVisual single buffered\n");
                exitHVR(VISUAL_FAILED);
            }
        }
        if(m_iNRendBuffs == 2) {
            viHVR = glXChooseVisual(dpyHVR, DefaultScreen(dpyHVR), attributeListDbl);
            if (viHVR == NULL) {
                fprintf(error_log,"could not glXChooseVisual double buffered\n");
                exitHVR(VISUAL_FAILED);
            }
        }    
    }
        
    cxHVR = glXCreateContext(dpyHVR, viHVR, 0, GL_TRUE);
    cmapHVR = XCreateColormap(dpyHVR, RootWindow(dpyHVR, viHVR->screen),
                                viHVR->visual, AllocNone);
    swaHVR.colormap = cmapHVR;
    swaHVR.border_pixel = 0;
    swaHVR.event_mask = StructureNotifyMask;
    winHVR  =  XCreateWindow(dpyHVR,  RootWindow(dpyHVR, viHVR->screen), 0, 0, m_ShvrIn.iWidth,
                          m_ShvrIn.iHeight, 0, viHVR->depth, InputOutput, viHVR->visual,
                          CWBorderPixel|CWColormap|CWEventMask, &swaHVR);
    
    if(!Renderbuffer) {
        XMapWindow(dpyHVR, winHVR);
        XIfEvent(dpyHVR, &eventHVR, WaitForNotify, (char*)winHVR);
    }

    glXMakeCurrent(dpyHVR, winHVR, cxHVR);

    int err = glewInit();
    if (GLEW_OK != err) {
        fprintf(error_log,(char*)glewGetErrorString(err));
        exitHVR(GLEW_FAILED);
    }

    ucGPix = (unsigned char *)malloc(3*m_ShvrIn.iWidth*m_ShvrIn.iHeight);

    if(Renderbuffer) {
        rb_width = m_ShvrIn.iWidth;
        rb_height = m_ShvrIn.iHeight;        
        glGenFramebuffersEXT(MAXFBO, fbo);
        FBOs_created = TRUE;
        new_FBO();
    }

    init(); // xx dhp xx
    iGLwindowIsMade = 1;

#endif
#ifdef USE_WGL
/************************************************************
 *  Here's how to create an OpenGL window using wgl         *
 ************************************************************/
    int depth = 0, width, height;
    BYTE  colortype;
    DWORD buffertype;
    char s[4096];
    HWND CreateOpenGLWindow(char* title, int x, int y, int width, int height, 
           BYTE type, DWORD flags, int depth, HWND hLocalWnd);
    void deleteGLPallet();
    void  setstr(char*, char*);
    DWORD dhpReadFromFileNoBuf(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes);

    *s = '\0';
    strcat(s,"HVR - SERVER");

    colortype = PFD_TYPE_RGBA;
//    colortype = PFD_TYPE_COLORINDEX;

    buffertype = PFD_DOUBLEBUFFER;
//    buffertype = 0;
    depth = 32;

    if(iGLwindowIsMade == 0) {

        width = m_ShvrIn.iWidth;
        height = m_ShvrIn.iHeight;

        hWnd = CreateOpenGLWindow(s, 0, 0, width, height, colortype, buffertype, depth, m_hLocalWnd);
        if (hWnd == NULL) {
/*
            fp = fopen("C:/DEBUG_HVR.txt", "a+");
            fprintf(fp, "Derived handle of OpenGL window is NULL\n");
            fprintf(fp, "-->Will try to create a Top Level OpenGL window\n");
            fclose(fp);
*/
            // Try again, this time with top level window
            hWnd = CreateOpenGLWindow(s, 0, 0, width, height, colortype,
                buffertype, depth, NULL);

            if (hWnd == NULL) {
/*                fp = fopen("C:/DEBUG_HVR.txt", "a+");
                fprintf(fp, "Handle of Top Level OpenGL window is still NULL\n");
                fprintf(fp, "-->Exit Rendering thread\n");
                fclose(fp);
*/
                return;
            }
        }

        hDC = GetDC(hWnd);
        if(hDC == NULL) {
//            fp = fopen("C:/DEBUG_HVR.txt", "a+");
//            fprintf(fp, "hDC from GetDC(hWnd) is NULL\n");
//            fprintf(fp, "-->will exit\n");
//            fclose(fp);
            return;
        }

        hRC = wglCreateContext(hDC);
        if(hRC == NULL) {
//            fp = fopen("C:/DEBUG_HVR.txt", "a+");
//            fprintf(fp, "hRC from wglCreateContext(hDC) is NULL\n");
//            fprintf(fp, "-->will exit\n");
//            fclose(fp);
            return;
        }

        if(wglMakeCurrent(hDC, hRC) == FALSE) {
//            fp = fopen("C:/DEBUG_HVR.txt", "a+");
//            fprintf(fp, "wglMakeCurrent(hDC, hRC) failed\n");
//            fprintf(fp, "-->will exit\n");
//            fclose(fp);
            return;
        }

        ShowWindow(hWnd, SW_SHOW);

        UpdateWindow(hWnd);

        init();

        iGLwindowIsMade = 1;
    } else {
        wglMakeCurrent(hDC, NULL);
        wglDeleteContext(hRC);
        if(m_hLocalWnd == NULL) { DestroyWindow(hWnd); }
        deleteGLPallet();

        width = m_ShvrIn.iWidth;
        height = m_ShvrIn.iHeight;
        hWnd = CreateOpenGLWindow(s, 0, 0, width, height, colortype, buffertype, depth, m_hLocalWnd);
        if (hWnd == NULL) {
//            fp = fopen("C:/DEBUG_HVR.txt", "a+");
//            fprintf(fp, "Derived handle of OpenGL window is NULL\n");
//            fprintf(fp, "-->Will try to create a Top Level OpenGL window\n");
//            fclose(fp);

            // Try again, this time with top level window
            hWnd = CreateOpenGLWindow(s, 0, 0, width, height, colortype,
                buffertype, depth, NULL);

            if (hWnd == NULL) {
//                fp = fopen("C:/DEBUG_HVR.txt", "a+");
//                fprintf(fp, "Handle of Top Level OpenGL window is still NULL\n");
//                fprintf(fp, "-->Exit Rendering thread\n");
//                fclose(fp);
                return;
            }
        }

        hDC = GetDC(hWnd);
        hRC = wglCreateContext(hDC);
        wglMakeCurrent(hDC, hRC);

        ShowWindow(hWnd, SW_SHOW);
        UpdateWindow(hWnd);

        init();

        isubcur = -1;         /* current sub-brick loaded into bobsub */
        sprintf(m_cOldHVname, "none");
    }

    // If hvr_server is being driven remotely, turn cursor off
    if(m_hLocalWnd == NULL) { ShowCursor(FALSE); }
#endif    
    if(iSpawnRenderThread == 0) return;
    
    while(iGLwindowIsMade == 1) {    
        if(iDoThis == 0) {
            _sleep(10);
        } else {
            if(iDoThis == 1) {
                draw_scene();
                iStopRendering = 0;
                iDoThis = 0;
            }
            if(iDoThis == 2) {
                // Shut down OpenGL environment IF last client
                iNClients--;
                if(iNClients == 0) {
#ifdef USE_WGL
                    void deleteGLPallet();
                    wglMakeCurrent(hDC, NULL);
                    wglDeleteContext(hRC);
                    if(m_hLocalWnd == NULL) { DestroyWindow(hWnd); }
                    deleteGLPallet();
#endif
                    isubcur = -1;         /* current sub-brick loaded into bobsub */
                    sprintf(m_cOldHVname, "none");
                    iGLwindowIsMade = 0;
                    dhpReadFromFileNoBuf(lpBuffer_of_iCache[0], "close", 0, 0);
                }
                iDoThis = 0;
            }
            if(iDoThis == 3) {
                DoSetOption();
                iDoThis = 0;
            }
            if(iDoThis == 4) {
                DoPutPix();
                iDoThis = 0;
            }
            if(iDoThis == 5) {
                DoGetPix();
                iDoThis = 0;
            }
        }

//        if(my_work->DoRender()) { my_work->draw_scene(); }
//        else                    { SleepEx(10, FALSE); }
    }
#endif
}

#ifdef USE_GLUT
void Chvr_work::GLUTdraw()
{
    DWORD dhpReadFromFileNoBuf(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes);

    // Initialize window if not already done so
    if(iGLwindowIsMade == 0) {
//        init();
        iGLwindowIsMade = 1;
        isubcur = -1;         /* declare that there is no current sub-brick loaded into bobsub */
        sprintf(m_cOldHVname, "none");
    }

    // Do something if called in responce to a request
    int iGLUTwork = 1;
    while(iGLUTwork == 1) {
        while(iDoThis == 0) { _sleep(10) ; }
        if(iDoThis == 1) {
            draw_scene();
            //iGLUTwork = 0;
            iStopRendering = 0;
            iDoThis = 0;
        }
        if(iDoThis == 2) {
            isubcur = -1;         /* declare that there is no current sub-brick loaded into bobsub */
            sprintf(m_cOldHVname, "none");
            iGLwindowIsMade = 0;
            dhpReadFromFileNoBuf(lpBuffer_of_iCache[0], "close", 0, 0);
            iDoThis = 0;
            iGLUTwork = 0;
        }
        if(iDoThis == 3) {
            DoSetOption();
            iDoThis = 0;
        }
        if(iDoThis == 4) {
            DoPutPix();
            iDoThis = 0;
        }
        if(iDoThis == 5) {
            DoGetPix();
            iDoThis = 0;
        }
    }
}
#endif

STDMETHODIMP Chvr_work::StartUp()
{    
#ifdef linux
    pthread_t thread_render, thread_io;
#endif
    /* Register that there is one more client for HVR server */
    iNClients++;

    /* If redering thread is already up then no need to
       start rendering or IO threads, so return */
    if(iGLwindowIsMade == 1) { return S_OK; }
#ifdef linux
    savethis = this;
#endif
    if(iSpawnRenderThread == 0) {
        /* Initialize OpenGL context and return */
        OpenAWindow((void *) this);
    } else {
        /* Start a rendering thread */
#ifdef linux
        pthread_create(&thread_render, NULL, (void*(*)(void*))OpenAWindow, (void *) this);
#else
        _beginthread(OpenAWindow, 0, (void *) this);
#endif
    }

    /* Wait untill window is created */
    while(iGLwindowIsMade != 1) { _sleep(10); }

    // Start IO threads
    int i;
    for(i=0; i<iN_IO_Threads; i++) { 
#ifdef linux
        pthread_create(&thread_io, NULL, (void*(*)(void*))StartIOThread, (void *) this);
#else
         _beginthread(StartIOThread, 0, (void *) this);
#endif
    }
    
    return S_OK;
}
/* subtile io stuff */
char *seps = " ,;\t\n";
int nsub, node0;
__int64 tile_off64t[MAXNODES];
int tile_nrep[MAXNODES];
int tile_nxsub[MAXNODES], tile_nysub[MAXNODES], tile_nzsub[MAXNODES];
int tile_ixoff[MAXNODES], tile_iyoff[MAXNODES], tile_izoff[MAXNODES];
int tile_parent[MAXNODES];
int tile_maxerr[MAXNODES];
int ncPath;
char tile_file[MAXNODES][64], tile_path[64], tile_full[256];
char line[256];
int tile_children[MAXNODES][32];
float fLimits[6];
/* subtile io stuff */

/* Cg Debugging Method */
#ifdef DEBUG_CG
void CheckCgError(void) 
{   
    CGerror error = cgGetError();

    if (error != CG_NO_ERROR)
    {
        fprintf(error_log,"CG error: %s\n", cgGetErrorString(error));
        exitHVR(CG_FAILED);
    }
}
#endif

#ifdef INTELLIGENT_OVERWRITE
int NextSlot(int tex_slot, int cache_size)
{
    int slot = tex_slot;
    
    slot++;
    if(slot == cache_size)
        slot = 0;

    return slot;
}

void setPrevCount(int cache_size)
{
    prev_count = count - 1;
    if (prev_count < 0)
        prev_count = cache_size - 1;
}
#endif

void Chvr_work::draw_scene()
{
    float eye[3], up[3], torq[3], eps, eps1;
    int i, j, joff, k, ic, ib, jb, kb, ijk, nslices, ntiles, ididread;
    int isub, jsub, ixoff, iyoff, izoff, itick, jlutmax;
    int i1, i2, iLength, fd;
    int nxsub, nysub, nzsub, nr, iprint, iTexBndry;
    int nxsb2, nysb2, nzsb2;
    int isoutside_n, isoutside_f, isoutside_t;
    int isoutside_b, isoutside_l, isoutside_r;
    // int fRBits, fGBits, fBBits, fABits;
    // float alphafac;

    float facXtoN, facYtoN, facZtoN, fVoxels;
    float subvolsiz[3], subminpos[3], fulminpos[3];
    float submaxpos[3], submintex[3], submaxtex[3];
    double aa[4];
    float alpha, alpha0, alpha_d;
    float xx, yy, zz, dxx, dyy, dzz, axx, ayy, azz, dz, dx, dy;
    int   idir, ids0, idsd;
    // int isBehind;
    //float vvvmin, vvvmax, vvvcent, vhhmin, vhhmax, vhhcent;
    float vhh, vnn, vvv;
    float vnormi;
    float xxv, yyv, zzv;
    // float dsmax, dsmin, sfinish,ss, dsfar, dsnear, fac;
    float edgedist, edgemin, edgemax;
    float ss,ds,sfinish,hdistfac,vdistfac, alphamin;
    float eyepos[3],cubesz[3];
    char old_amap[1024] = "default";
    // char old_hvname[1024] = "none";
    DWORD dwGot;
    char str[256], str_tmp[256];

    GLdouble dleft,dright,dbottom,dtop, dnearclip,dfarclip;

/* frustum stuff */
    void crossprod(float vec1[3], float vec2[3], float vec3[3]);
/* frustum stuff */

/* AuxClip stuff */
    int iClipThis, iClipThis0;
    float px0, px1, py0, py1, pz0, pz1, px0c, px1c, py0c, py1c, pz0c, pz1c;
    float xx0c, xx1c, yy0c, yy1c, zz0c, zz1c, zz0, zz1, fNvoxIn;
/* AuxClip stuff */

/* subtile stuff */
    float fminx,fminy,fminz,distmax, tile_dist[MAXNODES];
    int ippp, iccc, nccc, nlist, list_tiles[MAXNODES], tile_did[MAXNODES];
    int iNTileRender, iTileRender[MAXNODES];
    int inode, ilist, nlistnew, need_check, list_new[MAXNODES];
    int ixx, iyy, izz, jparent, nrj;
    float delxmin, delx, dely, delz, xmin, xmax, ymin, ymax, zmin, zmax;
    float distmin, delxtol, delxsize;
/* subtile stuff */

    int isallblack, length;
    char hostname[256]=" ", allblackfile[256];

/* subtile io stuff */
      FILE *fpp, *fpdhp;
//      FILE *fplim;
/* subtile io stuff */

/* IO stuff */
    int nbytes;
//    int nbin;
//    FILE *fdd;
    __int64 offset64t;
    int iOpac, iRGB, iALP;
    char tile_fileO[256];
    unsigned char *bobin;
/* IO stuff */

/* multiple pannel stuff */
    int   ip;
    int pan_idir[NPMAX];
    float d, dhorz, dvert, pcn, pan_cenmagi[NPMAX];
    float pan_eye[NPMAX][3],pan_up[NPMAX][3],pan_center[NPMAX][3];
    float pan_lookat[NPMAX][3],pan_torq[NPMAX][3],pan_dlef[NPMAX];
    float pan_drit[NPMAX],pan_dbot[NPMAX],pan_dtop[NPMAX];
    float pan_lefpar[NPMAX][3],pan_ritpar[NPMAX][3],pan_toppar[NPMAX][3];
    float pan_botpar[NPMAX][3],pan_lef_o[NPMAX][3],pan_rit_o[NPMAX][3];
    float pan_top_o[NPMAX][3], pan_bot_o[NPMAX][3];
/* multiple pannel stuff */

/* slice stuff */
    int islice0, islice1, isliceoff0, isliceoff1;
    float dx1, dy1, dz1, ppv0, ppv1, wt0;
    unsigned long lrgba0[256], lrgba1[256];
    int islicesize, islice, isliceoff;
    float xx0, xx1, yy0, yy1;
    float fZero, fOne, fMOne, xxv0, xxv1, yyv0, yyv1;
    int ixyz, ipm, iFace[3][2], iDrawEmAll;
    float x45p, x45m, y45p, y45m, z45p, z45m, px0u, py0u, pz0u, px1u, py1u, pz1u, deltadir;
    float xx0u, yy0u, zz0u, xx1u, yy1u, zz1u;

//    GLfloat fCurRastPos[4];
//    float fSMid, fLowLef[3], fLowRit[3], fTopLef[3], fTopRit[3];
/* slice stuff */

/* Statistics stuff (stats) */
    int ilo, ihi, jlo, jhi, klo, khi, iloc, ihic, jloc, jhic, kloc, khic;
    int i0, iPDF[256], iCross[256*256];
    long lBlack, lWhite, lGreen, lCrossLUT[256];
    int   ind[10] = {    0,     1,    33,    65,    97,   129,   161,   193,   225, 256  };
    double red[10] = {255.9,   0.0,   0.0,   0.0, 160.0, 160.0,   0.0,   0.0,   0.0, 255.9};
    double grn[10] = {255.9,   0.0, 160.0,   0.0,   0.0, 160.0, 160.0, 255.9,   0.0,   0.0};
    double blu[10] = {255.9,   0.0,   0.0, 160.0,   0.0,   0.0, 160.0,   0.0, 255.9,   0.0};
    float fRed, fGrn, fBlu;
    int jMax, iMaxCounts, iTotalCounts;
    double fDelX, fPDFfactor, f10PDFmax;
    float xPDFmin, xPDFmax;
    int i10PDFmax;
//HFONT hfnt, hfntPrev; 
//PLOGFONT plf;
//RECT rc;
/* Statistics stuff (stats) */

/* subvolume edges stuff */
    int isedge, nedges, nverts, ix, iy, iz, jx, jy, jz, iside, jdir;
    float edges[100][2][3], tvert[6][3], pos0[3], range_side_dir[2][3];
    float dot, ssmin, ssmax;
    float bound[2][3];
/* subvolume edges stuff */

/* prototyping functions */
    int match( char *token, char *word, int tokenlength );
    float dotproduct( const float*, const float* );
    void vnormalize( float v[3] );
    int section(float pos0[3], float norm0[3], float norm1[3], float norm2[3],
                float edges[100][2][3], int nedges, float tvert[6][3] );
    int   strdif(char*, char*);
    void  setstr(char*, char*);
    int readCmap( char *clrname, char *alpname, float ftable[256][4] );
    void TposeYZ( unsigned char *a1, unsigned char *a2, int *nx2, int *ny2, int *nz2,
                                                        int *nx,  int *ny,  int *nz  );
    void TposeXZ( unsigned char *a1, unsigned char *a2, int nx2, int ny2, int nz2,
                                                        int nx,  int ny,  int nz  );
    DWORD dhpReadFromFileNoBuf(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes);
    DWORD  dhpReadFromFileSync(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes);
    DWORD  dhpReadFromOpenFile(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes,
    int iFile, int iClose);
    void annotation(int iBefore, float eyepos[3]);
    void drawsurfaces(int isub, float ss1, float ss2, float eye[3]);
    void genvectorlist(int node0, int isub, float *subminpos, float *submaxpos);
    void setvectors();
/* prototyping functions */

    int iDoAnnotationBeforeRendering;
    SIZE_T dwNeedSize, dwMaxNeededSize;
    FILE *fp;
    #ifdef IDEBUG
    printf("At top of draw_scene\n");
    #endif
    #ifdef WINDOWS
    swprintf(wcErrorStatus, L"OK");
    #else
    swprintf(wcErrorStatus, "OK");
    #endif
    unsigned char *ucDebug;
    int iDebug;

    status = SUCCESS;

    /* Copy from GUI input struct to hvr use struct
    Set all iNew flags to the status of NOT new. */
    CopyHvrInputs();

    if(m_iTimersOn == 1) {
        tsEvent.SetOutputFileName(m_sTimerFile);
        tsEvent.StampTime("Begin Rendering");
    }

    if(m_iTimersOn == 1) { tsEvent.StampTime("Begin Setup"); }

    GLuint dataTexture, RGBATexture, LUTTexture;
    
    /* Resident Textures */
    GLubyte *texData = NULL;
    GLint resident = GL_FALSE;
    int texsize = 0;
    int texmem_used = 0;

    /* Frame Specific TCS Efficiency Counters */
    int nBricksInFrustum = 0; /* total # of bricks in frustum */
    int nBricksRead = 0; /* # of bricks read from disk */
    int nBricksRendered = 0; /* # of bricks rendered */
    int nBricksLoaded = 0; /* # of bricks loaded into texture memory and rendered */
    int nBricksFromCache = 0; /* # of bricks rendered from texture memory (no load) */
    int pctFrustumVolume = 0; /* frustum's % of total volume (by # of bricks) */
    int pctRenderedFromCache = 0; /* % of bricks rendered which were from texture cache */
    int pctCacheUsed = 0; /* % of bricks in cache which were used for rendering */
    int pctTCSEfficient = 0; /* % efficiency of Texture Caching Scheme */

    // fprintf(stderr, ">>>DHP<<<: at top of draw_scene\n");

    if(TexScheme == 1)
        texmem = 0;
    
    /* Cg Initialization */
    if(m_bColorTableIsAvailable)
    {    
        TexMode++;
        
        /* Do all of this only once */
        if(!CgInit)
        {           
            #ifdef DEBUG_TEX
                #ifdef WRITE_STD_OUT
                    fprintf(error_log,"Initializing Cg\n");
                #else
                    printf("Initializing Cg\n");
                #endif
            #endif
            
            /* Destroy Cg Context if it already exists */
            if(cgIsContext(Context)) {
                cgDestroyContext(Context);
                #ifdef DEBUG_CG
                CheckCgError();
                #endif
            }
            
            /* Create Cg Context */
            Context = cgCreateContext();
            #ifdef DEBUG_CG
            CheckCgError();
            #endif

            /* Get Optimal Profile, Options */
            fProfile = cgGLGetLatestProfile(CG_GL_FRAGMENT);
            cgGLSetOptimalOptions(fProfile);
            #ifdef DEBUG_CG
            CheckCgError();
            #endif
            
            /* Create Cg Programs */
            if(Precompile) {
                if(Tex3d)
                    fProgram = cgCreateProgram(Context,CG_OBJECT,CGobject,fProfile,"main",NULL);
                else
                    fProgram = cgCreateProgram(Context,CG_OBJECT,CGobject_2d,fProfile,"main",NULL);
            }
            else {
                if(Tex3d)
                    fProgram = cgCreateProgram(Context,CG_SOURCE,CGsource,fProfile,"main",NULL);
                else
                    fProgram = cgCreateProgram(Context,CG_SOURCE,CGsource_2d,fProfile,"main",NULL);
            }
            #ifdef DEBUG_CG
            CheckCgError();
            #endif

            /* Load Cg Programs */
            cgGLLoadProgram(fProgram);
            #ifdef DEBUG_CG
            CheckCgError();
            #endif

            /* Get Handles to Cg Parameters */
            textureHandle = cgGetNamedParameter(fProgram, "data");
            textureHandle2 = cgGetNamedParameter(fProgram, "LUT");
            #ifdef DEBUG_CG
            CheckCgError();
            #endif

            #ifdef DEBUG_TEX
                #ifdef WRITE_STD_OUT
                    fprintf(error_log,"Creating 1D LUT Texture\n");
                #else
                    printf("Creating 1D LUT Texture\n");
                #endif
            #endif
            
            /* Create and Initialize 1D Texture (for color LUT) */
            glDeleteTextures(1, &LUTTexture);
            glGenTextures(1, &LUTTexture);
            glBindTexture(GL_TEXTURE_1D, LUTTexture);
            glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP);
            glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
            glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        }
        
        #ifdef BGRA_LUT
        /* Create Array in BGRA format instead of RGBA */
        for(i=0; i<256; i++) 
        {    m_lrgba[i] =              (unsigned long)(255.99*frgba[i][2]) +
                (unsigned long)256 * ((unsigned long)(255.99*frgba[i][1]) +
                (unsigned long)256 * ((unsigned long)(255.99*frgba[i][0]) +
                (unsigned long)256 *  (unsigned long)(255.99*frgba[i][3])));
        }
        #endif

        /* Bind 1D Texture Object, Set Cg Texture Parameter */
        glBindTexture(GL_TEXTURE_1D,LUTTexture);
        cgGLSetTextureParameter(textureHandle2, LUTTexture);

        /* Load Color LUT/1D Texture into Texture Memory */
        #ifdef BGRA_LUT
        if(!CgInit) {
            glTexImage1D(GL_TEXTURE_1D,0,GL_RGBA8,256,0,GL_BGRA,GL_UNSIGNED_BYTE,m_lrgba);
            CgInit = TRUE;
            #ifdef DEBUG_TEX
            printf("Loading LUT into graphics memory\n");
            #endif
        }
        else {
            #ifdef DEBUG_TEX
            printf("Updating LUT data in graphics memory\n");
            #endif
            glTexSubImage1D(GL_TEXTURE_1D,0,0,256,GL_BGRA,GL_UNSIGNED_BYTE,m_lrgba);
        }
        #else
        if(!CgInit) {    
            glTexImage1D(GL_TEXTURE_1D,0,GL_RGBA8,256,0,GL_RGBA,GL_FLOAT,frgba);
            CgInit = TRUE;
            #ifdef DEBUG_TEX
                #ifdef WRITE_STD_OUT
                    fprintf(error_log,"Loading LUT into graphics memory\n");
                #else
                    printf("Loading LUT into graphics memory\n");
                #endif
            #endif
        }
        else {
            #ifdef DEBUG_TEX
                #ifdef WRITE_STD_OUT
                    fprintf(error_log,"Loading LUT into graphics memory\n");
                #else
                    printf("Updating LUT data in graphics memory\n");
                #endif
            #endif
            glTexSubImage1D(GL_TEXTURE_1D,0,0,256,GL_RGBA,GL_FLOAT,frgba);
        }
        #endif
    }

    iprint = 0;

    if(Renderbuffer) {
        /* If size of image has changed, create new FBO
           with Renderbuffer object of the new size */
        if((rb_width != m_ShvrUse.iWidth) || (rb_height != m_ShvrUse.iHeight)) {
            size_change = TRUE;
            #ifdef linux
            free(ucGPix);
            ucGPix = (unsigned char *)malloc(3*m_ShvrUse.iWidth*m_ShvrUse.iHeight);
            #else
            VirtualFree(ucGPix, 3*rb_width*rb_height, MEM_DECOMMIT | MEM_RELEASE);
            ucGPix = (unsigned char *)VirtualAlloc(NULL,3*m_ShvrUse.iWidth*m_ShvrUse.iHeight,
                                                       MEM_COMMIT, PAGE_READWRITE);
            #endif
            rb_width = m_ShvrUse.iWidth;
            rb_height = m_ShvrUse.iHeight;
            new_FBO();
        }
        else
            size_change = FALSE;
    }

    if(ResidentTextures)
        TexMode++;
    if(Tex3d)
        TexMode+=2;
    
    if(TexMode != Prev_TexMode) {
        loopcount = 0;
        TexScheme = 0;
    }
    
    /* Force re-read of all data if doing cross-correlation */
    if(strcmp(m_sStats, "cross") == 0) { sprintf(m_cOldHVname, "none"); }

    /* Read in, or reset VECTOR data, as needed */
    setvectors();

    dwMaxNeededSize = 0;
    if(strdif(m_cOldHVname, m_ShvrUse.sHVname) == 1 || dwMaxNeededSize < 10) {
    max_tex_needed = 0;
    // if(strdif(m_cOldHVname, m_ShvrUse.sHVname) != 987654321)
      /* If new data set, clear vid memory and 
         re-initialize textures below */

#ifdef AUTO_CACHE
      if(AutoCaching) {
          if(ResidentTextures) {
              #ifdef DEBUG_TEX
              printf("Turning texture caching OFF\n");
              #endif
              ResidentTextures = FALSE;
          }
      }
#endif

      /* dhp: hvfile */
      FlushCacheBuffers();
      setstr(m_cOldHVname, m_ShvrUse.sHVname);
      fpp = fopen(m_ShvrUse.sHVname, "r");
      if(fpp == NULL) {
        if(error_log != NULL) {
          fprintf(error_log,"FAILED TO OPEN HV FILE: %s\n",m_ShvrUse.sHVname);
        }
        status = HV_FAILED;
        tile_nrep[0] = 1;
        tile_nxsub[0]  = 64;
        tile_nysub[0]  = 64;
        tile_nzsub[0]  = 64;
        tile_ixoff[0]  =  0;
        tile_iyoff[0]  =  0;
        tile_izoff[0]  =  0;
        tile_parent[0] = -1;
        tile_maxerr[0] =  0;
        m_iNBoundaries = 0;
        for(i=0; i<128; i++) { m_iNBoundariesRep[i] = 0; }
        nsub = 1;
        node0 = 0;
        isubcur = 0;
        int ith = 2;
        bobsubZ = bobsubZ0;
        for(k=0; k<tile_nzsub[0]; k++) { kb=0; if(k<ith || k>tile_nzsub[0]-ith-1) { kb = 1; }
        for(j=0; j<tile_nysub[0]; j++) { jb=0; if(j<ith || j>tile_nysub[0]-ith-1) { jb = 1; }
        for(i=0; i<tile_nxsub[0]; i++) { ib=0; if(i<ith || i>tile_nxsub[0]-ith-1) { ib = 1; }
            ijk = i + tile_nxsub[0] * (j + tile_nysub[0] * k);
            if(kb+jb+ib > 1) { 
                bobsubZ[ijk] = (unsigned char)(127 + 
                    (int)((float)128 * ((float)i/(float)tile_nxsub[0] +
                                        (float)j/(float)tile_nysub[0] +
                                        (float)k/(float)tile_nzsub[0]   ) / (float)3));
            } else {
                bobsubZ[ijk] = (unsigned char)0;
            }
        }}}
        for(ijk=0; ijk<tile_nxsub[0]*tile_nysub[0]*tile_nzsub[0]; ijk++) {
            bobsubY[ijk] = bobsubZ[ijk];
        }
        bSetBobX = FALSE;
        bSetBobY = FALSE;

        fLimits[0] = -(float)1.0 * m_fScaleFactors[0];
        fLimits[1] =  (float)1.0 * m_fScaleFactors[0];
        fLimits[2] = -(float)1.0 * m_fScaleFactors[1];
        fLimits[3] =  (float)1.0 * m_fScaleFactors[1];
        fLimits[4] = -(float)1.0 * m_fScaleFactors[2];
        fLimits[5] =  (float)1.0 * m_fScaleFactors[2];
      } else {
        for(i=0; i<128; i++) { m_iNBoundariesRep[i] = 0; }
        nsub = 0;
        node0 = -1;
        while (node0 < 0) {
            fgets(line,200,fpp);
            i = sscanf(line, "%s %lld %d %d %d %d %d %d %d %d %d",
               tile_file[nsub],    &tile_off64t[nsub], &tile_nrep[nsub],
               &tile_nxsub[nsub],  &tile_nysub[nsub],  &tile_nzsub[nsub],
               &tile_ixoff[nsub],  &tile_iyoff[nsub],  &tile_izoff[nsub],
               &tile_parent[nsub], &tile_maxerr[nsub]                     );

            if(i < 11) {
              fclose(fpp);
              if(error_log != NULL) {
                fprintf(error_log,"BAD HV FILE: %s\n",m_ShvrUse.sHVname);
              }
              status = HV_FILE_BAD;
              write(STDOUT,&status,4);
#ifdef DEBUG_NEW_TREE
fprintf(stderr, "XXX bad hv file\n");
#endif
              return;
            }

            if(tile_parent[nsub] < 0) { node0 = nsub; }
            if(tile_ixoff[nsub]%tile_nxsub[nsub] != 0) { m_iNBoundariesRep[tile_nrep[nsub]] = 1; }
/*
Since only tile boundaries of 0 or 1 are used, the above should work in general

This, below, worked when tiles filled the full domain at all resolutions
            if( m_iNBoundariesRep[tile_nrep[nsub]] < -tile_ixoff[nsub] ) {
                m_iNBoundariesRep[tile_nrep[nsub]] = -tile_ixoff[nsub];
            }
*/

            /* Get Max memory needy for any brick in this hv file.
               This wiil get used below to determine if bob arrays
               need to be resized */
            nxsub = tile_nxsub[nsub];
            nysub = tile_nysub[nsub];
            nzsub = tile_nzsub[nsub];
            nxsb2 = 2;   while(nxsb2 < nxsub) { nxsb2 *= 2; }
            nysb2 = 2;   while(nysb2 < nysub) { nysb2 *= 2; }
            nzsb2 = 2;   while(nzsb2 < nzsub) { nzsb2 *= 2; }
            dwNeedSize = (SIZE_T)(nxsb2 * nysb2 * nzsb2);

            if(max_tex_needed < max_tex_mem+10) { max_tex_needed += (int)dwNeedSize; }
            
            /* Resident Textures */
            if(dwMaxNeededSize < dwNeedSize)
            {   /* Get Max Tex Dimensions */
                dwMaxNeededSize = dwNeedSize;
                max_nxsb2 = nxsb2;
                max_nysb2 = nysb2;
                max_nzsb2 = nzsb2;
            }
            nsub++;
        }
#ifdef DEBUG_NEW_TREE
fprintf(stderr, "node0 = %d\n", node0);
#endif

        // Allow user to RENAME an HV file
        // IF 1st tilefile is an HV file (i.e., file name has ".hv" extention)
        // THEN replase all tile_file names with the HV meta data file name

        ncPath = 0;
        iLength = strlen(m_ShvrUse.sHVname);
        for(i=0; i<iLength; i++) {
            if(m_ShvrUse.sHVname[i] ==  '/') { ncPath = i+1; }
            if(m_ShvrUse.sHVname[i] == '\\') { ncPath = i+1; }
        }

        length = strlen(tile_file[0]);
        if(strcmp(&tile_file[0][length-3],                      ".hv") == 0 &&
           strcmp(&tile_file[0][length-3], &m_ShvrUse.sHVname[ncPath]) != 0    ) {
          // fprintf(stderr, "OLD NAME: %s\n", tile_file[0]);
          // fprintf(stderr, "NEW NAME: %s\n", &m_ShvrUse.sHVname[ncPath]);
          // fprintf(stderr, "NEW FULL PATH: %s\n", m_ShvrUse.sHVname);
          for(i=0; i<nsub; i++) { sprintf(tile_file[i], "%s", &m_ShvrUse.sHVname[ncPath]); }
        }

        hv_count = 0;
        loopcount = 0; 
        TexScheme = 0;
#ifdef NEWTEX
        if(old_max_nxsb2 != max_nxsb2 || old_max_nysb2 != max_nysb2 || old_max_nzsb2 != max_nzsb2)
            reload = TRUE;
        else
            reload = FALSE;

        old_max_nxsb2 = max_nxsb2;
        old_max_nysb2 = max_nysb2;
        old_max_nzsb2 = max_nzsb2;
#endif

        // Set default limits
        /* modelview aspect for the entire mesh is based on X size. */
        m_iNBoundaries = m_iNBoundariesRep[tile_nrep[node0]];
        Nx = tile_nrep[node0] * (tile_nxsub[node0] - 2*m_iNBoundaries);  /* These will need to be generalized */
        Ny = tile_nrep[node0] * (tile_nysub[node0] - 2*m_iNBoundaries);  /* to be based on highest res tiles. */
        Nz = tile_nrep[node0] * (tile_nzsub[node0] - 2*m_iNBoundaries);

        fLimits[0] =  -(float)1.0           * m_fScaleFactors[0];
        fLimits[1] =   (float)1.0           * m_fScaleFactors[0];
        fLimits[2] = -((float)Ny/(float)Nx) * m_fScaleFactors[1];
        fLimits[3] =  ((float)Ny/(float)Nx) * m_fScaleFactors[1];
        fLimits[4] = -((float)Nz/(float)Nx) * m_fScaleFactors[2];
        fLimits[5] =  ((float)Nz/(float)Nx) * m_fScaleFactors[2];

        /* Set limits to those specified in hv file */
        if(match(fgets(line,200,fpp), "XYZ_Limits", 10) == 1) {
            strncpy(str_tmp, strtok( line,seps ), 64);
            fLimits[0] = (float)atof( strtok(NULL,seps) ) * m_fScaleFactors[0];
            fLimits[1] = (float)atof( strtok(NULL,seps) ) * m_fScaleFactors[0];
            fLimits[2] = (float)atof( strtok(NULL,seps) ) * m_fScaleFactors[1];
            fLimits[3] = (float)atof( strtok(NULL,seps) ) * m_fScaleFactors[1];
            fLimits[4] = (float)atof( strtok(NULL,seps) ) * m_fScaleFactors[2];
            fLimits[5] = (float)atof( strtok(NULL,seps) ) * m_fScaleFactors[2];
            /*fplim = fopen("hvr_limits.txt", "w");
              fprintf(fplim, "Limits X: %f %f\n", fLimits[0], fLimits[1]);
              fprintf(fplim, "Limits Y: %f %f\n", fLimits[2], fLimits[3]);
              fprintf(fplim, "Limits Z: %f %f\n", fLimits[4], fLimits[5]);
              fclose(fplim);*/
        }
        fclose(fpp);
        isubcur = -1;         /* current sub-brick loaded into bobsub */

        /* Check and fix tree structure if needed */
        if(tile_nzsub[node0] > 1) { for(i=0; i<nsub; i++) { if(i != node0) {
            nr = tile_nrep[i];
            ixx = nr*(tile_ixoff[i]+tile_nxsub[i]/2);
            iyy = nr*(tile_iyoff[i]+tile_nysub[i]/2);
            izz = nr*(tile_izoff[i]+tile_nzsub[i]/2);
            jparent = -1;
            for(j=0; j<nsub; j++) { if(tile_nrep[j] == 2*nr) {
                nrj = tile_nrep[j];
                if((nrj*tile_ixoff[j]-ixx)*(nrj*(tile_ixoff[j]+tile_nxsub[j])-ixx) < 0
                && (nrj*tile_iyoff[j]-iyy)*(nrj*(tile_iyoff[j]+tile_nysub[j])-iyy) < 0
                && (nrj*tile_izoff[j]-izz)*(nrj*(tile_izoff[j]+tile_nzsub[j])-izz) < 0) {
                    jparent = j;
                }
            }}
            if(jparent < 0) {
#ifdef DEBUG_NEW_TREE
                fprintf(stderr, "XXX jparent<0\n");
                fprintf(stderr, "File = %s\n", m_ShvrUse.sHVname);
                fprintf(stderr, "no parent for node %i\n", i);
                fprintf(stderr, "no parent for node %i\n", i);

                fprintf(stderr, "nr=%d  i[xyz]off= %d %d %d\n",
                    nr,tile_ixoff[i],tile_iyoff[i],tile_izoff[i] );
#endif
                status = HV_FILE_BAD;
                return;
            }
            if(jparent != tile_parent[i]) {
                /*fpp = fopen("C:/DEBUG_HVR.txt", "a+");
                fprintf(fpp, "File = %s\n", m_ShvrUse.sHVname);
                fprintf(fpp, "Incorrect parent for node %i\n", i);
                fprintf(fpp, "nr=%d  i[xyz]off= %d %d %d\n",
                    nr,tile_ixoff[i],tile_iyoff[i],tile_izoff[i] );

                fprintf(fpp, "Incorrect parent in file is:\n");
                j = tile_parent[i];
                fprintf(fpp, "node=%d   nr=%d  i[xyz]off= %d %d %d\n",
                    j, nr,tile_ixoff[j],tile_iyoff[j],tile_izoff[j] );

                fprintf(fpp, "Correct parent according to subdomains is:\n");
                j = jparent;
                fprintf(fpp, "node=%d   nr=%d  i[xyz]off= %d %d %d\n",
                    j, nr,tile_ixoff[j],tile_iyoff[j],tile_izoff[j] );

                fclose(fpp);*/
                tile_parent[i] = jparent;
            }
        }}}
      }
    }
#ifdef AUTO_CACHE
    else {
        if(AutoCaching) {
            if(!ResidentTextures) {
                // printf("Turning texture caching ON\n");
                loopcount = 0; 
                TexScheme = 0;
                ResidentTextures = TRUE;
            }
        }
    }
#endif

    // There appears to be a bug dhp: 2009 Apr 24
    if(max_tex_needed > max_tex_mem) { ResidentTextures = FALSE; }

    // fprintf(stderr, ">>>DHP<<<: Turning ResidentTextures OFF\n");
    // ResidentTextures = FALSE;


    /* Resize bob arrays if needed */
    if(dwMaxSize < dwMaxNeededSize) {
#ifdef linux
        free(bobsubX);
        free(bobsubY);
        free(bobsubZ0);
        free(bobOX);
        free(bobOY);
        free(bobOZ);

        dwMaxSize = dwMaxNeededSize;

        bobsubX  = (unsigned char *)malloc( dwMaxSize );
        bobsubY  = (unsigned char *)malloc( dwMaxSize );
        bobsubZ0 = (unsigned char *)malloc( dwMaxSize );
        bobOX    = (unsigned char *)malloc( dwMaxSize );
        bobOY    = (unsigned char *)malloc( dwMaxSize );
        bobOZ    = (unsigned char *)malloc( dwMaxSize );
#else
        VirtualFree(bobsubX  , dwMaxSize, MEM_DECOMMIT | MEM_RELEASE);
        VirtualFree(bobsubY  , dwMaxSize, MEM_DECOMMIT | MEM_RELEASE);
        VirtualFree(bobsubZ0 , dwMaxSize, MEM_DECOMMIT | MEM_RELEASE);
        VirtualFree(bobOX    , dwMaxSize, MEM_DECOMMIT | MEM_RELEASE);
        VirtualFree(bobOY    , dwMaxSize, MEM_DECOMMIT | MEM_RELEASE);
        VirtualFree(bobOZ    , dwMaxSize, MEM_DECOMMIT | MEM_RELEASE);

        dwMaxSize = dwMaxNeededSize;

        bobsubX  = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
        bobsubY  = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
        bobsubZ0 = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
        bobOX    = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
        bobOY    = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
        bobOZ    = (unsigned char *)VirtualAlloc( NULL, dwMaxSize, MEM_COMMIT, PAGE_READWRITE);
#endif
        bBobsAllocated = TRUE;
    }

    /* Test for an independent opacity field: switch is RGB in name of hv file */
    /* 01234567890 v72RGB_wa0123.hv */
    if(    tile_file[0][3] == 'R' &&
        tile_file[0][4] == 'G' &&
        tile_file[0][5] == 'B'        ) { iOpac = 1; } else { iOpac = 0; }

   /************************************************************************
    *  Generate directory (path) to specified HV file                      *
    ncPath = 0;
    iLength = strlen(m_ShvrUse.sHVname);
    for(i=0; i<iLength; i++) {
        if(m_ShvrUse.sHVname[i] ==  '/') { ncPath = i+1; }
        if(m_ShvrUse.sHVname[i] == '\\') { ncPath = i+1; }
    }
    ************************************************************************/
    for(i=0; i<ncPath; i++) { tile_path[i] = m_ShvrUse.sHVname[i]; }
    tile_path[ncPath] = '\0';

   /************************************************************************
    *  When current directory changes, look to see if there are other dirs *
    *  Put paths to these other directories into cOtherDirs                *
    ************************************************************************/
    if(strcmp(tile_path, cLastDir) != 0) {

        sprintf(cLastDir, tile_path);

        FILE *fpOtherDirs;
        char cOtherDirFile[256];
        sprintf(cOtherDirFile, "%sotherdirs.txt", tile_path);
        
        /*FILE *fp;
        fp = fopen("C:/DEBUG_HVR.txt", "a+");
        fprintf(fp, "otherdir file = %s\n", cOtherDirFile);
        fclose(fp);*/

        fpOtherDirs = fopen(cOtherDirFile, "r");
        if(fpOtherDirs != NULL) {
            /* Scan file for other directoies; */
            int iNewDir = 1;
            while(fscanf(fpOtherDirs, "%s", cOtherDirs[iNewDir]) > 0) {
                iNewDir++;
            }
            iNewDir--;

            /* if the new number of directories to read from is different from
               the current number of io threads, then must stop currently running
               io threads and start up a new set. */
            if(iNewDir+1 != iN_IO_Threads) {
                
                /*fp = fopen("C:/DEBUG_HVR.txt", "a+");
                fprintf(fp, "iNewDir = %d   iN_IO_Threads = %d\n",
                             iNewDir,       iN_IO_Threads);
                for(i=1; i <= iNewDir; i++) {
                    fprintf(fp, "%d -->%s<--\n", i, cOtherDirs[i]);
                }
                fclose(fp);*/

                /* Tell io therads to kill themselfs and wait for them to die! */
                iIO_Thread_Is_Up = 2;
                while(iIO_Thread_Is_Up != 0) { _sleep(50); }

                /*fp = fopen("C:/DEBUG_HVR.txt", "a+");
                fprintf(fp, "IO theads are killed\n");
                fclose(fp);*/

                /* Set new number of io threads and start them up */
                iN_IO_Threads = iNewDir+1;
                
                for(i=0; i<iN_IO_Threads; i++) {
#ifdef WINDOWS
                    _beginthread(StartIOThread, 0, (void *) this);
#else
                    pthread_t thread_io;
                    pthread_create(&thread_io, NULL, (void*(*)(void*))StartIOThread, (void *) this);
#endif
                }

                /* Wait for IO threads to all start up */
                while(iIO_Thread_Is_Up != 1) { _sleep(50); }
                
                /*fp = fopen("C:/DEBUG_HVR.txt", "a+");
                fprintf(fp, "IO are restarted\n");
                fclose(fp);*/

            }
        }
    }

    nbytes = m_ShvrUse.iHeight * m_ShvrUse.iWidth * 3;

    if((pannels_initted == 0) || size_change) {
      for(ip=0; ip<npannel; ip++) { rgbpannel[ip]  = (char*)malloc(nbytes); }
      pannels_initted = 1;
    }

    for(ip=0; ip<npannel; ip++) {
      for(k=0; k<nbytes; k++) { rgbpannel[ip][k] = 0; }
    }
    ipcur = -1;

   /************************************************************************
    *  Generate horizontal, vertical, and alpha scalings                   *
    ************************************************************************/
    vdistfac = (float)tan(0.5*Pid180 * m_ShvrUse.fFovAngle);
    hdistfac = vdistfac / m_ShvrUse.fAspectRatio;
    alpha0 = opacity * (float)128.0;

    dnearclip = m_ShvrUse.fNearClip;
    dfarclip  = m_ShvrUse.fFarrClip;
    //dx = hdistfac * dnearclip;
    //dy = vdistfac * dnearclip;

   /************************************************************************
    * Right handed ONB  which power-wall frame is attached to :            *
    * (eye[], up[], torq[])                                                *
    ************************************************************************/
    for(i=0; i<3; i++) {
        eye[i]    = m_ShvrUse.fCent[i] - m_ShvrUse.fEye[i];    /* direction eye looks in */
        up[i]     = m_ShvrUse.fUp[i];
        eyepos[i] = m_ShvrUse.fEye[i];
    }
    vnormalize(eye);

    /* Insist that up is normal to eye */
    d = dotproduct(eye,up);
    for (i=0; i<3; ++i) { up[i] = up[i]- eye[i]*d; }
    vnormalize( up );

    /* Construct the cross vector 'torque' */
    torq[0] = eye[1]*up[2] - eye[2]*up[1];
    torq[1] = eye[2]*up[0] - eye[0]*up[2];
    torq[2] = eye[0]*up[1] - eye[1]*up[0];

    /*fpp = fopen("rend_ONB.out", "w");
    fprintf(fpp, "NEW ONE\n");
    fprintf(fpp, "eye  = %f %f %f\n", eye[0], eye[1], eye[2]);
    fprintf(fpp, "up   = %f %f %f\n", up[0],  up[1],  up[2]);
    fprintf(fpp, "torq = %f %f %f\n", torq[0],torq[1],torq[2]);
      for(ip=0; ip<npannel; ip++) {
        fprintf(fpp, "ip = %d\n", ip);
        fprintf(fpp, "pw_normal= %f %f %f\n",
                pw_normal[ip][0],pw_normal[ip][1],pw_normal[ip][2]);
        fprintf(fpp, "pw_up    = %f %f %f\n",
                pw_up[ip][0],pw_up[ip][1],pw_up[ip][2]);
        fprintf(fpp, "pw_center= %f %f %f\n",
                pw_center[ip][0],pw_center[ip][1],pw_center[ip][2]);
      }

    fclose(fpp); */

    for(ip=0; ip<npannel; ip++) {
      for(i=0; i<3; i++) {
        pan_eye[ip][i]    = m_ShvrUse.fPanNorm[0] * torq[i] + 
                            m_ShvrUse.fPanNorm[1] *  eye[i] +
                            m_ShvrUse.fPanNorm[2] *   up[i];

        pan_up[ip][i]     = m_ShvrUse.fPanUp[0] * torq[i] + 
                            m_ShvrUse.fPanUp[1] *  eye[i] +
                            m_ShvrUse.fPanUp[2] *   up[i];

        pan_center[ip][i] = m_ShvrUse.fPanCent[0] * torq[i] + 
                            m_ShvrUse.fPanCent[1] *  eye[i] +
                            m_ShvrUse.fPanCent[2] *   up[i];

        pan_lookat[ip][i] = eyepos[i] + pan_eye[ip][i];
      }
      
      /* orientation dependent only on eye */
      /*dxx = eye[0];
      dyy = eye[1];
      dzz = eye[2];*/

      /* Normalize panel eye, up, and torq vectors */
      vnormalize(pan_eye[ip]);
      vnormalize(pan_up[ip]);
      crossprod(pan_eye[ip], pan_up[ip], pan_torq[ip]);
      vnormalize(pan_torq[ip]);

      /* Default orientation of rendering planes:
         used if to much of the subcube is behind you */
      dxx = pan_eye[ip][0];
      dyy = pan_eye[ip][1];
      dzz = pan_eye[ip][2];

      if(dxx > 0.) { axx = dxx; } else { axx = -dxx; }
      if(dyy > 0.) { ayy = dyy; } else { ayy = -dyy; }
      if(dzz > 0.) { azz = dzz; } else { azz = -dzz; }
      pan_idir[ip] = 2;                                     /* view along z-direction */
      if(ayy > axx  &&  ayy > azz) { pan_idir[ip] = 1; }    /* view along y-direction */
      if(axx > ayy  &&  axx > azz) { pan_idir[ip] = 0; }    /* view along x-direction */

     /***************************************************************
      * pan_center is position of the center (relative to the eye) of the pannel
      *   to be used in generating horizontal and vertical offset of
      *   pannel window
      * pan_lookat is a point in the direction normal to the pannel
      *   to be used in glLookAt
      * NOTE: the pannel must be centered on the normal to the eyeposition
      *       for pan_center & pan_lookat to be the same
      * NOTE: pan_center is NOT normalized
      *       ss * pan_dlef ... are shifts on a plane at a distance ss
      *       pan_cenmagi[ip] is the correction factor.
      ***************************************************************/
      pan_cenmagi[ip] = (float)1. / dotproduct(pan_eye[ip], pan_center[ip]);
      dhorz = dotproduct(pan_center[ip], pan_torq[ip]);
      dvert = dotproduct(pan_center[ip], pan_up[ip]);

      pcn = (float)1. / pan_cenmagi[ip];
      pan_dlef[ip] = (dhorz - m_ShvrUse.fHalfWidth ) * pan_cenmagi[ip];
      pan_drit[ip] = (dhorz + m_ShvrUse.fHalfWidth ) * pan_cenmagi[ip];
      pan_dbot[ip] = (dvert - m_ShvrUse.fHalfHeight) * pan_cenmagi[ip];
      pan_dtop[ip] = (dvert + m_ShvrUse.fHalfHeight) * pan_cenmagi[ip];

      /* fpp = fopen("rend_ONB.out", "a+");
      fprintf(fpp, "ip = %d\n", ip);
      fprintf(fpp, "Pid180= %f, fovangle=%f Aspect=%f\n", Pid180, m_ShvrUse.fFovAngle, m_ShvrUse.fAspectRatio);
      fprintf(fpp, "pan_drit= %f, dhorz=%f\n", pan_drit[ip], dhorz);
      fprintf(fpp, "half_width= %f, pan_cenmagi[ip]=%f\n", m_ShvrUse.fHalfWidth, pan_cenmagi[ip]);
      fprintf(fpp, "pan_up     = %f %f %f\n",pan_up[ip][0],pan_up[ip][1],pan_up[ip][2]);
      fprintf(fpp, "pan_torq   = %f %f %f\n",pan_torq[ip][0],pan_torq[ip][1],pan_torq[ip][2]);
      fprintf(fpp, "pan_center = %f %f %f\n",pan_center[ip][0],pan_center[ip][1],pan_center[ip][2]);
      fclose(fpp); */
    }

    /* modelview aspect for the entire mesh is based on X size. */
    m_iNBoundaries = m_iNBoundariesRep[tile_nrep[node0]];
    Nx = tile_nrep[node0] * (tile_nxsub[node0] - 2*m_iNBoundaries);  /* These will need to be generalized */
    Ny = tile_nrep[node0] * (tile_nysub[node0] - 2*m_iNBoundaries);  /* to be based on highest res tiles. */
    Nz = tile_nrep[node0] * (tile_nzsub[node0] - 2*m_iNBoundaries);

    cubesz[0]= (float)0.5 * (fLimits[1] - fLimits[0]);
    cubesz[1]= (float)0.5 * (fLimits[3] - fLimits[2]);
    cubesz[2]= (float)0.5 * (fLimits[5] - fLimits[4]);

    for(ip=0; ip<npannel; ip++) {
     /************************************************************************
      * Generate outward orthogonal surface vectors for 4 planes of the      *
      * viewing frustum.  They do not need to be normalized.                 *
      * These are used below to determine if a tile is OUTSIDE the frustum   *
      * Power wall change:  In principle, need to do these 4 planes for each *
      * pannel of a power wall.  In practice, it is actually easiest to, in  *
      * fact, do each of the pannels individuallys.  The condition on which  *
      * a tile is read in is simply that any part of the tile is in any one  *
      * of the pannels viewing frustum.                                      * 
      ************************************************************************/
      for(i=0; i<3; i++) {
        pan_lefpar[ip][i] = pan_eye[ip][i] + pan_dlef[ip]*pan_torq[ip][i];
        pan_ritpar[ip][i] = pan_eye[ip][i] + pan_drit[ip]*pan_torq[ip][i];
        pan_toppar[ip][i] = pan_eye[ip][i] + pan_dtop[ip]*pan_up[ip][i];
        pan_botpar[ip][i] = pan_eye[ip][i] + pan_dbot[ip]*pan_up[ip][i];
      }
                                            /* outward orthogonal vector at */
      crossprod(pan_up[ip],     pan_lefpar[ip], pan_lef_o[ip]); /*     left */
      crossprod(pan_ritpar[ip], pan_up[ip],     pan_rit_o[ip]); /*    right */ 
      crossprod(pan_torq[ip],   pan_toppar[ip], pan_top_o[ip]); /*      top */
      crossprod(pan_botpar[ip], pan_torq[ip],   pan_bot_o[ip]); /*   bottom */

      /* fpp = fopen("rend_ONB.out", "a+");
      fprintf(fpp, "ip = %d\n", ip);
      fprintf(fpp, "pan_ritpar= %f %f %f\n", pan_ritpar[ip][0],pan_ritpar[ip][1],pan_ritpar[ip][2]);
      fprintf(fpp, "pan_up    = %f %f %f\n", pan_up[ip][0],pan_up[ip][1],pan_up[ip][2]);
      fprintf(fpp, "pan_rit_o = %f %f %f\n", pan_rit_o[ip][0],pan_rit_o[ip][1],pan_rit_o[ip][2]);
      fclose(fpp); */
    }

    if(iprint == -1) { exitHVR(IPRINT_FAILED); }

/***************************************************************************
 * INPUT: color map information                                            *
 ***************************************************************************/
    if(strdif(m_ShvrUse.sCmapname, m_cOldCmap) == 1) {
      setstr(m_cOldCmap, m_ShvrUse.sCmapname);
      if(strdif(m_ShvrUse.sCmapname, "default") == 1) {
          ididread = readCmap( m_ShvrUse.sCmapname, m_ShvrUse.sAmapname, frgba );

          if(ididread < 0) {
              /*fpp = fopen("C:/DEBUG_HVR.txt", "a+");
              fprintf(fpp, "ididread = %d\n", ididread);
              fprintf(fpp, "cmap: -->%s<--\n", m_ShvrUse.sCmapname);
              fprintf(fpp, "amap: -->%s<--\n", m_ShvrUse.sAmapname);
              for(i=0; i<256; i++) {
                  fprintf(fpp, "%3d %f %f %f %f\n", i, frgba[i][0], frgba[i][1], frgba[i][2], frgba[i][3]);
              }
              fclose(fpp);
                MessageBox(hWnd, "Can NOT open alpha/color maps", "LALA", MB_OK); */
          }
          if(ididread < 0) {
#ifdef DEBUG_NEW_TREE
fprintf(stderr, "XXX ididread<0\n");
#endif
            return;
         }
      }

      if(!m_bColorTableIsAvailable)
      {
            for(i=0; i<256; i++) 
            {
                m_lrgba[i] =              (unsigned long)(255.99*frgba[i][0]) +
                    (unsigned long)256 * ((unsigned long)(255.99*frgba[i][1]) +
                    (unsigned long)256 * ((unsigned long)(255.99*frgba[i][2]) +
                    (unsigned long)256 *  (unsigned long)(255.99*frgba[i][3])));
            }
      }
    }

    /* This clear is as late as possible */
    glClear(GL_COLOR_BUFFER_BIT);

/***************************************************************************
 * INPUT: tile information                                                 *
 ***************************************************************************/
    for(i=0; i<nsub; i++) { for(nccc=0; nccc<32; nccc++) {
        tile_children[i][nccc] = -1;
    }}
    for(i=0; i<nsub; i++) { 
      ippp = tile_parent[i];
      nccc = 0;
      while(tile_children[ippp][nccc] >= 0  &&  ippp >= 0) { nccc++; }
      tile_children[ippp][nccc  ] =  i;
      tile_children[ippp][nccc+1] = -1;
      if(ippp >= MAXNODES  ||  nccc >= 32) {
          fp = fopen("C:\\HVR_bug_report.txt", "a+");
          fprintf(fp, "Array bounds problem in generating list of children for each node\n");
          fprintf(fp, "nsub=%d i=%d ippp=%d nccc=%d\n", nsub, i, ippp, nccc);
          fclose(fp);
      }
    }

/*****************************************************************************
 * GENERATE: distance**2 to each tile (rendering is done farthest to nearest *
 *           under each node                                                 *
 * distance = distance from eye to closest point JUST interior to tile   .   *
 *****************************************************************************/
    fminx = fLimits[0] - eyepos[0];
    fminy = fLimits[2] - eyepos[1];
    fminz = fLimits[4] - eyepos[2];
    delx    = (fLimits[1] - fLimits[0]) / (float)Nx;
    dely    = (fLimits[3] - fLimits[2]) / (float)Ny;
    delz    = (fLimits[5] - fLimits[4]) / (float)Nz;
    facXtoN = (fLimits[1] - fLimits[0]) / (float)Nx;
    facYtoN = (fLimits[3] - fLimits[2]) / (float)Ny;
    facZtoN = (fLimits[5] - fLimits[4]) / (float)Nz;

    for(i=0; i<nsub; i++) {
      nr = tile_nrep[i];
      nxsub = tile_nxsub[i];
      nysub = tile_nysub[i];
      nzsub = tile_nzsub[i];
      ixoff = tile_ixoff[i];
      iyoff = tile_iyoff[i];
      izoff = tile_izoff[i];
      m_iNBoundaries = m_iNBoundariesRep[nr];

      /*xx = fminx + (float)(nr*(2*tile_ixoff[i]+tile_nxsub[i]))/(float)Nx;
      yy = fminy + (float)(nr*(2*tile_iyoff[i]+tile_nysub[i]))/(float)Nx;
      zz = fminz + (float)(nr*(2*tile_izoff[i]+tile_nzsub[i]))/(float)Nx;*/

      xmin = fminx + (float)((ixoff+m_iNBoundaries)*nr) * facXtoN;
      ymin = fminy + (float)((iyoff+m_iNBoundaries)*nr) * facYtoN;
      zmin = fminz + (float)((izoff+m_iNBoundaries)*nr) * facZtoN;
  
      xmax = xmin + (float)((nxsub-2*m_iNBoundaries)*nr) * facXtoN;
      ymax = ymin + (float)((nysub-2*m_iNBoundaries)*nr) * facYtoN;
      zmax = zmin + (float)((nzsub-2*m_iNBoundaries)*nr) * facZtoN;

      xx = 0.; if(xmin > 0.) { xx = delx + xmin; }  if(xmax < 0.) { xx = delx - xmax; }
      yy = 0.; if(ymin > 0.) { yy = dely + ymin; }  if(ymax < 0.) { yy = dely - ymax; }
      zz = 0.; if(zmin > 0.) { zz = delz + zmin; }  if(zmax < 0.) { zz = delz - zmax; }

      tile_dist[i] = xx*xx + yy*yy + zz*zz;
      tile_did[i] = 0;
    }

/*****************************************************************************
 * GENERATE: ordered list of tiles                                           *
 *****************************************************************************/
    delxmin = 0.;
    if(viewres > 0.) { delxmin = (float)1.01 * viewres / ((float)Nx/(fLimits[1] - fLimits[0])); }

    nlist = 1;
    list_tiles[0] = node0;
    need_check = 1;

    while(need_check == 1) {
      nlistnew = 0;
      need_check = 0;
      for(ilist = 0; ilist<nlist; ilist++) {
        inode = list_tiles[ilist];

        distmin = (float)sqrt(tile_dist[inode]);
        if(distmin < dnearclip) { distmin = (float)dnearclip; }

        delxtol  = distmin * angtol;
        delxsize =  (float)tile_nrep[inode] / ((float)Nx/(fLimits[1] - fLimits[0]));

        if(tile_maxerr[inode] <= tolerance || delxsize < delxtol
                                           || delxsize < delxmin) {
          list_new[nlistnew] = inode;
          nlistnew++;
        } else {
          need_check = 1;
          nccc = 0;
          while(tile_children[inode][nccc] >= 0) { nccc++; }
          isub = 1;
          while(isub >= 0) {
            distmax = -10000000.;
            isub = -1;
            for(iccc = 0; iccc < nccc; iccc++) {
              i = tile_children[inode][iccc];
              if(tile_did[i] == 0) {
                if(distmax < tile_dist[i]) { isub=i; distmax = tile_dist[i]; }
              }
            }
            if(isub >= 0) {
              list_new[nlistnew] = isub;
              nlistnew++;
              tile_did[isub] = 1;
            }
          }
        }
      }
      for(i=0; i<nlistnew; i++) { list_tiles[i] = list_new[i]; }
      nlist = nlistnew;
    }
#ifdef DEBUG_NEW_TREE
fprintf(stderr, "after tree scan: nlist = %d\n", nlist);
#endif

/************************************************************************
 * THIS IS A TEST OF IO SPEED:  READ IN ORDER                           *
 ************************************************************************/
    /*if(viewres < 1.5) {
        nlist = 0;
        for(i=0; i<nsub; i++) { if(tile_nrep[i] == 1) { list_tiles[nlist] = i;  nlist++; }}
    }*/

/************************************************************************
 * Initialize statistics                                                *
 ************************************************************************/
    if(strcmp(m_sStats, "cross") == 0) {
        for(i=0; i<256*256; i++) { iCross[i] = 0; }
    }
    if(strcmp(m_sStats, "pdf") == 0) {
        for(i=0; i<256; i++) { iPDF[i] = 0; }
    }

    /* Set flag for doing annotation 1 time up front */
    iDoAnnotationBeforeRendering = 1;

/***************************************************************************
 *    hvr: subroutine "subvolrend" starts here                             *
 *    INPUTS: ixoff, iyoff, izoff, nxsub, nysub, nzsub, Nx, Ny, Nz         *
 *                                                                         *
 *    ixoff = x-offset of subvolume from left/min-edge of full volume      *
 *    iyoff = y-offset of subvolume from left/min-edge of full volume      *
 *    izoff = z-offset of subvolume from left/min-edge of full volume      *
 *    nxsub = x-dim size of sub-volume (not nec. power of 2, but <= 256)   *
 *    nysub = y-dim size of sub-volume (not nec. power of 2, but <= 256)   *
 *    nzsub = z-dim size of sub-volume (not nec. power of 2, but <= 256)   *
 *    Nx    = x-dim of full volume (not nec. power of 2, unlimitted size)  *
 *    Ny    = y-dim of full volume (not nec. power of 2, unlimitted size)  *
 *    Nz    = z-dim of full volume (not nec. power of 2, unlimitted size)  *
 *                                                                         *
 *    NOTE: ALL INPUTS ARE IN ZONES (int); ALL ZONES ARE CUBES             *
 ***************************************************************************/

/***************************************************************************
 *    CALCULATE: size and position of subvolume (in modelview space)       *
 *    CONVENTIONS: 1) x-dim always goes from (-1, 1);                      *
 *                 2) all zones are cubes                                  *
 *    ==>  y-dim goes from (-Ny/Nx, +Ny/Nx)                                *
 *    ==>  z-dim goes from (-Nz/Nx, +Nz/Nx)                                *
 *                                                                         *
 *    fulminpos[3] = xyz coordinates of min-vertex of full volume          *
 *    subminpos[3] = xyz coordinates of min-vertex of sub volume           *
 *    subvolsiz[3] = xyz size of subvolume                                 *
 *    subtexsiz[3] = xyz size of power of 2 texture subvolume in which     *
 *                   input subvolume is imbedded                           *
 *                                                                         *
 *                                                                         *
 *    +------------------------+ -  +Ny/Nx                                 *
 *    |                        | ^                                         *
 *    |                        | |                                         *
 *    |                        | |                                         *
 *    |    |<--- A --->|       | |     A = subtexsiz[0] = 2**n always      *
 *    |                        | |     B = subtexsiz[1] = 2**n always      *
 *    |    +-----------+ -     | |     C = subvolsiz[0]                    *
 *    |    |           | ^     | |     D = subvolsiz[1]                    *
 *    |    +-------+   | |     |       E = subminpos[0]                    *
 *    |    |   C   |   |       | Ny    F = subminpos[1]                    *
 *    |    |       |   | B     |       G = fulminpos[0] = -1 always        *
 *    |    |      D|   |       | |     H = fulminpos[1] = -Ny/Nx always    *
 *    |    |       |   | |     | |                                         *
 *    |    |       |   | v     | |                                         *
 *    |    +-------+---+ -     | |                                         *
 *    |  (E,F)                 | |                                         *
 *    |                        | |                                         *
 *    |                        | v                                         *
 *    +------------------------+ -  -Ny/Nx                                 *
 *  (G,H)                                                                  *
 *    |<--------- Nx --------->|                                           *
 *   -1                       +1                                           *
 *                                                                         *
 ***************************************************************************/

/**************************************************************************************
 *  Cache IO: loop over tiles in this list, request IO for those in viewing frustum   *
 **************************************************************************************/
    if(ResidentTextures) {
        if (loopcount == 0) {
            /* If new HV file, initialize tex-brick arrays */
            for(i=0; i<MAXTEX; i++) {    
                bricktex[i] = -1;
                texbrick[i] = -1;
            }
        }
    }
    
    if(nlist > 1) { isubcur = -1; }  /* Only reuse current tile of single tile renderings */
    ResetCacheStatus();
    iNTileRender = 0;
    nBricksInVolume = nlist;

//fprintf(stderr, "XXX: before jsub loop\n");

    for(jsub=0; jsub<nlist; jsub++) {
      if(iStopRendering == 1) {
            iStopRendering = 0;
            fprintf(error_log,"STOP RENDERING\n");fflush(error_log);
            isubcur = -1;
            while(iIO_Read != 0) { _sleep(10); }
#ifdef DEBUG_NEW_TREE
fprintf(stderr, "XXX iStopRendering==1\n");
#endif
            return;
      }
      isub = list_tiles[jsub];

      nr = tile_nrep[isub];

/*
#ifdef DEBUG_TEX
      fprintf(error_log,"jsub =%4d   isub =%4d    nrep =%3d\n", jsub, isub, nr);
#endif
*/

      nxsub = tile_nxsub[isub];
      nysub = tile_nysub[isub];
      nzsub = tile_nzsub[isub];
      ixoff = tile_ixoff[isub];
      iyoff = tile_iyoff[isub];
      izoff = tile_izoff[isub];
      offset64t = tile_off64t[isub];
      nbytes    = nxsub * nysub * nzsub;
      m_iNBoundaries = m_iNBoundariesRep[nr];

      fulminpos[0] = fLimits[0];
      fulminpos[1] = fLimits[2];
      fulminpos[2] = fLimits[4];

      facXtoN = (fLimits[1]-fLimits[0]) / (float)Nx;
      facYtoN = (fLimits[3]-fLimits[2]) / (float)Ny;
      facZtoN = (fLimits[5]-fLimits[4]) / (float)Nz;

/* was this (in texture request loop)
      subminpos[0] = fulminpos[0] + (float)((ixoff+m_iNBoundaries)*nr) * facXtoN;
      subminpos[1] = fulminpos[1] + (float)((iyoff+m_iNBoundaries)*nr) * facYtoN;
      subminpos[2] = fulminpos[2] + (float)((izoff+m_iNBoundaries)*nr) * facZtoN;
      subvolsiz[0] = (float)((nxsub-2*m_iNBoundaries)*nr) * facXtoN;
      subvolsiz[1] = (float)((nysub-2*m_iNBoundaries)*nr) * facYtoN;
      subvolsiz[2] = (float)((nzsub-2*m_iNBoundaries)*nr) * facZtoN;
*/
/* tex fix */
      fNvoxIn = (float)m_iNBoundaries*nr;
      if(fNvoxIn < (float)0.5) { fNvoxIn = (float)0.5 * (float)nr; }
      subminpos[0] = fulminpos[0] + ((float)(ixoff*nr) + fNvoxIn) * facXtoN;
      subminpos[1] = fulminpos[1] + ((float)(iyoff*nr) + fNvoxIn) * facYtoN;
      subminpos[2] = fulminpos[2] + ((float)(izoff*nr) + fNvoxIn) * facZtoN;
  
      subvolsiz[0] = ((float)(nxsub*nr)-2.0*fNvoxIn) * facXtoN; subvolsiz[1] = ((float)(nysub*nr)-2.0*fNvoxIn) * facYtoN;
      subvolsiz[2] = ((float)(nzsub*nr)-2.0*fNvoxIn) * facZtoN;
      m_iNBoundaries = m_iNBoundariesRep[nr];

      for(i=0; i<3; i++) { submaxpos[i] = subminpos[i] + subvolsiz[i]; }

      /* Generate Texture mem dims s.t.: smallest power of 2 just containing subtile */
      nxsb2 = 2;   while(nxsb2 < nxsub) { nxsb2 *= 2; }
      nysb2 = 2;   while(nysb2 < nysub) { nysb2 *= 2; }
      nzsb2 = 2;   while(nzsb2 < nzsub) { nzsb2 *= 2; }
      iTexBndry = 0;

      /* Use texture boundary of 1, if applicable */
      /* if(nxsb2/2 == nxsub-2  &&  nysb2/2 == nysub-2  &&  nzsb2/2 == nzsub-2) {
          // In this case, can use iTexBndry = 1 with a perfect fit
          iTexBndry = 1;
          nxsb2 = nxsub;
          nysb2 = nysub;
          nzsb2 = nzsub;
      }*/

     /*********************************************************************
      * START of ip loop -- LOOP over pannels                             *
      *********************************************************************/
      for(ip=0; ip < npannel; ip++) {

          /*This was here, but not used for getting list of brick to read in
           Completely changed for rendering, see rendering loop below
           idir = pan_idir[ip];
           dxx = pan_eye[ip][0];
           dyy = pan_eye[ip][1];
           dzz = pan_eye[ip][2];*/

         /*********************************************************************
          * Derive min and max distances for edges of each tile               *
          * As a byproduct, determine if tile is OUTSIDE the viewing frustum  *
          * A tile is OUTSIDE of a convex faceted volume if all 8 vertices    *
          * of the tile are on the outer side of any one plane tangent to     *
          * a facet of that volume.  Viewing frustums are convex.             *
          *********************************************************************/
          edgemin = 10000000.;
          edgemax = 0.;
          isoutside_n = 1;  /* near   */
          isoutside_f = 1;  /* far    */
          isoutside_t = 1;  /* top    */
          isoutside_b = 1;  /* bottom */
          isoutside_l = 1;  /* left   */
          isoutside_r = 1;  /* right  */
          for(i=0; i<2; i++) { xx = subminpos[0]+(float)i*subvolsiz[0]-eyepos[0];
          for(j=0; j<2; j++) { yy = subminpos[1]+(float)j*subvolsiz[1]-eyepos[1];
          for(k=0; k<2; k++) { zz = subminpos[2]+(float)k*subvolsiz[2]-eyepos[2];
            edgedist = (xx*pan_eye[ip][0]+yy*pan_eye[ip][1]+zz*pan_eye[ip][2]);
            if(edgemin > edgedist) { edgemin = edgedist; }
            if(edgemax < edgedist) { edgemax = edgedist; }

            if(edgedist > m_ShvrUse.fNearClip - softdistance   ) { isoutside_n = 0; }
            if(edgedist < m_ShvrUse.fFarrClip + softdistance   ) { isoutside_f = 0; }
            if(xx*pan_top_o[ip][0]+yy*pan_top_o[ip][1]
                                  +zz*pan_top_o[ip][2] < 0. ) { isoutside_t = 0; }
            if(xx*pan_bot_o[ip][0]+yy*pan_bot_o[ip][1]
                                  +zz*pan_bot_o[ip][2] < 0. ) { isoutside_b = 0; }
            if(xx*pan_lef_o[ip][0]+yy*pan_lef_o[ip][1]
                                  +zz*pan_lef_o[ip][2] < 0. ) { isoutside_l = 0; }
            if(xx*pan_rit_o[ip][0]+yy*pan_rit_o[ip][1]
                                  +zz*pan_rit_o[ip][2] < 0. ) { isoutside_r = 0; }
          }}}

        /******************************************************************************
         *    Test for Object oriented clipping planes                                *
         ******************************************************************************/
          if(iTexBndry == 1) {
                xx0 = (float)0.0;
                xx1 = (float)1.0;
                yy0 = (float)0.0;
                yy1 = (float)1.0;
                zz0 = (float)0.0;
                zz1 = (float)1.0;
          } else {
                /*xx0 =               (float)1.0  / (float)(nxsb2-1);
                xx1 = ((float)nxsub - (float)2.0) / (float)(nxsb2-1);
                yy0 =                 (float)1.0  / (float)(nysb2-1);
                yy1 = ((float)nysub - (float)2.0) / (float)(nysb2-1);
                zz0 =                 (float)1.0  / (float)(nzsb2-1);
                zz1 = ((float)nzsub - (float)2.0) / (float)(nzsb2-1);
                tex fix */
                fNvoxIn = (float)m_iNBoundaries;
                if(fNvoxIn < (float)0.5) { fNvoxIn = (float)0.5; }
                if(ResidentTextures) {
                    submintex[0] =                 fNvoxIn  / (float)max_nxsb2;
                    submaxtex[0] = ((float)nxsub - fNvoxIn) / (float)max_nxsb2;
                    submintex[1] =                 fNvoxIn  / (float)max_nysb2;
                    submaxtex[1] = ((float)nysub - fNvoxIn) / (float)max_nysb2;
                    submintex[2] =                 fNvoxIn  / (float)max_nzsb2;
                    submaxtex[2] = ((float)nzsub - fNvoxIn) / (float)max_nzsb2;
                } else {
                    submintex[0] =                 fNvoxIn  / (float)nxsb2;
                    submaxtex[0] = ((float)nxsub - fNvoxIn) / (float)nxsb2;
                    submintex[1] =                 fNvoxIn  / (float)nysb2;
                    submaxtex[1] = ((float)nysub - fNvoxIn) / (float)nysb2;
                    submintex[2] =                 fNvoxIn  / (float)nzsb2;
                    submaxtex[2] = ((float)nzsub - fNvoxIn) / (float)nzsb2;
                }
          }

          iClipThis0 = 0;

          xx0 = submintex[0];
          xx1 = submaxtex[0];
          yy0 = submintex[1];
          yy1 = submaxtex[1];
          zz0 = submintex[2];
          zz1 = submaxtex[2];

          px0 = subminpos[0];
          px1 = submaxpos[0];
          py0 = subminpos[1];
          py1 = submaxpos[1];
          pz0 = subminpos[2];
          pz1 = submaxpos[2];
          if(m_ShvrUse.ifauxclip[0] == 2) {
                px0c = m_ShvrUse.auxclip[0][3];
                px1c = m_ShvrUse.auxclip[1][3];
                py0c = m_ShvrUse.auxclip[2][3];
                py1c = m_ShvrUse.auxclip[3][3];
                pz0c = m_ShvrUse.auxclip[4][3];
                pz1c = m_ShvrUse.auxclip[5][3];
                xx0c = submintex[0] + (submaxtex[0]-submintex[0])*(px0c-subminpos[0])/subvolsiz[0];
                xx1c = submintex[0] + (submaxtex[0]-submintex[0])*(px1c-subminpos[0])/subvolsiz[0];
                yy0c = submintex[1] + (submaxtex[1]-submintex[1])*(py0c-subminpos[1])/subvolsiz[1];
                yy1c = submintex[1] + (submaxtex[1]-submintex[1])*(py1c-subminpos[1])/subvolsiz[1];
                zz0c = submintex[2] + (submaxtex[2]-submintex[2])*(pz0c-subminpos[2])/subvolsiz[2];
                zz1c = submintex[2] + (submaxtex[2]-submintex[2])*(pz1c-subminpos[2])/subvolsiz[2];
                if(xx0c > xx0) { xx0 = xx0c;    px0 = px0c; }
                if(xx1c < xx1) { xx1 = xx1c;    px1 = px1c; }
                if(yy0c > yy0) { yy0 = yy0c;    py0 = py0c; }
                if(yy1c < yy1) { yy1 = yy1c;    py1 = py1c; }
                if(zz0c > zz0) { zz0 = zz0c;    pz0 = pz0c; }
                if(zz1c < zz1) { zz1 = zz1c;    pz1 = pz1c; }
                if(xx0 >= xx1) { iClipThis0 = 1; }
                if(yy0 >= yy1) { iClipThis0 = 1; }
                if(zz0 >= zz1) { iClipThis0 = 1; }
          }

        /********************************************************************************
         *    SKIP this tile if it is outside viewing frustum OR object oriented clips  *
         *    this is end of jsub (tile) loop if tile is outside                        *
         ********************************************************************************/
          if(   isoutside_n == 1 || isoutside_f == 1 || isoutside_t == 1
             || isoutside_b == 1 || isoutside_l == 1 || isoutside_r == 1 || iClipThis0 == 1) {
          } else {
            /* Submit request to read in this tile */
            iTileRender[iNTileRender] = isub;
            iNTileRender++;

            if(ResidentTextures) { 
                /* Don't request a brick read if its corresponding
                   texture is already in texture memory */
#ifdef INTELLIGENT_OVERWRITE
                if(bricktex[isub] >= 0)            
                    isubcur = isub;
#else
                if(TexScheme == 1) {
                    if(bricktex[isub] >= 0)            
                        isubcur = isub;
                }
#endif
            }

            if(isubcur != isub) {
                sprintf(tile_full, "%s%s\0", tile_path, tile_file[isub]);
                RequestCacheIO(isub, tile_full, offset64t, (DWORD)nbytes);
/*
#ifdef DEBUG_TEX
      fprintf(error_log,"Requesting  isub = %d\n", isub);
      fprintf(error_log,"offset64t        = %lld\n", offset64t);
      fprintf(error_log,"nbytes           = %d\n", nbytes);
      fprintf(error_log,"hvfile           = %s\n", tile_full);
#endif
*/
            }
          }
      }  /* End of ipannel loop */
    }    /* End of jsub loop */

//fprintf(stderr, "XXX: after  jsub loop\n");

#ifdef DEBUG_NEW_TREE
fprintf(stderr, "after read queue jsub loop: iNTileRender = %d\n", iNTileRender);
#endif

    /* Make sure that the glClear is done even if nothing is rendered */
    if(iNTileRender <= 0) {
      glClear(GL_COLOR_BUFFER_BIT);
      glFlush();
      glFinish();
    }

    #ifdef IDEBUG
    printf("draw_scene: after assignment of reads\n");
    #endif
    
    /* All read requests for this pass are in
       set IO thread going only if there were any actual requests generated */
    if(iNRequests > 0) { iIO_Read = 1; }
    /* TexScheme 0 => FIFO queue */
    /* TexScheme 1 => Single-texture overwrite mode */
    if(ResidentTextures) {
        /* Get size of each filled brick */
        nBricksInFrustum = iNTileRender;

#ifndef INTELLIGENT_OVERWRITE
#ifndef FORCE_TEXSCHEME        
        /* Set or Change TexMem Management Scheme if necessary */
        if((iNTileRender*texsize)>=max_tex_mem) {
            if(TexScheme==0) {
                #ifdef DEBUG_TEX
                printf("TexScheme: single-tex overwrite mode\n");
                #endif
                TexScheme = 1;
                loopcount = 0;
            }
        }
        else {
            if(TexScheme==1) {
                #ifdef DEBUG_TEX
                printf("TexScheme: FIFO queue mode\n");
                #endif
                TexScheme = 0;
                loopcount = 0;
            }
        }
#else
        TexScheme = FORCE_TEXSCHEME;
#endif
#endif
    }
    
    /* Reload texture cache if scheme was deemed inefficient */
#ifndef INTELLIGENT_OVERWRITE
    if(reload) {
        loopcount = 0;
        reload = FALSE;
    }
#endif
    
    /* Texture Initialization */
    if(loopcount==0) {   
#ifdef NEWTEX
        bool mem_not_full = TRUE;
        int texcount = 0; /* number of textures loaded */
        int tex_mem = 0; /* amount of texture memory used */
#endif
        
        if(Tex3d) {
            if(glIsEnabled(GL_TEXTURE_2D))
                glDisable(GL_TEXTURE_2D);
            glEnable(GL_TEXTURE_3D);
        }
        else {
            if(glIsEnabled(GL_TEXTURE_3D))
                glDisable(GL_TEXTURE_3D);
            glEnable(GL_TEXTURE_2D);
        }
        
        texsize = max_nxsb2*max_nysb2*max_nzsb2;
        
#ifdef NEWTEX
        if (reload) {
            /* Delete all texture objects previously used */
            glDeleteTextures(MAXTEX, texture);
            glDeleteTextures(1, &dataTexture);
            glDeleteTextures(1, &RGBATexture);
            
            /* Make dummy texture data */    
            texData = (GLubyte *)calloc(texsize,sizeof(unsigned char));

            /* create texture objects */
            glGenTextures(MAXTEX, texture);
            
            /* set texture object properties */
            glEnable(GL_TEXTURE_3D);
            for(i=0;i<MAXTEX;i++) {
                glBindTexture(GL_TEXTURE_3D, texture[i]);    
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP);
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP);
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP);
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
            }
        
            while(mem_not_full) {
                glBindTexture(GL_TEXTURE_3D, texture[texcount]);
                glTexImage3D(GL_TEXTURE_3D,0,1,max_nxsb2,max_nysb2,max_nzsb2,0,GL_LUMINANCE,GL_UNSIGNED_BYTE,texData);
                /* 2nd bind forces upload */
                glBindTexture(GL_TEXTURE_3D, texture[texcount]);
                /* Check if all textures created so far are resident */
                /* in texture memory. If any are not, break out of loop */
                resident = glAreTexturesResident(texcount+1,texture,residences);
                if(resident == GL_FALSE) {
                    /* A texture was overwritten, texture memory is full */
                    mem_not_full = FALSE;
                } else {
                    /* Texture memory is not full, incrememnt count */
                    texcount++;
                }
#ifdef DEBUG_TEX
                if(texcount>MAX_TEX) { /* Prevent infinite loop if resident function fails */
                    fprintf(stderr, "Error, resident detection failed\n");
                    mem_not_full = 0;
                }
#endif
                nBricksFitInCache = texcount;
                tex_mem = (texsize*texcount) / MEGABYTE;

                /* DHP: Each new instance of hvr fills no more than tex_mem_avail */
                if(tex_mem >= tex_mem_avail) { mem_not_full = FALSE; }
            }
            reload = FALSE;
        }
#else
        nBricksFitInCache = max_tex_mem / texsize;
#endif

        #ifdef DEBUG_TEX
            #ifdef WRITE_STD_OUT
                fprintf(error_log,"Flushing texture memory\n");fflush(error_log);
            #else
                printf("Flushing texture memory\n");
            #endif
        #endif

#ifndef NEWTEX
        /* Delete all texture objects previously used */
        glDeleteTextures(MAXTEX, texture);
        glDeleteTextures(1, &dataTexture);
        glDeleteTextures(1, &RGBATexture);
#endif
        
        /* Initialize counters and markers */
        count = 0;
        end_count = 0;
        nBricksLoaded = 0; 
        nBricksRendered = 0;
        nBricksInCache = 0;
        nBricksRead = 0;
        resident = GL_FALSE;
#ifdef INTELLIGENT_OVERWRITE
        prev_count = -1;
        all_in_use = FALSE;
#else
        texmem = 0;
        texmem_previous = 0;
        texmem_full = FALSE;
        new_brick = TRUE;
#endif

        /* Resident Texture Initialization */
        if(ResidentTextures) {    
            #ifdef DEBUG_TEX
                #ifdef WRITE_STD_OUT
                    fprintf(error_log,"Creating and initializing texture objects\n");fflush(error_log);
                #else
                    printf("Creating and initializing texture objects\n");
                #endif
            #endif
            
#ifndef NEWTEX
            /* Create and initialize texture objects */
            glGenTextures(nsub, texture);
            for(i=0; i<nsub; i++) {    
                glBindTexture(GL_TEXTURE_3D, texture[i]);
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP);
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP);
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP);
            }
#endif

#ifdef INTELLIGENT_OVERWRITE
            /* Load dummy texture object 0 into texture memory
               to reserve space for overwriting later */
#ifdef PBO
            glGenBuffers(1, &texBuffer);
            glBindBuffer(GL_PIXEL_UNPACK_BUFFER_EXT, texBuffer);
            glBufferData(GL_PIXEL_UNPACK_BUFFER_EXT, texsize, NULL, GL_STREAM_DRAW);

            /*for(i=0; i<nBricksFitInCache; i++) {
                glBindTexture(GL_TEXTURE_3D, texture[i]);
                glTexImage3D(GL_TEXTURE_3D,0,1,max_nxsb2,max_nysb2,max_nzsb2,0,GL_LUMINANCE,GL_UNSIGNED_BYTE,NULL);
            }*/
#else
#ifndef NEWTEX  
            glBindTexture(GL_TEXTURE_3D, texture[0]);
            glTexImage3D(GL_TEXTURE_3D,0,1,max_nxsb2,max_nysb2,max_nzsb2,0,GL_LUMINANCE,GL_UNSIGNED_BYTE,NULL);
#endif
#endif
            count = 1;
#else
            if(TexScheme == 1) {
                /* Load dummy texture object 0 into texture memory
                   to reserve space for overwriting later */
                glBindTexture(GL_TEXTURE_3D, texture[0]);
                glTexImage3D(GL_TEXTURE_3D,0,1,max_nxsb2,max_nysb2,max_nzsb2,0,GL_LUMINANCE,GL_UNSIGNED_BYTE,NULL);
                count = 1;
            }
#endif
        }
        else
        {    /* Create and Initialize Single Texture object */
            if(Tex3d) {
                if(m_bColorTableIsAvailable) {
                    #ifdef DEBUG_TEX
                        #ifdef WRITE_STD_OUT
                            fprintf(error_log,"Creating single 3d paletted data texture\n");
                            fflush(error_log);
                        #else
                            printf("Creating single 3d paletted data texture\n");
                        #endif
                    #endif
                    glGenTextures(1, &dataTexture);
                    glBindTexture(GL_TEXTURE_3D, dataTexture);
                }
                else {
                    #ifdef DEBUG_TEX
                        #ifdef WRITE_STD_OUT
                            fprintf(error_log,"Creating single 3d RGBA data texture\n");fflush(error_log);
                        #else
                            printf("Creating single 3d RGBA data texture\n");
                        #endif
                    #endif
                    glGenTextures(1, &RGBATexture);
                    glBindTexture(GL_TEXTURE_3D, RGBATexture);    
                }
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP);
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP);
                glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP);
            }
            else {
                if(m_bColorTableIsAvailable) {
                    #ifdef DEBUG_TEX
                        #ifdef WRITE_STD_OUT
                            fprintf(error_log,"Creating single 2d paletted data texture\n");fflush(error_log);
                        #else
                            printf("Creating single 2d paletted data texture\n");
                        #endif
                    #endif
                    glGenTextures(1, &dataTexture);
                    glBindTexture(GL_TEXTURE_2D, dataTexture);
                }
                else {
                    #ifdef DEBUG_TEX
                        #ifdef WRITE_STD_OUT
                            fprintf(error_log,"Creating single 2d RGBA data texture\n");fflush(error_log);
                        #else
                            printf("Creating single 2d RGBA data texture\n");
                        #endif
                    #endif
                    glGenTextures(1, &RGBATexture);
                    glBindTexture(GL_TEXTURE_2D, RGBATexture);    
                }
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
                if(m_bColorTableIsAvailable) {
                    glBindTexture(GL_TEXTURE_2D, dataTexture);
                    cgGLSetTextureParameter(textureHandle,dataTexture);
                }
                else
                    glBindTexture(GL_TEXTURE_2D,RGBATexture);
            }
        }
    }

//fprintf(stderr, "XXX: after texture initialization if block\n");
#ifdef DEBUG_NEW_TREE
fprintf(stderr, "after texture initialization if block\n");
// return
#endif

    if(m_iTimersOn == 1) { tsEvent.StampTime("End   Setup"); }


/*********************************************************************************
 *  RENDER: loop over tiles in this list and render those in the viewing frustum  *
 *********************************************************************************/
    ntiles  = 0;
    nslices = 0;
    fVoxels = 0.0;

//fprintf(stderr, "XXX: before render loop: iNTileRender=%d\n", iNTileRender);

    for(jsub=0; jsub<iNTileRender; jsub++) {
      //if(m_iTimersOn == 1) { tsEvent.StampTime("Begin jsubRenderloop"); }
#ifdef DEBUG_NEW_TREE
fprintf(stderr, "in rendering loop: jsub = %d\n", jsub);
#endif

      if(iStopRendering == 1) {
            iStopRendering = 0;
            isubcur = -1;
            while(iIO_Read != 0) { _sleep(10); }

#ifdef DEBUG_NEW_TREE
fprintf(stderr, "in iStopRendering if-block\n");
#endif

            write(STDOUT, &status, 4);
            return;
      }

      isub = iTileRender[jsub];

//fprintf(stderr, "XXX: jsub=%4d    isub=%4d\n", jsub, isub);

      nr = tile_nrep[isub];
      nxsub = tile_nxsub[isub];
      nysub = tile_nysub[isub];
      nzsub = tile_nzsub[isub];
      ixoff = tile_ixoff[isub];
      iyoff = tile_iyoff[isub];
      izoff = tile_izoff[isub];
      offset64t = tile_off64t[isub];
      nbytes    = nxsub * nysub * nzsub;
      m_iNBoundaries = m_iNBoundariesRep[nr];

      fVoxels += (float)(nxsub*nysub*nzsub);

      fulminpos[0] = fLimits[0];
      fulminpos[1] = fLimits[2];
      fulminpos[2] = fLimits[4];

      facXtoN = (fLimits[1]-fLimits[0]) / (float)Nx;
      facYtoN = (fLimits[3]-fLimits[2]) / (float)Ny;
      facZtoN = (fLimits[5]-fLimits[4]) / (float)Nz;

/* dhp: issue with root node texture translation */
/* was this (in render loop)
      subminpos[0] = fulminpos[0] + (float)((ixoff+m_iNBoundaries)*nr) * facXtoN;
      subminpos[1] = fulminpos[1] + (float)((iyoff+m_iNBoundaries)*nr) * facYtoN;
      subminpos[2] = fulminpos[2] + (float)((izoff+m_iNBoundaries)*nr) * facZtoN;
      subvolsiz[0] = (float)((nxsub-2*m_iNBoundaries)*nr) * facXtoN;
      subvolsiz[1] = (float)((nysub-2*m_iNBoundaries)*nr) * facYtoN;
      subvolsiz[2] = (float)((nzsub-2*m_iNBoundaries)*nr) * facZtoN;
*/
/* tex fix */
      fNvoxIn = (float)m_iNBoundaries*nr;
      if(fNvoxIn < (float)0.5) { fNvoxIn = (float)0.5 * (float)nr; }
      subminpos[0] = fulminpos[0] + ((float)(ixoff*nr) + fNvoxIn) * facXtoN;
      subminpos[1] = fulminpos[1] + ((float)(iyoff*nr) + fNvoxIn) * facYtoN;
      subminpos[2] = fulminpos[2] + ((float)(izoff*nr) + fNvoxIn) * facZtoN;
  
      subvolsiz[0] = ((float)(nxsub*nr)-2.0*fNvoxIn) * facXtoN;
      subvolsiz[1] = ((float)(nysub*nr)-2.0*fNvoxIn) * facYtoN;
      subvolsiz[2] = ((float)(nzsub*nr)-2.0*fNvoxIn) * facZtoN;

      for(i=0; i<3; i++) { submaxpos[i] = subminpos[i] + subvolsiz[i]; }

      /* Generate Texture mem dims s.t.: smallest power of 2 just containing subtile */
      nxsb2 = 2;   while(nxsb2 < nxsub) { nxsb2 *= 2; }
      nysb2 = 2;   while(nysb2 < nysub) { nysb2 *= 2; }
      nzsb2 = 2;   while(nzsb2 < nzsub) { nzsb2 *= 2; }
      iTexBndry = 0;

      /* Use texture boundary of 1, if applicable
      if(nxsb2/2 == nxsub-2  &&  nysb2/2 == nysub-2  &&  nzsb2/2 == nzsub-2) {
          // In this case, can use iTexBndry = 1 with a perfect fit
          iTexBndry = 1;
          nxsb2 = nxsub;
          nysb2 = nysub;
          nzsb2 = nzsub;
      }*/

     /*********************************************************************
      * START of ip loop -- LOOP over pannels                             *
      *********************************************************************/
      for(ip=0; ip < npannel; ip++) {
//fprintf(stderr, "XXX: ip=%d\n", ip);
          /* Generate rendering plane orientation
          Function of view frustum and sub-brick
          Construct vector from eye to centroid of that part of the brick which is
               inside the view frustum.
          To do this 1) project verices of brick onto window at unit distance 
                     2) crop projection to view frustum
                     3) generate centroid of cropped image
                     4) base idir on max component of this eye to centroid vector

          For each vertice of brick: get horiz & vert limits of bounding box
          on a window which is at unit distance along panel normal

          vhhmin = (float)1.0;
          vhhmax = (float)0.0;
          isBehind = 1;
          for(i=0; i<2; i++) { xx = subminpos[0]+(float)i*subvolsiz[0]-eyepos[0];
          for(j=0; j<2; j++) { yy = subminpos[1]+(float)j*subvolsiz[1]-eyepos[1];
          for(k=0; k<2; k++) { zz = subminpos[2]+(float)k*subvolsiz[2]-eyepos[2];
            vhh = xx*pan_torq[ip][0] +yy*pan_torq[ip][1] +zz*pan_torq[ip][2]; 
            vnn = xx* pan_eye[ip][0] +yy* pan_eye[ip][1] +zz* pan_eye[ip][2]; 
            vvv = xx*  pan_up[ip][0] +yy*  pan_up[ip][1] +zz*  pan_up[ip][2]; 

            if(vnn > m_ShvrUse.fNearClip - softdistance) {
                // Found vertex which is in front of you
                isBehind = 0;

                // Input vertex into limits of bounding box
                vhh = vhh / vnn;
                vvv = vvv / vnn;
                if(vhhmin > vhhmax) {
                    vhhmin = vhh;
                    vhhmax = vhh;
                    vvvmin = vvv;
                    vvvmax = vvv;
                } else {
                    if(vhhmin > vhh) { vhhmin = vhh; }
                    if(vhhmax < vhh) { vhhmax = vhh; }
                    if(vvvmin > vvv) { vvvmin = vvv; }
                    if(vvvmax < vvv) { vvvmax = vvv; }
                }
            }
          }}}

          if(isBehind == 1) {
              // If no vertex was in front of you, just use panel info
              idir = pan_idir[ip];
              dxx = pan_eye[ip][0];
              dyy = pan_eye[ip][1];
              dzz = pan_eye[ip][2];
          } else {
              // Crop limits of bounding box to view frustum
              if(vhhmin < pan_dlef[ip]) { vhhmin = pan_dlef[ip]; }
              if(vhhmax > pan_drit[ip]) { vhhmax = pan_drit[ip]; }
              if(vvvmin < pan_dbot[ip]) { vvvmin = pan_dbot[ip]; }
              if(vvvmax > pan_dtop[ip]) { vvvmax = pan_dtop[ip]; }

              // Generate normalized eye to centroid vector
              vhhcent = (float)0.5 * (vhhmin + vhhmax);
              vvvcent = (float)0.5 * (vvvmin + vvvmax);
              dxx = pan_eye[ip][0] + vhhcent*pan_torq[ip][0] + vvvcent*pan_up[ip][0];
              dyy = pan_eye[ip][1] + vhhcent*pan_torq[ip][1] + vvvcent*pan_up[ip][1];
              dzz = pan_eye[ip][2] + vhhcent*pan_torq[ip][2] + vvvcent*pan_up[ip][2];

              vnormi = (float)(1.0 / sqrt(dxx*dxx + dyy*dyy + dzz*dzz));
              dxx = dxx * vnormi;
              dyy = dyy * vnormi;
              dzz = dzz * vnormi;
          }*/

          /* Direction to center of subvolume */
          dxx = subminpos[0]+(float)0.5*subvolsiz[0]-eyepos[0];
          dyy = subminpos[1]+(float)0.5*subvolsiz[1]-eyepos[1];
          dzz = subminpos[2]+(float)0.5*subvolsiz[2]-eyepos[2];

          /* Direction to center of subvolume in coordinate attached to panel */
          vhh = dxx*pan_torq[ip][0] + dyy*pan_torq[ip][1] + dzz*pan_torq[ip][2]; 
          vnn = dxx* pan_eye[ip][0] + dyy* pan_eye[ip][1] + dzz* pan_eye[ip][2]; 
          vvv = dxx*  pan_up[ip][0] + dyy*  pan_up[ip][1] + dzz*  pan_up[ip][2]; 

          if(vnn <= (float)0.0) {
              /* Use panel normal if center of cube is behind your head */
              dxx = pan_eye[ip][0];
              dyy = pan_eye[ip][1];
              dzz = pan_eye[ip][2];
          } else {
              /* Limit direction to be within view frustum */
              vhh = vhh / vnn;
              vvv = vvv / vnn;
              if(vhh < pan_dlef[ip]) { vhh = pan_dlef[ip]; }
              if(vhh > pan_drit[ip]) { vhh = pan_drit[ip]; }
              if(vvv < pan_dbot[ip]) { vvv = pan_dbot[ip]; }
              if(vvv > pan_dtop[ip]) { vvv = pan_dtop[ip]; }
              dxx = pan_eye[ip][0] + vhh*pan_torq[ip][0] + vvv*pan_up[ip][0];
              dyy = pan_eye[ip][1] + vhh*pan_torq[ip][1] + vvv*pan_up[ip][1];
              dzz = pan_eye[ip][2] + vhh*pan_torq[ip][2] + vvv*pan_up[ip][2];

              /* Normalize resulting direction */
              vnormi = (float)(1.0 / sqrt(dxx*dxx + dyy*dyy + dzz*dzz));
              dxx = dxx * vnormi;
              dyy = dyy * vnormi;
              dzz = dzz * vnormi;
          }

          /* Base idir on max absolute value of coordinate
             of bounded center-of-subvolume vector */
          if(dxx > 0.) { axx = dxx; } else { axx = -dxx; }
          if(dyy > 0.) { ayy = dyy; } else { ayy = -dyy; }
          if(dzz > 0.) { azz = dzz; } else { azz = -dzz; }
          idir = 2;                                     // view along z-direction
          if(ayy > axx  &&  ayy > azz) { idir = 1; }    // view along y-direction
          if(axx > ayy  &&  axx > azz) { idir = 0; }    // view along x-direction

     /*********************************************************************
      * Derive min and max distances for edges of each tile               *
      * As a byproduct, determine if tile is OUTSIDE the viewing frustum  *
      * A tile is OUTSIDE of a convex faceted volume if all 8 vertices    *
      * of the tile are on the outer side of any one plane tangent to     *
      * a facet of that volume.  Viewing frustums are convex.             *
      *********************************************************************/
      edgemin = 10000000.;
      edgemax = 0.;
      isoutside_n = 1;  /* near   */
      isoutside_f = 1;  /* far    */
      isoutside_t = 1;  /* top    */
      isoutside_b = 1;  /* bottom */
      isoutside_l = 1;  /* left   */
      isoutside_r = 1;  /* right  */
      for(i=0; i<2; i++) { xx = subminpos[0]+(float)i*subvolsiz[0]-eyepos[0];
      for(j=0; j<2; j++) { yy = subminpos[1]+(float)j*subvolsiz[1]-eyepos[1];
      for(k=0; k<2; k++) { zz = subminpos[2]+(float)k*subvolsiz[2]-eyepos[2];
        edgedist = (xx*pan_eye[ip][0]+yy*pan_eye[ip][1]+zz*pan_eye[ip][2]);
        if(edgemin > edgedist) { edgemin = edgedist; }
        if(edgemax < edgedist) { edgemax = edgedist; }

        if(edgedist > m_ShvrUse.fNearClip - softdistance   ) { isoutside_n = 0; }
        if(edgedist < m_ShvrUse.fFarrClip + softdistance   ) { isoutside_f = 0; }
        if(xx*pan_top_o[ip][0]+yy*pan_top_o[ip][1]
                              +zz*pan_top_o[ip][2] < 0. ) { isoutside_t = 0; }
        if(xx*pan_bot_o[ip][0]+yy*pan_bot_o[ip][1]
                              +zz*pan_bot_o[ip][2] < 0. ) { isoutside_b = 0; }
        if(xx*pan_lef_o[ip][0]+yy*pan_lef_o[ip][1]
                              +zz*pan_lef_o[ip][2] < 0. ) { isoutside_l = 0; }
        if(xx*pan_rit_o[ip][0]+yy*pan_rit_o[ip][1]
                              +zz*pan_rit_o[ip][2] < 0. ) { isoutside_r = 0; }
      }}}

/******************************************************************************
 *    Test for Object oriented clipping planes                                *
 ******************************************************************************/
//fprintf(stderr, "XXX: iTexBndry=%d\n", iTexBndry);
      if(iTexBndry == 1) {
            xx0 = (float)0.0;
            xx1 = (float)1.0;
            yy0 = (float)0.0;
            yy1 = (float)1.0;
            zz0 = (float)0.0;
            zz1 = (float)1.0;
      } else {
            /*xx0 =                 (float)1.0  / (float)(nxsb2-1);
            xx1 = ((float)nxsub - (float)2.0) / (float)(nxsb2-1);
            yy0 =                 (float)1.0  / (float)(nysb2-1);
            yy1 = ((float)nysub - (float)2.0) / (float)(nysb2-1);
            zz0 =                 (float)1.0  / (float)(nzsb2-1);
            zz1 = ((float)nzsub - (float)2.0) / (float)(nzsb2-1);
            tex fix */
            fNvoxIn = (float)m_iNBoundaries;
            if(fNvoxIn < (float)0.5) { fNvoxIn = (float)0.5; }
            /*if(ResidentTextures) {
                submintex[0] =                 fNvoxIn  / (float)max_nxsb2;
                submaxtex[0] = ((float)nxsub - fNvoxIn) / (float)max_nxsb2;
                submintex[1] =                 fNvoxIn  / (float)max_nysb2;
                submaxtex[1] = ((float)nysub - fNvoxIn) / (float)max_nysb2;
                submintex[2] =                 fNvoxIn  / (float)max_nzsb2;
                submaxtex[2] = ((float)nzsub - fNvoxIn) / (float)max_nzsb2;
            }
            else {*/
                submintex[0] =                 fNvoxIn  / (float)nxsb2;
                submaxtex[0] = ((float)nxsub - fNvoxIn) / (float)nxsb2;
                submintex[1] =                 fNvoxIn  / (float)nysb2;
                submaxtex[1] = ((float)nysub - fNvoxIn) / (float)nysb2;
                submintex[2] =                 fNvoxIn  / (float)nzsb2;
                submaxtex[2] = ((float)nzsub - fNvoxIn) / (float)nzsb2;
            //}
      }

      iClipThis0 = 0;

      xx0 = submintex[0];
      xx1 = submaxtex[0];
      yy0 = submintex[1];
      yy1 = submaxtex[1];
      zz0 = submintex[2];
      zz1 = submaxtex[2];

      px0 = subminpos[0];
      px1 = submaxpos[0];
      py0 = subminpos[1];
      py1 = submaxpos[1];
      pz0 = subminpos[2];
      pz1 = submaxpos[2];

      ilo = 0; ihi = nxsub;
      jlo = 0; jhi = nysub;
      klo = 0; khi = nzsub;

      if(m_ShvrUse.ifauxclip[0] == 2) {
            px0c = m_ShvrUse.auxclip[0][3];
            px1c = m_ShvrUse.auxclip[1][3];
            py0c = m_ShvrUse.auxclip[2][3];
            py1c = m_ShvrUse.auxclip[3][3];
            pz0c = m_ShvrUse.auxclip[4][3];
            pz1c = m_ShvrUse.auxclip[5][3];

            xx0c = submintex[0] + (submaxtex[0]-submintex[0])*(px0c-subminpos[0])/subvolsiz[0];
            xx1c = submintex[0] + (submaxtex[0]-submintex[0])*(px1c-subminpos[0])/subvolsiz[0];
            yy0c = submintex[1] + (submaxtex[1]-submintex[1])*(py0c-subminpos[1])/subvolsiz[1];
            yy1c = submintex[1] + (submaxtex[1]-submintex[1])*(py1c-subminpos[1])/subvolsiz[1];
            zz0c = submintex[2] + (submaxtex[2]-submintex[2])*(pz0c-subminpos[2])/subvolsiz[2];
            zz1c = submintex[2] + (submaxtex[2]-submintex[2])*(pz1c-subminpos[2])/subvolsiz[2];

            if(xx0c > xx0) { xx0 = xx0c;    px0 = px0c; }
            if(xx1c < xx1) { xx1 = xx1c;    px1 = px1c; }
            if(yy0c > yy0) { yy0 = yy0c;    py0 = py0c; }
            if(yy1c < yy1) { yy1 = yy1c;    py1 = py1c; }
            if(zz0c > zz0) { zz0 = zz0c;    pz0 = pz0c; }
            if(zz1c < zz1) { zz1 = zz1c;    pz1 = pz1c; }
            if(xx0 >= xx1) { iClipThis0 = 1; }
            if(yy0 >= yy1) { iClipThis0 = 1; }
            if(zz0 >= zz1) { iClipThis0 = 1; }

            iloc = (int)((float)nxsub * (px0c-subminpos[0])/subvolsiz[0]);
            ihic = (int)((float)nxsub * (px1c-subminpos[0])/subvolsiz[0]);
            jloc = (int)((float)nysub * (py0c-subminpos[1])/subvolsiz[1]);
            jhic = (int)((float)nysub * (py1c-subminpos[1])/subvolsiz[1]);
            kloc = (int)((float)nzsub * (pz0c-subminpos[2])/subvolsiz[2]);
            khic = (int)((float)nzsub * (pz1c-subminpos[2])/subvolsiz[2]);
            if(iloc > ilo) { ilo = iloc; }
            if(ihic < ihi) { ihi = ihic; }
            if(jloc > jlo) { jlo = jloc; }
            if(jhic < jhi) { jhi = jhic; }
            if(kloc > klo) { klo = kloc; }
            if(khic < khi) { khi = khic; }
      }

/********************************************************************************
 *    SKIP this tile if it is outside viewing frustum OR object oriented clips  *
 *    this is end of jsub (tile) loop if tile is outside                        *
 ********************************************************************************/
//fprintf(stderr, "XXX: before inside/outside if test\n");
      if(   isoutside_n == 1 || isoutside_f == 1 || isoutside_t == 1
         || isoutside_b == 1 || isoutside_l == 1 || isoutside_r == 1 || iClipThis0 == 1) {
      } else {
//fprintf(stderr, "XXX: in else if block\n");
            /* Resident Textures */
            if(ResidentTextures) { 
                /* Don't read in brick if its corresponding
                   texture is already in texture memory */
                if(bricktex[isub] >= 0)    {        
                    isubcur = isub;
                }
            }
     /************************************************************
      * Read in subtile isub using DIRECT IO (if not already in) *
      ************************************************************/
    if(isubcur != isub) {
//fprintf(stderr, "XXX: begin: isubcur=%d   isub=%4d\n", isubcur, isub);
        ntiles++;
        nBricksRead++; /* Resident Textures */
        
        /* Wait for IO to finish and get pointer to buffer */
        if(m_iTimersOn == 1) { tsEvent.StampTime("Begin GetCacheBuffer"); }

//fprintf(stderr, "XXX: before lpGetCacheBuffer\n");
        bobsubZ = (unsigned char *)lpGetCacheBuffer(isub);
//fprintf(stderr, "XXX: after  lpGetCacheBuffer\n");

//fprintf(stderr, "XXX: 000\n");
        if(m_iTimersOn == 1) { sprintf(str, "End   GetCacheBuffer : %d bytes", nzsub*nysub*nxsub);
                               tsEvent.StampTime(str); }

        /* Most fast implementations of OpenGL require power of 2 dimensions
           If tile is NOT a power of 2 in x or y:
             1) suffle memory pointers
             2) copy into a format which IS power of 2 in x and y */

//fprintf(stderr, "XXX: 000\n");
        if(ResidentTextures) {
//fprintf(stderr, "XXX: 100\n");
            /* If brick is smaller than the standard, max-sized brick,
               fill it out with zeros to make it standard sized */
            if(nxsub != max_nxsb2  ||  nysub != max_nysb2 || nzsub != max_nzsb2) {
                if(m_iTimersOn == 1) { tsEvent.StampTime("Begin OddBrickFill"); }

                /* Ensure bobsubZ0 contains all zeros */
                // dhp1: this was taking a LONG time
                // for(i=0;i<texsize;i++) { bobsubZ0[i] = 0; }

                bobin = bobsubZ;
                bobsubZ = bobsubZ0;
                ic = 0;

                for(k=0; k<nzsub; k++) {
                for(j=0; j<nysub; j++) { joff = max_nxsb2*(j + max_nysb2 * k);
                    memcpy(&bobsubZ[joff], &bobin[ic], nxsub);
                    ic += nxsub;
                }}

                if(m_iTimersOn == 1) { sprintf(str, "End   OddBrickFill : %d bytes", nzsub*nysub*nxsub);
                                           tsEvent.StampTime(str); }
            }
//fprintf(stderr, "XXX: 109\n");
        } else {
//fprintf(stderr, "XXX: 200\n");
            if(nxsb2 != nxsub  ||  nysb2 != nysub) {
                if(m_iTimersOn == 1) { tsEvent.StampTime("Begin OddBrickFillNR"); }
                bobin = bobsubZ;
                bobsubZ = bobsubZ0;
                ic = 0;
                
                for(k=0; k<nzsub; k++) {
                for(j=0; j<nysub; j++) { joff = nxsb2*(j + nysb2 * k);
                    memcpy(&bobsubZ[joff], &bobin[ic], nxsub);
                    ic += nxsub;
                }}

                if(m_iTimersOn == 1) { sprintf(str, "End   OddBrickFillNR : %d bytes", nzsub*nysub*nxsub);
                                           tsEvent.StampTime(str); }
            }
//fprintf(stderr, "XXX: 209\n");
        }

        sprintf(tile_full, "%s%s\0", tile_path, tile_file[isub]);
        if(iOpac == 1) {
            sprintf(tile_fileO, "%s", tile_file[isub]);
            tile_fileO[ 3] = 'A';
            tile_fileO[ 4] = 'L';
            tile_fileO[ 5] = 'P';
            tile_fileO[ 6] = 'H';
            tile_fileO[ 7] = 'A';
            tile_fileO[ 8] = '_';
            tile_fileO[ 9] = '_';
            tile_fileO[10] = '_';
            sprintf(m_sTileO, "%s%s\0", tile_path, tile_fileO);
        }
//fprintf(stderr, "XXX: 700\n");

        if(iOpac == 1  ||  strcmp(m_sStats, "cross") == 0) {
            if(ResidentTextures) {
                if(max_nxsb2 == nxsub  &&  max_nysb2 == nysub) {
                    if(m_iBufferedIO == 1) { dwGot = dhpReadFromFileNoBuf(bobOZ, m_sTileO, offset64t, (DWORD)nbytes); }
                    else                   { dwGot = dhpReadFromFileSync( bobOZ, m_sTileO, offset64t, (DWORD)nbytes); }
                } else {
                    if(m_iBufferedIO == 1) { dwGot = dhpReadFromFileNoBuf(bobOY, m_sTileO, offset64t, (DWORD)nbytes); } 
                    else                   { dwGot = dhpReadFromFileSync( bobOY, m_sTileO, offset64t, (DWORD)nbytes); }
                    ic = 0;
                    for(k=0; k<nzsub; k++) {
                    for(j=0; j<nysub; j++) {
                    for(i=0; i<nxsub; i++) {
                        bobOZ[i + max_nxsb2*(j + max_nysb2 * k)] = bobOY[ic];
                        ic++;
                    }}}
                }
            } else {
                if(nxsb2 == nxsub  &&  nysb2 == nysub) {
                if(m_iBufferedIO == 1) { dwGot = dhpReadFromFileNoBuf(bobOZ, m_sTileO, offset64t, (DWORD)nbytes); }
                else                   { dwGot = dhpReadFromFileSync( bobOZ, m_sTileO, offset64t, (DWORD)nbytes); }
            } else {
                if(m_iBufferedIO == 1) { dwGot = dhpReadFromFileNoBuf(bobOY, m_sTileO, offset64t, (DWORD)nbytes); } 
                else                   { dwGot = dhpReadFromFileSync( bobOY, m_sTileO, offset64t, (DWORD)nbytes); }
                ic = 0;
                for(k=0; k<nzsub; k++) {
                for(j=0; j<nysub; j++) {
                for(i=0; i<nxsub; i++) {
                    bobOZ[i + nxsb2*(j + nysb2 * k)] = bobOY[ic];
                    ic++;
                }}}
            }
            }
        }
//fprintf(stderr, "XXX: 900\n");

        bSetBobX = FALSE;
        bSetBobY = FALSE;
        isubcur = isub;
        ireload = 1;

//fprintf(stderr, "XXX: end  : isubcur=%d   isub=%4d\n", isubcur, isub);

      }
//fprintf(stderr, "XXX: A\n");
/*****************************************************************************
 * BEGIN multiple pannel stuff                                               *
 *****************************************************************************/
//fprintf(stderr, "XXX: B\n");
      if(ip != ipcur) {
//fprintf(stderr, "XXX: C\n");

       /****************************************************
        *  set up frustum for pannel ip                    *
        ****************************************************/
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();

        dleft   = dnearclip * pan_dlef[ip];
        dright  = dnearclip * pan_drit[ip];
        dbottom = dnearclip * pan_dbot[ip];
        dtop    = dnearclip * pan_dtop[ip];
        glFrustum( dleft, dright, dbottom, dtop, dnearclip, dfarclip );
        gluLookAt(m_ShvrUse.fEye[0], m_ShvrUse.fEye[1], m_ShvrUse.fEye[2],
                  pan_lookat[ip][0], pan_lookat[ip][1], pan_lookat[ip][2],
                  pan_up[ip][0],     pan_up[ip][1],     pan_up[ip][2]);

        xcenteye = m_ShvrUse.fEye[0] - pan_lookat[ip][0];
        ycenteye = m_ShvrUse.fEye[1] - pan_lookat[ip][1];
        zcenteye = m_ShvrUse.fEye[2] - pan_lookat[ip][2];
        vnormi = 1.0 / sqrt(xcenteye*xcenteye +
                            ycenteye*ycenteye +
                            zcenteye*zcenteye  );
        xcenteye *= vnormi;
        ycenteye *= vnormi;
        zcenteye *= vnormi;

       /*************************************************************
        *  swap out old pannel (if nec.) and swap in new pannel ip  *
        *************************************************************/
        glMatrixMode(GL_TEXTURE);
        if(Tex3d)
            glDisable(GL_TEXTURE_3D);
        else
            glDisable(GL_TEXTURE_2D);
        glDisable(GL_BLEND);
        if(ipcur > -1) {
            glReadPixels(0,0,m_ShvrUse.iWidth,m_ShvrUse.iHeight,
                GL_RGB,GL_UNSIGNED_BYTE,rgbpannel[ipcur]);
        }

        glDrawPixels(m_ShvrUse.iWidth, m_ShvrUse.iHeight,
            GL_RGB,GL_UNSIGNED_BYTE,rgbpannel[ip]);
        ipcur = ip;
        glEnable(GL_BLEND);
        if(Tex3d)
            glEnable(GL_TEXTURE_3D);
        else
            glEnable(GL_TEXTURE_2D);
      }
//fprintf(stderr, "XXX: D\n");
/*****************************************************************************
 * END   multiple pannel stuff                                               *
 *****************************************************************************/

     /***********************************************************************
      *  SET scale and offset of texture memory for subvolume               *
      *  ASSUMES glLoadIdentity sets limits of                              *
      *          --> subtexture <--  to (0,1)x(0,1)x(0,1)                   *
      *          which get mapped to                                        *
      *          (subminpos[i], subminpos[i]+subvolsiz[i])x for(i=0,1,2)    *
      ***********************************************************************/

//fprintf(stderr, "XXX: E1\n");
      while ( glGetError() != GL_NO_ERROR ) ;
//fprintf(stderr, "XXX: E2\n");

     /*********************************************************************
      *    THIS IS WHERE ALL ALPHA BLENDING IS DONE                       *
      *********************************************************************/

     /************************************************************************
      * Collect statistics                                                   *
      ************************************************************************/
      if(strcmp(m_sStats, "cross") == 0) {
            for(k=klo; k<khi; k++) {
            for(j=jlo; j<jhi; j++) {
            for(i=ilo; i<ihi; i++) {
                if(ResidentTextures) {
                    ijk = i + max_nxsb2*(j + max_nysb2 * k);
                }
                else {
                    ijk = i + nxsb2*(j + nysb2 * k);
                }
                iCross[(int)bobsubZ[ijk] + 256*(int)bobOZ[ijk]]++;
            }}}
      }
      if(strcmp(m_sStats, "pdf") == 0) {
            for(k=klo; k<khi; k++) {
            for(j=jlo; j<jhi; j++) {
            for(i=ilo; i<ihi; i++) {
                if(ResidentTextures) {
                    iPDF[(int)bobsubZ[i + max_nxsb2*(j + max_nysb2 * k)]]++;
                } else {
                    iPDF[(int)bobsubZ[i + nxsb2*(j + nysb2 * k)]]++;
                }
            }}}
      }

     /************************************************************************
      * Calculate vector from eye to brick for correct rendering orientation *
      ************************************************************************/

      // do 1st time only
      if(iDoAnnotationBeforeRendering == 1) {
        annotation(1, eyepos);
        iDoAnnotationBeforeRendering = 0;
      }

/* 3d textures */
//fprintf(stderr, "XXX: F\n");
    if(Tex3d) {

//fprintf(stderr, "XXX: start 3d texture code\n");

        if(m_iTimersOn == 1) { tsEvent.StampTime("Begin 3dTex_Load"); }
// dhp1
        /* Scale and shift texture to correspond to subvolume */
       fNvoxIn = (float)m_iNBoundaries;
       if(fNvoxIn < (float)0.5) { fNvoxIn = (float)0.5; }
        glMatrixMode(GL_TEXTURE);
        glLoadIdentity();
        if(ResidentTextures) {
            glScalef(((float)nxsub-2.0*fNvoxIn)/((float)(max_nxsb2)*subvolsiz[0]),
                     ((float)nysub-2.0*fNvoxIn)/((float)(max_nysb2)*subvolsiz[1]),
                     ((float)nzsub-2.0*fNvoxIn)/((float)(max_nzsb2)*subvolsiz[2]) );
        } else {
            glScalef(((float)nxsub-2.0*fNvoxIn)/((float)(nxsb2)*subvolsiz[0]),
                     ((float)nysub-2.0*fNvoxIn)/((float)(nysb2)*subvolsiz[1]),
                     ((float)nzsub-2.0*fNvoxIn)/((float)(nzsb2)*subvolsiz[2]) );
        }
          glTranslatef(0.5*subvolsiz[0]/((float)nxsub-2.0*fNvoxIn)-subminpos[0],
                       0.5*subvolsiz[1]/((float)nysub-2.0*fNvoxIn)-subminpos[1],
                       0.5*subvolsiz[2]/((float)nzsub-2.0*fNvoxIn)-subminpos[2] );

        /* Resident Textures */
        if(ResidentTextures) {    
            /* count = texture object number */
            /* isub = brick number */
#ifdef INTELLIGENT_OVERWRITE
            /* There are three distinct situations to consider: */
            /* 1. Brick is not in cache, cache is not full */
            /* 2. Brick is not in cache, cache is full */
            /* 3. Brick is in cache */

            if(count == 0) count = 1;
            
//fprintf(stderr, "XXX: bricktex[isub]=%d\n", bricktex[isub]);

            if(bricktex[isub] < 0) {
                /* Situations 1 & 2 */
                
                if(nBricksInCache < nBricksFitInCache) {
                    /* Situation 1: Brick is not in cache, cache is not full */
                    /* Place brick in next available texture slot */
                    
                    /* Associate brick with texture object */
                    bricktex[isub] = count; 
                    texbrick[count] = isub;            
                    
                    /* Bind appopriate texture object */
                    glBindTexture(GL_TEXTURE_3D, texture[bricktex[isub]]);
                    cgGLSetTextureParameter(textureHandle, texture[bricktex[isub]]);

#ifdef PBO
                    glBindBuffer(GL_PIXEL_UNPACK_BUFFER_EXT, texBuffer);
                    glBufferData(GL_PIXEL_UNPACK_BUFFER_EXT, texsize, NULL, GL_STREAM_DRAW);
                    pboMemory = glMapBuffer(GL_PIXEL_UNPACK_BUFFER_EXT, GL_WRITE_ONLY);
                    memcpy(pboMemory, bobsubZ, texsize);
                    if(!glUnmapBuffer(GL_PIXEL_UNPACK_BUFFER_EXT)) {
                        printf ("glUnmapBuffer error\n");
                        exitHVR(1);
                    }

                    /* Overwrite existing texture */
                    glTexSubImage3D(GL_TEXTURE_3D,0,0,0,0,max_nxsb2,max_nysb2,max_nzsb2,GL_LUMINANCE,GL_UNSIGNED_BYTE,BUFFER_OFFSET(0));
#else
#ifdef NEWTEX
                    /* Overwrite existing texture */
                    glTexSubImage3D(GL_TEXTURE_3D,0,0,0,0,max_nxsb2,max_nysb2,max_nzsb2,GL_LUMINANCE,GL_UNSIGNED_BYTE,bobsubZ);
#else
                    /* Load new texture object into card */
                    glTexImage3D(GL_TEXTURE_3D,0,1,max_nxsb2,max_nysb2,max_nzsb2,0,GL_LUMINANCE,GL_UNSIGNED_BYTE,bobsubZ);
#endif

#endif
                    /* Increment Texture Slot Pointer */
                    count = NextSlot(count,nBricksFitInCache);
                    
                    /* Update stats */
                    nBricksLoaded++;
                    nBricksInCache++; 
                    texmem += texsize;
                }
            //  else if(nBricksInCache = nBricksFitInCache)   it was this: ack!!
                else if(nBricksInCache == nBricksFitInCache) {
                    /* Situation 2: Brick is not in cache, cache is full */
                    /* Find texture slot to overwrite, place brick there */

                    int tex_slot = -1;

                    if(!all_in_use && prev_count != -1){
                        if(prev_count == 1 && count == 1)
                            prev_count = nBricksFitInCache - 1;
                        
                        while(count != prev_count) {
                            int brick_count = 0;
                            int brick = -1;
                            bool inFrustum = FALSE;

                            /* Get number of Brick which is in current texture slot */
                            brick = texbrick[count];

                            /* Find out if this brick is in frustum or not */
                            for(brick_count=0; brick_count<nBricksInFrustum; brick_count++) {
                                if(brick == iTileRender[brick_count]) {
                                    inFrustum = TRUE;
                                    all_in_use = FALSE;
                                    break;
                                }
                            }
                            
                            if(inFrustum) {
                                /* Brick is in frustum, don't overwrite it */
                                count = NextSlot(count,nBricksFitInCache);
                            }
                            else {
                                /* Brick is not in frustum, ok to overwrite */
                                tex_slot = count;
                                break;
                            }
                        }

                        /* If no slots were available for overwriting, overwrite slot 0 */
                        if(tex_slot == -1) {
                            all_in_use = TRUE;
                            count = 0;

                            /* Bind appopriate texture object */
                            glBindTexture(GL_TEXTURE_3D, texture[0]);
                            cgGLSetTextureParameter(textureHandle, texture[0]);
                        }
                        else
                        {
                            /* Old brick is no longer in cache */
                            if(texbrick[count] >= 0)
                                bricktex[texbrick[count]] = -1;

                            /* Associate brick with texture object */
                            bricktex[isub] = count;
                            texbrick[count] = isub;

                            /* Bind appopriate texture object */
                            glBindTexture(GL_TEXTURE_3D, texture[bricktex[isub]]);
                            cgGLSetTextureParameter(textureHandle, texture[bricktex[isub]]);
                        }

                    }
                    else {
                        /* Bind appopriate texture object */
                        glBindTexture(GL_TEXTURE_3D, texture[0]);
                        cgGLSetTextureParameter(textureHandle, texture[0]);
                        count = 0;
                    }

#ifdef PBO
                    glBindBuffer(GL_PIXEL_UNPACK_BUFFER_EXT, texBuffer);
                    glBufferData(GL_PIXEL_UNPACK_BUFFER_EXT, texsize, NULL, GL_STREAM_DRAW);
                    pboMemory = glMapBuffer(GL_PIXEL_UNPACK_BUFFER_EXT, GL_WRITE_ONLY);
                    memcpy(pboMemory, bobsubZ, texsize);
                    if(!glUnmapBuffer(GL_PIXEL_UNPACK_BUFFER_EXT)) {
                        printf ("glUnmapBuffer error\n");
                        exitHVR(1);
                    }

                    /* Overwrite existing texture */
                    glTexSubImage3D(GL_TEXTURE_3D,0,0,0,0,max_nxsb2,max_nysb2,max_nzsb2,GL_LUMINANCE,GL_UNSIGNED_BYTE,BUFFER_OFFSET(0));
#else
                    /* Overwrite existing texture */
                    glTexSubImage3D(GL_TEXTURE_3D,0,0,0,0,max_nxsb2,max_nysb2,max_nzsb2,GL_LUMINANCE,GL_UNSIGNED_BYTE,bobsubZ);
#endif
                    /* Increment Texture Slot Pointer */
                    count = NextSlot(count,nBricksInCache);

                    nBricksLoaded++;
                }
                else {
                    /* ERROR: Cache over-filled */
                    fprintf(error_log,"TCS ERROR: Cache over-filled:  ");
                    fprintf(error_log,"%d bricks in Cache, %d fit in cache.\n",nBricksInCache,nBricksFitInCache);
                    exitHVR(CACHE_FAILED);
                }
            } else {
                /* Situation 3: Brick is already in cache */

                /* Bind appopriate texture object */
                glBindTexture(GL_TEXTURE_3D, texture[bricktex[isub]]);
                cgGLSetTextureParameter(textureHandle, texture[bricktex[isub]]);

                #ifdef CHECK_RESIDENCY
                glGetTexParameteriv(GL_TEXTURE_3D,GL_TEXTURE_RESIDENT,&resident);            
                if(resident == GL_FALSE) {
                    fprintf(error_log,"TEXTURE RESIDENCY ERROR");
                    exitHVR(CACHE_FAILED);
                }
                #endif
            }

            nBricksRendered++;
#else
            if(TexScheme == 0) 
            {
                /* Texture Memory Management */
                if(bricktex[isub]<0) {    
                    new_brick = TRUE; /* non-resident brick */
                    if(texmem_full) {    
                        if(texbrick[count]>=0) {
                            /* dis-associate existing brick with texture */                
                            bricktex[texbrick[count]] = -1;
                        }
                    }
                    bricktex[isub] = count; /* associate brick with texture */
                    texbrick[count] = isub; /* associate texture with brick */
                    count++; /* keeps track of location in tex memory */
                }
                else
                    new_brick = FALSE; /* brick is resident in texture memory */

                /* Bind appopriate texture object */
                glBindTexture(GL_TEXTURE_3D, texture[bricktex[isub]]);
                cgGLSetTextureParameter(textureHandle, texture[bricktex[isub]]);
                
                /* Keep track of tex memory */
                texmem_previous = texmem;

                if(m_iTimersOn == 1) { tsEvent.StampTime("Begin 3dTexLoad_BOB"); }

                if(new_brick) {    
                    /* New, non-resident texture, need to load into card */
                    if(texmem_full) {
                        /* Memory is full, overwrite existing texture */
                        glTexSubImage3D(GL_TEXTURE_3D,0,0,0,0,max_nxsb2,max_nysb2,max_nzsb2,GL_LUMINANCE,GL_UNSIGNED_BYTE,bobsubZ);
                    }
                    else {
                        /* Load new texture into card */
                        glTexImage3D(GL_TEXTURE_3D,0,1,max_nxsb2,max_nysb2,max_nzsb2,0,GL_LUMINANCE,GL_UNSIGNED_BYTE,bobsubZ);
                        nBricksInCache++;
                    }
                    nBricksLoaded++;
                    texmem += texsize;
                }
                else {    
                    /* Texture is resident, no load necessary */
                    #ifdef CHECK_RESIDENCY
                    glGetTexParameteriv(GL_TEXTURE_3D,GL_TEXTURE_RESIDENT,&resident);            
                    if(resident == GL_FALSE) {
                        fprintf(error_log,"TEXTURE RESIDENCY ERROR");
                        exitHVR(CACHE_FAILED);
                    }
                    #endif
                }

                if(m_iTimersOn == 1) { tsEvent.StampTime("End   3dTexLoad_BOB"); }
            }
            else if(TexScheme == 1)
            {
                /* Texture Memory Management 
                   Bind appopriate texture object */
                if(bricktex[isub]<0)
                {    new_brick = TRUE; /* new, non-resident brick */
                    if(texmem_full)
                    {   /* texture memory is full */
                        overwrite = TRUE;
                        glBindTexture(GL_TEXTURE_3D, texture[0]);
                        cgGLSetTextureParameter(textureHandle, texture[0]);
                    }
                    else
                    {    /* texture memory is NOT full */
                        bricktex[isub] = count;
                        glBindTexture(GL_TEXTURE_3D, texture[bricktex[isub]]);
                        cgGLSetTextureParameter(textureHandle, texture[bricktex[isub]]);
                        count++;
                    }
                }
                else
                {    /* texture is resident in texture memory */
                    new_brick = FALSE;
                    if(texmem_full)
                        overwrite = FALSE;
                    glBindTexture(GL_TEXTURE_3D, texture[bricktex[isub]]);
                    cgGLSetTextureParameter(textureHandle, texture[bricktex[isub]]);
                }
                
                texmem_previous = texmem;
                if(m_iTimersOn == 1) { tsEvent.StampTime("Begin 3dTexLoad_BOB"); }
                
                /* New, non-resident texture, need to load into card */
                if(texmem_full) {
                    if(overwrite) {
                        /* Memory is full, overwrite existing texture */
                        glTexSubImage3D(GL_TEXTURE_3D,0,0,0,0,max_nxsb2,max_nysb2,max_nzsb2,GL_LUMINANCE,GL_UNSIGNED_BYTE,bobsubZ);
                        nBricksLoaded++;
                        texmem += texsize;
                    }
                    else {
                        /* Texture is resident, no load necessary */
                        #ifdef CHECK_RESIDENCY
                        glGetTexParameteriv(GL_TEXTURE_3D,GL_TEXTURE_RESIDENT,&resident);            
                        if(resident == GL_FALSE) {
                            fprintf(error_log,"TEXTURE RESIDENCY ERROR");
                            exitHVR(CACHE_FAILED);
                        }
                        #endif
                    }
                }
                else {
                    /* Texture memory is not full */
                    if(new_brick) {
                        /* Current texture is not resident, mem is not full.
                           Load new texture into texture memory */
                        glTexImage3D(GL_TEXTURE_3D,0,1,max_nxsb2,max_nysb2,max_nzsb2,0,GL_LUMINANCE,GL_UNSIGNED_BYTE,bobsubZ);
                        nBricksLoaded++;
                        nBricksInCache++;
                        texmem += texsize;
                    }
                    else {
                        /* Texture is resident, no load necessary */
                        #ifdef CHECK_RESIDENCY
                        glGetTexParameteriv(GL_TEXTURE_3D,GL_TEXTURE_RESIDENT,&resident);            
                        if(resident == GL_FALSE) {
                            fprintf(error_log,"TEXTURE RESIDENCY ERROR");
                            exitHVR(CACHE_FAILED);
                        }
                        #endif
                    }
                }

                if(m_iTimersOn == 1) { tsEvent.StampTime("End   3dTexLoad_BOB"); }
            }
#endif
        } else { /* Don't use resident textures */

//fprintf(stderr, "XXX: not useing resident textures\n");

            if(m_bColorTableIsAvailable) {
                if(m_iTimersOn == 1) { tsEvent.StampTime("Begin 3dTexLoad_BOB"); }
                glBindTexture(GL_TEXTURE_3D, dataTexture);
                cgGLSetTextureParameter(textureHandle, dataTexture);
                glTexImage3D(GL_TEXTURE_3D,0,1,nxsb2,nysb2,nzsb2,0,GL_LUMINANCE,GL_UNSIGNED_BYTE,bobsubZ);
                nBricksLoaded++;
                texmem += nbytes;
                if(m_iTimersOn == 1) { tsEvent.StampTime("End   3dTexLoad_BOB"); }
            } else {
                if(m_iTimersOn == 1) { tsEvent.StampTime("Begin 3dTexLoad_RGBA"); }
                glBindTexture(GL_TEXTURE_3D, RGBATexture);
                for(i=0; i<nxsb2*nysb2*nzsb2; i++) { sol[i] = m_lrgba[bobsubZ[i]]; }
                glTexImage3D(GL_TEXTURE_3D,0,4,nxsb2,nysb2,nzsb2,0,GL_RGBA,GL_UNSIGNED_BYTE,sol);
                nBricksLoaded++;
                texmem += nbytes*4;
                if(m_iTimersOn == 1) { tsEvent.StampTime("End   3dTexLoad_RGBA"); }
            }
        }

        if(m_iTimersOn == 1) { tsEvent.StampTime("End   3dTex_Load"); }
        if(m_iTimersOn == 1) { tsEvent.StampTime("Begin 3dTex_befDraw"); }
        
        /* DERIVE edges of subvolume for section calculation */

        // XYZ ranges of subvolume
        for(jdir=0; jdir<3; jdir++) {
        for(iside=0; iside<2; iside++) {
          range_side_dir[iside][jdir] = subminpos[jdir] + (float)iside * subvolsiz[jdir];
        }}

        // Impose aux clip planes
        if(m_ShvrUse.ifauxclip[0] == 2) {
          px0c = m_ShvrUse.auxclip[0][3];
          px1c = m_ShvrUse.auxclip[1][3];
          py0c = m_ShvrUse.auxclip[2][3];
          py1c = m_ShvrUse.auxclip[3][3];
          pz0c = m_ShvrUse.auxclip[4][3];
          pz1c = m_ShvrUse.auxclip[5][3];
          if(range_side_dir[0][0] < px0c) { range_side_dir[0][0] = px0c; }
          if(range_side_dir[0][1] < py0c) { range_side_dir[0][1] = py0c; }
          if(range_side_dir[0][2] < pz0c) { range_side_dir[0][2] = pz0c; }
          if(range_side_dir[1][0] > px1c) { range_side_dir[1][0] = px1c; }
          if(range_side_dir[1][1] > py1c) { range_side_dir[1][1] = py1c; }
          if(range_side_dir[1][2] > pz1c) { range_side_dir[1][2] = pz1c; }
        }

        /* XYZ coordinates of 8 edges for this subvolume */
        nedges = 0;
        for(ix=0; ix<2; ix++) { for(jx=ix; jx<2; jx++) {
        for(iy=0; iy<2; iy++) { for(jy=iy; jy<2; jy++) {
        for(iz=0; iz<2; iz++) { for(jz=iz; jz<2; jz++) {
          isedge = 0;
          if(jx >  ix  &&  jy == iy  &&  jz == iz) { isedge = 1; }
          if(jx == ix  &&  jy >  iy  &&  jz == iz) { isedge = 1; }
          if(jx == ix  &&  jy == iy  &&  jz >  iz) { isedge = 1; }
          if(isedge == 1) {
            edges[nedges][0][0] = range_side_dir[ix][0];
            edges[nedges][0][1] = range_side_dir[iy][1];
            edges[nedges][0][2] = range_side_dir[iz][2];
            edges[nedges][1][0] = range_side_dir[jx][0];
            edges[nedges][1][1] = range_side_dir[jy][1];
            edges[nedges][1][2] = range_side_dir[jz][2];
/*
            edges[nedges][0][0] = subminpos[0] + (float)ix * subvolsiz[0];
            edges[nedges][0][1] = subminpos[1] + (float)iy * subvolsiz[1];
            edges[nedges][0][2] = subminpos[2] + (float)iz * subvolsiz[2];
            edges[nedges][1][0] = subminpos[0] + (float)jx * subvolsiz[0];
            edges[nedges][1][1] = subminpos[1] + (float)jy * subvolsiz[1];
            edges[nedges][1][2] = subminpos[2] + (float)jz * subvolsiz[2];
*/
            nedges++;
          }
        }} }} }}

        /* Find range of ss offsets from eyepoint to
           rendering planes that intersect the subbrick */
        ssmin = dfarclip;
        ssmax = dnearclip;
        for(ix=0; ix<2; ix++) {
        for(iy=0; iy<2; iy++) {
        for(iz=0; iz<2; iz++) {
            dot = (subminpos[0] + (float)ix * subvolsiz[0] - eyepos[0]) * pan_eye[ip][0]
                + (subminpos[1] + (float)iy * subvolsiz[1] - eyepos[1]) * pan_eye[ip][1]
                + (subminpos[2] + (float)iz * subvolsiz[2] - eyepos[2]) * pan_eye[ip][2];
            if(ssmin > dot) { ssmin = dot; }
            if(ssmax < dot) { ssmax = dot; }
        }}}

        if(ssmin < dnearclip) { ssmin = dnearclip; }
        if(ssmax >  dfarclip) { ssmax =  dfarclip; }

        // ds  = subvolsiz[0]/(float)((nxsub-2*m_iNBoundaries) * nplanes);
        // alpha = alpha0 * (ds/(2.*cubesz[0]));
        ds  = (fLimits[1] - fLimits[0]) / (float)(Nx * nplanes);
        alpha = alpha0 * ds;

/*      dhp cludge: 3dtex min render plane spacing based on alphamin
        alphamin = 0.125f;
        if(alpha < alphamin) {
          ds = ds * alphamin / alpha;
          alpha = alpha0 * (ds/(2.*cubesz[0]));
        }
*/
        glColor4f(1.0, 1.0, 1.0, alpha);
        
        if(m_bColorTableIsAvailable) {
            /* Enable execution of Cg Fragment program (AAA) */
            cgGLBindProgram(fProgram);
            cgGLEnableProfile(fProfile);
            #ifdef DEBUG_CG
            CheckCgError();
            #endif
            /* Enable Cg Texture Parameters */
            cgGLEnableTextureParameter(textureHandle);
            cgGLEnableTextureParameter(textureHandle2);
            #ifdef DEBUG_CG
            CheckCgError();
            #endif
        }        

        if(m_iTimersOn == 1) { tsEvent.StampTime("End   3dTex_befDraw"); }
        if(m_iTimersOn == 1) { tsEvent.StampTime("Begin 3dTex_draw"); }
        ss = dfarclip;
        while(ss > ssmax) { ss -= ds; }
        while(ss > ssmin) {
            for(i=0; i<3; i++) { pos0[i] = eyepos[i] + ss * pan_eye[ip][i]; }
            nverts = section(pos0, pan_eye[ip], pan_torq[ip], pan_up[ip],
                              edges, nedges, tvert                        );
        
            if(nverts > 2) {

                 glBegin(GL_POLYGON);
                 for(i=0; i<nverts; i++) {
                   glTexCoord3fv(tvert[i]);
                   glVertex3fv(tvert[i]);
                 }
                 glEnd();

/* --------------------------------------------------------- */
      if(nvecobjs[isub] < 0) {
         // Generate list of vector objects inside this subvolume
         genvectorlist(node0, isub, subminpos, submaxpos);
       }

      if(nvecobjs[isub] > 0  &&  ss-ds > 0.0f) {   // VECTORS start

      /* Draw section of vectors between planes ss-ds & ss */
      if(m_bColorTableIsAvailable) {
          /* Disable execution of Cg Fragment Program (BBB) */
          cgGLDisableProfile(fProfile);
          cgGLDisableTextureParameter(textureHandle);
          cgGLDisableTextureParameter(textureHandle2);
      }

        glDisable(GL_BLEND);
        glDisable(GL_TEXTURE_3D);
        glEnable(GL_TEXTURE_2D);

        glColor4f(1.0, 1.0, 1.0, 1.0);

        glMatrixMode(GL_PROJECTION);
/*
        glPushMatrix();
        dleft   = (ss-ds) * pan_dlef[ip];
        dright  = (ss-ds) * pan_drit[ip];
        dbottom = (ss-ds) * pan_dbot[ip];
        dtop    = (ss-ds) * pan_dtop[ip];
        glLoadIdentity();
        glFrustum( dleft, dright, dbottom, dtop, (ss-ds), ss);
        gluLookAt(m_ShvrUse.fEye[0], m_ShvrUse.fEye[1], m_ShvrUse.fEye[2],
                  pan_lookat[ip][0], pan_lookat[ip][1], pan_lookat[ip][2],
                  pan_up[ip][0],     pan_up[ip][1],     pan_up[ip][2]);
*/

        glEnable(GL_CLIP_PLANE0);
        glEnable(GL_CLIP_PLANE1);
        glEnable(GL_CLIP_PLANE2);
        glEnable(GL_CLIP_PLANE3);
        glEnable(GL_CLIP_PLANE4);
        glEnable(GL_CLIP_PLANE5);

        aa[0]=  1.0; aa[1]=  0.0; aa[2]=  0.0; aa[3]= -subminpos[0];
        glClipPlane(GL_CLIP_PLANE0, aa);
        aa[0]= -1.0; aa[1]=  0.0; aa[2]=  0.0; aa[3]=  submaxpos[0];
        glClipPlane(GL_CLIP_PLANE1, aa);

        aa[0]=  0.0; aa[1]=  1.0; aa[2]=  0.0; aa[3]= -subminpos[1];
        glClipPlane(GL_CLIP_PLANE2, aa);
        aa[0]=  0.0; aa[1]= -1.0; aa[2]=  0.0; aa[3]=  submaxpos[1];
        glClipPlane(GL_CLIP_PLANE3, aa);

        aa[0]=  0.0; aa[1]=  0.0; aa[2]=  1.0; aa[3]= -subminpos[2];
        glClipPlane(GL_CLIP_PLANE4, aa);
        aa[0]=  0.0; aa[1]=  0.0; aa[2]= -1.0; aa[3]=  submaxpos[2];
        glClipPlane(GL_CLIP_PLANE5, aa);

        // xcenteye
        drawsurfaces(isub, ss-ds, ss, m_ShvrUse.fEye);

        glDisable(GL_CLIP_PLANE0);
        glDisable(GL_CLIP_PLANE1);
        glDisable(GL_CLIP_PLANE2);
        glDisable(GL_CLIP_PLANE3);
        glDisable(GL_CLIP_PLANE4);
        glDisable(GL_CLIP_PLANE5);

        glDisable(GL_TEXTURE_2D);
        glEnable(GL_TEXTURE_3D);
        glEnable(GL_BLEND);

/*
        glPopMatrix();
*/

        glColor4f(1.0, 1.0, 1.0, alpha);

        if(m_bColorTableIsAvailable) {
            /* Enable execution of Cg Fragment program (AAA) */
            cgGLBindProgram(fProgram);
            cgGLEnableProfile(fProfile);
            #ifdef DEBUG_CG
            CheckCgError();
            #endif
            /* Enable Cg Texture Parameters */
            cgGLEnableTextureParameter(textureHandle);
            cgGLEnableTextureParameter(textureHandle2);
            #ifdef DEBUG_CG
            CheckCgError();
            #endif
        }        
        }      // VECTORS: end
/* --------------------------------------------------------- */
                 nslices++;
            }
            ss -= ds;
        }
        if(m_iTimersOn == 1) { tsEvent.StampTime("End   3dTex_draw"); }
        if(m_iTimersOn == 1) { tsEvent.StampTime("Begin 3dTex_aftDraw"); }

/* --------------------------------------------------------- */
// BBB
        /* Disable execution of Cg Fragment Program */
        if(m_bColorTableIsAvailable) {
            cgGLDisableProfile(fProfile);
            cgGLDisableTextureParameter(textureHandle);
            cgGLDisableTextureParameter(textureHandle2);
        }
/* --------------------------------------------------------- */
        
#ifdef PBO
    //glBindBuffer(GL_PIXEL_UNPACK_BUFFER_EXT, 0);
#endif

#ifdef INTELLIGENT_OVERWRITE

#else
        if(ResidentTextures) {    
            nBricksRendered++;
            if(!texmem_full) 
            {    /* Determine if texture memory is full */
                if(texmem_previous <= max_tex_mem && texmem >= max_tex_mem)
                {    /* Texture memory is now full */
                    texmem_full = TRUE;
                    if(TexScheme == 0) {
                        end_count = count;
                        count = 0; /* Reset texture counter */
                        texmem = 0; /* Reset texture memory counter */
                    }
                    else if(TexScheme == 1)
                        overwrite = TRUE;
                }
            }
            else {    
                if(TexScheme == 0) {
                    /* If memory is full and counter has cylced through
                       to end of memory again, reset counter */
                    if(count == end_count) {
                        count = 0; /* Reset texture counter */
                        texmem = 0; /* Reset texture memory counter */
                    }
                }
                else if(TexScheme == 1)
                    overwrite = TRUE;
            }
        }
#endif
        
        if(m_iTimersOn == 1) { tsEvent.StampTime("End   3dTex_aftDraw"); }
/* End of 3D texture block */
    } else {
/* Start of 2D texture block */ 
        
        glMatrixMode(GL_TEXTURE);
        glLoadIdentity();
            
        if(m_bColorTableIsAvailable) {
            /* Enable execution of Cg Fragment program */
            cgGLBindProgram(fProgram);
            cgGLEnableProfile(fProfile);
            #ifdef DEBUG_CG
            CheckCgError();
            #endif
            /* Enable Cg Texture Parameters */
            cgGLEnableTextureParameter(textureHandle);
            cgGLEnableTextureParameter(textureHandle2);
            #ifdef DEBUG_CG
            CheckCgError();
            #endif
        }
        
    /***************************************************************************************
      * Determin which of 6 rendering orientations need bee done                            *
      * The 6 rendering orientations are geometrically equivalent to the 6 faces of a cube  *
      ***************************************************************************************/
      // Start by finding how many of the vertices of the subcube are in each face
      for(ixyz=0; ixyz<3; ixyz++) { for(ipm=0; ipm<2; ipm++) { iFace[ixyz][ipm] = 0; }}

      for(i=0; i<2; i++) { xx = subminpos[0]+(float)i*subvolsiz[0]-eyepos[0]; axx = xx; if(xx < 0.0) { axx = -xx; }
      for(j=0; j<2; j++) { yy = subminpos[1]+(float)j*subvolsiz[1]-eyepos[1]; ayy = yy; if(yy < 0.0) { ayy = -yy; }
      for(k=0; k<2; k++) { zz = subminpos[2]+(float)k*subvolsiz[2]-eyepos[2]; azz = zz; if(zz < 0.0) { azz = -zz; }
          if(axx/delx > ayy/dely) {
              if(axx/delx > azz/delz) { ixyz = 0; if(xx > 0.0) { ipm = 1; } else { ipm = 0; } }
              else                    { ixyz = 2; if(zz > 0.0) { ipm = 1; } else { ipm = 0; } }
          } else {
              if(ayy/dely > azz/delz) { ixyz = 1; if(yy > 0.0) { ipm = 1; } else { ipm = 0; } }
              else                    { ixyz = 2; if(zz > 0.0) { ipm = 1; } else { ipm = 0; } }
          }
          iFace[ixyz][ipm]++;
      }}}

      // Any face that has one or more verteces certainly needs to be drawn.
      // If any oposite pair of faces have vertices, then the subcube spans at least some of the
      //    faces in between: in this case just draw them all -- there will not be many subcubes like this
      iDrawEmAll = 0;
      for(ixyz=0; ixyz<3; ixyz++) { if(iFace[ixyz][0] > 0  &&  iFace[ixyz][1] > 0) { iDrawEmAll = 1; } }
      for(ixyz=0; ixyz<3; ixyz++) { for(ipm=0; ipm<2; ipm++) { iFace[ixyz][ipm] += iDrawEmAll; }}

      /******************************
       * Calculate the common alpha *
       ******************************/
      dx  = subvolsiz[0]/(float)((nxsub-2*m_iNBoundaries) * nplanes);
      dy  = subvolsiz[1]/(float)((nysub-2*m_iNBoundaries) * nplanes);
      dz  = subvolsiz[2]/(float)((nzsub-2*m_iNBoundaries) * nplanes);
      alpha_d = alpha0 * (float)sqrt(dx*dx + dy*dy + dz*dz) / ((float)2.0 * cubesz[0]);

      for(ipm=0; ipm<2; ipm++) {
     /**************************************************************
      * Render along X-axis where needed                           *
      **************************************************************/
      if(iFace[0][ipm] > 0) {
        if(bSetBobX == FALSE) {
            if(m_iTimersOn == 1) { tsEvent.StampTime("Begin TposeYZ"); }
            TposeYZ( bobsubZ, bobsubX, &nxsb2, &nysb2, &nzsb2, &nxsub, &nysub, &nzsub );
            if(iOpac == 1) {
                TposeYZ( bobOZ, bobOX, &nxsb2, &nysb2, &nzsb2, &nxsub, &nysub, &nzsub );
            }
            bSetBobX = TRUE;
            if(m_iTimersOn == 1) { tsEvent.StampTime("End   TposeYZ"); }
        }
        if(m_iTimersOn == 1) {
            if(m_bColorTableIsAvailable) { tsEvent.StampTime("Begin Bob_RenderingYZ"); }
            else                         { tsEvent.StampTime("Begin Bol_RenderingYZ"); }
        }

        /*if(iTexBndry == 1) {
            yy0 = (float)0.0;
            yy1 = (float)1.0;
            zz0 = (float)0.0;
            zz1 = (float)1.0;
        } else {
            yy0 =                 (float)1.0  / (float)(nysb2-1);
            yy1 = ((float)nysub - (float)2.0) / (float)(nysb2-1);
            zz0 =                 (float)1.0  / (float)(nzsb2-1);
            zz1 = ((float)nzsub - (float)2.0) / (float)(nzsb2-1);
        }*/

        if(ipm == 0) { ids0 = 0;                                             idsd =  1; }
        else          { ids0 = ((nxsub-2*m_iNBoundaries) * (int)nplanes) - 1; idsd = -1; }

        islicesize = nysb2*nzsb2;
        dx  = subvolsiz[0]/(float)((nxsub-2*m_iNBoundaries) * nplanes);
        dx1 = subvolsiz[0]/(float) (nxsub-2*m_iNBoundaries);
        // axx,ayy,azz is already normalized
        // alpha = (float)alpha0 * ((float)sqrt(axx*axx+ayy*ayy+azz*azz)/(axx)) * (dx/((float)2.*cubesz[0]));
        // alpha = (float)alpha0 * (dx/((float)2.*cubesz[0]*axx));
        alpha = (float)alpha0 * (dx/((float)2.*cubesz[0]));
        glColor4d((double)1.0, (double)1.0, (double)1.0, (double)alpha_d);

        for(islice = 0;  islice < ((nxsub-2*m_iNBoundaries) * nplanes);  islice++) {
        xxv = subminpos[0] + dx * ((float)0.5 + (float)(ids0 + idsd*islice));

        iClipThis = iClipThis0;
        if(m_ShvrUse.ifauxclip[0] == 2  &&  xxv < m_ShvrUse.auxclip[0][3]) { iClipThis = 1; }
        if(m_ShvrUse.ifauxclip[1] == 2  &&  xxv > m_ShvrUse.auxclip[1][3]) { iClipThis = 1; }
        if(iClipThis == 0) {

            islice0  = (int)((float)0.5 + (xxv - subminpos[0]) / dx1) + m_iNBoundaries - 1;
            islice1 = islice0 + 1;
            if(islice0 <      0) { islice0 = 0; }
            if(islice1 >= nxsub) { islice1 = nxsub - 1; }

            if(m_bColorTableIsAvailable  &&  iOpac == 0) {
                if(islice0 == islice1  ||  nplanes == 1) {
                    isliceoff = islicesize * islice0;
                } else {
                    isliceoff0 = islicesize * islice0;
                    isliceoff1 = islicesize * islice1;
                    ppv0 = subminpos[0] + dx1 * ((float)0.5 + (float)(islice0-m_iNBoundaries));
                    ppv1 = subminpos[0] + dx1 * ((float)0.5 + (float)(islice1-m_iNBoundaries));
                    wt0 = (ppv1 - xxv) / (ppv1 - ppv0);
                    if(wt0 > 0.5) { isliceoff = isliceoff0; } else { isliceoff = isliceoff1; }
                }
         
                /* Cg - Load texture data into texture memory */
                glTexImage2D(GL_TEXTURE_2D, 0, 1, nysb2, nzsb2, iTexBndry,
                             GL_LUMINANCE, GL_UNSIGNED_BYTE, &bobsubX[isliceoff]);

            } else {
                if(islice0 == islice1  ||  nplanes == 1) {
                    isliceoff = islicesize * islice0;
                    if(iOpac == 0) {
                        for(i=0; i<nysb2*nzsb2; i++) { sol[i] = m_lrgba[bobsubX[isliceoff+i]]; }
                    } else {
                        wt0 = (float)255.99;
                        for(i=0; i<nysb2*nzsb2; i++) {
                            iRGB = bobsubX[isliceoff+i];
                            iALP = bobOX[isliceoff+i];
                            sol[i] =  (unsigned long)(wt0*frgba[iRGB][0])
                                  + (((unsigned long)(wt0*frgba[iRGB][1]))<<8 )
                                  + (((unsigned long)(wt0*frgba[iRGB][2]))<<16)
                                  + (((unsigned long)(wt0*frgba[iALP][3]))<<24);
                        }
                    }
                } else {
                    isliceoff0 = islicesize * islice0;
                    isliceoff1 = islicesize * islice1;
                    ppv0 = subminpos[0] + dx1 * ((float)0.5 + (float)(islice0-m_iNBoundaries));
                    ppv1 = subminpos[0] + dx1 * ((float)0.5 + (float)(islice1-m_iNBoundaries));
                    wt0 = (float)255.99 * (ppv1 - xxv) / (ppv1 - ppv0);
                    for(i=0; i<256; i++) {
                        lrgba0[i] =   (unsigned long)(wt0*frgba[i][0])
                                  + (((unsigned long)(wt0*frgba[i][1]))<<8 )
                                  + (((unsigned long)(wt0*frgba[i][2]))<<16)
                                  + (((unsigned long)(wt0*frgba[i][3]))<<24);
                        lrgba1[i] = m_lrgba[i] - lrgba0[i];
                    }
                    for(i=0; i<nysb2*nzsb2; i++) {
                        sol[i] = lrgba0[bobsubX[isliceoff0+i]] + lrgba1[bobsubX[isliceoff1+i]];
                    }
                }
                glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,nysb2,nzsb2,iTexBndry,
                    GL_RGBA,GL_UNSIGNED_BYTE,sol);
            }

            // Find the limits of the rendering planes along 45 degree planes relative to the mesh
            // Only render those peices of each rendering plane which are at the best orientation
            deltadir = (float)idsd * (eyepos[0] - xxv);
            if(deltadir > 0.0) {
                y45p = eyepos[1] + deltadir * dely / delx;
                y45m = eyepos[1] - deltadir * dely / delx;
                z45p = eyepos[2] + deltadir * delz / delx;
                z45m = eyepos[2] - deltadir * delz / delx;

                if(py0 > y45m) { py0u = py0; } else { py0u = y45m; }
                if(py1 < y45p) { py1u = py1; } else { py1u = y45p; }
                if(pz0 > z45m) { pz0u = pz0; } else { pz0u = z45m; }
                if(pz1 < z45p) { pz1u = pz1; } else { pz1u = z45p; }

                if(py0u < py1u  &&  pz0u <pz1u) {
                    yy0u = submintex[1] + (submaxtex[1]-submintex[1])*(py0u-subminpos[1])/subvolsiz[1];
                    yy1u = submintex[1] + (submaxtex[1]-submintex[1])*(py1u-subminpos[1])/subvolsiz[1];
                    zz0u = submintex[2] + (submaxtex[2]-submintex[2])*(pz0u-subminpos[2])/subvolsiz[2];
                    zz1u = submintex[2] + (submaxtex[2]-submintex[2])*(pz1u-subminpos[2])/subvolsiz[2];

                    glBegin(GL_QUADS);
                        glTexCoord2f(yy0u, zz0u);    glVertex3f(xxv, py0u, pz0u);
                        glTexCoord2f(yy0u, zz1u);    glVertex3f(xxv, py0u, pz1u);
                        glTexCoord2f(yy1u, zz1u);    glVertex3f(xxv, py1u, pz1u);
                        glTexCoord2f(yy1u, zz0u);    glVertex3f(xxv, py1u, pz0u);
                    glEnd();
                    nslices++;
                }
            }
        }}     /* end of islice:for-loop and (iClipThis == 1):if-block */

        if(m_iTimersOn == 1) {
            if(m_bColorTableIsAvailable) { tsEvent.StampTime("End   Bob_RenderingYZ"); }
            else                         { tsEvent.StampTime("End   Bol_RenderingYZ"); }
        }
      }

     /**************************************************************
      * Render along Y-axis if needed                              *
      **************************************************************/
      if(iFace[1][ipm] > 0) {
        if(bSetBobY == FALSE) {
            if(m_iTimersOn == 1) { tsEvent.StampTime("Begin TposeXZ"); }
            TposeXZ( bobsubZ, bobsubY, nxsb2, nysb2, nzsb2, nxsub, nysub, nzsub );
            if(iOpac == 1) {
                TposeXZ( bobOZ, bobOY, nxsb2, nysb2, nzsb2, nxsub, nysub, nzsub  );
            }
            bSetBobY = TRUE;
            if(m_iTimersOn == 1) { tsEvent.StampTime("End   TposeXZ"); }
        }

        if(m_iTimersOn == 1) {
            if(m_bColorTableIsAvailable) { tsEvent.StampTime("Begin Bob_RenderingXZ"); }
            else                         { tsEvent.StampTime("Begin Bol_RenderingXZ"); }
        }
        if(ipm == 0) { ids0 = 0;                                             idsd =  1; }
        else          { ids0 = ((nysub-2*m_iNBoundaries) * (int)nplanes) - 1; idsd = -1; }

        islicesize = nzsb2*nxsb2;
        dy  = subvolsiz[1]/(float)((nysub-2*m_iNBoundaries) * nplanes);
        dy1 = subvolsiz[1]/(float) (nysub-2*m_iNBoundaries);
        // axx,ayy,azz is already normalized
        // alpha = (float)alpha0 * ((float)sqrt(axx*axx+ayy*ayy+azz*azz)/(ayy)) * (dy/((float)2.*cubesz[0]));
        // alpha = (float)alpha0 * (dy/((float)2.*cubesz[0]*ayy));
        alpha = (float)alpha0 * (dy/((float)2.*cubesz[0]));
        glColor4d((double)1.0, (double)1.0, (double)1.0, (double)alpha_d);

        for(islice = 0;  islice < ((nysub-2*m_iNBoundaries) * nplanes);  islice++) {
        yyv = subminpos[1] + dy * ((float)0.5 + (float)(ids0 + idsd*islice));

        iClipThis = iClipThis0;
        if(m_ShvrUse.ifauxclip[2] == 2  &&  yyv < m_ShvrUse.auxclip[2][3]) { iClipThis = 1; }
        if(m_ShvrUse.ifauxclip[3] == 2  &&  yyv > m_ShvrUse.auxclip[3][3]) { iClipThis = 1; }
        if(iClipThis == 0) {

            islice0  = (int)((float)0.5 + (yyv - subminpos[1]) / dy1) + m_iNBoundaries - 1;
            islice1 = islice0 + 1;
            if(islice0 <      0) { islice0 = 0; }
            if(islice1 >= nysub) { islice1 = nysub - 1; }

            if(m_bColorTableIsAvailable  &&  iOpac == 0) {
                if(islice0 == islice1  ||  nplanes == 1) {
                    isliceoff = islicesize * islice0;
                } else {
                    isliceoff0 = islicesize * islice0;
                    isliceoff1 = islicesize * islice1;
                    ppv0 = subminpos[1] + dy1 * ((float)0.5 + (float)(islice0-m_iNBoundaries));
                    ppv1 = subminpos[1] + dy1 * ((float)0.5 + (float)(islice1-m_iNBoundaries));
                    wt0 = (ppv1 - yyv) / (ppv1 - ppv0);
                    if(wt0 > 0.5) { isliceoff = isliceoff0; } else { isliceoff = isliceoff1; }
                }

                /* Cg - Load texture data into texture memory */
                glTexImage2D(GL_TEXTURE_2D, 0, 1, nxsb2, nzsb2, iTexBndry,
                             GL_LUMINANCE, GL_UNSIGNED_BYTE, &bobsubY[isliceoff]);

            } else {
                if(islice0 == islice1  ||  nplanes == 1) {
                    isliceoff = islicesize * islice0;
                    if(iOpac == 0) {
                        for(i=0; i<nxsb2*nzsb2; i++) { sol[i] = m_lrgba[bobsubY[isliceoff+i]]; }
                    } else {
                        wt0 = (float)255.99;
                        for(i=0; i<nxsb2*nxsb2; i++) {
                            iRGB = bobsubY[isliceoff+i];
                            iALP = bobOY[isliceoff+i];
                            sol[i] =  (unsigned long)(wt0*frgba[iRGB][0])
                                  + (((unsigned long)(wt0*frgba[iRGB][1]))<<8 )
                                  + (((unsigned long)(wt0*frgba[iRGB][2]))<<16)
                                  + (((unsigned long)(wt0*frgba[iALP][3]))<<24);
                        }
                    }
                } else {
                    isliceoff0 = islicesize * islice0;
                    isliceoff1 = islicesize * islice1;
                    ppv0 = subminpos[1] + dy1 * ((float)0.5 + (float)(islice0-m_iNBoundaries));
                    ppv1 = subminpos[1] + dy1 * ((float)0.5 + (float)(islice1-m_iNBoundaries));
                    wt0 = (float)255.99 * (ppv1 - yyv) / (ppv1 - ppv0);
                    for(i=0; i<256; i++) {
                        lrgba0[i] =   (unsigned long)(wt0*frgba[i][0])
                                  + (((unsigned long)(wt0*frgba[i][1]))<<8 )
                                  + (((unsigned long)(wt0*frgba[i][2]))<<16)
                                  + (((unsigned long)(wt0*frgba[i][3]))<<24);
                        lrgba1[i] = m_lrgba[i] - lrgba0[i];
                    }

                    for(i=0; i<nxsb2*nzsb2; i++) {
                        sol[i] = lrgba0[bobsubY[isliceoff0+i]] + lrgba1[bobsubY[isliceoff1+i]];
                    }
                } 
                glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,nxsb2,nzsb2,iTexBndry,
                      GL_RGBA,GL_UNSIGNED_BYTE,sol);
            };

            // Find the limits of the rendering planes along 45 degree planes relative to the mesh
            // Only render those peices of each rendering plane which are at the best orientation
            deltadir = (float)idsd * (eyepos[1] - yyv);
            if(deltadir > 0.0) {
                x45p = eyepos[0] + deltadir * delx / dely;
                x45m = eyepos[0] - deltadir * delx / dely;
                z45p = eyepos[2] + deltadir * delz / dely;
                z45m = eyepos[2] - deltadir * delz / dely;

                if(px0 > x45m) { px0u = px0; } else { px0u = x45m; }
                if(px1 < x45p) { px1u = px1; } else { px1u = x45p; }
                if(pz0 > z45m) { pz0u = pz0; } else { pz0u = z45m; }
                if(pz1 < z45p) { pz1u = pz1; } else { pz1u = z45p; }

                if(px0u < px1u  &&  pz0u <pz1u) {
                    xx0u = submintex[0] + (submaxtex[0]-submintex[0])*(px0u-subminpos[0])/subvolsiz[0];
                    xx1u = submintex[0] + (submaxtex[0]-submintex[0])*(px1u-subminpos[0])/subvolsiz[0];
                    zz0u = submintex[2] + (submaxtex[2]-submintex[2])*(pz0u-subminpos[2])/subvolsiz[2];
                    zz1u = submintex[2] + (submaxtex[2]-submintex[2])*(pz1u-subminpos[2])/subvolsiz[2];

                    glBegin(GL_QUADS);
                        glTexCoord2f(xx0u, zz0u);    glVertex3f(px0u, yyv, pz0u);
                        glTexCoord2f(xx0u, zz1u);    glVertex3f(px0u, yyv, pz1u);
                        glTexCoord2f(xx1u, zz1u);    glVertex3f(px1u, yyv, pz1u);
                        glTexCoord2f(xx1u, zz0u);    glVertex3f(px1u, yyv, pz0u);
                    glEnd();
                    nslices++;
                }
            }
        }}     /* end of islice:for-loop and (iClipThis == 1):if-block */

        if(m_iTimersOn == 1) {
            if(m_bColorTableIsAvailable) { tsEvent.StampTime("End   Bob_RenderingXZ"); }
            else                         { tsEvent.StampTime("End   Bol_RenderingXZ"); }
        }
      }

     /**************************************************************
      * Render along Z-axis where needed                           *
      **************************************************************/
      if(iFace[2][ipm] > 0) {
        if(m_iTimersOn == 1) {
            if(m_bColorTableIsAvailable) { tsEvent.StampTime("Begin Bob_RenderingXY"); }
            else                         { tsEvent.StampTime("Begin Bol_RenderingXY"); }
        }

        if(ipm == 0) { ids0 = 0;                                             idsd =  1; }
        else          { ids0 = ((nzsub-2*m_iNBoundaries) * (int)nplanes) - 1; idsd = -1; }

        islicesize = nxsb2*nysb2;
        //islicesize = nxsub*nysub;
        dz  = subvolsiz[2]/(float)((nzsub-2*m_iNBoundaries) * nplanes);
        dz1 = subvolsiz[2]/(float) (nzsub-2*m_iNBoundaries);
        // alpha = (float)alpha0 * ((float)sqrt(axx*axx+ayy*ayy+azz*azz)/(azz)) * (dz/((float)2.*cubesz[0]));
        // alpha = (float)alpha0 * (dz/((float)2.*cubesz[0]*azz));
        alpha = (float)alpha0 * (dz/((float)2.*cubesz[0]));
        glColor4d((double)1.0, (double)1.0, (double)1.0, (double)alpha_d);

        for(islice = 0;  islice < ((nzsub-2*m_iNBoundaries) * nplanes);  islice++) {
        zzv = subminpos[2] + dz * ((float)0.5 + (float)(ids0 + idsd*islice));

        iClipThis = iClipThis0;
        if(m_ShvrUse.ifauxclip[4] == 2  &&  zzv < m_ShvrUse.auxclip[4][3]) { iClipThis = 1; }
        if(m_ShvrUse.ifauxclip[5] == 2  &&  zzv > m_ShvrUse.auxclip[5][3]) { iClipThis = 1; }
        if(iClipThis == 0) {

            islice0  = (int)((float)0.5 + (zzv - subminpos[2]) / dz1) + m_iNBoundaries - 1;
            islice1 = islice0 + 1;
            if(islice0 <      0) { islice0 = 0; }
            if(islice1 >= nzsub) { islice1 = nzsub - 1; }

            if(m_bColorTableIsAvailable  &&  iOpac == 0) {
                if(islice0 == islice1  ||  nplanes == 1) {
                    isliceoff = islicesize * islice0;
                } else {
                    isliceoff0 = islicesize * islice0;
                    isliceoff1 = islicesize * islice1;
                    ppv0 = subminpos[2] + dz1 * ((float)0.5 + (float)(islice0-m_iNBoundaries));
                    ppv1 = subminpos[2] + dz1 * ((float)0.5 + (float)(islice1-m_iNBoundaries));
                    wt0 = (ppv1 - zzv) / (ppv1 - ppv0);
                    if(wt0 > 0.5) { isliceoff = isliceoff0; } else { isliceoff = isliceoff1; }
                }
                
                /* Cg - Load texture data into texture memory */
                glTexImage2D(GL_TEXTURE_2D, 0, 1, nxsb2, nysb2, iTexBndry,
                             GL_LUMINANCE, GL_UNSIGNED_BYTE, &bobsubZ[isliceoff]);
                
            } else {
                if(islice0 == islice1  ||  nplanes == 1) {
                    isliceoff = islicesize * islice0;
                    if(iOpac == 0) {
                        for(i=0; i<nxsb2*nysb2; i++) { sol[i] = m_lrgba[bobsubZ[isliceoff+i]]; }
                    } else {
                        wt0 = (float)255.99;
                        for(i=0; i<nxsb2*nysb2; i++) {
                            iRGB = bobsubZ[isliceoff+i];
                            iALP = bobOZ[isliceoff+i];
                            sol[i] =  (unsigned long)(wt0*frgba[iRGB][0])
                                  + (((unsigned long)(wt0*frgba[iRGB][1]))<<8 )
                                  + (((unsigned long)(wt0*frgba[iRGB][2]))<<16)
                                  + (((unsigned long)(wt0*frgba[iALP][3]))<<24);
                        }
                    }
                } else {
                    isliceoff0 = islicesize * islice0;
                    isliceoff1 = islicesize * islice1;
                    ppv0 = subminpos[2] + dz1 * ((float)0.5 + (float)(islice0-m_iNBoundaries));
                    ppv1 = subminpos[2] + dz1 * ((float)0.5 + (float)(islice1-m_iNBoundaries));
                    wt0 = (float)255.99 * (ppv1 - zzv) / (ppv1 - ppv0);

                    for(i=0; i<256; i++) {
                        lrgba0[i] =   (unsigned long)(wt0*frgba[i][0])
                                  + (((unsigned long)(wt0*frgba[i][1]))<<8 )
                                  + (((unsigned long)(wt0*frgba[i][2]))<<16)
                                  + (((unsigned long)(wt0*frgba[i][3]))<<24);
                        lrgba1[i] = m_lrgba[i] - lrgba0[i];
                    }
                    for(i=0; i<nxsb2*nysb2; i++) {
                        sol[i] = lrgba0[bobsubZ[isliceoff0+i]] + lrgba1[bobsubZ[isliceoff1+i]];
                    }
                }
                
                glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,nxsb2,nysb2,iTexBndry,
                             GL_RGBA,GL_UNSIGNED_BYTE,sol);
            }

            // Find the limits of the rendering planes along 45 degree planes relative to the mesh
            // Only render those peices of each rendering plane which are at the best orientation
            deltadir = (float)idsd * (eyepos[2] - zzv);
            if(deltadir > 0.0) {
                x45p = eyepos[0] + deltadir * delx / delz;
                x45m = eyepos[0] - deltadir * delx / delz;
                y45p = eyepos[1] + deltadir * dely / delz;
                y45m = eyepos[1] - deltadir * dely / delz;

                if(px0 > x45m) { px0u = px0; } else { px0u = x45m; }
                if(px1 < x45p) { px1u = px1; } else { px1u = x45p; }
                if(py0 > y45m) { py0u = py0; } else { py0u = y45m; }
                if(py1 < y45p) { py1u = py1; } else { py1u = y45p; }
                
                if(py0u < py1u  &&  px0u <px1u) {
                    xx0u = submintex[0] + (submaxtex[0]-submintex[0])*(px0u-subminpos[0])/subvolsiz[0];
                    xx1u = submintex[0] + (submaxtex[0]-submintex[0])*(px1u-subminpos[0])/subvolsiz[0];
                    yy0u = submintex[1] + (submaxtex[1]-submintex[1])*(py0u-subminpos[1])/subvolsiz[1];
                    yy1u = submintex[1] + (submaxtex[1]-submintex[1])*(py1u-subminpos[1])/subvolsiz[1];

                    glBegin(GL_QUADS);
                        glTexCoord2f(xx0u, yy0u);    glVertex3f(px0u, py0u, zzv);
                        glTexCoord2f(xx0u, yy1u);    glVertex3f(px0u, py1u, zzv);
                        glTexCoord2f(xx1u, yy1u);    glVertex3f(px1u, py1u, zzv);
                        glTexCoord2f(xx1u, yy0u);    glVertex3f(px1u, py0u, zzv);
                    glEnd();
                    nslices++;
                }
            }
        }}     /* end of islice:for-loop and (iClipThis == 1):if-block */

        /* Disable execution of Cg Fragment Program */
        if(m_bColorTableIsAvailable) {
            cgGLDisableProfile(fProfile);
            cgGLDisableTextureParameter(textureHandle);
            cgGLDisableTextureParameter(textureHandle2);
        }
        
        if(m_iTimersOn == 1) {
            if(m_bColorTableIsAvailable) { tsEvent.StampTime("End   Bob_RenderingXY"); }
            else                         { tsEvent.StampTime("End   Bol_RenderingXY"); }
        }
      }
      }  /* end of ipm do loop */
      }    /* End of 2D texture block */
      } /* end of isoutside = 0 case: tile is inside frustum */

    } /* end of ip loop (loop over pannels)    */

    /* Free cache buffer */
    iUsed_of_iCache[iCache_of_iTile[isub]] = 0;
    //if(m_iTimersOn == 1) { tsEvent.StampTime("End   jsubRenderloop"); }

//fprintf(stderr, "XXX: at end of render loop\n");

    } /* end of jsub "render" loop */
//fprintf(stderr, "XXX: after jsub render loop\n");

    if(m_iTimersOn == 1) { tsEvent.StampTime("Begin Output"); }

#ifdef DEBUG_NEW_TREE
fprintf(stderr, "after rendering jsub loop\n");
// return
#endif

#ifdef PBO
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER_EXT, 0);
#endif

#ifdef DEBUG_TEX
    #ifdef WRITE_STD_OUT
        fprintf(error_log,"done\n");fflush(error_log);
    #else
        printf("done\n");
    #endif
#endif

#ifdef INTELLIGENT_OVERWRITE
    setPrevCount(nBricksFitInCache);
    all_in_use = FALSE;
#endif

    if(m_iShowColorBar > 0) { colorbar(); }


#ifdef DO_ANNOTATION
    /* Annotation routine */      //   this was at 5261 - 5117
    annotation(0, eyepos);

    /* Pass color LUT onto display if desired */
    if(m_iShowColorBar > 0) {
        fZero = (float)0.0;
        fOne  = (float)1.0;
        fMOne = (float)(-1.0);

        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        dleft     = (GLdouble)(-1.0) * (GLdouble)m_fScreenRatio;
        dright    = (GLdouble)( 1.0) * (GLdouble)m_fScreenRatio;
        dbottom   = (GLdouble)(-1.0) * (GLdouble)m_fScreenRatio;
        dtop      = (GLdouble)( 1.0) * (GLdouble)m_fScreenRatio;
        dnearclip = (GLdouble)( 0.5);
        dfarclip  = (GLdouble)( 1.5);
        glFrustum( dleft, dright, dbottom, dtop, dnearclip, dfarclip );
        gluLookAt(fZero, fZero, fZero,  fZero, fZero, fMOne,  fZero, fOne, fZero);

        glColor4d((double)1.0, (double)1.0, (double)1.0, (double)1.0);

        xx0 =  (float)1.8;
        if(m_iNTicks > 0) {
            yy0  = (float)(-1.6);
            yy1  = (float)(-1.9);
            eps  =  (float)0.01;
            eps1 =  (float)0.05;
            jlutmax = 16;
        } else {
            yy0  = (float)(-1.8);
            yy1  = (float)(-1.9);
            eps  =  (float)0.01;
            eps1 =  (float)0.01;
            jlutmax = 34;
        }
        i1 = (int)((float)128 * eps1/(xx0+eps1));
        i2 = 258 - i1;
        //        if(m_iNTicks > 0) { eps1 = 0.1; }
        // background
        /*for(i=0; i<258; i++) {
        for(j=0; j<34; j++) {
              sol[i+258*j] = (((unsigned long)((float)100.00))<<24);

        }}*/

        for(i=0; i<256; i++) {
        for(j=0; j<32; j++) {
              sol[i+256*j] = (((unsigned long)((float)100.00))<<24);

        }}

        zzv   = (float)(-1.0);
        xxv0  = -xx0-eps1;
        xxv1  =  xx0+eps1;
        yyv0  =  yy0+eps;
        yyv1  =  yy1-eps;

        //glTexImage2D(GL_TEXTURE_3D,0,GL_RGBA,258,34,1,GL_RGBA,GL_UNSIGNED_BYTE,sol);
        glTexImage2D(GL_TEXTURE_3D,0,GL_RGBA,256,32,0,GL_RGBA,GL_UNSIGNED_BYTE,sol);

        glBegin(GL_QUADS);
                glTexCoord2f(fZero, fZero);    glVertex3f(xxv0, yyv0, zzv);
                glTexCoord2f(fOne , fZero);    glVertex3f(xxv1, yyv0, zzv);
                glTexCoord2f(fOne , fOne );    glVertex3f(xxv1, yyv1, zzv);
                glTexCoord2f(fZero, fOne );    glVertex3f(xxv0, yyv1, zzv);
        glEnd();

        /* Color LUT */
        for(i=0; i<256; i++) { k = i;
if(i%16 == 0) {
fprintf(stderr, "irgba= %d %f %f %f %f\n", k,frgba[k][0],frgba[k][1],frgba[k][2],frgba[k][3]);
}
        for(j=0; j<32; j++) {
            if(j < jlutmax) {
                //sol[i+258*j] = (unsigned long)((float)255.99 * frgba[k][0])
                sol[i+256*j] =  (unsigned long)((float)255.99 * frgba[k][0])
                            + (((unsigned long)((float)255.99 * frgba[k][1]))<<8 )
                            + (((unsigned long)((float)255.99 * frgba[k][2]))<<16)
                            + (((unsigned long)((float)255.99              ))<<24);
            } else {
                //sol[i+258*j] = (unsigned long)0;
                sol[i+256*j] = (unsigned long)0;
            }
        }}
        if(m_iNTicks > 0) {
            for(j=jlutmax; j<24; j++) {
            for(itick=0; itick<m_iNTicks; itick++) {
                //sol[1+m_iTicks[itick]+258*j] = (unsigned long)((float)255.99)
                sol[m_iTicks[itick]+256*j] =        (unsigned long)((float)255.99)
                                                + (((unsigned long)((float)255.99))<<8 )
                                                + (((unsigned long)((float)255.99))<<16)
                                                + (((unsigned long)((float)255.99))<<24);
            }}
        }
        xxv0  = -xx0;
        xxv1  =  xx0;
        yyv0  =  yy0;
        yyv1  =  yy1;

        glTexImage2D(GL_TEXTURE_3D,0,GL_RGBA,258,34,1,GL_RGBA,GL_UNSIGNED_BYTE,sol);
        glBegin(GL_QUADS);
                glTexCoord2f(fZero, fZero);    glVertex3f(xxv0, yyv0, zzv);
                glTexCoord2f(fOne , fZero);    glVertex3f(xxv1, yyv0, zzv);
                glTexCoord2f(fOne , fOne );    glVertex3f(xxv1, yyv1, zzv);
                glTexCoord2f(fZero, fOne );    glVertex3f(xxv0, yyv1, zzv);
        glEnd();

        glFlush();
        glFinish();

        if(m_iNTicks > 0) {
            if(iFirst_Test == 1) {
#ifdef WINDOWS
                //SelectObject (hDC, GetStockObject (SYSTEM_FONT)); 
                SelectObject (hDC, CreateFont(0, 0, 0, 0, FW_DONTCARE,
                                     FALSE, FALSE, FALSE,
                                     ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
                                     DEFAULT_QUALITY, FF_ROMAN, NULL) );
                wglUseFontBitmaps (hDC, 0, 255, 1000); 
                glListBase (1000);
                iFirst_Test = 0;
#else
                myCreateFont();
#endif
            }

            glMatrixMode(GL_TEXTURE);
            glDisable(GL_TEXTURE_3D);
            glDisable(GL_BLEND);

            for(itick=0; itick<m_iNTicks; itick++) {
                xx1 = -xx0 - eps1 + (float)2. * xx0 * (float)m_iTicks[itick]/(float)256.;
                glRasterPos3f(xx1, yy1, -1.);
                glCallLists (strlen(m_sTicks[itick]), GL_UNSIGNED_BYTE, m_sTicks[itick]);
            }

            glFlush();
            glFinish();
            glEnable(GL_BLEND);
            glEnable(GL_TEXTURE_3D);
        }

        glFlush();
        glFinish();
    }
#endif

#ifdef WRITE_STD_OUT
    /* Prepare the standard output for writing binary data. */
    /* This must be done before writing anything to stdout. */
    SET_BINARY_MODE(STDOUT);
    if(iWriteImageToSTDout == 1) {
        glReadPixels(0,0,m_ShvrUse.iWidth,m_ShvrUse.iHeight,GL_RGB,GL_UNSIGNED_BYTE,ucGPix);
        glFlush();
        glFinish();
    }
    #ifdef DEBUG_TEX
        fprintf(stderr,"printing image...");fflush(stderr);
    #endif
    write(STDOUT,&status,4);
    if(iWriteImageToSTDout == 1  &&  status == SUCCESS) {
        #ifdef TEST_ALLBLACK
        isallblack = 1;
        for(i=0; i<3*m_ShvrUse.iWidth*m_ShvrUse.iHeight; i++) {
          if(ucGPix[i] != 0) { isallblack = 0; }
        }
        if(isallblack == 1) {
          if(hostname[0] == ' ') { gethostname(hostname, 256); }
          sprintf(allblackfile, "/mnt/global/dhp/%s_hvr_allblack", hostname);
          fpdhp = fopen(allblackfile,"w");
          fprintf(fpdhp, "this frame is all black\n");
          fclose(fpdhp);
        }
        #endif
/* dhp1: write image to a file */
        if(strdif(m_cImageName, "none") == 0) {
          write(STDOUT,ucGPix,3*m_ShvrUse.iWidth*m_ShvrUse.iHeight);
        } else {
          fd = open(m_cImageName, O_RDWR|O_CREAT,  0644 );
          write(fd, ucGPix,3*m_ShvrUse.iWidth*m_ShvrUse.iHeight);
          close(fd);
          fprintf(stderr, "Wrote image to file: %s\n", m_cImageName);
        }
    }
    #ifdef DEBUG_TEX
        fprintf(stderr,"done\n");fflush(stderr);
    #endif
#endif
        
    /* Analyze Texture Caching Scheme Efficiency */
    nBricksFromCache = nBricksRendered - nBricksLoaded;
    pctRenderedFromCache = 100*(int)((float)nBricksFromCache/(float)nBricksRendered);
    pctCacheUsed = 100*(int)((float)nBricksFromCache/(float)nBricksInCache);

    if (nBricksRendered <= nBricksInCache)
        pctTCSEfficient = pctRenderedFromCache;
    else
        pctTCSEfficient = pctCacheUsed;

#ifdef DEBUG_TCS
    FILE *logfile;
    texmem_used = nBricksLoaded*texsize;

    /* error checking */
    if(nBricksRendered != nBricksInFrustum) {
        fprintf(stderr, "WARNING: # of bricks rendered (%d) does not equal # of bricks in frustum (%d)\n",
            nBricksRendered,nBricksInFrustum);
    }

    /* stat reporting */
    if(hv_count==0) {
        logfile = fopen("log.txt","w");
        fprintf(logfile,"%d Bricks in Volume (%d MB)\n",nBricksInVolume,nBricksInVolume*texsize/MEGABYTE);
        fprintf(logfile,"%d Bricks in Initial Frustum (%d MB)\n",nBricksInFrustum,nBricksInFrustum*texsize/MEGABYTE);
        fprintf(logfile,"Minimum TCS Efficiency Level: %d%%\n",TCS_Eff_Min);
        fprintf(logfile,"----------------------------------\n\n");    
        fflush(logfile);
        fclose(logfile);
    }
    else {
        logfile = fopen("log.txt","a");
        fprintf(logfile,"%d MB, ",nBricksRendered,nBricksRendered*texsize/MEGABYTE);
        fprintf(logfile,"%d%% from cache. ",pctRenderedFromCache);
        fprintf(logfile,"%d MB in Cache, %d %% in Frustum.  [%d%%]",
            nBricksInCache*texsize/MEGABYTE, pctCacheUsed, pctTCSEfficient);
        if(TexScheme==1 && (pctCacheUsed<TCS_Eff_Min) && nBricksFromCache!=0)
            fprintf(logfile,"-- RELOADING CACHE");
        fprintf(logfile,"\n");
        fflush(logfile);
        fclose(logfile);
    }
#endif
    
    /* Reload texture cache if caching scheme
       drops below desired efficiency level */
    if (TexScheme==1 && loopcount>0 && (pctCacheUsed<TCS_Eff_Min)) {
        TexScheme = 0;
        //reload = TRUE;
#ifdef PRINT_TCS_EFF
        fprintf(stderr,"Reloading texture cache  (TCS Efficiency: %d%%)\n",pctTCSEfficient);fflush(stderr);
#endif
    }
    else {
        loopcount++;
        hv_count++;
#ifdef PRINT_TCS_EFF
        fprintf(stderr,"TCS Efficiency: %d %%\n",pctTCSEfficient);fflush(stderr);
#endif
    }

    Prev_TexMode = TexMode;
    TexMode = 0;

#ifdef TEST_FBO
    if(Renderbuffer) {
        if(test_count == 0) {
            FILE* output;
            output = fopen("renderbuffer.raw","wb");
            printf("printing image\n");
            glReadPixels(0,0,m_ShvrUse.iWidth,m_ShvrUse.iHeight,GL_RGB,GL_UNSIGNED_BYTE,ucGPix);
            fwrite(ucGPix,3,m_ShvrUse.iWidth*m_ShvrUse.iHeight,output);
            fclose(output);
            glFlush();
            glFinish();
        }
        else if(test_count == 1) {
            FILE* output;
            output = fopen("renderbuffer2.raw","wb");
            printf("printing 2nd image\n");
            glReadPixels(0,0,m_ShvrUse.iWidth,m_ShvrUse.iHeight,GL_RGB,GL_UNSIGNED_BYTE,ucGPix);
            fwrite(ucGPix,3,m_ShvrUse.iWidth*m_ShvrUse.iHeight,output);
            fclose(output);
            glFlush();
            glFinish();
        }
        else if(test_count == 2) {
            FILE* output;
            output = fopen("renderbuffer3.raw","wb");
            printf("printing 3rd image\n");
            glReadPixels(0,0,m_ShvrUse.iWidth,m_ShvrUse.iHeight,GL_RGB,GL_UNSIGNED_BYTE,ucGPix);
            fwrite(ucGPix,3,m_ShvrUse.iWidth*m_ShvrUse.iHeight,output);
            fclose(output);
            glFlush();
            glFinish();
        }
        test_count++;
    }
#endif

    /* Need these to make sure all bricks get rendered */
    glFlush();
    glFinish();

    /* Close any open files after loop over jsub */
    if(m_iBufferedIO != 1) {        
    }

    /* Annotation routine */      //   5261 - 5117   was here
// 5261

/************************************************************************
 * Plot Statistics                                                      *
 ************************************************************************/
    if(strcmp(m_sStats, "cross") == 0) { 
        /* Set Background of entire frame to white, reset color of clear color to black */
        glClearColor(1.0, 1.0, 1.0, 1.0);
        glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
        glClearColor(0.0, 0.0, 0.0, 1.0);

        /* Set the viewing geometry */
        fZero = (float)0.0;
        fOne  = (float)1.0;
        fMOne = (float)(-1.0);
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        dleft     = (GLdouble)(-1.0);
        dright    = (GLdouble)( 1.0);
        dbottom   = (GLdouble)(-1.0);
        dtop      = (GLdouble)( 1.0);
        dnearclip = (GLdouble)( 0.5);
        dfarclip  = (GLdouble)( 1.5);
        glFrustum( dleft, dright, dbottom, dtop, dnearclip, dfarclip );
        gluLookAt(fZero, fZero, fZero,  fZero, fZero, fMOne,  fZero, fOne, fZero);

        /* Plot labels 1st, just because ... */
        xxv0  = (float)(-1.68);
        xxv1  = (float)( 1.99);
        yyv0  = (float)(-1.78);
        yyv1  = (float)( 1.90);
        zzv   = (float)(-1.00);

        /* Generate Text font */
        glColor3d(0.0, 0.0, 0.0);
        if(iFirst_Test == 1) {
#ifdef WINDOWS
            SelectObject (hDC, CreateFont(0, 0, 0, 0, FW_DONTCARE,
                          FALSE, FALSE, FALSE,
                          ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
                          DEFAULT_QUALITY, FF_ROMAN, NULL) );
            wglUseFontBitmaps (hDC, 0, 255, 1000); 
            glListBase (1000);
            iFirst_Test = 0;
#else
            myCreateFont();
#endif
        }
        /* DO the labels */
        glMatrixMode(GL_TEXTURE);
        glDisable(GL_TEXTURE_3D);
        glDisable(GL_BLEND);

        for(itick=0; itick<m_iXNTicks; itick++) {
            xx1 = xxv0 - (float)0.07 + (xxv1-xxv0) * (float)m_iXTicks[itick]/(float)256.;
            glRasterPos3f(xx1, (float)(-1.9), -1.);
            glCallLists (strlen(m_sXTicks[itick]), GL_UNSIGNED_BYTE, m_sXTicks[itick]);
        }

        for(itick=0; itick<m_iYNTicks; itick++) {
            yy1 = yyv0 - (float)0.03 + (yyv1-yyv0) * (float)m_iYTicks[itick]/(float)256;
            glRasterPos3f((float)(-1.99), yy1, -1.);
            glCallLists (strlen(m_sYTicks[itick]), GL_UNSIGNED_BYTE, m_sYTicks[itick]);
        }
        glEnable(GL_BLEND);
        glEnable(GL_TEXTURE_3D);

        glFlush();
        glFinish();

        /* Generate RGBA LUT for counts
           fpp = fopen("DEBUG_CROSS.txt", "w");
           fprintf(fpp, "IRGB = %3d    %f  %f  %f\n", ind[0], red[0], grn[0], blu[0]);*/
        for(j=1; j<10; j++) {
            //fprintf(fpp, "IRGB = %3d    %f  %f  %f\n", ind[j], red[j], grn[j], blu[j]);
            i0 = ind[j-1];
            i1 = ind[j];
            for(i=i0; i<i1; i++) {
                fRed = (float)red[j-1] + ((float)(red[j]-red[j-1])/(float)(i1-i0)) * (float)(i-i0);
                fGrn = (float)grn[j-1] + ((float)(grn[j]-grn[j-1])/(float)(i1-i0)) * (float)(i-i0);
                fBlu = (float)blu[j-1] + ((float)(blu[j]-blu[j-1])/(float)(i1-i0)) * (float)(i-i0);
                lCrossLUT[i] =     (unsigned long)(         fRed)
                             + (((unsigned long)(         fGrn))<<8 )
                             + (((unsigned long)(         fBlu))<<16)
                             + (((unsigned long)((float)255.99))<<24);
            }
        }

        /* Map counts into RGB display array */
        iMaxCounts = 0;
        for(i=0; i<256*256; i++) {
            if(iCross[i] < 1) { j = 0; }
            else              { j = 1 + (int)(32.0*log((double)iCross[i])/log(10.0)); }
            if(j > 255)       { j = 255; }
            sol[i] = lCrossLUT[ j ];

            /*if(iCross[i] > 0  &&  iMaxCounts < 20  &&  i > 256) {
                fprintf(fpp, "%3d  %3d   iCross=%5d   sol=%ld\n", i%256, i/256, iCross[i], sol[i]);
                iMaxCounts++;
            }*/
        }

        //fprintf(fpp, "m_sTileO = %s\n",m_sTileO);
        //fclose(fpp);

        /* Apply Bounding Box and Tick Marks
           Assume that X and Y ticks are already set */
        lBlack =        (unsigned long)((float)  0.00)
                    + (((unsigned long)((float)  0.00))<<8 )
                    + (((unsigned long)((float)  0.00))<<16)
                    + (((unsigned long)((float)255.99))<<24);
        for(i=0; i<256; i++) {
            // Black bounding box
            if(i < 1  ||  i > 254) { for(j=0; j<512; j++) {sol[i+256*j] = lBlack; }}
            sol[i+256*  0] = lBlack;
            sol[i+256*255] = lBlack;

            // Black Horzontal Ticks
            if(i < 8) {
                for(itick=0; itick<m_iYNTicks; itick++) {
                    sol[i+256*m_iYTicks[itick]] = lBlack;
                }
            }
        }
        // Black Vertical Ticks
        for(itick=0; itick<m_iXNTicks; itick++) {
            for(j=0; j<8; j++) { sol[m_iXTicks[itick]+256*j] = lBlack; }
        }

        // Plot It

        // Plot the RGBA map
        glColor4d((double)1.0, (double)1.0, (double)1.0, (double)1.0);
        glTexImage2D(GL_TEXTURE_3D,0,GL_RGBA,256,256,0,GL_RGBA,GL_UNSIGNED_BYTE,sol);
        glBegin(GL_QUADS);
                glTexCoord2f(fZero, fZero);    glVertex3f(xxv0, yyv0, zzv);
                glTexCoord2f(fOne , fZero);    glVertex3f(xxv1, yyv0, zzv);
                glTexCoord2f(fOne , fOne );    glVertex3f(xxv1, yyv1, zzv);
                glTexCoord2f(fZero, fOne );    glVertex3f(xxv0, yyv1, zzv);
        glEnd();

        // Draw Labels 

        /*plf = (PLOGFONT) LocalAlloc(LPTR, sizeof(LOGFONT)); 
        lstrcpy(plf->lfFaceName, "Arial"); 
        plf->lfWeight = FW_NORMAL; 
        GetClientRect(hWnd, &rc); 
         SetBkMode(hDC, TRANSPARENT); 
         plf->lfEscapement = 900;
        hfnt = CreateFontIndirect(plf);
        SelectObject(hDC, hfnt);
        TextOut(hDC, rc.right / 100, rc.bottom, m_sXTicks[0], lstrlen(m_sXTicks[0]));
        DeleteObject(hfnt);
         SetBkMode(hDC, OPAQUE); 
         LocalFree((LOCALHANDLE) plf); */

        glFlush();
        glFinish();

        /* Make sure that tick marks will be current the next time */
        m_iXNTicks = 0;
        m_iYNTicks = 0;
    }

    if(strcmp(m_sStats, "pdf") == 0) {

        /* Set colors as RGBA longs */
        lBlack =        (unsigned long)((float)  0.00)
                    + (((unsigned long)((float)  0.00))<<8 )
                    + (((unsigned long)((float)  0.00))<<16)
                    + (((unsigned long)((float)255.99))<<24);


        lWhite =        (unsigned long)((float)255.99)
                    + (((unsigned long)((float)255.99))<<8 )
                    + (((unsigned long)((float)255.99))<<16)
                    + (((unsigned long)((float)255.99))<<24);

        lGreen =        (unsigned long)((float)  0.00)
                    + (((unsigned long)((float)160.00))<<8 )
                    + (((unsigned long)((float)  0.00))<<16)
                    + (((unsigned long)((float)255.99))<<24);

        // Get max counts
        iMaxCounts = 0;
        iTotalCounts = 0;
        for(i=0; i<256; i++) {
            iTotalCounts += iPDF[i];
            if(iMaxCounts < (int)iPDF[i]) { iMaxCounts = iPDF[i]; }
        }

        // SET X and Y ticks
        if(m_iNTicks > 0) {
            m_iXNTicks = m_iNTicks;
            for(itick=0; itick<m_iXNTicks; itick++) {
                m_iXTicks[itick] = m_iTicks[itick];
                sprintf(m_sXTicks[itick], "%s", m_sTicks[itick]);
            }
        } else {
            m_iXNTicks = 5;
            for(itick=0; itick<m_iXNTicks; itick++) {
                m_iXTicks[itick] = 64*itick;
                sprintf(m_sXTicks[itick], "%d", 64*itick);
            }
        }

        sscanf(m_sXTicks[           0], "%f", &xPDFmin);
        sscanf(m_sXTicks[m_iXNTicks-1], "%f", &xPDFmax);
        fDelX = (double)(xPDFmax - xPDFmin) / (double)(m_iXTicks[m_iXNTicks-1]-m_iXTicks[0]);


        if(fDelX < 0.) { fDelX = -fDelX; }
        fPDFfactor = 1.0 / (fDelX * (double)iTotalCounts);
        i10PDFmax = (int)(100.9 + log((double)iMaxCounts * fPDFfactor) / log(10.)) - 100;
        f10PDFmax = (float)i10PDFmax;

        /*fpp = fopen("DEBUG_PDF.txt", "w");
          fprintf(fpp, "xPDFmax=%f xPDFmin=%f fDelX=%f\n", xPDFmax, xPDFmin, fDelX);
          fprintf(fpp, "i10PDFmax=%d  Log_10(iMaxC*fPDFfac)=%f\n", i10PDFmax,
                  (float)(log( (double)iMaxCounts*fPDFfactor)/log(10.) )  );
          fclose(fpp);*/

        m_iYNTicks = 5;
        for(itick=0; itick<m_iYNTicks; itick++) {
            m_iYTicks[itick] = 100 + 100 * itick;
            sprintf(m_sYTicks[itick], "%d", itick - 4 + i10PDFmax);
        }

        /* Generate pdf as a raster */
        for(i=0; i<256; i++) {
            // White background
            for(j=0; j<512; j++) { sol[i+256*j] = lWhite; }

            // Green PDF
            if(iPDF[i] > 0) {
                jMax = (int)(m_iYTicks[m_iYNTicks-1] + 100.0*(log((double)iPDF[i]*fPDFfactor)/log(10.) - f10PDFmax));
            } else {
                jMax = 0;
            }
            for(j=0; j<jMax; j++) { sol[i+256*j] = lGreen; }

            // Black bounding box
            if(i < 1  ||  i > 254) { for(j=0; j<512; j++) {sol[i+256*j] = lBlack; }}
            sol[i+256*  0] = lBlack;
            sol[i+256*  1] = lBlack;
            sol[i+256*510] = lBlack;
            sol[i+256*511] = lBlack;

            // Black Horzontal Ticks
            if(i < 8) {
                for(itick=0; itick<m_iYNTicks; itick++) {
                    sol[i+256*m_iYTicks[itick]] = lBlack;
                }
            }
        }
        // Black Vertical Ticks
        for(itick=0; itick<m_iXNTicks; itick++) {
            for(j=0; j<16; j++) { sol[m_iXTicks[itick]+256*j] = lBlack; }
        }


        fZero = (float)0.0;
        fOne  = (float)1.0;
        fMOne = (float)(-1.0);

        // Set the viewing geometry
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        dleft     = (GLdouble)(-1.0);
        dright    = (GLdouble)( 1.0);
        dbottom   = (GLdouble)(-1.0);
        dtop      = (GLdouble)( 1.0);
        dnearclip = (GLdouble)( 0.5);
        dfarclip  = (GLdouble)( 1.5);
        glFrustum( dleft, dright, dbottom, dtop, dnearclip, dfarclip );
        gluLookAt(fZero, fZero, fZero,  fZero, fZero, fMOne,  fZero, fOne, fZero);

        xxv0  = (float)(-1.78);
        xxv1  = (float)( 1.90);
        yyv0  = (float)(-1.78);
        yyv1  = (float)( 1.90);
        zzv   = (float)(-1.00);

        //====================================================================================
        // Plot it
        // Set Background of entire frame to white, reset color of clear color to black 
        glClearColor(1.0, 1.0, 1.0, 1.0);
        glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
        glClearColor(0.0, 0.0, 0.0, 1.0);

        // Lay down the RGBA map
        glColor4d((double)1.0, (double)1.0, (double)1.0, (double)1.0);
        glTexImage2D(GL_TEXTURE_3D,0,GL_RGBA,256,512,0,GL_RGBA,GL_UNSIGNED_BYTE,sol);
        glBegin(GL_QUADS);
                glTexCoord2f(fZero, fZero);    glVertex3f(xxv0, yyv0, zzv);
                glTexCoord2f(fOne , fZero);    glVertex3f(xxv1, yyv0, zzv);
                glTexCoord2f(fOne , fOne );    glVertex3f(xxv1, yyv1, zzv);
                glTexCoord2f(fZero, fOne );    glVertex3f(xxv0, yyv1, zzv);
        glEnd();

        // Generate Text font
        glColor3d(0.0, 0.0, 0.0);

        if(iFirst_Test == 1) {
#ifdef WINDOWS
            SelectObject (hDC, CreateFont(0, 0, 0, 0, FW_DONTCARE,
                          FALSE, FALSE, FALSE,
                          ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
                          DEFAULT_QUALITY, FF_ROMAN, NULL) );
            wglUseFontBitmaps (hDC, 0, 255, 1000); 
            glListBase (1000);

        //=============================================================================
        //    plf = (PLOGFONT) LocalAlloc(LPTR, sizeof(LOGFONT)); 
        //    lstrcpy(plf->lfFaceName, "Arial"); 
        //    plf->lfWeight = FW_NORMAL;
        //    plf->lfEscapement = 900; 
        //    hfnt = CreateFontIndirect(plf); 
        //    SelectObject(hDC, hfnt); 

        //            SelectObject (hDC, CreateFont(100, 100, 700, 700, FW_DONTCARE,
        //                          FALSE, FALSE, FALSE,
        //                          ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
        //                          DEFAULT_QUALITY, FF_ROMAN, NULL) );
        //            wglUseFontBitmaps (hDC, 0, 255, 1500); 
        //            glListBase (1500);
        //=============================================================================
            iFirst_Test = 0;
#else
            myCreateFont();
#endif
        }

        // DO the labels
        glMatrixMode(GL_TEXTURE);
        glDisable(GL_TEXTURE_3D);
        glDisable(GL_BLEND);

        for(itick=0; itick<m_iXNTicks; itick++) {
            xx1 = xxv0 - (float)0.07 + (xxv1-xxv0) * (float)m_iXTicks[itick]/(float)256.;
            glRasterPos3f(xx1, (float)(-1.9), -1.);
            glCallLists (strlen(m_sXTicks[itick]), GL_UNSIGNED_BYTE, m_sXTicks[itick]);
        }

        for(itick=0; itick<m_iYNTicks; itick++) {
            yy1 = yyv0 - (float)0.03 + (yyv1-yyv0) * (float)m_iYTicks[itick]/(float)512;
            glRasterPos3f((float)(-1.9), yy1, -1.);
            glCallLists (strlen(m_sYTicks[itick]), GL_UNSIGNED_BYTE, m_sYTicks[itick]);
        }
        sprintf(m_sYTicks[m_iYNTicks], "Log PDF");
        yy1 = yyv0 + (yyv1-yyv0) * (float)470/(float)512.;
        glRasterPos3f(xxv0+(float)(0.05), yy1, -1.);
        glCallLists (strlen(m_sYTicks[m_iYNTicks]), GL_UNSIGNED_BYTE, m_sYTicks[m_iYNTicks]);

        glEnable(GL_BLEND);
        glEnable(GL_TEXTURE_3D);

        glFlush();
        glFinish();

        /* Make sure that tick marks will be current the next time */
        m_iXNTicks = 0;
        m_iYNTicks = 0;
    }                   // END PDF block

    sprintf(m_sStats, "none");      /* Do not get stuck in cross-correlation mode */

    if(m_iNRendBuffs == 2  &&  m_iAutoSwapBuffers == 1) {
        mySwapBuffers();
        glFlush();
        glFinish();
    }
    m_iRender = 0;        /* will stop rendering upon return */
    
    /* glGetIntegerv(GL_ACCUM_RED_BITS,   &fRBits);
    glGetIntegerv(GL_ACCUM_GREEN_BITS, &fGBits);
    glGetIntegerv(GL_ACCUM_BLUE_BITS,  &fBBits);
    glGetIntegerv(GL_ACCUM_ALPHA_BITS, &fABits);

    fpp = fopen("AccumBits.txt", "w");
    fprintf(fpp, "Number of Red    bits = %d\n", (int)fRBits);
    fprintf(fpp, "Number of Green  bits = %d\n", (int)fGBits);
    fprintf(fpp, "Number of Blue   bits = %d\n", (int)fBBits);
    fprintf(fpp, "Number of Alpha  bits = %d\n", (int)fABits);
    fclose(fpp); */

    if(m_iTimersOn == 1) { tsEvent.StampTime("End   Output"); }
    if(m_iTimersOn == 1) {
        tsEvent.StampTime("End   Rendering");
        tsEvent.WriteToFile();
    }

    /* FILE *fp;
       fp = fopen("C:/DEBUG_HVR.txt", "a+");
       fprintf(fp, "At end of draw_scene iNTileRender = %d\n", iNTileRender);
       fclose(fp); */

#ifdef DEBUG_NEW_TREE
fprintf(stderr, "XXX at end of draw_scene\n");
#endif
    return;
}

/* ================================================================ *
 * ROUTINE section calculates an ordered set of N points            *
 * of the edges intersecting a plane specified by pos0 and norm0    *
 * ================================================================ */

int section(float pos0[3], float norm0[3], float norm1[3], float norm2[3],
            float edges[100][2][3], int nedges, float tvert[6][3]         )
{
    float dotsumi, dx[2][3], dot[2], verts[6][3], dvcx, dvcy, magi;
    int i, j, jmin, k, npoints;
    float posc[3], clock[9], clocksav[9], cmin;

    npoints = 0;
    for(k=0; k<nedges; k++) {
      for(j=0; j<2; j++) {
        for(i=0; i<3; i++) { dx[j][i] = edges[k][j][i] - pos0[i]; }
        dot[j] = dx[j][0]*norm0[0] + dx[j][1]*norm0[1] + dx[j][2]*norm0[2];
      }
      if(dot[0]*dot[1] < 0.) {
        dotsumi = 1. / (dot[0] - dot[1]);
        for(i=0; i<3; i++) { 
          verts[npoints][i] = pos0[i]
                            + dotsumi * (dot[0]*dx[1][i] - dot[1]*dx[0][i]);
        }
        npoints++;
      }
    }

    if(npoints < 3) {    /* NO SECTION FOUND */
      return(0);

    } else {             /* SECTION FOUND, SORT VERTICES IN CLOCKWISE ORDER */

      for(i=0; i<3; i++) { 
        posc[i] = 0.;
        for(k=0; k<npoints; k++) { posc[i] += verts[k][i]; }
        posc[i] /= (float)npoints;
      }

      for(k=0; k<npoints; k++) {
        dvcx = norm1[0]*(verts[k][0]-posc[0])
             + norm1[1]*(verts[k][1]-posc[1])
             + norm1[2]*(verts[k][2]-posc[2]) ;
        dvcy = norm2[0]*(verts[k][0]-posc[0])
             + norm2[1]*(verts[k][1]-posc[1])
             + norm2[2]*(verts[k][2]-posc[2]) ;
        magi = 1. / sqrt(dvcx*dvcx + dvcy*dvcy);
        dvcx *= magi;
        dvcy *= magi;
        if(dvcx >= 0. && dvcy >  0.) { clock[k] =     (dvcx    ) -(dvcy-1.0); }
        if(dvcx >  0. && dvcy <= 0.) { clock[k] = 2. -(dvcx-1.0) -(dvcy    ); }
        if(dvcx <= 0. && dvcy <  0.) { clock[k] = 4. -(dvcx    ) +(dvcy+1.0); }
        if(dvcx <  0. && dvcy >= 0.) { clock[k] = 6. +(dvcx+1.0) +(dvcy    ); }
      }

      for(k=0; k<npoints; k++) {
        cmin = 10.;
        for(j=0; j<npoints; j++) {
          if(clock[j] < cmin) { jmin=j; cmin=clock[j]; }
        }
        for(i=0; i<3; i++) { tvert[k][i] = verts[jmin][i]; }
        clocksav[k] = clock[jmin];
        clock[jmin] = 12.0;
      }

      return(npoints);
    }
}

/*----------------------------------------
 *      TODO
 * merge tube/point coords into 3 [xyz] long arrays
 * array of beginning & ending indexs for each tube
 * generate beginning & ending indexs for each level of blocktree
 * for each level, associate line strips w/ subvolumes
 */

/*-------------------------------------------------------
 * VECTOR ROUTINES
 */

void setvectors()
{
  FILE *fp;
  char cline[256], ckey[256];
  int iobj, nvec, ic, j;
  float dv0x,dv0y,dv0z,dv0mag,dv1x,dv1y,dv1z,dot01,dv1mag, pi;
  int jcur, jm1, jp1, isset;

  if(strcmp(cVecFile, cVecFileOld) == 0) { return; }
  sprintf(cVecFileOld, "%s", cVecFile);

  ic = 0;
  nvecobj = 0;
  ntotvecobj = nvecobj;
  for(j=0; j<MAXNODES; j++) { nvecobjs[j] = -1; }

  if(strcmp(cVecFile,"none") == 0) { return; }

  fp = fopen(cVecFile, "r");
  if(fp == NULL) { return; }

  fgets(cline, 256, fp);
  sscanf(cline, "%s", ckey);
  if(strcmp(ckey,"#ntubes") == 0) {
    sscanf(cline, "%s %d", ckey, &nvecobj);

    if(nvecobj > MAXOBJ) { nvecobj = MAXOBJ; }
    for(iobj=0; iobj<nvecobj; iobj++) {
      fscanf(fp, "%s %d", ckey, &nvec);

      iptstart[iobj] = ic;
      for(j=0; j<nvec; j++) {
        fscanf(fp, "%f %f %f %f", &xpt[ic],&ypt[ic],&zpt[ic],&vpt[ic]);
        ic++;
      }
      // ic 1 past last pt of strip: j<iptend[iobj] will be used
      iptend[iobj] = ic;
    }
  }
  fclose(fp);

  ntotvecobj = nvecobj;

  vecvalmin = vpt[0];
  vecvalmax = vpt[0];
  for(j=0; j<iptend[ntotvecobj-1]; j++) {
    if(vecvalmin > vpt[j]) { vecvalmin = vpt[j]; }
    if(vecvalmax < vpt[j]) { vecvalmax = vpt[j]; }
  }

  pi = acos(-1.0);
  for(j=0; j<ntubefacets; j++) {
    ccos[j] = cos(2.0*pi*(float)j / (float)ntubefacets);
    ssin[j] = sin(2.0*pi*(float)j / (float)ntubefacets);
  }

 /*********************************************************
  * Construct two unit normal vectors for each line strip
  */
  for(iobj=0; iobj<ntotvecobj; iobj++) {

    isset = 0;

    for(j=iptstart[iobj]; j<iptend[iobj]; j++) {
      mynormalsareset[j] = 0;

      jm1 = j-1;
      jp1 = j+1;
      if(jm1 < iptstart[iobj]) { jm1=j; }
      if(jp1 > iptend[iobj]-1) { jp1=j; }

      dv0x = xpt[jp1] - xpt[jm1];
      dv0y = ypt[jp1] - ypt[jm1];
      dv0z = zpt[jp1] - zpt[jm1];
      dv0mag = sqrt(dv0x*dv0x + dv0y*dv0y + dv0z*dv0z);
      if(dv0mag <= 0.0) { dv0mag = 1.0; }                          // modified: 2008 Nov 24
      if(dv0mag > 0.0) {
        dv0x /= dv0mag;   // Unit vector tangent to line-strip
        dv0y /= dv0mag;
        dv0z /= dv0mag;

        if(isset == 0) {
          if(dv0x < -0.7 || 0.7 < dv0x) { dv1x = 0.0; dv1y = 1.0; dv1z = 0.0; }
          else                          { dv1x = 1.0; dv1y = 0.0; dv1z = 0.0; }
          isset = 1;
        }
        dot01 = dv0x*dv1x + dv0y*dv1y + dv0z*dv1z;
        dv1x -= dot01 * dv0x;
        dv1y -= dot01 * dv0y;
        dv1z -= dot01 * dv0z;

        dv1mag = sqrt(dv1x*dv1x + dv1y*dv1y + dv1z*dv1z);
        if(dv1mag <= 0.0) { dv1mag = 1.0; }                        // modified: 2008 Nov 24
        if(dv1mag > 0.0) {
          dv1x /= dv1mag;   // 1st unit normal to line-strip
          dv1y /= dv1mag;
          dv1z /= dv1mag;
          xundir1[j] = dv1x;                   // 1st unit normal
          yundir1[j] = dv1y;
          zundir1[j] = dv1z;

          xundir2[j] = dv0y*dv1z - dv0z*dv1y;  // 2nd unit normal
          yundir2[j] = dv0z*dv1x - dv0x*dv1z;
          zundir2[j] = dv0x*dv1y - dv0y*dv1x;

          mynormalsareset[j] = 1;
        }
      }
    }

    // Make sure all normals are set
    jcur = -1;
    j = iptstart[iobj];
    while(jcur < 0  &&  j<iptend[iobj]) {
      if(mynormalsareset[j] == 1) { jcur = j; } 
      j++;
    }
    if(jcur < 0) {
      fprintf(stderr, "can not set normals\n");
      vecradius = 0.0;
      return;
    }
    for(j=iptstart[iobj]; j<iptend[iobj]; j++) {
      if(mynormalsareset[j] == 0) {
        xundir1[j] = xundir1[jcur];
        yundir1[j] = yundir1[jcur];
        zundir1[j] = zundir1[jcur];
        xundir2[j] = xundir2[jcur];
        yundir2[j] = yundir2[jcur];
        zundir2[j] = zundir2[jcur];

        //fprintf(stderr, "iobj=%2d   j=%3d\n", iobj, j);
      } 
      jcur = j;
    }
  }
}

void genvectorlist(int node0, int isub, float *subminpos, float *submaxpos)
{
  int j, iobj;
  int iwasinside,iaminside;
  float xmin, xmax, ymin, ymax, zmin, zmax;

  nvecobjs[isub] = 0;
  if(nvecobj == 0) { return; }

  if(isub == node0) {
    // Volume of root node contains all vector objects
    nvecobjs[isub] = nvecobj;
    istartvecobj[isub] = 0;
    return;
  }

  // Find all vector objects that overlap this subvolume

  xmin = subminpos[0] - vecradius;
  ymin = subminpos[1] - vecradius;
  zmin = subminpos[2] - vecradius;
  xmax = submaxpos[0] + vecradius;
  ymax = submaxpos[1] + vecradius;
  zmax = submaxpos[2] + vecradius;

  istartvecobj[isub] = ntotvecobj;
  for(iobj=0; iobj<nvecobj; iobj++) {
    iwasinside = 0;
    for(j=iptstart[iobj]; j<iptend[iobj]; j++) {
      iaminside = 1;
      if(xpt[j] < xmin  ||  xpt[j] > xmax) { iaminside = 0; }
      if(ypt[j] < ymin  ||  ypt[j] > ymax) { iaminside = 0; }
      if(zpt[j] < zmin  ||  zpt[j] > zmax) { iaminside = 0; }

      if(iwasinside == 0  &&  iaminside == 1) { // start a new vector object
        if(j > iptstart[iobj]) { iptstart[ntotvecobj] = j-1; }
        else                   { iptstart[ntotvecobj] = j; }
      }
      if(iwasinside == 1  &&  iaminside == 0) { // finish the new vector object
        iptend[ntotvecobj] = j+1;
        ntotvecobj++;
        if(ntotvecobj > MAXOBJ) {
          fprintf(stderr, "ntotvecobj > MAXOBJ\n");
          nvecobjs[isub] = ntotvecobj - istartvecobj[isub];
          return;
        }
      }
      iwasinside = iaminside;
    }
    if(iwasinside == 1) { // finish the new vector object
      iptend[ntotvecobj] = iptend[iobj];
      ntotvecobj++;
      if(ntotvecobj > MAXOBJ) {
        fprintf(stderr, "ntotvecobj > MAXOBJ\n");
        nvecobjs[isub] = ntotvecobj - istartvecobj[isub];
        return;
      }
    }
  }
  nvecobjs[isub] = ntotvecobj - istartvecobj[isub];
}

void drawsurfaces(int isub, float ss1, float ss2, float eye[3])
{
  int iobj, j, iobjstart, iobjend;
  float red2[2], grn2[2], blu2[2], xx, yy, zz;
  int i, ii, j1, j2, jj;
  float red0=0.0, grn0=0.5, blu0=0.0;
  float dred=1.0, dgrn=0.5, dblu=1.0;
  float red, grn, blu;
  float rangeinv, frac;
  float xounv,younv,zounv,facillum,dot, dot0, dot1;
  float xounm[2][20],younm[2][20],zounm[2][20],dotm[2][20];
  int iistart, iiend, i0, i1;
  int ired0, igrn0, iblu0, ired1, igrn1, iblu1;
  float dotav, dotavlast;
  static float radmin=1.0f;
  float radnow;

  sscanf(cVecLoClr, "#%02x%02x%02x", &ired0, &igrn0, &iblu0);
  sscanf(cVecHiClr, "#%02x%02x%02x", &ired1, &igrn1, &iblu1);
  red0 = (float)ired0 / 255.0f;
  grn0 = (float)igrn0 / 255.0f;
  blu0 = (float)iblu0 / 255.0f;
  dred = (float)ired1 / 255.0f - red0;
  dgrn = (float)igrn1 / 255.0f - grn0;
  dblu = (float)iblu1 / 255.0f - blu0;

  rangeinv = 1.0;
  if(vecvalmax > vecvalmin) { rangeinv = 1.0 / (vecvalmax-vecvalmin); }

  iobjstart = istartvecobj[isub];
  iobjend   = istartvecobj[isub] + nvecobjs[isub];

  // glColor3f(1.0, 1.0, 1.0);
  for(iobj=iobjstart; iobj<iobjend; iobj++) {
    if(vecradius == 0.0) {
      glBegin(GL_LINE_STRIP);
      for(j=iptstart[iobj]; j<iptend[iobj]; j++) {
        frac = (vpt[j]-vecvalmin) * rangeinv;
        red = red0+frac*dred;
        grn = grn0+frac*dgrn;
        blu = blu0+frac*dblu;
        glColor3f(red, grn, blu);
        glVertex3f(xpt[j], ypt[j], zpt[j]);
      }
      glEnd();
    } else {
      j1 = iptstart[iobj];
      dot0 = (eye[0]-xpt[j1])*xcenteye
           + (eye[1]-ypt[j1])*ycenteye
           + (eye[2]-zpt[j1])*zcenteye;

      for(j=iptstart[iobj]; j<iptend[iobj]-1; j++) {
        
// HERE
        j1 = j + 1;
        dot1 = (eye[0]-xpt[j1])*xcenteye
             + (eye[1]-ypt[j1])*ycenteye
             + (eye[2]-zpt[j1])*zcenteye;
        if((ss1-vecradius <= dot0  &&  dot0 <= ss2+vecradius) ||
           (ss1-vecradius <= dot1  &&  dot1 <= ss2+vecradius)   ) {

/*
        if(radnow < dot1) {
          radnow = sqrt((eye[0]-xpt[j1])*(eye[0]-xpt[j1])
                 +      (eye[1]-ypt[j1])*(eye[1]-ypt[j1])
                 +      (eye[2]-zpt[j1])*(eye[2]-zpt[j1]));

          fprintf(stderr,"  \n");
          fprintf(stderr, "Rad=%f  dot1=%f ss1=%f  ss2=%f  vecrad=%f\n",
                   radnow, dot1, ss1, ss2, vecradius);
          fprintf(stderr,"EYE=%f %f %f\n", eye[0],eye[1],eye[2]);
          fprintf(stderr,"PT =%f %f %f\n", xpt[j1],ypt[j1],zpt[j1]);
          radmin = dot1;
        }
*/

        for(jj=0; jj<2; jj++) {
          j1 = j + jj;
          j2 = j1 % 2;

          // Vertex color
          frac = (vpt[j1]-vecvalmin) * rangeinv;
          red2[jj] = red0+frac*dred;
          grn2[jj] = grn0+frac*dgrn;
          blu2[jj] = blu0+frac*dblu;

          // XYZ-components of Outward Unit Normal Vectors
          for(i=0; i < ntubefacets; i++) {
            xounm[j2][i] = ccos[i]*xundir1[j1] + ssin[i]*xundir2[j1];
            younm[j2][i] = ccos[i]*yundir1[j1] + ssin[i]*yundir2[j1];
            zounm[j2][i] = ccos[i]*zundir1[j1] + ssin[i]*zundir2[j1];
            dotm[j2][i] = xounm[j2][i]*xcenteye
                        + younm[j2][i]*ycenteye
                        + zounm[j2][i]*zcenteye;
          }
        }

        // Generate range of edges bounding quads that face you
        dotavlast = 0.25 *(dotm[0][0]+dotm[1][0] +
                           dotm[0][1]+dotm[1][1]  );
        ii = 1;
        iistart = -1;
        iiend   = -1;
        while(iiend < 0) {
          i0 =  ii    % ntubefacets;
          i1 = (ii+1) % ntubefacets;
          dotav = 0.25 *(dotm[0][i0]+dotm[1][i0] +
                         dotm[0][i1]+dotm[1][i1]  );
          if(dotav >= 0.0 && dotavlast <  0.0) { iistart = ii; }
          if(dotav < 0.0 && dotavlast >= 0.0 && iistart >= 0) { iiend   = ii; }

          dotavlast = dotav;
          ii++;
        }

        glBegin(GL_QUAD_STRIP);
        for(ii=iistart; ii <= iiend; ii++) {
          i = ii % ntubefacets;
          for(jj=0; jj<2; jj++) {
            j1 = j + jj;
            j2 = j1 % 2;

            xounv = xounm[j2][i];
            younv = younm[j2][i];
            zounv = zounm[j2][i];

            // Vertex on surface of tube
            xx = xpt[j1] + vecradius * xounv;
            yy = ypt[j1] + vecradius * younv;
            zz = zpt[j1] + vecradius * zounv;

            // Lighting model: diffusive, 70% lit from the eye, 30% ambient
            dot = xounv*xcenteye + younv*ycenteye + zounv*zcenteye;
            dot = dotm[j2][i];
            if(dot < 0.0) { facillum = 0.3; }
            else          { facillum = 0.3 + 0.7*dot; }

            red = facillum * red2[jj];
            grn = facillum * grn2[jj];
            blu = facillum * blu2[jj];

            glColor3f(red, grn, blu);
            glVertex3f(xx, yy, zz);
          }
        }
        glEnd();
        }
        dot0 = dot1;
      }
    }
  }
}

void Chvr_work::colorbar()
{
  float fZero, fOne, fMOne;
  GLdouble dleft,dright,dbottom,dtop,dnearclip,dfarclip;
  float epsx,epsy,zzv,xxv0,xxv1,yyv0,yyv1, yyvm, xxvm;
  float xcb0,xcb1,xfb0,xfb1,ycb0,ycb1,yfb0,yfb1,delchecker;
  int ix, iy;
  int knot, idoindex[256], i, k, ind;
  double RED, GRN, BLU, ALP, val0, val1, dfrac, dOne, dLow;

  // glDisable(GL_BLEND);
  glDisable(GL_TEXTURE_3D);
  glEnable(GL_TEXTURE_2D);

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();

  dleft     = (GLdouble)(-1.0) * (GLdouble)m_fScreenRatio;
  dright    = (GLdouble)( 1.0) * (GLdouble)m_fScreenRatio;
  dbottom   = (GLdouble)(-1.0) * (GLdouble)m_fScreenRatio;
  dtop      = (GLdouble)( 1.0) * (GLdouble)m_fScreenRatio;
  dnearclip = (GLdouble)( 0.5);
  dfarclip  = (GLdouble)( 1.5);
  glFrustum( dleft, dright, dbottom, dtop, dnearclip, dfarclip );

  fZero = (float)0.0;
  fOne  = (float)1.0;
  fMOne = (float)(-1.0);
  gluLookAt(fZero, fZero, fZero,  fZero, fZero, fMOne,  fZero, fOne, fZero);

  // Note: at z = -1, full XY visible is [-2,2]x[-2,2]
  zzv   = (float)(-1.0);

  // Color bar (cb), and full box (fb), ranges
  xcb0 = -1.8;
  xcb1 =  1.8;
  xfb0 = -1.8;
  xfb1 =  1.8;
  epsx =  0.01;

  if(m_iNTicks > 0) {
    ycb0 = -1.7;
    ycb1 = -1.6;
    yfb0 = -1.9;
    yfb1 = -1.6;
    epsy =  0.05;
  } else {
    ycb0 = -1.9;
    ycb1 = -1.8;
    yfb0 = -1.9;
    yfb1 = -1.8;
    epsy =  0.01;
  }
  if(m_iSelALPknot >= 0) { epsy = 0.05; }

  xxv0  = xfb0-epsx;
  xxv1  = xfb1+epsx;
  yyv0  = yfb0-epsy;
  yyv1  = yfb1+epsy;

  // Draw background of colorbar+annotation region: black
  glColor4d((double)0.0, (double)0.0, (double)0.0, (double)1.0);
  glBegin(GL_QUADS);
    glVertex3f(xxv0, yyv0, zzv);
    glVertex3f(xxv1, yyv0, zzv);
    glVertex3f(xxv1, yyv1, zzv);
    glVertex3f(xxv0, yyv1, zzv);
  glEnd();

  // Draw checkboard in colorbar region
  delchecker = 0.025;
  glColor4d((double)1.0, (double)1.0, (double)1.0, (double)1.0);
  iy = 0;
  for(yyv0=ycb0; yyv0<ycb1; yyv0 += delchecker) {
    iy++;
    yyv1 = yyv0 + delchecker;  if(yyv1 > ycb1) { yyv1 = ycb1; }
    ix = 0;
    for(xxv0=xcb0; xxv0<xcb1; xxv0 += delchecker) {
      ix++;
      if((ix+iy)%2 == 0) {
        xxv1 = xxv0 + delchecker;  if(xxv1 > xcb1) { xxv1 = xcb1; }
        glBegin(GL_QUADS);
          glVertex3f(xxv0, yyv0, zzv);
          glVertex3f(xxv1, yyv0, zzv);
          glVertex3f(xxv1, yyv1, zzv);
          glVertex3f(xxv0, yyv1, zzv);
        glEnd();
      }
    }
  }

  // Blend (color,alpha) map onto checker board
  yyv0 = ycb0;
  yyv1 = ycb1;

  for(i=0; i<256; i++) { idoindex[i] = 0; }
  for(i=0; i<m_iNalpNots; i++) { idoindex[m_iALPind[i]] = 1; }
  for(i=0; i<m_iNrgbNots; i++) { idoindex[m_iRGBind[i]] = 1; }

  glBegin(GL_QUAD_STRIP);
  for(ind=0; ind<256; ind++) { if(idoindex[ind] == 1) {

   // Interpolate X coordinate from index
   xxv0 = xcb0 + (xcb1-xcb0) * (float)ind / 255.0f;

    // Interpolate RGB values from index
    for(k=0; k<m_iNrgbNots-1; k++) { if(m_iRGBind[k] <= ind) { knot=k; }}
    dfrac = ((double)ind               - (double)m_iRGBind[knot])
          / ((double)m_iRGBind[knot+1] - (double)m_iRGBind[knot]);

    val0 = (double)m_fRED[knot  ];
    val1 = (double)m_fRED[knot+1];
    RED  = val0 + (val1-val0) * dfrac;

    val0 = (double)m_fGRN[knot  ];
    val1 = (double)m_fGRN[knot+1];
    GRN  = val0 + (val1-val0) * dfrac;

    val0 = (double)m_fBLU[knot  ];
    val1 = (double)m_fBLU[knot+1];
    BLU  = val0 + (val1-val0) * dfrac;

    // Interpolate Alpha value from index
    for(k=0; k<m_iNalpNots-1; k++) { if(m_iALPind[k] <= ind) { knot=k; }}
    dfrac = ((double)ind               - (double)m_iALPind[knot])
          / ((double)m_iALPind[knot+1] - (double)m_iALPind[knot]);

    val0 = (double)m_fALP[knot  ];
    val1 = (double)m_fALP[knot+1];
    ALP  = val0 + (val1-val0) * dfrac;
   
    glColor4d(RED, GRN, BLU, ALP);
    glVertex3f(xxv0, yyv0, zzv);
    glVertex3f(xxv0, yyv1, zzv);
  }}
  glEnd();


  // Show knots if any are selected
  if(m_iSelALPknot >= 0) {
    dOne = (double)1.0;
    dLow = (double)0.5;

    // Show alpha knots below color bar
    yyv0 = ycb0-epsy;
    yyv1 = ycb0;
    yyvm = 0.5 * (yyv0 + yyv1);
    for(k=0; k<m_iNalpNots; k++) { 
      xxvm = xcb0 + (xcb1-xcb0) * (float)m_iALPind[k] / 255.0f;
      xxv0 = xxvm - 0.5*epsy;
      xxv1 = xxvm + 0.5*epsy;
      if(k == m_iSelALPknot) { glColor4d(dOne, dOne, dOne, dOne); }
      else                   { glColor4d(dLow, dLow, dLow, dOne); }
      glBegin(GL_QUADS);
        glVertex3f(xxvm, yyv0, zzv);
        glVertex3f(xxv1, yyvm, zzv);
        glVertex3f(xxvm, yyv1, zzv);
        glVertex3f(xxv0, yyvm, zzv);
      glEnd();
    }

    // Show color knots above color bar
    yyv0 = ycb1;
    yyv1 = ycb1+epsy;
    yyvm = 0.5 * (yyv0 + yyv1);
    for(k=0; k<m_iNrgbNots; k++) { 
      xxvm = xcb0 + (xcb1-xcb0) * (float)m_iRGBind[k] / 255.0f;
      xxv0 = xxvm - 0.5*epsy;
      xxv1 = xxvm + 0.5*epsy;
      if(k == m_iSelRGBknot) { glColor4d(dOne, dOne, dOne, dOne); }
      else                   { glColor4d(dLow, dLow, dLow, dOne); }
      glBegin(GL_QUADS);
        glVertex3f(xxvm, yyv0, zzv);
        glVertex3f(xxv1, yyvm, zzv);
        glVertex3f(xxvm, yyv1, zzv);
        glVertex3f(xxv0, yyvm, zzv);
      glEnd();
    }
  }

// ------------------

  glPopMatrix();
  glDisable(GL_TEXTURE_2D);
  glEnable(GL_TEXTURE_3D);
  //glEnable(GL_BLEND);

}

void annotation(int iBefore, float eyepos[3])
{
    if(iFloorAlp + iBoxAlp == 0) { return; }
    
    /* Make Sure that alpha is full on for fiducials */
    glColor4d((double)1.0, (double)1.0, (double)1.0, (double)1.0);

    // Work variables
    float fZero = (float)0.0;
    float fOne  = (float)1.0;
    float zzv   = fLimits[4];

    float fLimMin = (fLimits[1] - fLimits[0]);
    if(fLimMin > fLimits[3]-fLimits[2]) { fLimMin = fLimits[3] - fLimits[2]; }
    if(fLimMin > fLimits[5]-fLimits[4]) { fLimMin = fLimits[5] - fLimits[4]; }
    float fLimMax = (fLimits[1] - fLimits[0]);
    if(fLimMax < fLimits[3]-fLimits[2]) { fLimMax = fLimits[3] - fLimits[2]; }
    if(fLimMax < fLimits[5]-fLimits[4]) { fLimMax = fLimits[5] - fLimits[4]; }

    float fLwidth = (float)0.005 * fLimMax;

    float floorX0  = fLimits[0] - fFloorSize * (fLimits[1] - fLimits[0]);
    float floorX1  = fLimits[1] + fFloorSize * (fLimits[1] - fLimits[0]);
    float floorY0  = fLimits[2] - fFloorSize * (fLimits[3] - fLimits[2]);
    float floorY1  = fLimits[3] + fFloorSize * (fLimits[3] - fLimits[2]);


    float xxv0  = floorX0;
    float xxv1  = floorX1;
    float yyv0  = floorY0;
    float yyv1  = floorY1;
    int i, j;

    if(iFloorAlp > 0) {
        /* IF floor alpha is > 0, THEN do the floor */

        /* Do Floor before or after rendering of volume depending on postion of eyepoint */
        if( (iBefore == 0  &&  eyepos[2] < fLimits[4])  ||  (iBefore == 1  &&  eyepos[2] > fLimits[4]) ) {
            for(i=0; i<256*32; i++) {
                sol[i] =      (unsigned long)iFloorRed
                          + (((unsigned long)iFloorGrn)<<8 )
                          + (((unsigned long)iFloorBlu)<<16)
                          + (((unsigned long)iFloorAlp)<<24);
            }

            glColor4d((double)1.0, (double)1.0, (double)1.0, (double)1.0);

            glTexImage2D(GL_TEXTURE_3D,0,GL_RGBA,256,32,0,GL_RGBA,GL_UNSIGNED_BYTE,sol);
            glBegin(GL_QUADS);
            glTexCoord2f(fZero, fZero);    glVertex3f(xxv0, yyv0,zzv);
            glTexCoord2f(fOne , fZero);    glVertex3f(xxv1, yyv0,zzv);
            glTexCoord2f(fOne , fOne );    glVertex3f(xxv1, yyv1,zzv);
            glTexCoord2f(fZero, fOne );    glVertex3f(xxv0, yyv1,zzv);
            glEnd();

            for(i=0; i<32*256; i++) { sol[i] = ((unsigned long)255)<<24; }

            glTexImage2D(GL_TEXTURE_3D,0,GL_RGBA,256,32,0,GL_RGBA,GL_UNSIGNED_BYTE,sol);

            int i0 = 100 - (int)((float)100.0 - floorX0 / fTickSpacing);
            int i1 = (int)((float)100.0 + floorX1 / fTickSpacing) - 100;
            for(i=i0; i <= i1; i++) {
                xxv0 = fTickSpacing * (float)i;
                xxv1 = fTickSpacing * (float)i;

                glBegin(GL_QUADS);
                glTexCoord2f(fZero, fZero);    glVertex3f(xxv0-fLwidth, floorY0, zzv);
                glTexCoord2f(fOne , fZero);    glVertex3f(xxv1+fLwidth, floorY0, zzv);
                glTexCoord2f(fOne , fOne );    glVertex3f(xxv1+fLwidth, floorY1, zzv);
                glTexCoord2f(fZero, fOne );    glVertex3f(xxv0-fLwidth, floorY1, zzv);
                glEnd();

                /*glBegin(GL_QUADS);
                glTexCoord2f(fZero, fZero);    glVertex3f(xxv0, floorY0, zzv-fLwidth);
                glTexCoord2f(fOne , fZero);    glVertex3f(xxv1, floorY0, zzv+fLwidth);
                glTexCoord2f(fOne , fOne );    glVertex3f(xxv1, floorY1, zzv+fLwidth);
                glTexCoord2f(fZero, fOne );    glVertex3f(xxv0, floorY1, zzv-fLwidth);
                glEnd();*/
            }

            int j0 = 100 - (int)((float)100.0 - floorY0 / fTickSpacing);
            int j1 = (int)((float)100.0 + floorY1 / fTickSpacing) - 100;
            for(j=j0; j <= j1; j++) {
                yyv0 = fTickSpacing * (float)j  -  fLwidth;
                yyv1 = fTickSpacing * (float)j  +  fLwidth;

                glBegin(GL_QUADS);
                glTexCoord2f(fZero, fZero);    glVertex3f(floorX0, yyv0, zzv);
                glTexCoord2f(fOne , fZero);    glVertex3f(floorX1, yyv0, zzv);
                glTexCoord2f(fOne , fOne );    glVertex3f(floorX1, yyv1, zzv);
                glTexCoord2f(fZero, fOne );    glVertex3f(floorX0, yyv1, zzv);
                glEnd();
            }

            glFlush();
            glFinish();
        }
    }

    if(iBoxAlp > 0) {
        /* IF box alpha is > 0, THEN do the bounding box */

        for(i=0; i<256*32; i++) {
            sol[i] =      (unsigned long)iBoxRed
                      + (((unsigned long)iBoxGrn)<<8 )
                      + (((unsigned long)iBoxBlu)<<16)
                      + (((unsigned long)iBoxAlp)<<24);
        }
        glTexImage2D(GL_TEXTURE_3D,0,GL_RGBA,256,32,0,GL_RGBA,GL_UNSIGNED_BYTE,sol);

        float fBAmp;
        float vert[4][3], vc[4][3];
        //int ijk0, ijk1, ijk2;
        int id0, id1, id2, iv0, iv1, iid, iv, iSide, iDir;
        float dvc0, dvc1, fwd[3];
        if(iBefore == 1) { fBAmp = -(float)1.0; } else { fBAmp = (float)1.0; }
        for(iDir=0; iDir<3; iDir++) {
            id0 =  iDir;
            id1 = (iDir+1)%3;
            id2 = (iDir+2)%3;
            for(iSide = 0; iSide < 2; iSide++) {
                if(fBAmp*(float)(2*iSide-1)*(eyepos[id2] - fLimits[2*id2+iSide]) > (float)0.0) {

                    /* vc[0] -> vc[1] -> vc[2] -> vc[3] -> vc[0] goes around the cube face
                       on iSide in the iDir direction */
                    for(iv=0; iv<4; iv++) { vc[iv][id2] = fLimits[2*id2+iSide]; }
                    vc[0][id0] = fLimits[2*id0  ];   vc[0][id1] = fLimits[2*id1  ];
                    vc[1][id0] = fLimits[2*id0+1];   vc[1][id1] = fLimits[2*id1  ];
                    vc[2][id0] = fLimits[2*id0+1];   vc[2][id1] = fLimits[2*id1+1];
                    vc[3][id0] = fLimits[2*id0  ];   vc[3][id1] = fLimits[2*id1+1];

                    for(iv0=0; iv0<4; iv0++) {
                        iv1 = (iv0+1)%4;

                        dvc0 = vc[iv1][id0] - vc[iv0][id0];   if(dvc0 < (float)0.0) { dvc0 = -dvc0; }
                        dvc1 = vc[iv1][id1] - vc[iv0][id1];   if(dvc1 < (float)0.0) { dvc1 = -dvc1; }

                        if(dvc0 > dvc1) { fwd[id0] = (float)0.0;  fwd[id1] = fLwidth; fwd[id2] = (float)0.0; }
                        else            { fwd[id1] = (float)0.0;  fwd[id0] = fLwidth; fwd[id2] = (float)0.0; }

                        for(iid=0; iid<3; iid++) {
                            vert[0][iid] = vc[iv0][iid] - fwd[iid];
                            vert[1][iid] = vc[iv1][iid] - fwd[iid];
                            vert[2][iid] = vc[iv1][iid] + fwd[iid];
                            vert[3][iid] = vc[iv0][iid] + fwd[iid];
                        }

                        glBegin(GL_QUADS);
                        glTexCoord2f(fZero, fZero);    glVertex3f(vert[0][0],vert[0][1],vert[0][2]);
                        glTexCoord2f(fOne , fZero);    glVertex3f(vert[1][0],vert[1][1],vert[1][2]);
                        glTexCoord2f(fOne , fOne );    glVertex3f(vert[2][0],vert[2][1],vert[2][2]);
                        glTexCoord2f(fZero, fOne );    glVertex3f(vert[3][0],vert[3][1],vert[3][2]);
                        glEnd();
                    }
                }
            }
        }
        glFlush();
        glFinish();
    }
}


void crossprod(float vec1[3], float vec2[3], float vec3[3]) {
    vec3[0] = vec1[1] * vec2[2]  -  vec1[2] * vec2[1];
    vec3[1] = vec1[2] * vec2[0]  -  vec1[0] * vec2[2];
    vec3[2] = vec1[0] * vec2[1]  -  vec1[1] * vec2[0];
}

/***************************************************************************
 *  The integer function strdif(str1, str2) takes two string arguments.    *
 *  If the strings are identical (ending in '\0') strdif returns 0.        *
 *  If there are any differences, strdif returns 1.                        *
 ***************************************************************************/
int strdif(char* str1, char* str2)
{
    int i, idiff;
    i = 0;
    while(str1[i] == str2[i] && str1[i] != '\0') { i++; }
    idiff = 1;
    if(str1[i] == str2[i]) { idiff = 0; }
    return(idiff);
}
void setstr(char* str1, char* str2)
{
    int i;
    i = 0;
    while(str2[i] != '\0') { str1[i] = str2[i]; i++; }
    str1[i] = '\0';
}


void Chvr_work::init(void)
{   
    signal(SIGTERM,&SIGTERM_Handler);
    signal(SIGINT,&SIGTERM_Handler);
#ifdef LINUX
    signal(SIGQUIT,&SIGTERM_Handler);
    signal(SIGHUP,&SIGTERM_Handler);
#endif
    
    error_log = fopen("hvr_log.txt","w");
    if(error_log == NULL) {error_log  = stderr; }
    if(error_log == NULL) { exitHVR(LOG_FAILED); }
    else                  { status = SUCCESS; }
    
    int i;

    /* Initialize Texture Memory Management Array */
    if(ResidentTextures) {
        for(i=0; i<MAXTEX; i++)
            bricktex[i] = -1;
    }

    glClearColor(0.0, 0.0, 0.0, 1.0);

    glEnable(GL_TEXTURE_1D);
    glEnable(GL_BLEND);

    glDisable(GL_DEPTH_TEST);
    glDisable(GL_DITHER);
    glDisable(GL_FOG);
    glDisable(GL_LIGHTING);
    glDisable(GL_STENCIL_TEST);
    glDisable(GL_LOGIC_OP);

    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glPixelStorei(GL_PACK_ALIGNMENT, 1);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    /* glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST ); */

#ifndef USE_GLUT
    //glClear(GL_COLOR_BUFFER_BIT);
    //glFlush();
    //if(m_iNRendBuffs == 2) {  mySwapBuffers(); }
#endif
}
/*
bool Chvr_work::DoRender()
{
    bool bRet;
    bRet = FALSE;
    if(m_iRender == 1) bRet = TRUE;
    return bRet;

}
*/
STDMETHODIMP Chvr_work::SetDoRender()
{
    /*Only one thread should be running draw_scene at a time
    int ii;
    ii = InterlockedIncrement(&iNRendering);
    while(ii > 1) {
        ii = InterlockedDecrement(&iNRendering);
        _sleep(10);
        ii = InterlockedIncrement(&iNRendering);
    }
    m_iRender = 1;*/

    /* Copy THIS objects versions of inputs to global copy for rendering thread to use
       See CopyHvrInputs */
    memcpy(&global_ShvrIn.iNewHV, &m_ShvrIn.iNewHV, sizeof(HvrInput));

    if(iSpawnRenderThread == 0) {  // >>>>
        draw_scene();
        iStopRendering = 0;
        iDoThis = 0;
    } else {
        iDoThis = 1;
        //if(iUseOpenGL == 1) { glutPostRedisplay(); }
        if(m_iBlockOnRender == 1) { while(iDoThis == 1) { _sleep(10); } }
    }

    //    ii = InterlockedDecrement(&iNRendering);

    return S_OK;
}

STDMETHODIMP Chvr_work::Set_HVname(unsigned char *cHVnew)
{
    sprintf(m_ShvrIn.sHVname, "%s\0", cHVnew);

    // void  setstr(char*, char*);
    // setstr(hvname, cHVnew);

    return S_OK;
}


/*****************************************************************************
 *                        READ IN PANNEL INFO HERE                           *
 *---------------------------------------------------------------------------*
 * read in of configuration file goes here : configuration file gives rules  *
 * for generating [center? and (horiz, vert) offsets]                        *
 * for each pannel for each pannel given center?, eye?, up? from above       *
 *---------------------------------------------------------------------------*
 *              torq    eye     up                                           *
 *            i    0      1      2                                           *
 *  pw_normal[i]   0.     1.     0.    <-- typical example of central pannel *
 *      pw_up[i]   0.     0.     1.    <-- typical example of central pannel *
 *                                                                           *
 *---------------------------------------------------------------------------*
 * Example PW_CONFIGURATION for single pannel (1280x1024 per pannel)         *
 *       0. 1. 0.  0. 0. 1.  0. 1. 0.  0.625 0.5                             *
 *---------------------------------------------------------------------------*
 * Example PW_CONFIGURATION for flat 4 pannel (1600x1200 per pannel)         *
 *       0. 1. 0.  0. 0. 1.  -0.6667 1.0 -0.5  0.6667 0.5                    *
 *       0. 1. 0.  0. 0. 1.   0.6667 1.0 -0.5  0.6667 0.5                    *
 *       0. 1. 0.  0. 0. 1.  -0.6667 1.0  0.5  0.6667 0.5                    *
 *       0. 1. 0.  0. 0. 1.   0.6667 1.0  0.5  0.6667 0.5                    *
 *****************************************************************************/

STDMETHODIMP Chvr_work::AddFrustum(float *center, float *up, float *normal, int iWidth, int iHeight, float fPanWidth)
{
    int i;
    float dot;

    m_ShvrIn.iNewFrustum = 1;
    m_ShvrIn.iWidth  = iWidth;
    m_ShvrIn.iHeight = iHeight;
    m_ShvrIn.fAspectRatio = (float)iHeight / (float)iWidth;

    for(i=0; i<3; i++) { m_ShvrIn.fPanCent[i] = center[i]; }
    for(i=0; i<3; i++) { m_ShvrIn.fPanNorm[i] = normal[i]; }
    for(i=0; i<3; i++) { m_ShvrIn.fPanUp[i]   = up[i]; }

    m_ShvrIn.fHalfWidth = (float)0.5 * fPanWidth;
    m_ShvrIn.fHalfHeight = m_ShvrIn.fAspectRatio * m_ShvrIn.fHalfWidth;

    dot = center[0]*normal[0] + center[1]*normal[1] + center[2]*normal[2];
    if(dot < 0.) { dot = -dot; }
    m_ShvrIn.fFovAngle = (float)2.0 * (float)atan(m_ShvrIn.fHalfHeight/dot) / Pid180;

    npannel = 1;

    return S_OK;
}

STDMETHODIMP Chvr_work::SetFrustum(float *center, float *up, float *normal, int iWidth, int iHeight, float fPanWidth)
{
    int i;
    float dot;

    m_ShvrIn.iNewFrustum = 1;
    m_ShvrIn.iWidth  = iWidth;
    m_ShvrIn.iHeight = iHeight;
    m_ShvrIn.fAspectRatio = (float)iHeight / (float)iWidth;

    for(i=0; i<3; i++) { m_ShvrIn.fPanCent[i] = center[i]; }
    for(i=0; i<3; i++) { m_ShvrIn.fPanNorm[i] = normal[i]; }
    for(i=0; i<3; i++) { m_ShvrIn.fPanUp[i]   = up[i]; }

    m_ShvrIn.fHalfWidth = (float)0.5 * fPanWidth;
    m_ShvrIn.fHalfHeight = m_ShvrIn.fAspectRatio * m_ShvrIn.fHalfWidth;

    dot = center[0]*normal[0] + center[1]*normal[1] + center[2]*normal[2];
    if(dot < 0.) { dot = -dot; }
    m_ShvrIn.fFovAngle = (float)2.0 * (float)atan(m_ShvrIn.fHalfHeight/dot) / Pid180;

    npannel = 1;

    return S_OK;
}

void Chvr_work::SetDefaultFrustum()
{

    m_ShvrIn.iNewFrustum = 1;
    m_ShvrIn.iWidth  = 768;
    m_ShvrIn.iHeight = 768;
    m_ShvrIn.fAspectRatio = (float)m_ShvrIn.iWidth / (float)m_ShvrIn.iHeight;

    m_ShvrIn.fPanNorm[0] = 0.;    /* default single pannel settings */
    m_ShvrIn.fPanNorm[1] = 1.;
    m_ShvrIn.fPanNorm[2] = 0.;
    m_ShvrIn.fPanUp[0]   = 0.;
    m_ShvrIn.fPanUp[1]   = 0.;
    m_ShvrIn.fPanUp[2]   = 1.;
    m_ShvrIn.fPanCent[0] = 0.;
    m_ShvrIn.fPanCent[1] = 1.;
    m_ShvrIn.fPanCent[2] = 0.;
    
    m_ShvrIn.fFovAngle  = (float)35.3084;
    m_ShvrIn.fHalfHeight = (float)tan( Pid180* m_ShvrIn.fFovAngle * 0.5 );
    m_ShvrIn.fHalfWidth  = m_ShvrIn.fHalfHeight / m_ShvrIn.fAspectRatio;

    npannel = 1;
}
/*
    struct HvrInput {    // This structure contains all of the input needed for the hvr code
        int iNewHV;
        char sHVname[100];

        int iNewFrustum;
        int iWidth;
        int iHeight;
        float fPanCent[3];
        float fPanNorm[3];
        float fPanUp[3];
        float fHalfWidth;
        float fHalfHeight;
        float fFovAngle;
        float fAspectRatio;

        int iNewLUT;
        char sCmapname[100];
        char sAmapname[100];

        int iNewView;
        float fEye[3];
        float fCent[3];
        float fUp[3];
        float fNearClip;
        float fFarrClip;
    };
    HvrInput m_ShvrIn;     // GUI writes input here
    HvrInput m_ShvrUse;    // Rendering code uses ShvrUse, after copying info from ShvrIn
*/

void Chvr_work::SetDefaultView()
{
    m_ShvrIn.iNewView = 1;
    m_ShvrIn.fCent[0] = 0.;
    m_ShvrIn.fCent[1] = 0.;
    m_ShvrIn.fCent[2] = 0.;
    m_ShvrIn.fEye[0] = 0.0;
    m_ShvrIn.fEye[1] = 0.0;
    m_ShvrIn.fEye[2] = 10.0;
    m_ShvrIn.fUp[0] = 0.;
    m_ShvrIn.fUp[1] = 1.;
    m_ShvrIn.fUp[2] = 0.;
    m_ShvrIn.fNearClip = 0.5;
    m_ShvrIn.fFarrClip = 20.5;
}

void Chvr_work::CopyHvrInputs()
{
    /* Copy from GUI input struct to hvr use struct
       Set all iNew flags to the status of NOT new.
      BEGIN NEED THREAD SAFE!! */
    memcpy(&m_ShvrIn.iNewHV,  &global_ShvrIn.iNewHV, sizeof(HvrInput));
    memcpy(&m_ShvrUse.iNewHV, &global_ShvrIn.iNewHV, sizeof(HvrInput));
    opacity   = global_ShvrIn.opacity;
    tolerance = global_ShvrIn.tolerance;        /* max color level error allowed with no scan down  */
    angtol    = global_ShvrIn.angtol;            /* max angle allowed with no scan down              */
    viewres   = global_ShvrIn.viewres;            /* dx_brick > farclip * vrtolerance => no scan down */
    nplanes   = global_ShvrIn.nplanes;
    maxplanes = global_ShvrIn.maxplanes;

    /*    memcpy(&m_ShvrUse.iNewHV, &m_ShvrIn.iNewHV, sizeof(HvrInput)); */
    m_ShvrIn.iNewHV      = 0;
    m_ShvrIn.iNewLUT     = 0;
    m_ShvrIn.iNewView    = 0;
    m_ShvrIn.iNewFrustum = 0;
    // END NEED THREAD SAFE!!
}

STDMETHODIMP Chvr_work::IncrementEyePoint(int delx, int dely, int iMouseMode)
{
    void crossprod(float vec1[3], float vec2[3], float vec3[3]);
    void vnormalize( float v[3] );
    float dotproduct( const float*, const float* );
    float rotfacx, rotfacy;
    float fEC[3], fHoriz[3], fVert[3], fDelUp, fEN[3], fNewEN[3], fNormEN2, fNormNewEN2;
    float fNewEC[3], fNormEC2, fNormEC, fNormNewEC2;
    float fReNormFac, fNewDelUp, fDelCent, ffac;
    int i;

    for(i=0; i<3; i++) { fEC[i] = m_ShvrIn.fEye[i] - m_ShvrIn.fCent[i]; }
    vnormalize(m_ShvrIn.fUp);
    crossprod(fEC, m_ShvrIn.fUp, fHoriz);

    if(iMouseMode == 0  ||  iMouseMode == 3) {           /*  Polar or Track Ball Mode */
        rotfacx = (float)0.0100 * (float)delx;
        rotfacy = (float)0.0050 * (float)dely;
        vnormalize(fHoriz);
        crossprod(fHoriz, fEC, fVert);
        vnormalize(fVert);
        fNormEC2 = dotproduct(fEC, fEC);
        fNormEC  = (float)sqrt(fNormEC2);
        for(i=0; i<3; i++) { fNewEC[i] = fEC[i] + fNormEC * (rotfacx * fHoriz[i] + rotfacy * fVert[i]); }
        fNormNewEC2 = dotproduct(fNewEC, fNewEC);
        fReNormFac = (float)sqrt(fNormEC2 / fNormNewEC2);
        for(i=0; i<3; i++) { fNewEC[i] *= fReNormFac; }    /* Want norms of NewEC and old EC to be equal */
        for(i=0; i<3; i++) { m_ShvrIn.fEye[i] = m_ShvrIn.fCent[i] + fNewEC[i]; }

        if(iMouseMode == 3) {                             /* Track Ball Mode */
            crossprod(fNewEC, m_ShvrIn.fUp, fHoriz);
            crossprod(fHoriz, fNewEC, m_ShvrIn.fUp);
            vnormalize(m_ShvrIn.fUp);
        }
    }

    if(iMouseMode == 1) {                                 /* Cylinder Mode */
        rotfacx = (float)0.0100 * (float)delx;
        rotfacy = (float)0.0010 * (float)dely;
        fDelUp = dotproduct(m_ShvrIn.fUp, fEC);
        for(i=0; i<3; i++) { fEN[i]    = fEC[i] - fDelUp * m_ShvrIn.fUp[i]; }
        for(i=0; i<3; i++) { fNewEN[i] = fEN[i] + rotfacx * fHoriz[i]; }
        fNormEN2    = dotproduct(fEN, fEN);
        fNormNewEN2 = dotproduct(fNewEN, fNewEN);
        fReNormFac = (float)sqrt(fNormEN2 / fNormNewEN2);
        for(i=0; i<3; i++) { fNewEN[i] *= fReNormFac; }    /* Want norms of NewEN and old EN to be equal */
        fDelCent  = rotfacy * (float)sqrt(fNormEN2);
        fNewDelUp = fDelUp + fDelCent;
        for(i=0; i<3; i++) {
            m_ShvrIn.fEye[i]  = m_ShvrIn.fCent[i] + fNewEN[i] + fNewDelUp * m_ShvrIn.fUp[i];
            m_ShvrIn.fCent[i] = m_ShvrIn.fCent[i]             + fDelCent  * m_ShvrIn.fUp[i];
        }
    }

    if(iMouseMode == 2) {                                   /* Pan Mode */
        crossprod(fHoriz, fEC, fVert);
        fNormEC = (float)sqrt(dotproduct(fEC, fEC));
        vnormalize(fHoriz);
        vnormalize(fVert);

        rotfacx = (float)0.00065 * fNormEC * (float)delx;
        rotfacy = (float)0.00065 * fNormEC * (float)dely;
        ffac = (float)(1. / sqrt(dotproduct(fHoriz, fHoriz)));
        for(i=0; i<3; i++) {
            m_ShvrIn.fEye[i]  += rotfacx * fHoriz[i] + rotfacy * fVert[i];
            m_ShvrIn.fCent[i] += rotfacx * fHoriz[i] + rotfacy * fVert[i];
        }
    }


    return S_OK;
}

STDMETHODIMP Chvr_work::SetRenderQuality(int iPixErr, float fAngleError, float fSizeError)
{
    tolerance = iPixErr;        /* max color level error allowed with no scan down  */
    angtol    = fAngleError;    /* max angle allowed with no scan down              */
    viewres   = fSizeError;        /* dx_brick > farclip * vrtolerance => no scan down */
    return S_OK;
}

STDMETHODIMP Chvr_work::StopRendering()
{
    if(iDoThis == 1) {
        iStopRendering = 1;
        iIO_Read = 2;
    }
    return S_OK;
}

STDMETHODIMP Chvr_work::SaveKey(struct Renderparams *SKey)
{
    int i, j;

    SKey->nx = Nx;
    SKey->ny = Ny;
    SKey->nz = Nz;
    SKey->opacity = opacity;
    sprintf((char *)SKey->cmapname,"%s", m_ShvrIn.sCmapname);
    sprintf((char *)SKey->amapname,"%s", m_ShvrIn.sAmapname);
#ifdef WINDOWS    
    *SKey->bobname = NULL;
#else
    sprintf((char *)SKey->bobname,"");
#endif
    sprintf((char *)SKey->hvname, "%s", m_ShvrIn.sHVname);
    SKey->isectx = 0;
    SKey->isecty = 0;
    SKey->nsectx = 1;
    SKey->nsecty = 1;
    SKey->width = m_ShvrIn.iWidth;
    SKey->height = m_ShvrIn.iHeight;
    SKey->hpos = 0;
    SKey->vpos = 0;
    SKey->moviemode = 0;
    SKey->movieframes = 1;
    sprintf((char *)SKey->imagename, "Test_Image.pix");
    sprintf((char *)SKey->moviename, "Test_Movie.pix");
    SKey->fovangle = m_ShvrIn.fFovAngle;
    SKey->eyeX = m_ShvrIn.fEye[0];
    SKey->eyeY = m_ShvrIn.fEye[1];
    SKey->eyeZ = m_ShvrIn.fEye[2];
    SKey->centerX = m_ShvrIn.fCent[0];
    SKey->centerY = m_ShvrIn.fCent[1];
    SKey->centerZ = m_ShvrIn.fCent[2];
    SKey->upX = m_ShvrIn.fUp[0];
    SKey->upY = m_ShvrIn.fUp[1];
    SKey->upZ = m_ShvrIn.fUp[2];
    SKey->eyedirX = m_ShvrIn.fCent[0] - m_ShvrIn.fEye[0];
    SKey->eyedirY = m_ShvrIn.fCent[1] - m_ShvrIn.fEye[1];
    SKey->eyedirZ = m_ShvrIn.fCent[2] - m_ShvrIn.fEye[2];
    for(i=0; i<4; i++) { SKey->viewdefined[i] = 1; }
    for(i=0; i<6; i++) { SKey->ifauxclip[i] = m_ShvrIn.ifauxclip[i]; }
    for(i=0; i<6; i++) { for(j=0; j<4; j++) { SKey->auxclip[i][j] = m_ShvrIn.auxclip[i][j]; }}
    SKey->my_near = m_ShvrIn.fNearClip;
    SKey->my_far  = m_ShvrIn.fFarrClip;
    SKey->softdistance = 0.;
    SKey->nplanes = nplanes;
    SKey->maxplanes = maxplanes;
    SKey->boxcolor = -1;
    SKey->saveslabs = 0;
    SKey->movieframe = 0;
    SKey->nfpk = 10;
    SKey->ntpf = 0;
    SKey->nrpk = 0.;
    SKey->icoltolerance = tolerance;
    SKey->angletolerance = angtol;
    SKey->vrtolerance = viewres;

    return S_OK;
}

STDMETHODIMP Chvr_work::RestorKey(struct Renderparams *SKey)
{
    int i, j;

    sprintf(m_ShvrIn.sCmapname, "%s", SKey->cmapname);
    sprintf(m_ShvrIn.sAmapname, "%s", SKey->amapname);
    sprintf(m_ShvrIn.sHVname, "%s", SKey->hvname);
    m_ShvrIn.fFovAngle = SKey->fovangle;
    m_ShvrIn.fEye[0] = SKey->eyeX;
    m_ShvrIn.fEye[1] = SKey->eyeY;
    m_ShvrIn.fEye[2] = SKey->eyeZ;
    m_ShvrIn.fCent[0] = SKey->centerX;
    m_ShvrIn.fCent[1] = SKey->centerY;
    m_ShvrIn.fCent[2] = SKey->centerZ;
    m_ShvrIn.fUp[0] = SKey->upX;
    m_ShvrIn.fUp[1] = SKey->upY;
    m_ShvrIn.fUp[2] = SKey->upZ;
    m_ShvrIn.fNearClip = SKey->my_near;
    m_ShvrIn.fFarrClip = SKey->my_far;
    for(i=0; i<6; i++) { m_ShvrIn.ifauxclip[i] = SKey->ifauxclip[i]; }
    for(i=0; i<6; i++) { for(j=0; j<4; j++) { m_ShvrIn.auxclip[i][j] = SKey->auxclip[i][j]; }}
    sprintf(m_cImageName, "%s", SKey->imagename);

    opacity = SKey->opacity;
    tolerance = SKey->icoltolerance;                /* max color level error allowed with no scan down  */
    angtol    = SKey->angletolerance;                /* max angle allowed with no scan down              */
    viewres   = SKey->vrtolerance;                    /* dx_brick > farclip * vrtolerance => no scan down */
    nplanes   = SKey->nplanes;
    maxplanes = SKey->maxplanes;

    m_ShvrIn.opacity = opacity;
    m_ShvrIn.tolerance = tolerance;
    m_ShvrIn.angtol = angtol;
    m_ShvrIn.viewres = viewres;
    m_ShvrIn.nplanes = nplanes;
    m_ShvrIn.maxplanes = maxplanes;

    return S_OK;
}

STDMETHODIMP Chvr_work::ZoomEyePoint(int delx, int dely, int iMouseMode)
{
    float fMult, fEC[3];
    int i;

    fMult = (float)exp(-0.002 * (double)dely);

    for(i=0; i<3; i++) { fEC[i] = m_ShvrIn.fEye[i] - m_ShvrIn.fCent[i]; }
    for(i=0; i<3; i++) { m_ShvrIn.fEye[i] = m_ShvrIn.fCent[i] + fMult * fEC[i]; }

    if(iMouseMode == 2) {   // Pan Mode
        for(i=0; i<3; i++) { m_ShvrIn.fCent[i] = m_ShvrIn.fEye[i] - fEC[i]; }
    }

    return S_OK;
}

STDMETHODIMP Chvr_work::SetLUT(unsigned char *sAmap, unsigned char *sCmap)
{
    setstr(m_ShvrIn.sCmapname, (char *)sCmap);
    setstr(m_ShvrIn.sAmapname, (char *)sAmap);
    return S_OK;
}

STDMETHODIMP Chvr_work::CopyPixToFile(unsigned char *cFile)
{
    FILE *fdout;
    int nbytes, npix, i0, i1;
    char string[100], cOutFile[100];

    //    glReadBuffer(GL_FRONT);
    if(m_iNRendBuffs == 2) { mySwapBuffers(); }

    nbytes = m_ShvrUse.iHeight * m_ShvrUse.iWidth * 3;

    glReadPixels( 0, 0, m_ShvrUse.iWidth, m_ShvrUse.iHeight,
          GL_RGB, GL_UNSIGNED_BYTE, rgbpannel[0]);
    if(m_iNRendBuffs == 2) { mySwapBuffers(); }

    setstr(string, (char *)cFile);

    i0 = (int)string[0] - (int)'0';
    i1 = (int)string[1] - (int)'0';
    if(0 <= i0 && i0 <= 9 && 0 <= i1  &&  i1 <= 9) {
        sprintf(cOutFile, "%s_%c%c", m_cImageName, string[0], string[1]);
    } else {
        setstr(cOutFile, (char *)cFile);
    }

    fdout = fopen(cOutFile, "wb");
    if ( fdout == NULL ) { return S_OK; }

    npix = fwrite(rgbpannel[0], sizeof(char), nbytes, fdout);
    // if ( npix != nbytes ) { fclose(fdout); return (S_FAIL); }
    fclose(fdout);

    /*fp = fopen("DebugCopyPix.txt", "w");
    fprintf(fp, "wrote %d bytes into file %s\n", nbytes, cOutFile);
    fprintf(fp, "Height = %d  Width = %d\n", m_ShvrUse.iHeight, m_ShvrUse.iWidth);
    if(npix != nbytes) { fprintf(fp, "write failed, actually wrote: %d bytes\n", npix); }
    fclose(fp);*/

    return S_OK;
}

STDMETHODIMP Chvr_work::SetCurrentWorkingDir(unsigned char *cDir)
{
    char string[100];
    //FILE *fp;

#ifdef linux
    sprintf(string, "cd %s", (char *)cDir);
    system(string);
#else
    setstr(string, (char *)cDir);
    SetCurrentDirectory( string );
#endif
    sprintf(cLastDir, "");
    //fp = fopen("ImHere", "w");
    //fprintf(fp, "directory = %s\n", string);
    //fclose(fp);

    return S_OK;
}

STDMETHODIMP Chvr_work::SetOpacity(float fOpacity)
{
    opacity = fOpacity;
    return S_OK;
}

STDMETHODIMP Chvr_work::SetNearFarClip(float fNear, float fFar)
{
    m_ShvrIn.fNearClip = fNear;
    m_ShvrIn.fFarrClip = fFar;

    return S_OK;
}

DWORD dhpReadFromOpenHandle(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes, int iClose)
{
#ifdef linux
    int fd;
    int iIsOpen = 0;
    unsigned int ngot, nread;

    if((iClose > 0)  &&  (iIsOpen > 0)) {
        close(fd);
        iIsOpen = 0;
    }
    if(iClose > 1) { return 0; }
    if(iIsOpen == 0) {
        fd = open(lpFileName, O_RDONLY);
        if (fd <= 0) { return 0; }
        iIsOpen = 1;
    }
    ngot = lseek(fd, offset, SEEK_SET);
    ngot = read(fd, lpBuffer, nbytes);
    return ngot;
#else
    HANDLE hFile = NULL;     /* handle to read file */
    int iIsOpen = 0;
    __int64 shift32, shift8;
    LONG dwHigh, dwLow;
    unsigned long ngot, nread;
    OVERLAPPED    overlapped;
    BOOL fs;

    if((iClose > 0)  &&  (iIsOpen > 0)) {
        CloseHandle(hFile);
        iIsOpen = 0;
    }

    if(iClose > 1) { return 0; }

    if(iIsOpen == 0) {
        hFile = CreateFile(    lpFileName,
                            GENERIC_READ, 
                            FILE_SHARE_READ,
                            NULL, 
                            OPEN_EXISTING, 
                            FILE_FLAG_NO_BUFFERING | FILE_FLAG_OVERLAPPED, 
                            NULL                                            );

        if (hFile == INVALID_HANDLE_VALUE) { return 0; }

        iIsOpen = 1;
    }

    shift8  = (__int64)256;
    shift32 = shift8 * shift8 * shift8 * shift8;
    dwHigh  = (LONG)(offset / shift32);
    dwLow   = (LONG)(offset % shift32);
    overlapped.Offset = dwLow;
    overlapped.OffsetHigh = dwHigh;
    overlapped.hEvent = NULL;

    fs = ReadFile(hFile, lpBuffer, nbytes, &nread, &overlapped);
    GetOverlappedResult(hFile, &overlapped, &ngot, TRUE);

    return ngot;
#endif
}


#define MAXOPENFILES 8192
#ifdef WINDOWS
DWORD dhpReadFromOpenFile(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes,
                          int itcd, int iClose)
{
    HANDLE hFiles[MAXOPENFILES];
    int iFirst = 1, iIsOpen[MAXOPENFILES];
    int i;
    HANDLE hFile;       /* handle to read file */
    __int64 shift32, shift8;
    LONG dwHigh, dwLow;
    unsigned long ngot;
    //    DWORD lasterror;
    //    LPSTR lpsErrorMsg;

    shift8  = (__int64)256;
    shift32 = shift8 * shift8 * shift8 * shift8;
    dwHigh  = (LONG)(offset / shift32);
    dwLow   = (LONG)(offset % shift32);

    if(iFirst == 1) {
        for(i=0; i<MAXOPENFILES; i++) { iIsOpen[i] = 0; }
        iFirst = 0;
    }

    if(iClose == 2) {
        for(i=0; i<MAXOPENFILES; i++) {
            if(iIsOpen[i] != 0) {
                hFile = hFiles[i];
                CloseHandle(hFile);
                iIsOpen[i] = 0;
            }
        }
        return 0;
    }

    if(iIsOpen[itcd] == 0) {
        hFile = CreateFile(lpFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
                            FILE_ATTRIBUTE_NORMAL, NULL);

        if(hFile == INVALID_HANDLE_VALUE) {
            //lasterror = GetLastError();
            //printf("Got an invalid file handle for file %s\n", lpFileName);
            //FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM |
            //        FORMAT_MESSAGE_IGNORE_INSERTS, NULL, lasterror, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
            //        lpsErrorMsg, 0, NULL);
            //printf("Tried to open file    -->%s<--\n", lpFileName);
            //printf("OK, lets try that again\n");
            //printf("Error on CreateFile was: %ld\n", lasterror); //, lpsErrorMsg);
            //exit(0);
        }

        hFiles[itcd] = hFile;
        iIsOpen[itcd] = 1;
        //printf("----Openned file: itcd = %d\n", itcd);
    }
    hFile = hFiles[itcd];
    
    SetFilePointer(hFile, dwLow, &dwHigh, FILE_BEGIN);
    ReadFile(hFile, lpBuffer, nbytes, &ngot, NULL);

    if(ngot != nbytes) {
        //lasterror = GetLastError();
        //printf("----Only read %ld out of %ld bytes\n", *ngot, nbytes, lpFileName);
        //printf("----LastError was: %ld\n", lasterror);//, lpsErrorMsg);
        //printf("----Offset = %ld\n", offset);
    }
    
    if(iClose == 1) {
        CloseHandle(hFile);
        iIsOpen[itcd] = 0;
    }

    return ngot;
}
#endif

DWORD dhpReadFromFileSync(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes)
{
#ifdef linux
    DWORD ngot;
    int fd;
    fd = open(lpFileName, O_RDONLY);
    ngot = lseek(fd, offset, SEEK_SET);
    ngot = read(fd, lpBuffer, nbytes);
    close(fd);
#else
    HANDLE hFile;                        /* handle to read file */
    __int64 shift32, shift8;
    LONG dwHigh, dwLow;
    unsigned long ngot;

    shift8  = (__int64)256;
    shift32 = shift8 * shift8 * shift8 * shift8;
    dwHigh  = (LONG)(offset / shift32);
    dwLow   = (LONG)(offset % shift32);
    hFile = CreateFile(lpFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
        FILE_ATTRIBUTE_NORMAL, NULL);
    SetFilePointer(hFile, dwLow, &dwHigh, FILE_BEGIN);
    ReadFile(hFile, lpBuffer, nbytes, &ngot, NULL);
    CloseHandle(hFile);
#endif
    return ngot;
}

DWORD dhpReadFromFileNoBuf(LPVOID lpBuffer, LPCTSTR lpFileName, __int64 offset, DWORD nbytes)
{
#ifdef linux
    /* IF there is a linux high performance read api, THEN it should go here */
    int fd;
    char cLastFile[256] = "nofileopen";
    DWORD ngot;

    /* Protect against odd sized reads */
    if((int)nbytes%(64*1024) != 0) {
        ngot = dhpReadFromFileSync(lpBuffer, lpFileName, offset, nbytes);
        return ngot;
    }
    if(strdif((char *)lpFileName, cLastFile) != 0) {
        if(strdif((char *)lpFileName, "none") == 0) { return 0; }
        if(strdif(cLastFile, "nofileopen") != 0) {
            close(fd);
            sprintf(cLastFile, "nofileopen");
        }

        if(strdif((char *)lpFileName, "close") == 0) { return 0; }

        fd = open(lpFileName, O_RDONLY);
        sprintf(cLastFile, "%s", lpFileName);

        if (fd < 0) {
            sprintf(cLastFile, "nofileopen");
            return 0;
        }
    }
    ngot = lseek(fd, offset, SEEK_SET);
    ngot = read(fd, lpBuffer, nbytes);
    return ngot;
#else
    HANDLE hFile = NULL;                     /* handle to read file */
    char cLastFile[256] = "nofileopen";
    __int64 shift32, shift8;
    LONG dwHigh, dwLow;
    unsigned long ngot, nread;
    OVERLAPPED    overlapped;
    BOOL fs;

    // Protect against odd sized reads
    if((int)nbytes%(64*1024) != 0) {
        ngot = dhpReadFromFileSync(lpBuffer, lpFileName, offset, nbytes);
        return ngot;
    }

    if(strdif((char *)lpFileName, cLastFile) != 0) {
        if(strdif((char *)lpFileName, "none") == 0) { return 0; }

        if(strdif(cLastFile, "nofileopen") != 0) {
            CloseHandle(hFile);
            sprintf(cLastFile, "nofileopen");
        }

        if(strdif((char *)lpFileName, "close") == 0) { return 0; }

        hFile = CreateFile(    lpFileName,
                            GENERIC_READ, 
                            FILE_SHARE_READ,
                            NULL, 
                            OPEN_EXISTING, 
                            FILE_FLAG_NO_BUFFERING | FILE_FLAG_OVERLAPPED, 
                            NULL                                            );

        sprintf(cLastFile, "%s", lpFileName);

        if (hFile == INVALID_HANDLE_VALUE) {
            sprintf(cLastFile, "nofileopen");
            return 0;
        }
    }

    shift8  = (__int64)256;
    shift32 = shift8 * shift8 * shift8 * shift8;
    dwHigh  = (LONG)(offset / shift32);
    dwLow   = (LONG)(offset % shift32);
    overlapped.Offset = dwLow;
    overlapped.OffsetHigh = dwHigh;
    overlapped.hEvent = NULL;

    fs = ReadFile(hFile, lpBuffer, nbytes, &nread, &overlapped);
    GetOverlappedResult(hFile, &overlapped, &ngot, TRUE);

    return ngot;
#endif
}

#ifdef WINDOWS
int TwoBrickIO(unsigned char* bricks[2], int iNBytes, char* filename)
{
    int                fs;
    int                        iCurrentBuffer;
    unsigned long    bytesread, bytestransferred;
    int                bufno = -1;
    HANDLE            fd = NULL; 
    OVERLAPPED        overlapped;

    iCurrentBuffer = bufno;

    if (bufno >= 0)   /* Finish Previous Read & Close Handle to File */
    {
      GetOverlappedResult(fd, &overlapped, &bytestransferred, TRUE);
      CloseHandle(fd);
    }

    if (iNBytes > 0)  /* Open Handle to File and Initiate Read */
    {
        fd = CreateFile(filename,
                GENERIC_READ, 
                FILE_SHARE_READ,
                0, 
                OPEN_EXISTING, 
                FILE_FLAG_NO_BUFFERING | FILE_FLAG_OVERLAPPED, 
                0);

        if (fd == INVALID_HANDLE_VALUE)
        {
            fprintf(error_log,"Could not open a handle to file: %s\n", filename);
            exit(E_FAIL);
        }
        bufno = (bufno + 1)%2;
        overlapped.Offset = 0;
        overlapped.OffsetHigh = 0;
        overlapped.hEvent = NULL;
        fs = ReadFile(fd, bricks[bufno], iNBytes, &bytesread, &overlapped);
    }

    return(iCurrentBuffer); 
}
#endif

STDMETHODIMP Chvr_work::Set_hWnd(int i_hWnd)
{

    m_hLocalWnd = (HWND)i_hWnd;

    return S_OK;
}

STDMETHODIMP Chvr_work::SetLUTirgb(int N, int *ind, float *red, float *grn, float *blu)
{
    int i0, i1, i, j;
    float y0, yp;

    m_iNrgbNots = N;
    for(j=0; j<N; j++) {
        m_iRGBind[j] = ind[j];
        m_fRED[j]    = red[j];
        m_fGRN[j]    = grn[j];
        m_fBLU[j]    = blu[j];
    }

    for(j=1; j<N; j++) {
        i0 = ind[j-1];
        i1 = ind[j];

        yp = (red[j] - red[j-1])/(float)(i1-i0);
        y0 =  red[j-1];
        for(i=i0; i<i1; i++) { frgba[i][0] = y0 + yp * (float)(i-i0); }

        yp = (grn[j] - grn[j-1])/(float)(i1-i0);
        y0 =  grn[j-1];
        for(i=i0; i<i1; i++) { frgba[i][1] = y0 + yp * (float)(i-i0); }

        yp = (blu[j] - blu[j-1])/(float)(i1-i0);
        y0 =  blu[j-1];
        for(i=i0; i<i1; i++) { frgba[i][2] = y0 + yp * (float)(i-i0); }
    }
          i0     = ind[N-1];
    frgba[i0][0] = red[N-1];
    frgba[i0][1] = grn[N-1];
    frgba[i0][2] = blu[N-1];

    if(m_iAlphaIsSet == 0) {
        for(i=0; i <= i0; i++) {
            frgba[i][3] = (float)0.1 * frgba[i][0] + (float)0.6 * frgba[i][1] + (float)0.3 * frgba[i][2];
        }

        m_iNalpNots = m_iNrgbNots;
        for(i=0; i < m_iNrgbNots; i++) {
            m_iALPind[j] = m_iRGBind[j];
            m_fALP[j]    = (float)0.1 * m_fRED[j] +(float)0.6 * m_fGRN[j] +(float)0.3 * m_fBLU[j];
        }
    }
      
    /* Cg */
    if(!m_bColorTableIsAvailable)
    {
        for(i=0; i<256; i++) 
        {
            m_lrgba[i] =              (unsigned long)(255.99*frgba[i][0]) +
                (unsigned long)256 * ((unsigned long)(255.99*frgba[i][1]) +
                (unsigned long)256 * ((unsigned long)(255.99*frgba[i][2]) +
                (unsigned long)256 *  (unsigned long)(255.99*frgba[i][3])));
        }
    }

    m_iAlphaIsSet = 0;

    return S_OK;
}

STDMETHODIMP Chvr_work::SetLUTia(int N, int *ind, float *alpha)
{
    int i0, i1, i, j;
    float y0, yp;
    
    m_iNalpNots = N;
    for(j=0; j<N; j++) {
        m_iALPind[j] = ind[j];
        m_fALP[j]    = alpha[j];
    }

    for(j=1; j<N; j++) {
        i0 = ind[j-1];
        i1 = ind[j];
        yp = (alpha[j] - alpha[j-1])/(float)(i1-i0);
        y0 =  alpha[j-1];
        for(i=i0; i<i1; i++) { frgba[i][3] = y0 + yp * (float)(i-i0); }
    }
          i0     =   ind[N-1];
    frgba[i0][3] = alpha[N-1];

    m_iAlphaIsSet = 1;

    return S_OK;
}

STDMETHODIMP Chvr_work::GetLUT32irgb(int *N, int *ind, float *red, float *grn, float *blu)
{
    int j;

    *N = m_iNrgbNots;
    for(j=0; j<m_iNrgbNots; j++) {
        ind[j] = m_iRGBind[j];
        red[j] = m_fRED[j];
        grn[j] = m_fGRN[j];
        blu[j] = m_fBLU[j];
    }

    return S_OK;
}

STDMETHODIMP Chvr_work::GetLUT32ia(int *N, int *ind, float *alp)
{
    int j;

    *N = m_iNalpNots;
    for(j=0; j<m_iNalpNots; j++) {
        ind[j] = m_iALPind[j];
        alp[j] = m_fALP[j];
    }

    return S_OK;
}


STDMETHODIMP Chvr_work::SetOption(wchar_t *wcInKey, wchar_t *wcInValue)
{
    /* All of the gl stuff needs to be done by the rendering thread.
       Hence: 1) copy input strings to variables
              2) trigger the rendering thread to run DoSetOption
              3) wait for DoSetOption to finish */
    wcscpy(wcKey, wcInKey);
    wcscpy(wcValue, wcInValue);
    if(iSpawnRenderThread == 0  ||  iGLwindowIsMade == 0) {
        DoSetOption();
    } else {
        iDoThis = 3;
        //if(iUseOpenGL == 1) { glutPostRedisplay(); }
        while(iDoThis == 3) { _sleep(10); }
    }

    return S_OK;
}

#ifdef WINDOWS
void Chvr_work::DoSetOption()
{
    /* Offscreen Rendering */
    if(wcscmp(wcKey, L"Offscreen") == 0) {
        if(wcscmp(wcValue, L"TRUE") == 0)
            Renderbuffer = TRUE;
        else
            Renderbuffer = FALSE;
        SaveOption();
    }
    
    /* Resident Textures */
    if(wcscmp(wcKey, L"UseResTex") == 0) {
        if(wcscmp(wcValue, L"TRUE") == 0) {
            ResidentTextures = TRUE;
            #ifdef DEBUG_TEX
            printf("Texture Caching: ON\n");
            #endif
        }
        else {
            ResidentTextures = FALSE;
            #ifdef DEBUG_TEX
            printf("Texture Caching: OFF\n");
            #endif
        }
        SaveOption();
    }

#ifdef AUTO_CACHE
    /* Automatic Caching */
    if(wcscmp(wcKey, L"AutoCache") == 0) {
        if(wcscmp(wcValue, L"TRUE") == 0) {
            AutoCaching = TRUE;
            ResidentTextures = FALSE;
            #ifdef DEBUG_TEX
            printf("Automatic Texture Caching: ON\n");
            printf("Texture Caching OFF\n");
            #endif
        }
        else {
            AutoCaching = FALSE;
            #ifdef DEBUG_TEX
            printf("Automatic Texture Caching: OFF\n");
            #endif
        }
        SaveOption();
    }
#endif

    /* Paletted Textures */
    if(wcscmp(wcKey, L"UsePalet") == 0) {
        if(wcscmp(wcValue, L"TRUE") == 0) {
            m_bColorTableIsAvailable = TRUE;
            #ifdef DEBUG_TEX
            printf("Texture Format: Paletted\n");
            #endif
        }
        else {
            m_bColorTableIsAvailable = FALSE;
            CgInit = FALSE;
            #ifdef DEBUG_TEX
            printf("Texture Format: RGBA\n");
            #endif
        }
        SaveOption();
    }

    /* 3d Textures */
    if(wcscmp(wcKey, L"3dtextures") == 0) {
        if(wcscmp(wcValue, L"TRUE") == 0) {
            Tex3d = TRUE;
            CgInit = FALSE;
            #ifdef DEBUG_TEX
            printf("Texture Type: 3d\n");
            #endif
        }
        else {
            Tex3d = FALSE;
            CgInit = FALSE;
            #ifdef DEBUG_TEX
            printf("Texture Type: 2d\n");
            #endif
        }
        SaveOption();
    }

    if(wcscmp(wcKey, L"ScreenRatio") == 0) { swscanf(wcValue, L"%f", &m_fScreenRatio); }

    if(wcscmp(wcKey, L"TimersOn") == 0) { swscanf(wcValue, L"%d", &m_iTimersOn); }

    if(wcscmp(wcKey, L"TimerOutputFile") == 0) { swscanf(wcValue, L"%hs", m_sTimerFile); }

    if(wcscmp(wcKey, L"BufferedIO") == 0) {
        swscanf(wcValue, L"%d", &m_iBufferedIO);
        SaveOption();
    }

    if(wcscmp(wcKey, L"NRenderingBuffers") == 0) {
        swscanf(wcValue, L"%d", &m_iNRendBuffs);
        if(m_iNRendBuffs == 2) { 
            glDrawBuffer(GL_BACK);
            glReadBuffer(GL_BACK);
        }
        else { 
            glDrawBuffer(GL_FRONT);
            glReadBuffer(GL_FRONT);
        }
    }

    if(wcscmp(wcKey, L"ShowColorBar") == 0) {
        swscanf(wcValue, L"%d", &m_iShowColorBar);
        SaveOption();
    }

    if(wcscmp(wcKey, L"SelectKnots") == 0) {
        swscanf(wcValue, L"%d %d", &m_iSelALPknot, &m_iSelRGBknot);
        SaveOption();
    }

    if(wcscmp(wcKey, L"ClearTicks" ) == 0) {
        m_iNTicks  = 0;
        SaveOption();
        ClearOption(L"AddTick");
    }

    if(wcscmp(wcKey, L"ClearXTicks") == 0) {
        m_iXNTicks = 0;
        SaveOption();
        ClearOption(L"AddXTick");
    }

    if(wcscmp(wcKey, L"ClearYTicks") == 0) {
        m_iYNTicks = 0;
        SaveOption();
        ClearOption(L"AddYTick");
    }

    if(wcscmp(wcKey, L"AddTick") == 0) {
        swscanf(wcValue, L"%3d %hs", &m_iTicks[m_iNTicks], m_sTicks[m_iNTicks]);
        m_iNTicks++;
        AddOption();
    }

    if(wcscmp(wcKey, L"AddXTick") == 0) {
        swscanf(wcValue, L"%3d %hs", &m_iXTicks[m_iXNTicks], m_sXTicks[m_iXNTicks]);
        m_iXNTicks++;
        AddOption();
    }

    if(wcscmp(wcKey, L"AddYTick") == 0) {
        swscanf(wcValue, L"%3d %hs", &m_iYTicks[m_iYNTicks], m_sYTicks[m_iYNTicks]);
        m_iYNTicks++;
        AddOption();
    }

    if(wcscmp(wcKey, L"BlockOnRender") == 0) { swscanf(wcValue, L"%d", &m_iBlockOnRender); }

    if(wcscmp(wcKey, L"Stats") == 0) { swscanf(wcValue, L"%hs", m_sStats); }

    if(wcscmp(wcKey, L"Yfile") == 0) { swscanf(wcValue, L"%hs", m_sTileO); }

    if(wcscmp(wcKey, L"NCacheMB") == 0) {
        //int iMB_OLD = iNCacheMB;
        swscanf(wcValue, L"%d", &iNCacheMB);
        //if(iIO_Thread_Is_Up == 1) {
        //    SIZE_T  dwSizeOLD = (SIZE_T)iMB_OLD   * (SIZE_T)1024 * (SIZE_T)1024;
        //    SIZE_T  dwSizeNEW = (SIZE_T)iNCacheMB * (SIZE_T)1024 * (SIZE_T)1024;
        //    VirtualFree(BobCache, dwSizeOLD, MEM_DECOMMIT | MEM_RELEASE);
        //    BobCache = (unsigned char *)VirtualAlloc(NULL, dwSizeNEW,
        //                                             MEM_COMMIT, PAGE_READWRITE);
        //    FlushCacheBuffers();
        //}
        //SaveOption();
    }

    // INPUTS for bounding box and floor
    /*
       float fTickSpacing = (float)0.2;
       float fFloorSize = (float)5.0;
       int iFloorRed = 255;
       int iFloorGrn = 255;
       int iFloorBlu = 255;
       int iFloorAlp =   0;
       int iBoxRed =   0;
       int iBoxGrn = 128;
       int iBoxBlu =   0;
       int iBoxAlp = 200;
    */
    if(wcscmp(wcKey, L"Floor") == 0) {
        swscanf(wcValue, L"%hd %hd %hd %hd %f %f",
            &iFloorRed, &iFloorGrn, &iFloorBlu, &iFloorAlp, &fFloorSize, &fTickSpacing);
        SaveOption();
    }
    if(wcscmp(wcKey, L"BoundingBox") == 0) {
        swscanf(wcValue, L"%hd %hd %hd %hd", &iBoxRed, &iBoxGrn, &iBoxBlu, &iBoxAlp);
        SaveOption();
    }

    if(wcscmp(wcKey, L"AutoSwapBuffers") == 0) {
        swscanf(wcValue, L"%hd", &m_iAutoSwapBuffers);
    }

    if(wcscmp(wcKey, L"SwapBuffers") == 0) {
        mySwapBuffers();
        glFlush();
        glFinish();
    }

    if(wcscmp(wcKey, L"WriteImageToSTDout") == 0) {
        swscanf(wcValue, L"%hd", &iWriteImageToSTDout);
    }

    if(wcscmp(wcKey, L"PriorityRequest") == 0) { swscanf(wcValue, L"%hd", &iPriorityRequest); }

    if(wcscmp(wcKey, L"SetStatusRequest") == 0) {
        //swscanf(wcValue, L"%hs", m_cStatusRequest);

        sprintf(m_cStatusRequest, "%ls", wcValue);
    }

    if(wcscmp(wcKey, L"ScaleFactors") == 0) {
        swscanf(wcValue, L"%f %f %f", &m_fScaleFactors[0],
                                      &m_fScaleFactors[1],
                                      &m_fScaleFactors[2]);
        sprintf(m_cOldHVname, "none");
        SaveOption();
    }

    if(wcscmp(wcKey, L"ResetALLOptions") == 0) {
        initialize_vars();
        iOpNumber = 0;
    }
}

STDMETHODIMP Chvr_work::GetStatus(wchar_t **wcStatus)
{
    int strdif(char*, char*);

    /* this is the default : return status of RENDERING, always goes back to this */
    if(iDoThis == 0) {
        swprintf(wcLocalStatus, L"IDLE"); 
        if(wcErrorStatus[0] == L'E') {
            swprintf(wcLocalStatus, wcErrorStatus);
            swprintf(wcErrorStatus, L"OK");
        }
    } else {
        swprintf(wcLocalStatus, L"WORKING");
    }

    if(strdif(m_cStatusRequest, "PRIORITYREQUEST") == 0) {
        swprintf(wcLocalStatus, L"%hd", iPriorityRequest);
        sprintf(m_cStatusRequest, "RENDERING");
    }
    if(strdif(m_cStatusRequest, "IS_GL_WINDOW_UP") == 0) {
        if(iGLwindowIsMade == 1) { swprintf(wcLocalStatus, L"GL_WINDOW_IS_UP");   }
        else                     { swprintf(wcLocalStatus, L"GL_WINDOW_IS_DOWN"); }
        sprintf(m_cStatusRequest, "RENDERING");
    }
    if(strdif(m_cStatusRequest, "GET_ALL_PREVIOUS_OPTIONS") == 0) {
        int i, iNChar=0;
        for(i=0; i<iOpNumber; i++) {
            swprintf(&wcLocalStatus[iNChar], L"%s %s\n", wcOpKeys[i], wcOpVals[i]);
            iNChar += wcslen(wcOpKeys[i]) + 2 + wcslen(wcOpVals[i]);
        }
        sprintf(m_cStatusRequest, "RENDERING");
    }

    long lAllocSize = (wcslen(wcLocalStatus)+1) * sizeof(wchar_t);
    *wcStatus = static_cast<wchar_t *>(CoTaskMemAlloc(lAllocSize));
    wcscpy(*wcStatus, wcLocalStatus);

    return S_OK;
}

#else
void Chvr_work::DoSetOption()
{
    /* Offscreen Rendering */
    if(wcscmp(wcKey, "Offscreen") == 0) {
        if(wcscmp(wcValue, "TRUE") == 0)
            Renderbuffer = TRUE;
        else
            Renderbuffer = FALSE;
        SaveOption();
    }
    
    /* Resident Textures */
    if(wcscmp(wcKey, "UseResTex") == 0) {
        if(wcscmp(wcValue, "TRUE") == 0) {
            ResidentTextures = TRUE;
            #ifdef DEBUG_TEX
            printf("Texture Caching: ON\n");
            #endif
        }
        else {
            ResidentTextures = FALSE;
            #ifdef DEBUG_TEX
            printf("Texture Caching: OFF\n");
            #endif
        }
        SaveOption();
    }

#ifdef AUTO_CACHE
    /* Automatic Caching */
    if(wcscmp(wcKey, "AutoCache") == 0) {
        if(wcscmp(wcValue, "TRUE") == 0) {
            AutoCaching = TRUE;
            ResidentTextures = FALSE;
            #ifdef DEBUG_TEX
            printf("Automatic Texture Caching: ON\n");
            #endif
        }
        else {
            AutoCaching = FALSE;
            #ifdef DEBUG_TEX
            printf("Automatic Texture Caching: OFF\n");
            #endif
        }
        SaveOption();
    }
#endif

    /* Paletted Textures */
    if(wcscmp(wcKey, "UsePalet") == 0) {
        if(wcscmp(wcValue, "TRUE") == 0) {
            m_bColorTableIsAvailable = TRUE;
            #ifdef DEBUG_TEX
            printf("Texture Format: Paletted\n");
            #endif
        }
        else {
            m_bColorTableIsAvailable = FALSE;
            CgInit = FALSE;
            #ifdef DEBUG_TEX
            printf("Texture Format: RGBA\n");
            #endif
        }
        SaveOption();
    }

    /* 3d Textures */
    if(wcscmp(wcKey, "3dtextures") == 0) {
        if(wcscmp(wcValue, "TRUE") == 0) {
            Tex3d = TRUE;
            CgInit = FALSE;
            #ifdef DEBUG_TEX
            printf("Texture Type: 3d\n");
            #endif
        }
        else {
            Tex3d = FALSE;
            CgInit = FALSE;
            #ifdef DEBUG_TEX
            printf("Texture Type: 2d\n");
            #endif
        }
        SaveOption();
    }

    if(wcscmp(wcKey, "vecfile") == 0) { swscanf(wcValue, "%s", cVecFile); }
    if(wcscmp(wcKey, "vecradius") == 0) { swscanf(wcValue, "%f", &vecradius); }
    if(wcscmp(wcKey, "vecloclr") == 0) { swscanf(wcValue, "%s", cVecLoClr); }
    if(wcscmp(wcKey, "vechiclr") == 0) { swscanf(wcValue, "%s", cVecHiClr); }

    if(wcscmp(wcKey, "ScreenRatio") == 0) { swscanf(wcValue, "%f", &m_fScreenRatio); }

    if(wcscmp(wcKey, "TimersOn") == 0) { swscanf(wcValue, "%d", &m_iTimersOn); }

    if(wcscmp(wcKey, "TimerOutputFile") == 0) { swscanf(wcValue, "%hs", m_sTimerFile); }

    if(wcscmp(wcKey, "BufferedIO") == 0) {
        swscanf(wcValue, "%d", &m_iBufferedIO);
        SaveOption();
    }

    if(wcscmp(wcKey, "NRenderingBuffers") == 0) {
        swscanf(wcValue, "%d", &m_iNRendBuffs);
        if(m_iNRendBuffs == 2) { 
            glDrawBuffer(GL_BACK);
            glReadBuffer(GL_BACK);
            #ifdef DEBUG_TEX
            printf("Rendering to a doublebuffered window\n");
            #endif
        }
        else { 
            glDrawBuffer(GL_FRONT);
            glReadBuffer(GL_FRONT);
            #ifdef DEBUG_TEX
            printf("Rendering to a singlebuffered window\n");
            #endif
        }
    }

    if(wcscmp(wcKey, "ShowColorBar") == 0) {
        swscanf(wcValue, "%d", &m_iShowColorBar);
        SaveOption();
    }

    if(wcscmp(wcKey, "SelectKnots") == 0) {
        swscanf(wcValue, "%d %d", &m_iSelALPknot, &m_iSelRGBknot);
        SaveOption();
    }

    if(wcscmp(wcKey, "ClearTicks" ) == 0) {
        m_iNTicks  = 0;
        SaveOption();
        ClearOption("AddTick");
    }

    if(wcscmp(wcKey, "ClearXTicks") == 0) {
        m_iXNTicks = 0;
        SaveOption();
        ClearOption("AddXTick");
    }

    if(wcscmp(wcKey, "ClearYTicks") == 0) {
        m_iYNTicks = 0;
        SaveOption();
        ClearOption("AddYTick");
    }

    if(wcscmp(wcKey, "AddTick") == 0) {
        swscanf(wcValue, "%3d %hs", &m_iTicks[m_iNTicks], m_sTicks[m_iNTicks]);
        m_iNTicks++;
        AddOption();
    }

    if(wcscmp(wcKey, "AddXTick") == 0) {
        swscanf(wcValue, "%3d %hs", &m_iXTicks[m_iXNTicks], m_sXTicks[m_iXNTicks]);
        m_iXNTicks++;
        AddOption();
    }

    if(wcscmp(wcKey, "AddYTick") == 0) {
        swscanf(wcValue, "%3d %hs", &m_iYTicks[m_iYNTicks], m_sYTicks[m_iYNTicks]);
        m_iYNTicks++;
        AddOption();
    }

    if(wcscmp(wcKey, "BlockOnRender") == 0) { swscanf(wcValue, "%d", &m_iBlockOnRender); }

    if(wcscmp(wcKey, "Stats") == 0) { swscanf(wcValue, "%hs", m_sStats); }

    if(wcscmp(wcKey, "Yfile") == 0) { swscanf(wcValue, "%hs", m_sTileO); }

    if(wcscmp(wcKey, "NCacheMB") == 0) {
//        int iMB_OLD = iNCacheMB;
        swscanf(wcValue, "%d", &iNCacheMB);
//        if(iIO_Thread_Is_Up == 1) {
//            SIZE_T  dwSizeOLD = (SIZE_T)iMB_OLD   * (SIZE_T)1024 * (SIZE_T)1024;
//            SIZE_T  dwSizeNEW = (SIZE_T)iNCacheMB * (SIZE_T)1024 * (SIZE_T)1024;
//            VirtualFree(BobCache, dwSizeOLD, MEM_DECOMMIT | MEM_RELEASE);
//            BobCache = (unsigned char *)VirtualAlloc(NULL, dwSizeNEW,
//                                                     MEM_COMMIT, PAGE_READWRITE);
//            FlushCacheBuffers();
//        }
//        SaveOption();
    }

// INPUTS for bounding box and floor
/*
   float fTickSpacing = (float)0.2;
   float fFloorSize = (float)5.0;
   int iFloorRed = 255;
   int iFloorGrn = 255;
   int iFloorBlu = 255;
   int iFloorAlp =   0;
   int iBoxRed =   0;
   int iBoxGrn = 128;
   int iBoxBlu =   0;
   int iBoxAlp = 200;
*/
    if(wcscmp(wcKey, "Floor") == 0) {
        swscanf(wcValue, "%hd %hd %hd %hd %f %f",
            &iFloorRed, &iFloorGrn, &iFloorBlu, &iFloorAlp, &fFloorSize, &fTickSpacing);
        SaveOption();
    }
    if(wcscmp(wcKey, "BoundingBox") == 0) {
        swscanf(wcValue, "%hd %hd %hd %hd", &iBoxRed, &iBoxGrn, &iBoxBlu, &iBoxAlp);
        SaveOption();
    }

    if(wcscmp(wcKey, "AutoSwapBuffers") == 0) {
        swscanf(wcValue, "%hd", &m_iAutoSwapBuffers);
    }

    if(wcscmp(wcKey, "SwapBuffers") == 0) {
            mySwapBuffers();
       /* glFlush();
        glFinish();*/
    }

    if(wcscmp(wcKey, "WriteImageToSTDout") == 0) {
        swscanf(wcValue, "%hd", &iWriteImageToSTDout);
    }

    if(wcscmp(wcKey, "PriorityRequest") == 0) { swscanf(wcValue, "%hd", &iPriorityRequest); }

    if(wcscmp(wcKey, "SetStatusRequest") == 0) {
//        swscanf(wcValue, "%hs", m_cStatusRequest);

        sprintf(m_cStatusRequest, "%ls", wcValue);
    }

    if(wcscmp(wcKey, "ScaleFactors") == 0) {
        swscanf(wcValue, "%f %f %f", &m_fScaleFactors[0],
                                      &m_fScaleFactors[1],
                                      &m_fScaleFactors[2]);
        sprintf(m_cOldHVname, "none");
        SaveOption();
    }

    if(wcscmp(wcKey, "ResetALLOptions") == 0) {
        initialize_vars();
        iOpNumber = 0;
    }
}

STDMETHODIMP Chvr_work::GetStatus(wchar_t **wcStatus)
{
    int strdif(char*, char*);

    // this is the default : return status of RENDERING, always goes back to this
    if(iDoThis == 0) {
        swprintf(wcLocalStatus, "IDLE"); 
        if(wcErrorStatus[0] == L'E') {
            swprintf(wcLocalStatus, wcErrorStatus);
            swprintf(wcErrorStatus, "OK");
        }
    } else {
        swprintf(wcLocalStatus, "WORKING");
    }

    if(strdif(m_cStatusRequest, "PRIORITYREQUEST") == 0) {
        swprintf(wcLocalStatus, "%hd", iPriorityRequest);
        sprintf(m_cStatusRequest, "RENDERING");
    }
    if(strdif(m_cStatusRequest, "IS_GL_WINDOW_UP") == 0) {
        if(iGLwindowIsMade == 1) { swprintf(wcLocalStatus, "GL_WINDOW_IS_UP");   }
        else                     { swprintf(wcLocalStatus, "GL_WINDOW_IS_DOWN"); }
        sprintf(m_cStatusRequest, "RENDERING");
    }
    if(strdif(m_cStatusRequest, "GET_ALL_PREVIOUS_OPTIONS") == 0) {
        int i, iNChar=0;
        for(i=0; i<iOpNumber; i++) {
            swprintf(&wcLocalStatus[iNChar], "%s %s\n", wcOpKeys[i], wcOpVals[i]);
            iNChar += wcslen(wcOpKeys[i]) + 2 + wcslen(wcOpVals[i]);
        }
        sprintf(m_cStatusRequest, "RENDERING");
    }

#ifdef linux
    size_t lAllocSize = (wcslen(wcLocalStatus)+1) * sizeof(wchar_t);
    *wcStatus = (char *)malloc(lAllocSize);
#else
    long lAllocSize = (wcslen(wcLocalStatus)+1) * sizeof(wchar_t);
    *wcStatus = static_cast<wchar_t *>(CoTaskMemAlloc(lAllocSize));
#endif
    wcscpy(*wcStatus, wcLocalStatus);

    return S_OK;
}
#endif
STDMETHODIMP Chvr_work::GetPix(BinaryDataHVR *pData)
{
    int nbytes, n_padbytes;

    if(iSpawnRenderThread == 0) {  // >>>>
        DoGetPix();
    } else {
        // Set Rendering thread to run DoGetPix and wait for it to finish
        iDoThis = 5;
        while(iDoThis == 5) { _sleep(10); }
    }

    n_padbytes = (m_ShvrUse.iWidth*3)%4;
    if(n_padbytes!=0)
        nbytes = m_ShvrUse.iHeight * ((m_ShvrUse.iWidth*3)+(4-n_padbytes));
    else    
        nbytes = m_ShvrUse.iHeight * m_ShvrUse.iWidth * 3;
        
    pData->m_lSize = (long)nbytes;
#ifdef linux
    pData->m_pBuffer = (unsigned char *)malloc(pData->m_lSize);
#else
    pData->m_pBuffer = static_cast<unsigned char *>(CoTaskMemAlloc(pData->m_lSize));
#endif
    memcpy(pData->m_pBuffer, ucGPix, sizeof(unsigned char) * nbytes);

    return S_OK;
}

STDMETHODIMP Chvr_work::PutPix(long lSize, unsigned char *cPix)
{
    memcpy(ucGPix, cPix, sizeof(unsigned char)*3*m_ShvrUse.iWidth*m_ShvrUse.iHeight);

    if(iSpawnRenderThread == 0) {  // >>>>
        DoPutPix();
    } else {
        /* Set Rendering thread to run DoGetPix and wait for it to finish */
        iDoThis = 4;
        // if(iUseOpenGL == 1) { glutPostRedisplay(); }
        while(iDoThis == 4) { _sleep(10); }
    }
    return S_OK;
}

void Chvr_work::DoGetPix()
{    
#ifdef TEMP_TEST
    int i = 0;
    int size = m_ShvrUse.iWidth*m_ShvrUse.iHeight*3;
    for(i=0;i<size;i+=3){
        ucGPix[i] = 255;
        ucGPix[i+1] = 0;
        ucGPix[i+2] = 0;
    }
#else
    glReadPixels(0,0,m_ShvrUse.iWidth,m_ShvrUse.iHeight,GL_RGB,GL_UNSIGNED_BYTE,ucGPix);
#endif
    glFlush();
    glFinish();
}

void Chvr_work::DoPutPix()
{
    glMatrixMode(GL_TEXTURE);
    glDisable(GL_TEXTURE_3D);
    glDisable(GL_BLEND);
    glDrawPixels(m_ShvrUse.iWidth,m_ShvrUse.iHeight,GL_RGB,GL_UNSIGNED_BYTE,ucGPix);
    glFlush();
    glFinish();
    glEnable(GL_BLEND);
    glEnable(GL_TEXTURE_3D);
    if(m_iNRendBuffs == 2) { mySwapBuffers(); }
}

void Chvr_work::SaveOption()
{
    /* Recorded option for later retrieval: replace if already there */
    int i, iMatch = iOpNumber;
    for(i=0; i<iOpNumber; i++) {
        if(wcscmp(wcKey, wcOpKeys[i]) == 0) { iMatch = i; }
    }
    wcscpy(wcOpKeys[iMatch], wcKey);
    wcscpy(wcOpVals[iMatch], wcValue);
    if(iMatch == iOpNumber) { iOpNumber++; }
}

void Chvr_work::AddOption()
{
    /* Recored option for later retrieval: always add to list of options */
    if(iOpNumber < MAX_OPTIONS) {
        wcscpy(wcOpKeys[iOpNumber], wcKey);
        wcscpy(wcOpVals[iOpNumber], wcValue);
        iOpNumber++;
    }
}

void Chvr_work::ClearOption(wchar_t wcClearOp[256])
{
    /* Clear list of options of all occurances of wcClearOp */
    int i, j;
    i = 0;
    while(i < iOpNumber) {
        if(wcscmp(wcOpKeys[i], wcClearOp) == 0) {
            for(j=i+1; j<iOpNumber; j++) {
                wcscpy(wcOpKeys[j-1], wcOpKeys[j]);
                wcscpy(wcOpVals[j-1], wcOpVals[j]);
            }
            iOpNumber--;
        } else {
            i++;
        }
    }
}
